#!/opt/edge/venv/bin/python3.11

from edge import eventlog
from edge.webserver import cgi
from edge.webserver.origins import alarm


class _HandlerFactory:
    """
    GET   /api/v2/origins/alarm   /
    PUT   /api/v2/origins/alarm   /<origin>/<name>
    """
    @classmethod
    def create(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        method_to_handlers_map: cgi.CgiHandlerFactoryDict = {
            'GET': [
                cls._alarm__get,
            ],
            'PUT': [
                cls._alarm__put_origin_name
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    @classmethod
    def _alarm__get(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 0:
            raise cgi.CgiError404

        request.assert_query_present()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            origin = None
            name = None
            for name_value in request.query_params:
                name, value = name_value.split('=', 1)
                if name == 'name':
                    name = value
                elif name == 'origin':
                    origin = value

            if None in [origin, name]:
                raise cgi.CgiError400

            result = alarm.get_alarm(origin, name)
            rc = cgi.CGI_R200_SUCCESS if result else cgi.CGI_R404_NOT_FOUND
            return cgi.CgiResponse(status=rc)

        return handler

    @classmethod
    def _alarm__put_origin_name(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            origin = request.path_info[0]
            name = request.path_info[1]
            status = cgi.CGI_R200_SUCCESS if alarm.add_alarm(origin, name) else cgi.CGI_R404_NOT_FOUND
            return cgi.CgiResponse(status=status)

        return handler


if __name__ == '__main__':
    eventlog.open_log('cgi-alarm', 'debug')
    try:
        cgi.cgi_handle(_HandlerFactory.create)
    finally:
        eventlog.close_log()
