#!/opt/edge/venv/bin/python3.11

from edge import eventlog
from edge.webserver import cgi
from edge.webserver.sysconf.api_wiegand_pcb import WiegandPcbApi


class _HandlerFactory:
    """
    GET      /api/v2/sys/conf/wiegand-pcb.input-port   /<uiref>
    PUT      /api/v2/sys/conf/wiegand-pcb.input-port   /<uiref>

    GET      /api/v2/sys/conf/wiegand-pcb.input-port   /<uiref>/stats
    """

    @classmethod
    def create(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        method_to_handlers_map: cgi.CgiHandlerFactoryDict = {
            'GET': [
                cls._input_port__get_uiref,
                cls._input_port__get_uiref_stats
            ],
            'PUT': [
                cls._input_port__put_uiref
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    @classmethod
    def _input_port__get_uiref(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            uiconf_pub = WiegandPcbApi.load_input_unit(request.path_info[0])
            return cgi.CgiResponse(body=uiconf_pub)

        return handler

    @classmethod
    def _input_port__put_uiref(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        def handler() -> cgi.CgiResponse:
            new_uiconf = WiegandPcbApi.save_input_unit(request.path_info[0], request.body_jnode)
            return cgi.CgiResponse(body=new_uiconf)

        return handler

    @classmethod
    def _input_port__get_uiref_stats(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[1] != 'stats':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            uiconf_pub = WiegandPcbApi.load_input_unit_stats(request.path_info[0])
            return cgi.CgiResponse(body=uiconf_pub)

        return handler


if __name__ == '__main__':
    eventlog.open_log('cgi-wiegand-pcb-input-port', 'debug')
    try:
        cgi.cgi_handle(_HandlerFactory.create)
    finally:
        eventlog.close_log()
