import os
import inspect
import syslog
import traceback


def _get_final_entry_text(entry_level: str, entry_text: str) -> str:
    template = '[{entry_level}|{caller_file}:{caller_line}|{caller_class}::{caller_fn}] {entry_text}'

    caller_stack_frame = inspect.stack()[2][0]
    caller_file = os.path.basename(inspect.getfile(caller_stack_frame))
    caller_line = inspect.getlineno(caller_stack_frame)
    caller_fn = inspect.stack()[2][3]
    try:
        caller_class = caller_stack_frame.f_locals["self"].__class__.__name__
    except KeyError:
        template = '[{entry_level}|{caller_file}:{caller_line}|{caller_fn}] {entry_text}'

    return template.format(**locals())


def open_log(module_ident: str, log_level: str = 'info') -> None:
    edge_to_syslog_log_level_map = {
        'trace': syslog.LOG_DEBUG,
        'debug': syslog.LOG_DEBUG,
        'info': syslog.LOG_INFO,
        'warn': syslog.LOG_WARNING,
        'error': syslog.LOG_ERR,
        'none': syslog.LOG_DEBUG
    }

    syslog.openlog(ident=module_ident, logoption=syslog.LOG_PID, facility=syslog.LOG_DAEMON)
    syslog_log_level = edge_to_syslog_log_level_map[log_level]
    priority_mask = syslog.LOG_UPTO(syslog_log_level)
    syslog.setlogmask(priority_mask)


def close_log() -> None:
    syslog.closelog()


def log_d(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_DEBUG, _get_final_entry_text('d', entry_text))


def log_debug(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_DEBUG, _get_final_entry_text('d', entry_text))


def log_i(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_INFO, _get_final_entry_text('i', entry_text))


def log_info(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_INFO, _get_final_entry_text('i', entry_text))


def log_w(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_WARNING, _get_final_entry_text('w', entry_text))


def log_warning(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_WARNING, _get_final_entry_text('w', entry_text))


def log_e(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_ERR, _get_final_entry_text('e', entry_text))


def log_error(entry_text: str) -> None:
    syslog.syslog(syslog.LOG_ERR, _get_final_entry_text('e', entry_text))


def log_e_param(param_name: str) -> None:
    syslog.syslog(syslog.LOG_ERR, _get_final_entry_text('e', f'param {param_name} has invalid value'))


def log_x() -> None:
    log_entries = traceback.format_exc().split('\n')
    for log_entry in log_entries:
        syslog.syslog(syslog.LOG_ERR, _get_final_entry_text('e', log_entry))


def log_debug_last_exc() -> None:
    log_entries = traceback.format_exc().split('\n')
    for log_entry in log_entries:
        syslog.syslog(syslog.LOG_ERR, _get_final_entry_text('e', log_entry))
