from datetime import datetime
from typing import Final, Generic, List, Optional, TypeVar

from edge import eventlog
from edge.image import contrast
from edge.image.helpers import ImHelpers
from edge.image.improc import ImProc
from edge.image.profiles.live import LiveImProfile
from edge.image.profiles.overview import OverviewImProfile
from edge.image.profiles.plate import PlateImProfile
from edge.image.profiles.scene import SceneImProfile
from edge.image.profiles.thumb import ThumbImProfile
from edge.image.profiles.trainer import TrainerImProfile
from edge.image.profiles.vehicle import VehicleImProfile
from edge.image.types.footer import ImFooterVars
from edge.types.geoposition import Geoposition
from edge.types.image import ImRgba, ImJpeg
from edge.image.types.types import EventMeasImBuilderConf, LiveImConf

from edge.types.blob import MemBlob
from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectF
from edge.types.track import ITrack

from edge.support.bbox import mk_vehicle_bbox


class LiveImBuilder:
    def __init__(self, conf: LiveImConf):
        self._profile = LiveImProfile(conf)

    def build_scene(self, scene_dh_data: ImRgba, footer_vars: ImFooterVars) -> MemBlob:
        return self._profile.process(scene_dh_data, footer_vars)


class TrainerImBuilder:
    def __init__(self, wh_ratio: float = 4.0 / 3.0):
        self._wh_ratio = wh_ratio
        self._profile = TrainerImProfile()

    @staticmethod
    def build_scene(scene_dh_data: ImRgba) -> MemBlob:
        return MemBlob('scene', scene_dh_data.ts, ImHelpers.mk_jpeg2(scene_dh_data))

    def build_vehicle(self, scene_dh_data: ImRgba, plate_bbox: RectF) -> MemBlob:
        try:
            vehicle_bbox = mk_vehicle_bbox(plate_bbox, scene_dh_data.wh, self._wh_ratio)
        except ValueError as x:
            eventlog.log_d(f'fail: define vehicle-rect ({x})')
            vehicle_bbox = None

        return self._profile.process(scene_dh_data, vehicle_bbox)


class EventOverCamImBuilder:
    """
    Builder of event images which come from overview camera.
    """
    def __init__(self):
        self._profile = OverviewImProfile()

    def build_scene(self, scene_dh_data: ImJpeg) -> MemBlob:
        return self._profile.process(scene_dh_data)


class EventMeasCamImBuilder:
    """
    Builder of event images which come from measurement camera.
    """
    def __init__(self, conf: EventMeasImBuilderConf):

        self._plate_profile = PlateImProfile(conf.plate)
        self._thumb_profile = ThumbImProfile(conf.thumb)

        self._vehicle_profile = VehicleImProfile(conf.vehicle) if conf.vehicle else None
        self._vehicle_wh_ratio = conf.vehicle.wh_ratio if conf.vehicle else 4.0 / 3.0
        self._vehicle_schedule = contrast.create_schedule(conf.vehicle_contrast.shedule)
        self._vehicle_improc = ImProc(conf.vehicle_contrast.clache)

        self._scene_profile = SceneImProfile(conf.scene)
        self._scene_schedule = contrast.create_schedule(conf.scene_contrast.shedule)
        self._scene_improc = ImProc(conf.scene_contrast.clache)

    @property
    def scene_dh(self) -> int:
        return self._scene_profile.dh

    def build(self,
              scene_dh_data: ImRgba,
              scene_dt_utc: datetime,
              scene_geopos: Optional[Geoposition],
              footer_vars: ImFooterVars,
              track: Optional[ITrack],
              plate_quad: QuadI,
              vehicle: bool = False,
              scene: bool = False,
              plate: bool = False,
              thumb: bool = False,
              pixelate_bboxes: Optional[List[RectF]] = None,
              draw_bboxes: Optional[List[RectF]] = None
              ) -> List[MemBlob]:

        plate_bbox = RectF(*plate_quad.bbox_recti.ltrb)
        try:
            vehicle_bbox = mk_vehicle_bbox(plate_bbox, scene_dh_data.wh, self._vehicle_wh_ratio)
        except ValueError as x:
            eventlog.log_d(f'fail: define vehicle-rect ({x})')
            vehicle_bbox = None

        # strip plate invisible parts
        visible_plate_bbox = plate_bbox
        if not plate_bbox.get_visible_100(scene_dh_data.w, scene_dh_data.h):
            visible_plate_bbox = RectF(plate_bbox.fl, plate_bbox.ft, plate_bbox.fr, plate_bbox.fb)
            visible_plate_bbox.strip_invisible(scene_dh_data.w, scene_dh_data.h)

        # The order of calls is significant
        blobs = []

        # plate
        if plate:
            blobs.extend(self._plate_profile.process(
                scene_dh_data, visible_plate_bbox))

        # thumb
        if thumb:
            blobs.extend(self._thumb_profile.process(
                scene_dh_data, visible_plate_bbox, vehicle_bbox))

        improve_vehicle_contrast = self._vehicle_schedule.test_active(scene_dt_utc, scene_geopos) if vehicle else False
        improve_scene_contrast = self._scene_schedule.test_active(scene_dt_utc, scene_geopos) if scene else False

        vehicle_improc: Optional[ImProc] = None
        scene_improc: Optional[ImProc] = None
        if improve_vehicle_contrast and improve_scene_contrast:
            vehicle_improc = self._vehicle_improc
            scene_improc = self._scene_improc

        elif improve_vehicle_contrast:
            vehicle_improc = self._vehicle_improc

        elif improve_scene_contrast:
            scene_improc = self._scene_improc

        # vehicle
        if vehicle:
            blobs.extend(self._vehicle_profile.process(
                scene_dh_data, vehicle_bbox, pixelate_bboxes, draw_bboxes, footer_vars, plate_quad, vehicle_improc))

        # scene
        if scene:
            blobs.extend(self._scene_profile.process(
                scene_dh_data, visible_plate_bbox, vehicle_bbox, pixelate_bboxes, draw_bboxes, footer_vars, track,
                plate_quad, scene_improc))

        return blobs


TImBuilder = TypeVar('TImBuilder', bound=EventMeasCamImBuilder)


class ImBuilderEnv(Generic[TImBuilder]):
    """
    A typical forkwlow to create an evidience of an event is as follows:
        1) For each waypoint load all required images (scene) from the videosources.
        1.1) The allocated memory for the image (scene) should include extra space (dh) for its footer.
        2) For each waypoint build required event images (scene, vehicle, thumb, plate).

    Since all the images should be loaded (step 1) before its processing (step 2), the memory for the images
    should be allocated. The memory allocation/reallocation is done by the ImRgba class.

    The TImBuilder is a type of the builder which is used to build the event images.
    It holds the configuration for the event images

    So for each waypoint for each videosource there should be an instance of ImBuilderEnv class.
    """
    def __init__(self, builder: TImBuilder):
        self.scene_rgba: Final[ImRgba] = ImRgba()
        self.builder: Final[TImBuilder] = builder

    @property
    def scene_dh(self) -> int:
        return self.builder.scene_dh
