import datetime

from typing import Dict, List

from edge import eventlog
from edge.types.exceptions import JnodeError
from edge.image.types.types import (SceneImAnnoConf, VehicleImConf, VehicleImAnnoConf,
                                    SceneImConf, EventMeasImBuilderConf, ContrastScheduleConf, ContrastConf, ContrastClaheConf)
from edge.image.types.footer import ImFooterColConf, ImFooterRowConf, ImFooterConf

from edge.webserver.sysconf.uiconf.constroller import CtrlUiconf


def decode__contrast_shedule(shedule_jnode: List) -> ContrastScheduleConf:
    time_from = None
    time_to = None
    if len(shedule_jnode) == 0:
        mode = ContrastScheduleConf.MODE_NIGHT

    elif len(shedule_jnode) == 1:
        mode = shedule_jnode[0]
        if mode not in [ContrastScheduleConf.MODE_ALWAYS, ContrastScheduleConf.MODE_NEVER,
                        ContrastScheduleConf.MODE_NIGHT]:
            eventlog.log_e_param('#/mode')
            raise JnodeError

    elif len(shedule_jnode) == 2:
        mode = shedule_jnode[0]
        schedule_jnode = shedule_jnode[1]

        if mode != ContrastScheduleConf.MODE_TIME:
            eventlog.log_e_param('#/mode')
            raise JnodeError

        # deserialize #/time-from
        time_from_jnode = schedule_jnode.get('time-from')

        if not type(time_from_jnode) in [list, tuple] or len(time_from_jnode) != 2:
            eventlog.log_e_param('#/time-from')
            raise JnodeError

        hh, mm = time_from_jnode
        if not (int == type(hh) and 0 <= hh <= 23):
            eventlog.log_e_param('#/time-from/0')
            raise JnodeError

        if not (int == type(mm) and 0 <= mm <= 59):
            eventlog.log_e_param('#/time-from/1')
            raise JnodeError

        time_from = datetime.time(hh, mm)

        # deserialize #/time-to
        time_to_jnode = schedule_jnode['time-to']
        if not type(time_to_jnode) in [list, tuple] or len(time_to_jnode) != 2:
            eventlog.log_e_param('#/time-to')
            raise JnodeError

        hh, mm = time_to_jnode
        if not (int == type(hh) and 0 <= hh <= 23):
            eventlog.log_e_param('#/time-to/0')
            raise JnodeError

        if not (int == type(mm) and 0 <= mm <= 59):
            eventlog.log_e_param('#/time-to/1')
            raise JnodeError

        time_to = datetime.time(hh, mm)

    else:
        mode = ContrastScheduleConf.MODE_NEVER

    return ContrastScheduleConf(mode, time_from, time_to)


def decode__contrast_clahe(clache_jnode: Dict) -> ContrastClaheConf:
    clip_limit = clache_jnode.get('clip-limit', 10.0)
    grid_size = clache_jnode.get('grid-size')

    if grid_size is not None:
        if type(grid_size) != list:
            eventlog.log_e_param('#/grid-size')
            raise JnodeError
        elif len(grid_size) != 2:
            eventlog.log_e_param('#/grid-size')
            raise JnodeError
        elif type(grid_size[0]) != int:
            eventlog.log_e_param('#/grid-size/0')
            raise JnodeError
        elif type(grid_size[1]) != int:
            eventlog.log_e_param('#/grid-size/1')
            raise JnodeError

        grid_rows = grid_size[0]
        grid_cols = grid_size[1]

    else:
        grid_cols = 8
        grid_rows = 8

    return ContrastClaheConf(clip_limit, grid_cols, grid_rows)


def decode__contrast_conf(args_jnode: Dict) -> ContrastConf:
    """
    :raise JnodeError:
    """
    shedule_jnode = args_jnode.get('schedule', [])
    shedule = decode__contrast_shedule(shedule_jnode)

    clache_jnode = args_jnode.get('clahe', {})
    clache = decode__contrast_clahe(clache_jnode)
    return ContrastConf(shedule=shedule, clache=clache)


def decode__vehicle_im_conf(profile_jnode: Dict) -> VehicleImConf:
    """
    :raise JnodeError:
    """
    # deserialize #/width
    image_w = profile_jnode['width']
    if not (type(image_w) == int and 320 <= image_w <= 800):
        eventlog.log_e_param('#/width')
        raise JnodeError

    # deserialize #/footer
    footer_jnode = profile_jnode.get('footer')
    if dict != type(footer_jnode):
        eventlog.log_e_param('#/footer')
        raise JnodeError

    footer_args = decode__im_footer_conf(footer_jnode)

    # deserialize annotations
    anno_jnode = profile_jnode.get('annotations')
    anno = decode__vehicle_im_anno_conf(anno_jnode) if anno_jnode else VehicleImAnnoConf()

    return VehicleImConf(image_w, footer_args, anno)


def decode__vehicle_im_anno_conf(anno_jnode: Dict) -> VehicleImAnnoConf:
    """
    :raise JnodeError:
    """

    vehicle_jnode = anno_jnode.get('vehicle')
    if not vehicle_jnode:
        return VehicleImAnnoConf()

    plate = vehicle_jnode.get('plate', False)

    return VehicleImAnnoConf(plate)


def decode__scene_im_anno_conf(anno_jnode: Dict) -> SceneImAnnoConf:
    """
    :raise JnodeError:
    """

    vehicle_jnode = anno_jnode.get('vehicle')
    if not vehicle_jnode:
        return SceneImAnnoConf()

    track = vehicle_jnode.get('track', False)
    body = vehicle_jnode.get('body', False)
    plate = vehicle_jnode.get('plate', False)

    return SceneImAnnoConf(track, body, plate)


def decode__scene_im_conf(profile_jnode: Dict) -> SceneImConf:
    """
    :raise JnodeError:
    """
    # deserialize image-width
    image_w = profile_jnode['width']

    # deserialize footer-args
    footer_jnode = profile_jnode.get('footer')
    if not footer_jnode:
        eventlog.log_e('fail: param #/footer has invalid value')
        raise JnodeError

    footer = decode__im_footer_conf(footer_jnode)

    # deserialize annotations
    anno_jnode = profile_jnode.get('annotations')
    anno = decode__scene_im_anno_conf(anno_jnode) if anno_jnode else SceneImAnnoConf()

    return SceneImConf(image_w, footer, anno)


def decode__im_footer_col_conf(col_jnode: Dict) -> ImFooterColConf:
    """
    :raise JnodeError:
    """
    if dict != type(col_jnode):
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize #/width
    width = col_jnode.get('width')
    if not (int == type(width) and 1 <= width <= 65535):
        eventlog.log_e_param('#/width')
        raise JnodeError

    # deserialize #/align
    raw_align = col_jnode.get('align')
    if not type(raw_align) in [str]:
        eventlog.log_e_param('#/align')
        raise JnodeError

    if raw_align == 'left':
        align = -1
    elif raw_align == 'center':
        align = 0
    elif raw_align == 'right':
        align = +1
    else:
        eventlog.log_e_param('#/align')
        raise JnodeError

    # deserialize #/format
    template = col_jnode.get('format')
    if not (type(raw_align) in [str] and 0 <= len(template) <= 128):
        eventlog.log_e_param('#/format')
        raise JnodeError

    return ImFooterColConf(width, align, template)


def decode__im_footer_row_conf(row_jnode: Dict) -> ImFooterRowConf:
    """
    :raise JnodeError:
    """
    if dict != type(row_jnode):
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize #/enabled
    enabled = row_jnode.get('enabled')
    if bool != type(enabled):
        eventlog.log_e_param('#/enabled')
        raise JnodeError

    if enabled:
        cols_jnode = row_jnode.get('cols')
        if list != type(cols_jnode):
            eventlog.log_e_param('#/cols')
            raise JnodeError

        cols = list(map(decode__im_footer_col_conf, cols_jnode))

    else:
        cols = []

    return ImFooterRowConf(enabled, cols)


def decode__im_footer_conf(footer_jnode: Dict) -> ImFooterConf:
    """
    :raise JnodeError:
    """
    # deserialize #/font-height
    font_size = footer_jnode.get('font-height')
    if not (int == type(font_size) and 8 <= font_size <= 72):
        eventlog.log_e_param('#/font-height')
        raise JnodeError

    # deserialize #/rows
    rows_jnode = footer_jnode.get('rows')
    if not (list == type(rows_jnode) and len(rows_jnode) > 0):
        eventlog.log_e_param('#/rows')
        raise JnodeError

    rows = [row for row in map(decode__im_footer_row_conf, rows_jnode) if row.enabled]

    if len(rows) > 0:
        height = int((1.4 * font_size) * len(rows) + 0.4 * font_size)  # 120%
    else:
        # no active rows in footer
        height = 0

    return ImFooterConf(height, font_size, rows)


def decode__im_builder_conf(jnode: Dict) -> EventMeasImBuilderConf:
    """
    :raise JnodeError:
    """
    #
    scene_jnode = jnode.get('scene')
    if not scene_jnode:
        eventlog.log_e_param('#/scene')
        raise JnodeError

    scene_im = decode__scene_im_conf(scene_jnode)

    scene_footer = scene_jnode.get('footer')
    if not scene_footer:
        eventlog.log_e_param('#/scene/footer')
        raise JnodeError

    scene_contrast = decode__contrast_conf(CtrlUiconf.get_contrast_scene())

    #
    vehicle_jnode = jnode.get('vehicle')
    if vehicle_jnode:
        vehicle_im = decode__vehicle_im_conf(vehicle_jnode)
        vehicle_footer = vehicle_jnode.get('footer')
        if not vehicle_footer:
            eventlog.log_e_param('#/vehicle/footer')
            raise JnodeError
    else:
        vehicle_im = None
        vehicle_footer = None

    vehicle_contrast = decode__contrast_conf(CtrlUiconf.get_contrast_vehicle())

    return EventMeasImBuilderConf(
        scene_im=scene_im,
        scene_footer=scene_footer,
        scene_contrast=scene_contrast,
        vehicle_im=vehicle_im,
        vehicle_footer=vehicle_footer,
        vehicle_contrast=vehicle_contrast)
