from typing import Final, List, Optional, Tuple

from PIL import ImageFont, Image

from edge.image.helpers import ImHelpers
from edge.image.improc import ImProc
from edge.image.profiles import FONT_PATH
from edge.types.blob import MemBlob, JsonBlobMeta
from edge.image.types.footer import ImFooterVars
from edge.types.image import ImRgba, ImTrans
from edge.image.types.types import VehicleImConf

from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectF, RectI


class VehicleImProfile:
    def __init__(self, conf: VehicleImConf):
        self._conf: Final[VehicleImConf] = conf
        self._rgba_scaled_: Final[ImRgba] = ImRgba()

        self._font = ImageFont.truetype(FONT_PATH, conf.footer.font_size, encoding='unic')

    def _preprocess(self,
                    scene_dh_data: ImRgba,
                    vehicle_bbox: RectF,
                    improc: Optional[ImProc]) -> Tuple[ImRgba, ImTrans | None]:

        vehicle_bboxi = RectI(*vehicle_bbox.ltrb)
        scaled_im = ImHelpers.mk_crop_resize_image(
            scene_dh_data, vehicle_bboxi, self._conf.width, self._conf.height)

        dh_data: ImRgba = self._rgba_scaled_
        dh_data.reset(scaled_im.size[0], scaled_im.size[1], self._conf.footer.height, scene_dh_data.ts)
        dh_trans = ImHelpers.mk_crop_resize_trans(vehicle_bbox, self._conf.width, self._conf.height)

        if improc is None:
            dh_data.with_image(lambda im: im.paste(scaled_im, (0, 0)))
        else:
            improc.improve_contrast_rgba(scaled_im, dh_data)

        return dh_data, dh_trans

    def process(self, scene_dh_data: ImRgba,
                vehicle_bbox: Optional[RectF],
                pixelate_bboxes: Optional[List[RectF]],
                draw_bboxes: Optional[List[RectF]],
                footer_vars: ImFooterVars,
                plate_quad: QuadI,
                improc: Optional[ImProc]) -> List[MemBlob]:

        if not vehicle_bbox:
            return []

        dh_data, dh_trans = self._preprocess(scene_dh_data, vehicle_bbox, improc)

        # render footer
        footer_conf = self._conf.footer.eval(footer_vars)
        dh_data.with_footer_image(
            lambda im: ImHelpers.draw_footer(im, footer_conf, self._font, self._conf.footer.font_size))

        # render annotations
        def handler(im: Image.Image) -> bytes:
            if self._conf.anno.plate:
                ImHelpers.draw_poly(im, plate_quad, dh_trans)

            for pixelate_bbox in pixelate_bboxes or []:
                if dh_trans:
                    pixelate_bbox = dh_trans.map_rectf(pixelate_bbox)
                else:
                    pixelate_bbox = pixelate_bbox

                ImHelpers.pixelate_roi(im, RectI(*pixelate_bbox.ltrb))
                ImHelpers.draw_rect(im, pixelate_bbox, color='yellow')

            for draw_bbox in draw_bboxes or []:
                if dh_trans:
                    draw_bbox = dh_trans.map_rectf(draw_bbox)
                else:
                    draw_bbox = draw_bbox

                ImHelpers.draw_rect(im, draw_bbox, None)

            return ImHelpers.mk_jpeg(im)

        blob_data: bytes = dh_data.with_image(handler)
        blob_meta = JsonBlobMeta({
            'trans': dh_trans.serialize(),
            'size': [dh_data.w, dh_data.h + dh_data.dh, len(blob_data)]
        })

        return [MemBlob('vehicle', scene_dh_data.ts, blob_data, blob_meta)]
