import hashlib
import os
import subprocess

from typing import Callable, Dict, List, Optional, Tuple, TypedDict

try:
    from typing import NotRequired
except ImportError:
    from typing_extensions import NotRequired

from edge import eventlog
from edge.support import fs, net
from edge.types.geoposition import Geoposition
from edge.types.origin import Origin
from edge.types.location import Location
from edge.services.types.gnss import IGnssService
from edge.services.types.about import AboutServiceConf, IAboutService
from edge.webserver.sysconf.ph import CTRL_UICONF_PATH


SW_DIRPATH = '/opt/edge/share/dpkg/info'


class SwFileInfo(TypedDict):
    path: str
    md5: str


SwPackageAuthor = str
SwPackageName = str
SwPackageVer = str


class SwPackageInfo(TypedDict):
    package: Tuple[SwPackageAuthor, SwPackageName, SwPackageVer]
    files: NotRequired[List[SwFileInfo]]


class AboutError(RuntimeError):
    pass


def _attach_files_md5(sw_files_jnode: List[SwFileInfo]) -> List[SwFileInfo]:
    """
    :param sw_files_jnode:
        *.sw/#/files
    """
    blocksize = 65536

    for file_jnode in sw_files_jnode:
        hasher = hashlib.md5()
        file_path = file_jnode['path']
        with open(file_path, 'rb') as f:
            buf = f.read(blocksize)
            while len(buf) > 0:
                hasher.update(buf)
                buf = f.read(blocksize)

        file_jnode['md5'] = hasher.hexdigest()

    return sw_files_jnode


def get_hw_version() -> Optional[Dict]:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    return about_jnode.get('hw')


def get_hw_serial() -> Optional[str]:
    about_hw = get_hw_version()
    if about_hw is not None and len(about_hw) == 3:
        return about_hw[2]
    else:
        return None


def get_hw_name() -> Optional[str]:
    about_hw = get_hw_version()
    if about_hw is not None and len(about_hw) == 3:
        return about_hw[1]
    else:
        return None


def get_sw_name() -> Optional[dict]:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    return about_jnode.get('sw-name')


def get_sw_lang() -> Optional[List[str]]:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    return about_jnode.get('sw-lang')


def _get_sw_version_etc() -> Optional[str]:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    if about_jnode:
        return about_jnode.get('sw-ver')


def _get_sw_version_opt() -> str:
    with open('/opt/edge/semver', 'r') as text_file:
        return text_file.readline()


def get_sw_version(sw_files_dir: str) -> List[SwPackageInfo]:
    """
    :param sw_files_dir:
        /opt/edge/share/dpkg/info/
    """
    ctrl_sw_version: str = _get_sw_version_etc() or _get_sw_version_opt()
    ctrl_sw_packages: List[SwPackageInfo] = [{
        'package': ('VIT', 'edge', ctrl_sw_version)
    }]

    try:
        sw_file_names: List[str] = os.listdir(sw_files_dir)
    except os.error as x:
        eventlog.log_d('fail: os.listdir <{}>'.format(x))
        return ctrl_sw_packages

    # The default format string is “${binary:Package}\t${Version}\n”.
    try:
        cli_params = ['dpkg-query', '--show'] + [name[:-3] for name in sw_file_names if name.endswith('.sw')]
        shell_result = subprocess.run(cli_params, capture_output=True, text=True, encoding='ascii',
                                      check=True, timeout=10.0)

    except subprocess.TimeoutExpired as x:
        eventlog.log_e('fail: <{}> (timeout:{})'.format(x.cmd, x.timeout))
        raise AboutError

    except subprocess.CalledProcessError as x:
        eventlog.log_e('fail: <{}> (returncode:{})'.format(x.cmd, x.returncode))
        raise AboutError

    for line in shell_result.stdout.split('\n'):
        if not line:
            continue

        deb_name_arch, deb_version = line.split(None, 1)
        deb_name = deb_name_arch.split(':', 1)[0]

        sw_path = fs.join_path(sw_files_dir, deb_name + '.sw')
        sw_jnode = fs.load_json3(sw_path, default=dict())
        sw_files_jnode: List[Dict] = sw_jnode.get('files', [])

        if sw_files_jnode:
            ctrl_sw_packages.append({
                'package': ('VIT', deb_name, deb_version),
                'files': _attach_files_md5(sw_files_jnode)
            })
        else:
            ctrl_sw_packages.append({
                'package': ('VIT', deb_name, deb_version)
            })

    return ctrl_sw_packages


class _Service(IAboutService):
    def __init__(self, args: AboutServiceConf, geolocation_service: IGnssService):
        IAboutService.__init__(self)
        self._gpsd = geolocation_service

        def get_hw_info() -> Tuple[str, str, str]:
            jnode = fs.load_json3(args.server_about_path, default=dict())
            hw_jnode = jnode.get('hw') or list()
            return hw_jnode if len(hw_jnode) == 3 else ['', '', '']

        def mk_origin_factory() -> Callable[[Optional[Geoposition], str], Origin]:
            jnode = fs.load_json3(args.origin_about_path, default=dict())
            o_jnode = jnode.get('origin') or dict()

            uuid = o_jnode['uuid']
            ctime = o_jnode['time']
            place = o_jnode.get('place') or ''
            name = o_jnode.get('name') or ''

            return lambda geolocation, lane: Origin(uuid,
                                                    ctime,
                                                    name,
                                                    Location(place=place, lane=lane),
                                                    geolocation)

        self._hw_info = get_hw_info()
        self._mk_origin = mk_origin_factory()
        self._sw_info = get_sw_version(SW_DIRPATH)
        self._sw_name = get_sw_name() or 'EDGE'
        self._base_url = None

    @property
    def hw_info(self) -> Tuple[str, str, str]:
        return self._hw_info

    @property
    def sw_info(self) -> List[Dict]:
        return self._sw_info

    @property
    def sw_name(self) -> str:
        return self._sw_name

    @property
    def base_url(self) -> str:
        if self._base_url is None:
            try:
                endpoint = net.get_http_api_endpoint()
                self._base_url = f'{endpoint}/api/v2/origins/event/dist'
            except net.NetError:
                eventlog.log_w('fail: define <base_url>')

        return self._base_url

    def get_origin(self, ts: int, lane: str = '') -> Origin:
        geoposition = self._gpsd.get_geoposition(ts)
        return self._mk_origin(geoposition, lane)

    @property
    def hw_vendor(self) -> str:
        return self._hw_info[0]

    @property
    def hw_model(self) -> str:
        return self._hw_info[1]

    @property
    def hw_serial(self) -> str:
        return self._hw_info[2]


def init_service(args: AboutServiceConf, gpsd_service: IGnssService) -> IAboutService:
    return _Service(args, gpsd_service)


# noinspection PyUnusedLocal
def fini_service(service: IAboutService):
    service.destroy()
