import sortedcontainers
import time
import ujson

from twisted.internet.protocol import ReconnectingClientFactory, connectionDone, Protocol
from twisted.protocols.basic import LineOnlyReceiver
from typing import Optional, Tuple, Callable, Dict

from astral.location import Location, LocationInfo

from edge import eventlog
from edge.services import tz
from edge.services.types.gnss import GnssServiceConf, IGnssService
from edge.support import isotime
from edge.types.geoposition import Geoposition

_TPV_MODE_NONE = 0
_TPV_MODE_NO_FIX = 1
_TPV_MODE_2D_FIX = 2
_TPV_MODE_3D_FIX = 3


class _GeodataCounters:
    def __init__(self):
        self.mts: Optional[int] = None

        self.tpv_ts: Optional[int] = None
        self.tpv_lat: Optional[float] = None
        self.tpv_lon: Optional[float] = None
        self.tpv_alt: Optional[float] = None

    def inc(self, tpv_ts: Optional[int],
            tpv_lat: Optional[float],
            tpv_lon: Optional[float],
            tpv_alt: Optional[float]) -> None:

        self.mts = int(time.time())

        self.tpv_ts = tpv_ts
        self.tpv_lat = tpv_lat
        self.tpv_lon = tpv_lon
        self.tpv_alt = tpv_alt


class _GpsdCounters:
    MODE_0D_FIX = 0
    MODE_1D_FIX = 1
    MODE_2D_FIX = 2
    MODE_3D_FIX = 3

    def __init__(self):
        self.mode: int = 0
        self.mode_mts: Optional[int] = None

        self.store_count = int()
        self.geodata = _GeodataCounters()

    def inc_mode(self, mode):
        if self.mode != mode:
            self.mode = mode
            self.mode_mts = int(time.time())

    def inc_store_count(self, count):
        self.store_count += count


class _GeopositionLog:
    def __init__(self, max_entry_count: int, max_age_delta_usec: int, max_lookup_dist_usec: int):
        self._max_entry_count = max_entry_count
        self._max_age_delta_usec = max_age_delta_usec
        self._max_lookup_dist_usec = max_lookup_dist_usec

        self._ts_to_geodata_map = sortedcontainers.SortedDict()

    def add_record(self, ts: int, geoposition: Geoposition) -> None:
        self._ts_to_geodata_map[ts] = geoposition

        while len(self._ts_to_geodata_map) > self._max_entry_count:
            self._ts_to_geodata_map.popitem(index=0)

        while True:
            max_ts = self._ts_to_geodata_map.peekitem()[0]
            min_ts = self._ts_to_geodata_map.peekitem(0)[0]
            if max_ts - min_ts > self._max_age_delta_usec:
                self._ts_to_geodata_map.popitem(index=0)

            else:
                break

    def lookup_nearest(self, lookup_ts: int) -> Optional[Geoposition]:
        entry_count = len(self._ts_to_geodata_map)
        if not entry_count:
            return None

        # insert lookup_ts
        bisect_index = self._ts_to_geodata_map.bisect_right(lookup_ts)
        if 0 < bisect_index < entry_count:
            # exact-match with lookup_ts
            less_equal_ts_dist = lookup_ts - self._ts_to_geodata_map.iloc[bisect_index - 1]
            if less_equal_ts_dist == 0:
                return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1]

            # greater_ts is better
            greater_ts_dist = self._ts_to_geodata_map.iloc[bisect_index] - lookup_ts
            if less_equal_ts_dist < greater_ts_dist and less_equal_ts_dist <= self._max_lookup_dist_usec:
                return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1]

            # less_equal_ts is better
            return self._ts_to_geodata_map.peekitem(bisect_index)[1] \
                if greater_ts_dist <= self._max_lookup_dist_usec else None

        # push-back lookup_ts
        if bisect_index == entry_count:
            less_equal_ts_dist = lookup_ts - self._ts_to_geodata_map.iloc[entry_count - 1]
            return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1] \
                if less_equal_ts_dist <= self._max_lookup_dist_usec else None

        # push-front lookup_ts
        greater_ts_dist = self._ts_to_geodata_map.iloc[bisect_index] - lookup_ts
        if greater_ts_dist <= self._max_lookup_dist_usec:
            return self._ts_to_geodata_map.peekitem(bisect_index)[1]
        else:
            eventlog.log_w('fail: target-ts:{}usec, min-ts:{}usec, dt:{}sec'.format(
                lookup_ts, self._ts_to_geodata_map.peekitem(0)[0],
                (self._ts_to_geodata_map.peekitem(0)[0] - lookup_ts) / 1e6))
            return None

    def get_count(self) -> int:
        return len(self._ts_to_geodata_map)


class _GpsdProtocol(LineOnlyReceiver):
    delimiter = b'\n'

    def __init__(self, geodata_store: _GeopositionLog, stat_counters: Optional[_GpsdCounters] = None):
        self._geodata_store = geodata_store
        self._stat_counters = stat_counters

    def connectionMade(self):
        # noinspection PyBroadException
        try:
            self.sendLine(b'?WATCH={"enable":true,"json":true}')
        except Exception:
            eventlog.log_x()

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        pass

    def lineReceived(self, raw_reply: str) -> None:
        try:
            self._handle(raw_reply)

        except ValueError:
            eventlog.log_x()

    def _handle(self, raw_reply: str) -> None:
        reply_jnode = ujson.loads(raw_reply)
        if reply_jnode['class'] == 'TPV':
            # Time/date stamp in ISO8601 format, UTC.
            # May have a fractional part of up to .001sec precision. May be absent if mode is not 2 or 3.
            raw_tpv_time: Optional[str] = reply_jnode.get('time')
            tpv_time_usec = None
            tpv_time_sec = None
            if raw_tpv_time:
                tpv_time_usec = isotime.isostr_to_timestamp_usec(raw_tpv_time)
                tpv_time_sec = int(tpv_time_usec / 1e6)

            # Latitude in degrees: +/- signifies North/South. Present when mode is 2 or 3.
            tpv_lat = reply_jnode.get('lat')

            # Longitude in degrees: +/- signifies East/West. Present when mode is 2 or 3.
            tpv_lon = reply_jnode.get('lon')

            # Altitude in meters. Present if mode is 3.
            tpv_alt = reply_jnode.get('alt')

            # Speed over ground, meters per second.
            tpv_speed = reply_jnode.get('speed')

            # TODO: implement stub stat counters
            if self._stat_counters:
                self._stat_counters.geodata.inc(tpv_time_sec, tpv_lat, tpv_lon, tpv_alt)

            # NMEA mode: %d, 0=no mode value yet seen, 1=no fix, 2=2D, 3=3D.
            tpv_mode = reply_jnode['mode']
            if self._stat_counters:
                self._stat_counters.inc_mode(tpv_mode)

            if tpv_time_usec is not None and tpv_mode in [_TPV_MODE_2D_FIX, _TPV_MODE_3D_FIX]:
                self._geodata_store.add_record(tpv_time_usec, Geoposition(tpv_lat, tpv_lon, tpv_alt, tpv_speed))
                if self._stat_counters:
                    self._stat_counters.inc_store_count(self._geodata_store.get_count())


class _GpsdClientFactory(ReconnectingClientFactory):
    initialDelay = 1
    maxDelay = 15
    factor = 1.6180339887498948

    def __init__(self, geodata_store: _GeopositionLog, stat_counters: Optional[_GpsdCounters] = None):
        self._geodata_store = geodata_store
        self._stat_counters = stat_counters

    def buildProtocol(self, addr) -> Protocol:
        self.resetDelay()
        protocol = _GpsdProtocol(self._geodata_store, self._stat_counters)
        protocol.factory = self
        return protocol

    def clientConnectionFailed(self, connector, reason) -> None:
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, unused_reason) -> None:
        ReconnectingClientFactory.clientConnectionLost(self, connector, unused_reason)


class _BaseGeolocationService(IGnssService):
    def __init__(self, twisted_reactor,
                 endpoint: Tuple[str, int],
                 geodata_store: _GeopositionLog,
                 stat_counters: Optional[_GpsdCounters] = None,
                 azimuth: Optional[float] = None):

        IGnssService.__init__(self)

        self._azimuth = azimuth
        self._geodata_store = geodata_store

        gpsd_host, gpsd_port = endpoint
        client_factory = _GpsdClientFactory(self._geodata_store, stat_counters)
        twisted_reactor.connectTCP(gpsd_host, gpsd_port, client_factory)

    def get_geoposition(self, ts: int) -> Optional[Geoposition]:
        geo = self._geodata_store.lookup_nearest(ts)
        if geo is not None:
            geo.azimuth = self._azimuth
        return geo


class _GeolocationService(_BaseGeolocationService):
    def __init__(self, twisted_reactor, tz_service: tz.ITzService, args: GnssServiceConf):
        self._azimut = args.azimuth
        self._counters = _GpsdCounters() if bool(args.endpoint) else None
        self._tz_service = tz_service
        geodata_store = _GeopositionLog(args.max_entry_count, args.max_age_delta_usec, args.max_lookup_dist_usec)
        super().__init__(twisted_reactor, args.endpoint, geodata_store, self._counters, args.azimuth)

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        if self._counters is None:
            return

        stat_jnode = {
            'last-geoposition': {
                'azimuth': self._azimut,
            }
        }

        if self._counters.mode_mts is not None:
            stat_jnode['last-status'] = {
                'mtime': [
                    self._tz_service.map_ts_sec_to_isostr(self._counters.mode_mts),
                    int(self._counters.mode_mts)
                ],
                'fix-mode': self._counters.mode
            }

        if self._counters.geodata.mts is not None:
            geo_jnode = {
                'mtime': [
                    self._tz_service.map_ts_sec_to_isostr(self._counters.geodata.mts),
                    int(self._counters.geodata.mts)
                ]
            }

            if self._counters.geodata.tpv_ts is not None:
                geo_jnode['time'] = [
                    self._tz_service.map_ts_sec_to_isostr(self._counters.geodata.tpv_ts),
                    int(self._counters.geodata.tpv_ts)
                ]

            if None not in [self._counters.geodata.tpv_lat, self._counters.geodata.tpv_lon]:
                geo = Geoposition(self._counters.geodata.tpv_lat, self._counters.geodata.tpv_lon)

                location_info = LocationInfo()
                location_info.longitude = geo.longitude
                location_info.latitude = geo.latitude
                location_info.timezone = 'UTC'

                astral_loc = Location(location_info)
                # <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
                coordinates = '{}{:.6f}{}{:.6f}'.format(
                    'N' if geo.latitude >= 0 else 'S', abs(geo.latitude),
                    'E' if geo.longitude >= 0 else 'W', abs(geo.longitude))

                geo_jnode['latitude'] = geo.latitude
                geo_jnode['longitude'] = geo.longitude
                geo_jnode['coordinates'] = coordinates
                geo_jnode['sunset'] = self._tz_service.map_datetime_to_isostr(astral_loc.sunset())
                geo_jnode['sunrise'] = self._tz_service.map_datetime_to_isostr(astral_loc.sunrise())

            # add altitude
            if self._counters.geodata.tpv_alt is not None:
                geo_jnode['altitude'] = self._counters.geodata.tpv_alt

            stat_jnode['last-geoposition'] = geo_jnode

        json_handler({
            'type': 'geolocation',
            'name': 'geolocation.0',
            'stat': stat_jnode
        })


class _StubGeolocationService(IGnssService):
    def __init__(self, geoposition: Optional[Geoposition], tz_service: tz.ITzService):
        IGnssService.__init__(self)

        # http://catb.org/gpsd/gpsd_json.html
        self._geoposition = geoposition

        if geoposition:
            location_info = LocationInfo()
            location_info.longitude = geoposition.longitude
            location_info.latitude = geoposition.latitude
            location_info.timezone = 'UTC'

            self._astral_loc = Location(location_info)
        else:
            self._astral_loc = None

        self._tz_service = tz_service

    def get_geoposition(self, ts: int) -> Optional[Geoposition]:
        return self._geoposition

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        if self._geoposition is None:
            return

        # <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
        coordinates = '{}{:.6f}{}{:.6f}'.format(
            'N' if self._geoposition.latitude >= 0 else 'S', abs(self._geoposition.latitude),
            'E' if self._geoposition.longitude >= 0 else 'W', abs(self._geoposition.longitude))

        jnode = {
            'type': 'geolocation',
            'name': 'geolocation.0',
            'stat': {
                'last-geoposition': {
                    'mtime': self._tz_service.now_isostr_tssec(),
                    'coordinates': coordinates,
                    'latitude': self._geoposition.latitude,
                    'longitude': self._geoposition.longitude,
                    'sunset': self._tz_service.map_datetime_to_isostr(self._astral_loc.sunset()),
                    'sunrise': self._tz_service.map_datetime_to_isostr(self._astral_loc.sunrise())
                }
            }
        }

        if self._geoposition.altitude is not None:
            jnode['stat']['last-geoposition']['altitude'] = self._geoposition.altitude

        if self._geoposition.azimuth is not None:
            jnode['stat']['last-geoposition']['azimuth'] = self._geoposition.azimuth

        json_handler(jnode)


def init_service(twisted_reactor, args: GnssServiceConf, tz_service: tz.ITzService) -> IGnssService:
    if args.provider == args.PROVIDER_STATIC:
        service = _StubGeolocationService(args.geodata, tz_service)
    else:
        service = _GeolocationService(twisted_reactor, tz_service, args)

    return service
