from typing import Dict

from edge import eventlog
from edge.services.types.gnss import GnssServiceConf
from edge.support import fs
from edge.types.exceptions import JnodeError
from edge.types.geoposition import Geoposition


def decode__gnss_service_conf(jnode: Dict) -> GnssServiceConf:
    """
    :param jnode:
        /etc/edge/setupstore/admin/actions/pipeline/0/gnss/0.json#/
    :raises JnodeError:
    """
    rt_jnode = jnode.get('rt')
    if rt_jnode is None:
        return GnssServiceConf(endpoint=('localhost', 2947))

    az = rt_jnode.get('azimuth')
    jnode = rt_jnode.get('source')
    if jnode is None:
        eventlog.log_e_param('#/')
        raise JnodeError

    if not (type(jnode) in [tuple, list] and len(jnode) == 3):
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize geodata
    if jnode[0] == 'static':
        _, latitude, longitude = jnode
        if not (type(latitude) in [int, float] and -90.0 <= latitude <= +90.0):
            eventlog.log_e_param('#/1')
            raise JnodeError

        if not (type(longitude) in [int, float] and -180.0 <= longitude <= +180.0):
            eventlog.log_e_param('#/2')
            raise JnodeError

        return GnssServiceConf(provider=GnssServiceConf.PROVIDER_STATIC,
                               geodata=Geoposition(latitude, longitude, azimuth=az))

    if jnode[0] == 'gpsd':
        _, gpsd_host, gpsd_port = jnode
        if not (type(gpsd_host) in [str] and len(gpsd_host) <= 32):
            eventlog.log_e_param('#/1')
            raise JnodeError

        if not (type(gpsd_port) in [int] and 0 < gpsd_port <= 65535):
            eventlog.log_e_param('#/2')
            raise JnodeError

        return GnssServiceConf(provider=GnssServiceConf.PROVIDER_GPSD,
                               endpoint=(gpsd_host, gpsd_port))

    eventlog.log_e_param('#/0')
    raise JnodeError


def decode__gnss_service_conf_file(conf_path: str) -> GnssServiceConf:
    """
    :param conf_path:
        /etc/edge/setupstore/admin/actions/pipeline/0/gnss/0.json
    :raise JnodeError:
    """
    conf_jnode = fs.load_json3(conf_path, default=None)
    if not conf_jnode:
        eventlog.log_e_param('#/')
        raise JnodeError

    return decode__gnss_service_conf(conf_jnode)
