from typing import TypeVar, List, Optional, Callable

from edge.types.blob import FileBlob
from edge.types.image import ImRgba, ImJpeg
from edge.services.types.service import IService


class VsError(RuntimeError):
    """
    Generic Error.
    """
    pass


class VsSessionAbsentError(VsError):
    """
    Session is absent in the pool.
    """
    pass


class VsSessionExpiredError(VsError):
    """
    Session is expired.
    Session should be closed and opened again.
    """
    pass


class VsServiceConf:
    def __init__(self, store: int, tmp_path: str, chan: str = '0', log_level: str = 'info'):
        self.store = store
        self.tmp_path = tmp_path
        self.chan = chan
        self.log_level = log_level


class VsImageInfo:
    def __init__(self, ts: int):
        self._ts = ts

    @property
    def ts(self) -> int:
        """
        :returns:
            value in usec
        """
        return self._ts


TVsImageInfo = TypeVar('TVsImageInfo', bound=VsImageInfo)


class IVsService(IService):
    def export_image_rgba(self,
                          im_rgba: ImRgba,
                          im_infos: List[TVsImageInfo],
                          im_footer_h: int = 0,
                          bs_chan: Optional[str] = None) -> Optional[TVsImageInfo]:
        """
        :raise VsError:
        """
        raise NotImplementedError

    def export_image_jpeg(self,
                          im_jpeg: ImJpeg,
                          im_infos: List[TVsImageInfo],
                          bs_chan: Optional[str] = None
                          ) -> Optional[TVsImageInfo]:
        """
        :raise VsError:
        """
        raise NotImplementedError

    def export_video(self, min_ts: int, max_ts: int, max_w: int, max_h: int, rec_quality: int,
                     handler: Callable[[Optional[FileBlob]], None]) -> None:
        """
        :param min_ts:
            ms since 01.01.1970 00:00 UTC
        :param max_ts:
            ms since 01.01.1970 00:00 UTC
        :param max_w:
            valid values in range [16; 32752], multiple 16.
        :param max_h:
            valid values in range [16; 32752], multiple 16.
        :param rec_quality:
            Constant Bit Rate (CBR), Kbit/s. Valid values in range [100; 102400].

        :param handler:

        :raise VsSessionAbsentError:
        :raise VsError:
        """
        raise NotImplementedError
