import pytz

from datetime import datetime
from typing import Tuple, Union

from edge.services.types.tz import TzServiceConf, ITzService
from edge.support.isotime import datetime_to_isostr


class _TzService(ITzService):
    def __init__(self, conf: TzServiceConf):
        super().__init__()
        self._tz = conf.tz_info

    def map_ts_sec_to_isostr(self, ts_sec: int | float) -> str:
        """
        :param ts_sec:
            time in seconds since the epoch
        :returns:
            ISO8601
        """
        # 2018-01-23T15:49:47.124+0400
        dt_local = self.map_ts_sec_to_datetime_local(ts_sec)
        return datetime_to_isostr(dt_local)

    def map_ts_usec_to_isostr(self, ts_usec: int | float) -> str:
        return self.map_ts_sec_to_isostr(ts_usec / 1e6)

    def map_ts_sec_to_datetime_utc(self, ts_sec: int | float) -> datetime:
        """
        :param ts_sec:
            time in seconds since the epoch
        """
        return datetime.fromtimestamp(float(ts_sec), tz=pytz.utc)

    def map_ts_usec_to_datetime_utc(self, ts_usec: int | float) -> datetime:
        return self.map_ts_sec_to_datetime_utc(ts_usec / 1e6)

    def map_ts_sec_to_datetime_local(self, ts_sec: Union[int, float]) -> datetime:
        """
        :param ts_sec:
            time in seconds since the epoch
        """
        return datetime.fromtimestamp(float(ts_sec), tz=self._tz)

    def map_ts_usec_to_datetime_local(self, ts_usec: int | float) -> datetime:
        return self.map_ts_sec_to_datetime_local(ts_usec / 1e6)

    def map_datetime_to_isostr(self, dt: datetime, with_usecs: bool = False) -> str:
        dt2 = dt.astimezone(self._tz)
        return datetime_to_isostr(dt2, with_usecs)

    def now_isostr(self) -> str:
        """
        :returns:
            ISO8601
        """
        dt_local = datetime.now(self._tz)
        return datetime_to_isostr(dt_local)

    def now_isostr_tssec(self) -> Tuple[str, int]:
        """
        :returns:
            (ISO8601, time in seconds since the epoch)
        """
        dt_local = datetime.now(self._tz)
        dt_local_epoch = datetime(1970, 1, 1, tzinfo=self._tz)
        dt_local_seconds = (dt_local - dt_local_epoch).total_seconds()

        dt_local_isostr = datetime_to_isostr(dt_local)
        return dt_local_isostr, int(dt_local_seconds)


def init_service(conf: TzServiceConf) -> ITzService:
    return _TzService(conf)
