from typing import Tuple

from edge.types.shape.rect import RectF


def mk_plate_bbox(rel_plate: RectF, image_iw: int, image_ih: int) -> RectF:
    """
    :param rel_plate:
        Relative coords. Visible values are in range [0.0; 1.0).
    :param image_iw:
        Width, px
    :param image_ih:
        Weight, px

    :returns:
        Absolute coords. Visible values are in range [0.0; image_iw) and [0.0; image_ih).
    """
    fl = image_iw * rel_plate.fl
    ft = image_ih * rel_plate.ft
    fr = image_iw * rel_plate.fr
    fb = image_ih * rel_plate.fb

    # tweak plate rect for better view
    sym_margin_cm, sym_height_cm = 1.5, 8.0
    delta = (fb - ft) * sym_margin_cm / sym_height_cm
    return RectF(
        fl - 3 * delta,
        ft - 1 * delta,
        fr + 2 * delta,
        fb + 1 * delta)


def _upscale_plate_wh(plate_bbox: RectF, wh_ratio: float, target_w: float) -> Tuple[float, float]:
    """
    :param plate_bbox:
        absolute coords

    Vehicle's make and model would be recognizable in case of wh_ratio 4.0/3.0
    """
    # define the width of a vehicle as the width of the lane (typ. 2.75m)
    if plate_bbox.fw / plate_bbox.fh > 2.0:
        # single-line plates with w/h-ratio more then 2:1 (typ. 5:1)
        # suppose the average plate's height that was defined by ANPR is around 10cm
        # suppose the average plate's width that was defined by ANPR is around 50cm
        # car_w_height = (target_w / 0.10) * plate_bbox.fh
        car_w_width = (target_w / 0.50) * plate_bbox.fw
        # car_w = (car_w_height + car_w_width) / 2
        car_w = car_w_width

    else:
        # double-line plates with w/h-ratio 2:1 or less
        # suppose the average plate's height that was defined by ANPR is around 17cm
        # suppose the average plate's width that was defined by ANPR is around 22cm
        car_w_height = (target_w / 0.17) * plate_bbox.fh
        car_w_width = (target_w / 0.22) * plate_bbox.fw
        car_w = (car_w_height + car_w_width) / 2

    return car_w, car_w / wh_ratio


def mk_vehicle_bbox(plate_bbox: RectF, scene_wh: Tuple[int, int], wh_ratio: float = 4.0 / 3.0) -> RectF:
    """
    :param plate_bbox:
        absolute coords
    :param scene_wh:
    :param wh_ratio:

    :returns:
        absolute coords
    :raise ValueError:
    """
    center_fl, center_ft = plate_bbox.center
    width, height = _upscale_plate_wh(plate_bbox, wh_ratio, 2.75)
    bbox = RectF.create(center_fl, center_ft - height / 5.0, width, height)  # move bbox 20% up from plate center
    bbox.scale_max(plate_bbox.fw, plate_bbox.fh, *scene_wh)
    bbox.make_visible(*scene_wh)
    return bbox


def mk_mmr_bbox(plate_bbox: RectF, scene_wh: Tuple[int, int], min_visible_factor: float) -> RectF:
    """
    :param plate_bbox:
        absolute coords
    :param scene_wh:
    :param min_visible_factor:

    :returns:
        absolute coords
    :raise ValueError:
    """
    center_fl, center_ft = plate_bbox.center
    width, height = _upscale_plate_wh(plate_bbox, 1, 2.0)
    bbox = RectF.create(center_fl, center_ft - height / 4.0, width, height)  # move bbox 25% up from plate center
    if bbox.get_visible(*scene_wh) < min_visible_factor:
        raise ValueError

    bbox.scale_max(plate_bbox.fw, plate_bbox.fh, *scene_wh)
    bbox.make_visible(*scene_wh)
    return bbox
