import ctypes

from glob import glob
from typing import Dict, Optional, Self, Tuple

from edge import eventlog


class BsError(RuntimeError):
    pass


class BsSessionError(BsError):
    pass


class IBsCounters:
    def inc_blob(self, chan: str) -> None:
        raise NotImplementedError

    def inc_done(self, chan: str) -> None:
        raise NotImplementedError

    def inc_fail(self, chan: str) -> None:
        raise NotImplementedError

    def serialize(self) -> Dict:
        raise NotImplementedError


class StubBsCounters(IBsCounters):
    def __init__(self):
        IBsCounters.__init__(self)

    def inc_blob(self, chan: str) -> None:
        pass

    def inc_done(self, chan: str) -> None:
        pass

    def inc_fail(self, chan: str) -> None:
        pass

    def serialize(self) -> Dict:
        return dict()


class _Api:
    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2
    RC_CONN = -3

    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    LOG_FLAGS_NONE = 0
    LOG_FLAGS_STDERR = 1

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libblobstore-bsb*.so')[0])

    # int
    # BsbOpen(uint16_t aStore, uint8_t aLogLevel, void **aSession);
    open = _lib.BsbOpen
    open.argtypes = (ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_void_p))
    open.restype = ctypes.c_int

    # int
    # BsbSave(void *aSession, const char *aChannel, uint64_t aRootIdent, uint64_t aBlobTs, const uint8_t *aBlobData,
    #     uint32_t aBlobDataSize, const uint8_t *aBlobMeta, uint16_t aBlobMetaSize, int64_t *aBlobIdent);
    save = _lib.BsbSave
    save.argtypes = (
        ctypes.c_void_p, ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint64, ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16, ctypes.POINTER(ctypes.c_int64))
    save.restype = ctypes.c_int

    # int
    # BsbSave2(void *aSession, const char *aChannel, uint64_t aRootIdent, uint64_t aBlobTs, const char *aBlobPath,
    #     const uint8_t *aBlobMeta, uint16_t aBlobMetaSize, int64_t *aBlobIdent);
    save2 = _lib.BsbSave2
    save2.argtypes = (
        ctypes.c_void_p, ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16, ctypes.POINTER(ctypes.c_int64))
    save2.restype = ctypes.c_int

    # int
    # BsbLoad(void *aSession, const char *aChannel, int64_t aBlobIdent, uint8_t *aBlobData,
    #     uint32_t *aBlobDataSize, uint8_t *aBlobMeta, uint16_t *aBlobMetaSize);
    load = _lib.BsbLoad
    load.argtypes = (
        ctypes.c_void_p, ctypes.c_char_p, ctypes.c_int64, ctypes.POINTER(ctypes.c_uint8),
        ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint16))
    load.restype = ctypes.c_int

    # void
    # BsbClose(void *aSession);
    close = _lib.BsbClose
    close.argtype = ctypes.c_void_p

    # void
    # BsbLogOpen(const char *aName, int aMaxLevel, int aFlags);
    logOpen = _lib.BsbLogOpen
    logOpen.argtypes = (ctypes.c_char_p, ctypes.c_int, ctypes.c_int)

    # void
    # BsbLogClose();
    logClose = _lib.BsbLogClose


LOG_LEVEL_NAME_TO_CODE_MAP = {
    'none': _Api.LOG_N,
    'error': _Api.LOG_E,
    'warn': _Api.LOG_W,
    'info': _Api.LOG_I,
    'debug': _Api.LOG_D,
    'trace': _Api.LOG_T
}


def _check_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _Session
        """
        if self.closed:
            eventlog.log_e('fail: invalid session')
            raise BsSessionError
        else:
            return fn(self, *args, **kwargs)

    return proxy_fn


class Session:
    def __init__(self, counters, handle=None, closed=False):
        self._handle = handle
        self.closed = closed

        self._counters = counters

    @classmethod
    def open(cls, module_ref: str = '', counters: IBsCounters = StubBsCounters(),
             store: int = 65535, log_level: int = _Api.LOG_E) -> Self:
        """
        :raise BsSessionError:
        """
        handle = ctypes.c_void_p(0)
        rc = _Api.open(store, log_level, handle)
        _Api.logOpen(bytes(module_ref, 'ascii'), log_level, _Api.LOG_FLAGS_NONE)

        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsbOpen (%s)' % rc)
            raise BsSessionError

        return cls(counters=counters, handle=handle)

    @_check_session
    def save_blob(self, chan: str, blob_ts: int, blob_data: bytes, blob_meta: bytes | None) -> int:
        """
        :param chan:
        :param blob_ts:
            value in msec
        :param blob_data:
        :param blob_meta:

        :raise BsError:
        :raise BsSessionError:
        """
        data_uint8_p = ctypes.cast(ctypes.c_char_p(blob_data), ctypes.POINTER(ctypes.c_uint8))
        if blob_meta is None:
            meta_uint8_p = None
            meta_size = ctypes.c_uint16(0)
        else:
            meta_uint8_p = ctypes.cast(ctypes.c_char_p(blob_meta),
                                       ctypes.POINTER(ctypes.c_uint8))
            meta_size = len(blob_meta)

        ident = ctypes.c_int64()
        rc = _Api.save(
            self._handle, bytes(chan, 'ascii'), 0, ctypes.c_uint64(blob_ts),
            data_uint8_p, len(blob_data), meta_uint8_p, meta_size,
            ident)

        if rc == _Api.RC_CONN:
            self._counters.inc_fail(chan)
            eventlog.log_e('fail: BsbSave (kBS_CONN)')
            raise BsSessionError

        elif rc != _Api.RC_DONE:
            self._counters.inc_fail(chan)
            eventlog.log_e('fail: BsbSave (%s)' % rc)
            raise BsError

        self._counters.inc_done(chan)

        return ident.value

    @_check_session
    def get_blob_size(self, chan: str, blob_ref: str) -> int:
        """
        :raise BsError:
        :raise BsSessionError:
        """
        ident = ctypes.c_int64(int(blob_ref))
        blob_size = ctypes.c_uint32()
        blob_meta_size = ctypes.c_uint16()

        answer = _Api.load(
            self._handle, bytes(chan, 'ascii'), ident,
            ctypes.POINTER(ctypes.c_ubyte)(), blob_size,
            ctypes.POINTER(ctypes.c_ubyte)(), blob_meta_size)

        if answer == _Api.RC_CONN:
            eventlog.log_e(f'fail: BsbLoad (chan:{chan}, blob-ref:{hex(int(blob_ref))}, kBS_CONN)')
            raise BsSessionError

        elif answer != _Api.RC_DONE:
            eventlog.log_e(f'fail: BsbLoad (chan:{chan}, blob-ref:{hex(int(blob_ref))}, answer:{answer})')
            raise BsError

        return blob_size.value

    @_check_session
    def save_blob2(self, chan: str, blob_ts: int, blob_path: str, blob_meta: Optional[bytes] = None) -> int:
        """
        :param chan:
        :param blob_ts:
            value in msec
        :param blob_ts:
        :param blob_path:
        :param blob_meta:

        :raise BsError:
        :raise BsSessionError:
        """
        if blob_meta is None:
            meta_uint8_p = None
            meta_size = ctypes.c_uint16(0)
        else:
            meta_uint8_p = ctypes.cast(ctypes.c_char_p(blob_meta), ctypes.POINTER(ctypes.c_uint8))
            meta_size = len(blob_meta)

        ident = ctypes.c_int64()
        rc = _Api.save2(
            self._handle, bytes(chan, 'ascii'), 0, ctypes.c_uint64(blob_ts), bytes(blob_path, 'ascii'),
            meta_uint8_p, meta_size, ident)

        if rc == _Api.RC_CONN:
            self._counters.inc_fail(chan)
            eventlog.log_e('fail: BsbSave2 (kBS_CONN)')
            raise BsSessionError

        elif rc != _Api.RC_DONE:
            self._counters.inc_fail(chan)
            eventlog.log_e('fail: BsbSave2 (%s)' % rc)
            raise BsError

        self._counters.inc_done(chan)

        return ident.value

    @_check_session
    def load_blob(self, chan: str, blob_ts: int) -> Tuple[bytes, int, bytes, int]:
        """
        :param chan:
        :param blob_ts:
            value in msec

        :raise BsError:
        :raise BsSessionError:
        """
        buf_size = 512 * 1024
        buf_size_delta = 64 * 1024

        while True:
            blob_data = (ctypes.c_ubyte * buf_size)()
            blob_meta = (ctypes.c_ubyte * buf_size)()
            data_size = ctypes.c_uint32(buf_size)
            meta_size = ctypes.c_uint16(buf_size)

            rc = _Api.load(
                self._handle, bytes(chan, 'ascii'), ctypes.c_int64(blob_ts),
                ctypes.cast(blob_data, ctypes.POINTER(ctypes.c_ubyte)), ctypes.byref(data_size),
                ctypes.cast(blob_meta, ctypes.POINTER(ctypes.c_ubyte)), ctypes.byref(meta_size))

            if rc == _Api.RC_CONN:
                eventlog.log_e('fail: BsbLoad (kBS_CONN)')
                raise BsSessionError

            elif rc != _Api.RC_DONE:
                eventlog.log_e(f'fail: BsbLoad (answer:{rc})')
                raise BsError

            if data_size.value < buf_size and meta_size.value < buf_size:
                break
            else:
                eventlog.log_i(f'exec: resize buffer (old:{buf_size}, new:{buf_size + buf_size_delta}')
                buf_size += buf_size_delta

        return (
            ctypes.string_at(ctypes.cast(blob_data, ctypes.POINTER(ctypes.c_ubyte)), data_size.value), data_size.value,
            ctypes.string_at(ctypes.cast(blob_meta, ctypes.POINTER(ctypes.c_ubyte)), meta_size.value), meta_size.value)

    def close(self) -> None:
        _Api.logClose()
        _Api.close(self._handle)
        self.closed = True


def reopen_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _BsService
        """
        try_count = 0
        while try_count <= 3:
            try_count += 1
            try:
                if self._session.closed:
                    self._session = Session.open(*self._session_args)

                return fn(self, *args, **kwargs)

            except BsSessionError:
                self._session.close()

        raise BsError

    return proxy_fn
