from typing import List, Self, TextIO

from edge.types.exceptions import *


class CliRequest:
    def __init__(self, cli_path: str, cli_args: List[str]):
        self.path = cli_path
        self.args = cli_args

    @classmethod
    def create(cls) -> Self:
        import sys
        return cls.create2(sys.argv)

    @classmethod
    def create2(cls, argv: List[str]) -> Self:
        return cls(argv[0], argv[1:])

    @property
    def log_name(self) -> str:
        from os import path
        return path.splitext(path.basename(self.path))[0]

    @property
    def log_level(self) -> str:
        return 'debug'

    def assert_args(self, count: int) -> None:
        if len(self.args) != count:
            raise NotFoundError

    def assert_min_args(self, count: int) -> None:
        if len(self.args) < count:
            raise NotFoundError

    def assert_arg(self, index: int, value: str) -> None:
        if self.args[index] != value:
            raise NotFoundError


class CliResponse:
    EX_OK = 0         # sysexists.h / successful termination
    EX_USAGE = 64     # sysexists.h / command line usage error
    EX_SOFTWARE = 70  # sysexists.h / internal software error

    def __init__(self, cli_status: int, cli_stdout: str = '', cli_stderr: str = ''):
        self._status = cli_status
        self._stdout = cli_stdout
        self._stderr = cli_stderr

    @classmethod
    def done(cls, cli_stdout: str = '', cli_stderr: str = '') -> Self:
        return cls(cls.EX_OK, cli_stdout, cli_stderr)

    @classmethod
    def fail_external(cls, cli_stdout: str = '', cli_stderr: str = '') -> Self:
        return cls(cls.EX_USAGE, cli_stdout, cli_stderr)

    @classmethod
    def fail_internal(cls, cli_stdout: str = '', cli_stderr: str = '') -> Self:
        return cls(cls.EX_SOFTWARE, cli_stdout, cli_stderr)

    def render(self) -> int:
        import sys
        return self.render2(sys.stdout, sys.stderr)

    def render2(self, stdout_stream: TextIO, stderr_stream: TextIO) -> int:
        if self._stdout:
            stdout_stream.write(self._stdout)

        if self._stderr:
            stderr_stream.write(self._stderr)

        return self._status


class CliBoolResponse(CliResponse):
    def __init__(self, cli_status: int, cli_stdout: str = '', cli_stderr: str = ''):
        super().__init__(cli_status, cli_stdout, cli_stderr)

    @classmethod
    def done(cls, cli_stdout: str = 'DONE\n', cli_stderr: str = '') -> Self:
        return cls(cls.EX_OK, cli_stdout, cli_stderr)

    @classmethod
    def fail_external(cls, cli_stdout: str = 'FAIL\n', cli_stderr: str = '') -> Self:
        return cls(cls.EX_USAGE, cli_stdout, cli_stderr)

    @classmethod
    def fail_internal(cls, cli_stdout: str = 'FAIL\n', cli_stderr: str = '') -> Self:
        return cls(cls.EX_SOFTWARE, cli_stdout, cli_stderr)


class ICliEngine:
    def __init__(self):
        pass

    def usage(self) -> str:
        raise NotImplementedError

    def activate(self, request: CliRequest) -> CliResponse:
        raise NotImplementedError
