from typing import List, Optional

from edge.types.dir import Dir
from edge.types.track import LrTrackItem, DirTrackItem, RaTrackItem


def get_spot_dir_item_newer(items: List[LrTrackItem], item_index: int, max_dist_usec: int) -> Optional[LrTrackItem]:
    """
    Search of the first item of the track, which satisfies the conditions:
        1) the item is newer than the target item
        2) the item's quad center is not in the target items's quad
        3) the item's ts is not newer than (item.ts + max_dist_usec) msec
    """
    item: LrTrackItem = items[item_index]
    item_quad = item.plate_quad

    max_ts: int = item.ts + max_dist_usec
    for newer_index in range(item_index + 1, len(items), 1):
        newer_item = items[newer_index]

        if newer_item.ts > max_ts:
            return

        if newer_item.plate_quad.center not in item_quad:
            return newer_item


def get_spot_dir_item_older(items: List[LrTrackItem], item_index: int, max_dist_usec: int) -> Optional[LrTrackItem]:
    """
    Search of the first item of the track, which satisfies the conditions:
        1) the item is older than the target item
        2) the item's quad center is not in the target items's quad
        3) the item's ts is not older than (item.ts - max_dist_usec) msec
    """
    item: LrTrackItem = items[item_index]
    item_quad = item.plate_quad

    min_ts: int = item.ts - max_dist_usec
    for older_index in range(item_index - 1, -1, -1):
        older_item = items[older_index]

        if older_item.ts < min_ts:
            return

        if older_item.plate_quad.center not in item_quad:
            return older_item


def get_spot_dir_item(items: List[LrTrackItem], item_index: int, max_dist_usec: int) -> DirTrackItem:
    index_item: LrTrackItem = items[item_index]

    newer_item: Optional[LrTrackItem] = get_spot_dir_item_newer(items, item_index, max_dist_usec)
    if newer_item is None:
        return DirTrackItem(index_item.ts, Dir.CODE_U)

    older_item: Optional[LrTrackItem] = get_spot_dir_item_older(items, item_index, max_dist_usec)
    if older_item is None:
        return DirTrackItem(index_item.ts, Dir.CODE_U)

    newer_item_t: int = newer_item.plate_quad.center_t
    index_item_t: int = index_item.plate_quad.center_t
    older_item_t: int = older_item.plate_quad.center_t

    if abs(newer_item_t - older_item_t) < index_item.plate_quad.bbox_recti.ih:
        return DirTrackItem(index_item.ts, Dir.CODE_U)

    newer_dir = Dir.CODE_A if index_item_t < newer_item_t else Dir.CODE_R
    older_dir = Dir.CODE_A if older_item_t < index_item_t else Dir.CODE_R

    return DirTrackItem(index_item.ts, older_dir if older_dir == newer_dir else Dir.CODE_U)


def create_dir_track_lr_nonlinear(
        items: List[LrTrackItem],
        max_dist_usec: int,
        debouncer: bool = True) -> List[DirTrackItem]:

    spot_dirs: List[DirTrackItem] = [get_spot_dir_item(items, x, max_dist_usec) for x in range(0, len(items))]
    if not debouncer:
        return spot_dirs

    dir_items = list[DirTrackItem]()
    for index in range(0, len(spot_dirs) - 2):
        index_item = spot_dirs[index + 0]
        if index_item.dir == Dir.CODE_U:
            continue

        newer_item = spot_dirs[index + 1]
        if newer_item.dir == Dir.CODE_U:
            continue

        if index_item.dir == newer_item.dir:
            dir_items.append(index_item)

    return dir_items


def create_dir_track_lr_linear(items: List[LrTrackItem], linear_dir: Dir) -> List[DirTrackItem]:
    if linear_dir != Dir.CODE_U:
        return [DirTrackItem(item.ts, linear_dir) for item in items]

    return []


def get_dir_track_lr(
        items: List[LrTrackItem],
        linear_dir: Dir,
        linear_max_size: int = 10,
        linear_max_dur_usec: int = int(1000e3),
        nonlinear_max_dist_usec: int = int(500e3)) -> List[DirTrackItem]:

    if len(items) <= linear_max_size:
        return create_dir_track_lr_linear(items, linear_dir)

    dur_usec = items[-1].ts - items[0].ts
    if dur_usec <= linear_max_dur_usec:
        return create_dir_track_lr_linear(items, linear_dir)

    return create_dir_track_lr_nonlinear(items, max_dist_usec=nonlinear_max_dist_usec)


def get_dir_track_ra(items: List[RaTrackItem]) -> List[DirTrackItem]:
    return [
        DirTrackItem(item.frame_ts, item.speed_dir)
        for item in items if item.speed_dir in (Dir.CODE_A, Dir.CODE_R)
    ]
