from abc import ABCMeta, abstractmethod
from typing import Final, List, Optional

from PIL import Image

from edge import eventlog
from edge.image.helpers import ImHelpers
from edge.services.types.vs import IVsService, TVsImageInfo, VsError
from edge.types.exceptions import ParamsError
from edge.types.image import ImRgba, ImJpeg
from edge.types.imageref import ImageRef
from edge_rawutil.rawutil import RawUtil


class ImrefLoader:
    MAX_WIDTH = 65535
    MAX_HEIGHT = 65535

    def __init__(self,
                 max_target_width: int = MAX_WIDTH,
                 max_target_height: int = MAX_HEIGHT):

        self._rawutil = RawUtil.create(max_target_width, max_target_height, log_level='debug')

    def destroy(self):
        self._rawutil.destroy()

    def load_rgba(self, im_rgba: ImRgba, im_ref: ImageRef, dh: int) -> None:
        """
        :raise ValueError:
        """
        if im_ref.codec == 'JPEG':
            source_im = Image.open(im_ref.url)
            im_rgba.reset(source_im.size[0], source_im.size[1], dh, im_ref.ts)
            im_rgba.with_image(lambda target_im: target_im.paste(source_im))

        elif im_ref.codec in ['MONO8T', 'I420PLN', 'NV12PLN']:
            im_rgba.reset(im_ref.width, im_ref.height, dh, im_ref.ts)
            success = self._rawutil.handle_image3(
                im_ref.url, im_ref.codec, im_ref.width, im_ref.height, im_rgba.buffer)

            if not success:
                eventlog.log_w(f'fail: conv image {im_ref.codec} -> RGBA')
                raise ValueError

            _, w, h = success
            im_rgba.reset(w, h, dh, im_ref.ts)

        else:
            raise NotImplementedError

    def load_jpeg(self, im_jpeg: ImJpeg, im_ref: ImageRef) -> None:
        if im_ref.codec == 'JPEG' and im_ref.url is not None:
            wh = ImHelpers.get_jpeg_size(im_ref.url)
            im_jpeg.reset(wh[0], wh[1], wh[0] * wh[1], im_ref.ts)

            with open(im_ref.url, 'rb') as bin_file:
                file_size = bin_file.readinto(im_jpeg.buffer)

            im_jpeg.reset(wh[0], wh[1], file_size, im_ref.ts)

        else:
            raise NotImplementedError


class ImStats:
    def __init__(self):
        self.total: int = 0
        self.success_arch: int = 0
        self.success_file: int = 0
        self.failure: int = 0


class RgbaImLoader:
    def __init__(self, stats: ImStats, vs_service: IVsService, footer_h: int):
        self._stats: Final[ImStats] = stats
        self._imref_loader: Final[ImrefLoader] = ImrefLoader()
        self._vs: Final[IVsService] = vs_service

        self._footer_h: Final[int] = footer_h

    def destroy(self):
        self._imref_loader.destroy()

    def load(self,
             im_rgba: ImRgba,
             im_infos: List[TVsImageInfo],
             def_imref: Optional[ImageRef]) -> Optional[TVsImageInfo]:
        """
        :raises ParamsError:
        """
        self._stats.total += 1

        try:
            im_info = self._vs.export_image_rgba(im_rgba, im_infos, self._footer_h)
            if im_info is not None:
                self._stats.success_arch += 1
                return im_info
        except VsError:
            eventlog.log_w('fail: IVsService::export_image_rgba2')

        if def_imref and def_imref.url:
            try:
                self._imref_loader.load_rgba(im_rgba, def_imref, self._footer_h)
                self._stats.success_file += 1
                return None
            except ValueError as x:
                eventlog.log_w('fail: ImLoader::load_rgba (<{}>, {})'.format(def_imref, str(x)))
            except NotImplementedError:
                eventlog.log_w('fail: ImLoader::load_rgba (<{}>, unsupported format)'.format(def_imref))

        self._stats.failure += 1
        raise ParamsError


class IJpegImLoader(metaclass=ABCMeta):
    @abstractmethod
    def destroy(self) -> None:
        pass

    @abstractmethod
    def load(self,
             im_jpeg: ImJpeg,
             im_infos: List[TVsImageInfo],
             def_imref: Optional[ImageRef]) -> Optional[TVsImageInfo]:

        """
        :raises ParamsError:
        """
        pass


class StubJpegImLoader(IJpegImLoader):
    def destroy(self) -> None:
        pass

    def load(self,
             im_jpeg: ImJpeg,
             im_infos: List[TVsImageInfo],
             def_imref: Optional[ImageRef]) -> Optional[TVsImageInfo]:

        raise ParamsError


class JpegImLoader(IJpegImLoader):
    def __init__(self, stats: ImStats, vs_service: Optional[IVsService]):
        self._stats: Final[ImStats] = stats
        self._imref_loader: Final[ImrefLoader] = ImrefLoader()
        self._vs: Final[IVsService] = vs_service

    def destroy(self) -> None:
        self._imref_loader.destroy()

    def load(self,
             im_jpeg: ImJpeg,
             im_infos: List[TVsImageInfo],
             def_imref: Optional[ImageRef]) -> Optional[TVsImageInfo]:
        """
        :raises ParamsError:
        """
        self._stats.total += 1

        if self._vs is not None:
            try:
                im_info = self._vs.export_image_jpeg(im_jpeg, im_infos, '1')
                if im_info is not None:
                    self._stats.success_arch += 1
                    return im_info
            except VsError:
                eventlog.log_w('fail: IVsService::export_image_jpeg')

        if def_imref and def_imref.url:
            try:
                self._imref_loader.load_jpeg(im_jpeg, def_imref)
                self._stats.success_file += 1
                return None
            except ValueError as x:
                eventlog.log_w('fail: ImrefLoader::load_jpeg (<{}>, {})'.format(def_imref, str(x)))
            except NotImplementedError:
                eventlog.log_w('fail: ImrefLoader::load_jpeg (<{}>, unsupported format)'.format(def_imref))

        # eventlog.log_w('fail: load overview image')
        self._stats.failure += 1
        raise ParamsError
