from shapely.geometry.point import Point
from shapely.geometry.polygon import Polygon

from typing import Final, Optional, Tuple, List

from edge import eventlog
from edge.types.exceptions import JnodeError

from edge.types.shape.point import PointI
from edge.types.shape.quad import QuadI


class MaskRegion(list):
    def __init__(self, points: List[Tuple[int, int]]):
        super().__init__(points)

    def test_valid(self) -> bool:
        polymask = Polygon(self)
        return polymask.is_valid

    def normalize(self, image_w: int, image_h: int) -> None:
        raise NotImplementedError


class MaskConf:
    def __init__(self, region: MaskRegion, mask_wh: Tuple[int, int]):
        self.region: Final[MaskRegion] = region
        self.mask_wh = mask_wh


class Mask:
    def __init__(self, conf: MaskConf):
        self._polygon: Final[Polygon] = Polygon(conf.region)
        self._scene_iw, self._scene_ih = conf.mask_wh

    def check_point_abs(self, abs_point: PointI) -> bool:
        return self._polygon.contains(
            Point(abs_point.il, abs_point.it))

    def check_quad_abs(self, quad: QuadI) -> bool:
        abs_point = PointI(*quad.center)
        return self.check_point_abs(abs_point)


class IntersectionMask:
    STOPBAR_NONE = 0
    STOPBAR_PRE = 1
    STOPBAR_ON = 2
    STOPBAR_POST = 3

    _UP = 4
    _DOWN = 5
    _EQ = 6
    _ERROR = 6

    def __init__(self, conf: MaskConf):
        self._scene_iw, self._scene_ih = conf.mask_wh

        self._stopbar_edge_far: Final[Tuple[Point, Point]] = Point(*conf.region[0]), Point(*conf.region[1])
        self._stopbar_edge_near: Final[Tuple[Point, Point]] = Point(*conf.region[2]), Point(*conf.region[3])
        self._stopbar_polygon: Final[Polygon] = Polygon(conf.region)

    @staticmethod
    def _point_to_line_relation(point_to_check: Point, line_start: Point, line_end: Point) -> int:
        x1 = line_start.x
        y1 = line_start.y

        x2 = line_end.x
        y2 = line_end.y

        x = point_to_check.x
        y = point_to_check.y

        # img coordinate y are reversed
        if x2 == x1:
            return IntersectionMask._ERROR

        elif y < (y2 - y1) / (x2 - x1) * (x - x1) + y1:
            return IntersectionMask._UP

        elif y == (y2 - y1) / (x2 - x1) * (x - x1) + y1:
            return IntersectionMask._EQ

        else:
            return IntersectionMask._DOWN

    def check_point_abs(self, point_lt: Tuple[int, int], dir_r: bool = True) -> Optional[int]:
        """
        :param point_lt:
        :param dir_r:
            Flag, which indicates the receding movement direction of the vehicle.
        """
        pt = Point(*point_lt)

        if self._stopbar_polygon.contains(pt):
            return self.STOPBAR_ON

        relation = self._point_to_line_relation(pt, *self._stopbar_edge_far)

        if relation == self._UP:
            return self.STOPBAR_POST if dir_r else self.STOPBAR_PRE

        elif relation == self._DOWN:
            return self.STOPBAR_PRE if dir_r else self.STOPBAR_POST

    def check_point_rel(self, point_lt: Tuple[float, float], dir_r: bool = True) -> Optional[int]:
        """
        :param point_lt:
        :param dir_r:
            Flag, which indicates the receding movement direction of the vehicle.
        """
        point_il = int(point_lt[0] * self._scene_iw)
        point_it = int(point_lt[1] * self._scene_ih)
        return self.check_point_abs((point_il, point_it), dir_r)


def decode__mask_region(mask_jnode: List[Tuple[int, int]]) -> MaskRegion:
    """
    :raise JnodeError:
    """
    if type(mask_jnode) != list:
        eventlog.log_e_param('#/')
        raise JnodeError

    index = 0
    while index < len(mask_jnode):
        point_jnode = mask_jnode[index]
        if not (type(point_jnode) in [list, tuple] or len(point_jnode) != 2):
            eventlog.log_e_param(f'#/{index}')
            raise JnodeError

        x, y = point_jnode
        if not (int == type(x) and 0 <= x <= 65535):
            eventlog.log_e_param(f'#/{index}/0')
            raise JnodeError

        if not (int == type(y) and 0 <= y <= 65535):
            eventlog.log_e_param(f'#/{index}/1')
            raise JnodeError

        index += 1

    return MaskRegion(mask_jnode)


def decode__mask_region2(mask_jnode: List[int]) -> MaskRegion:
    """
    :raise JnodeError:
    """
    if type(mask_jnode) != list:
        eventlog.log_e_param('#/')
        raise JnodeError

    it = iter(mask_jnode)
    mask_jnode = list(zip(it, it))

    return decode__mask_region(mask_jnode)
