import os

from PIL import Image
from datetime import datetime, timedelta
from typing import Dict, Final, List, Optional, Tuple

from edge import eventlog
from edge.support import fs
from edge.types.imageref import ImageRef
from edge_rawutil import rawutil


class ImageFileQueue:
    def __init__(self, twisted_reactor,
                 target_codec: Optional[str],
                 root_dir: str,
                 min_file_count: int,
                 max_file_count: int,
                 log_level: str = 'info') -> None:

        self._root_dir: Final[str] = root_dir
        self._min_file_count: Final[int] = min_file_count
        self._max_file_count: Final[int] = max_file_count

        self._target_codec = target_codec

        self._rawutil = rawutil.RawUtil.create(rawutil.MAX_WIDTH, rawutil.MAX_HEIGHT, log_level)
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._rawutil.destroy)

        fs.silent_mkdir_r(self._root_dir)
        file_names = os.listdir(self._root_dir)

        file_names.sort(reverse=False, key=self._decode_ts)

        self._last_file_suffix = self._decode_ts(file_names.pop()) if file_names else 0
        self._expected_file_count = len(file_names)

    @staticmethod
    def _decode_ts(name: str) -> int:
        return int(name.split('-')[-1])

    @classmethod
    def _cleanup(cls, root_dir: str, min_file_count: int) -> int:
        file_names = os.listdir(root_dir)
        file_names.sort(reverse=False, key=cls._decode_ts)

        actual_file_count = len(file_names)
        for file_name in file_names:
            if actual_file_count <= min_file_count:
                break

            fs.silent_rm_r(fs.join_path(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self, next_name: int) -> str:
        self._last_file_suffix += 1
        next_path = fs.join_path(self._root_dir, f'{self._last_file_suffix}-{next_name}')
        return next_path

    def add_image(self, ts: Optional[int], rgba_image: Image.Image) -> ImageRef:
        next_filepath = self._get_next_filepath(ts)

        if self._target_codec == 'JPEG':
            fs.edit_dir(self._root_dir, fs.write_jpeg_file, rgba_image, next_filepath)

        elif self._target_codec == 'RGBA':
            fs.edit_dir(self._root_dir, fs.write_rgba_file, rgba_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        w, h = rgba_image.size
        return ImageRef(ts,
                        url=next_filepath,
                        codec=self._target_codec,
                        width=w,
                        height=h)

    def add_image_ref(self, image_ref: ImageRef) -> ImageRef:
        next_filepath = self._get_next_filepath(image_ref.ts)

        if self._target_codec is None:
            fs.edit_dir(self._root_dir, os.link, image_ref.url, next_filepath)

        elif image_ref.codec == 'JPEG' and self._target_codec == 'JPEG':
            fs.edit_dir(self._root_dir, os.link, image_ref.url, next_filepath)

        else:
            rgba_pil_image = self._rawutil.handle_image(
                image_ref.url, image_ref.codec, image_ref.width, image_ref.height)
            if not rgba_pil_image:
                raise RuntimeError('convert raw-image to rgba')

            if self._target_codec == 'JPEG':
                fs.edit_dir(self._root_dir, fs.write_jpeg_file, rgba_pil_image, next_filepath)

            elif self._target_codec == 'RGBA':
                fs.edit_dir(self._root_dir, fs.write_rgba_file, rgba_pil_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(image_ref.ts,
                        url=next_filepath,
                        codec=self._target_codec,
                        width=image_ref.width,
                        height=image_ref.height)

    def add_image_blob(self,
                       blob_bytes: bytes,
                       blob_ts: Optional[int] = None,
                       blob_format: Optional[str] = None) -> ImageRef:

        next_filepath = self._get_next_filepath(blob_ts)

        # write file to the queue
        def write_bin_file(content_bytes):
            with open(next_filepath, 'wb') as bin_file:
                bin_file.write(content_bytes)

        fs.edit_dir(self._root_dir, write_bin_file, blob_bytes)

        # cleanup the queue
        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(blob_ts,
                        url=next_filepath,
                        codec=blob_format)

    def del_image(self, image_ref) -> None:
        # assert image_ref.codec == 'JPEG' #todo : move to where called

        # noinspection PyBroadException
        try:
            os.unlink(image_ref.url)
            self._expected_file_count -= 1
        except Exception:
            eventlog.log_x()


class JsonFileQueue:
    def __init__(self, args) -> None:
        """
        :type args: QueueArgs
        """
        self._root_dir = args.endpoint
        self._min_file_count = args.min_file_count
        self._max_file_count = args.max_file_count

        fs.silent_mkdir_r(self._root_dir)
        file_name_list = list(map(str, os.listdir(self._root_dir)))
        file_name_list.sort(key=lambda name: int(name.split('-')[0]))

        self._last_file_suffix = int(file_name_list.pop().split('-')[0]) if file_name_list else 0
        self._expected_file_count = len(file_name_list)

    @staticmethod
    def _cleanup(root_dir: str, min_file_count: int) -> int:
        file_name_list = list(map(str, os.listdir(root_dir)))
        file_name_list.sort(key=lambda name: int(name.split('-')[0]))

        actual_file_count = len(file_name_list)
        for file_name in file_name_list:
            if actual_file_count <= min_file_count:
                break

            fs.silent_rm_r(fs.join_path(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self, next_filename: Optional[str] = None) -> str:
        self._last_file_suffix += 1
        if next_filename:
            next_filepath = fs.join_path(self._root_dir, f'{self._last_file_suffix}-{next_filename}')
        else:
            next_filepath = fs.join_path(self._root_dir, str(self._last_file_suffix))

        return next_filepath

    def add_jnode(self, jnode: Dict, next_filename: Optional[str] = None) -> None:
        next_filepath = self._get_next_filepath(next_filename)
        fs.edit_dir(self._root_dir, fs.write_json_file, jnode, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)


class BlobFileTree:
    def __init__(self, root_dir: str, min_df: float,  max_dir_files: int = 256) -> None:
        """
        :param root_dir:
            <root>/<min-datetime>/<file-seq>__<date>__<time>[__<vrp>].<file_ext>
            <root>/2023-05-02T164205/000014__2023-05-02__16-42-05__30P264RA.jpeg
            <root>/2023-05-02T164205/000015__2023-05-02__16-42-06.jpeg
        :param min_df:
        """
        self._root_dir: Final[str] = root_dir
        self._min_df: Final[float] = min_df
        self._max_dir_files: Final[int] = max_dir_files
        self._max_depth: Final[timedelta] = timedelta(days=1)

        self._last_blob_dir: Optional[datetime] = None
        self._last_blob_seq: int = 0

        last_blob: Tuple[datetime, int] | None = self._load_state(self._root_dir)
        if last_blob is not None:
            self._last_blob_dir, self._last_blob_seq = last_blob

    @staticmethod
    def _load_state(root_dir: str) -> Tuple[datetime, int] | None:
        fs.silent_mkdir_r(root_dir)
        dt_dir_names: List[str] = os.listdir(root_dir)
        if not dt_dir_names:
            eventlog.log_i(f'done: root <{root_dir}> is empty')
            return

        # define <root>/<max-datetime>
        max_dt_dir_name: str = max(dt_dir_names, key=lambda d_name: datetime.fromisoformat(d_name))
        max_dt_dir_path: str = os.path.join(root_dir, max_dt_dir_name)

        # define <root>/<max-datetime>/<max-file-sec>*
        file_names: List[str] = os.listdir(max_dt_dir_path)
        if not file_names:
            eventlog.log_i(f'done: max-file <{max_dt_dir_path}> is empty')
            return

        max_file_name = max(file_names, key=lambda f_name: int(f_name.split('__')[0]))

        max_blob_dir: datetime = datetime.fromisoformat(max_dt_dir_name)
        max_blob_seq: int = int(max_file_name.split('__')[0])

        eventlog.log_i(f'done: max-blob-dir:<{max_dt_dir_name}, max-blob-seq:{max_blob_seq}>')
        return max_blob_dir, max_blob_seq

    def _cleanup(self, min_dt: datetime) -> None:
        dir_names: List[str] = os.listdir(self._root_dir)
        if not dir_names:
            return

        dir_name_to_dt_list_asc: List[Tuple[str, datetime]] = [
            (d, datetime.fromisoformat(d)) for d in dir_names
        ]

        dir_name_to_dt_list_asc.sort(key=lambda d_dt: d_dt[1])
        for dir_name, dir_dt in dir_name_to_dt_list_asc:
            if dir_dt < min_dt:
                fs.silent_rm_r(os.path.join(self._root_dir, dir_name))
                eventlog.log_i(f'done: rm <{dir_name}/>')
                continue

            df = fs.get_df(self._root_dir)
            if df < self._min_df:
                eventlog.log_i('exec: rm <{}/>, df:{:.4}, min-df:{:.4}'.format(dir_name, df, self._min_df))
                fs.silent_rm_r(os.path.join(self._root_dir, dir_name))
                continue

            break

    def _get_next_filepath(self, blob_dt: datetime, blob_name: str) -> str:
        self._last_blob_seq += 1
        if self._last_blob_dir is None or self._last_blob_seq % self._max_dir_files == 0:
            self._cleanup(blob_dt - self._max_depth)
            self._last_blob_dir = blob_dt

        dir_dt: str = self._last_blob_dir.strftime('%Y-%m-%dT%H%M%S%z')
        file_dt: str = blob_dt.strftime('%Y-%m-%d__%H-%M-%S')
        file_name = '{:05}__{}{}'.format(self._last_blob_seq, file_dt, blob_name)

        return fs.join_path(self._root_dir, dir_dt, file_name)

    def add_blob(self, blob_dt: datetime, blob_name: Optional[str], blob_bytes: bytes) -> Optional[str]:
        next_filepath = self._get_next_filepath(blob_dt, blob_name)

        def write_bin_file(content_bytes):
            with open(next_filepath, 'wb') as bin_file:
                bin_file.write(content_bytes)

        fs.edit_dir(os.path.dirname(next_filepath), write_bin_file, blob_bytes)
        return next_filepath
