import math

from edge.types.shape.point import PointI
from typing import List


def linear_regression(points: List[PointI]) -> tuple[float, float]:
    """
    Y = a + bX
    b = slope = sum((Xi - X_mean) * (Yi - Y_mean)) / sum((Xi - X_mean)^2)
    a = intercept = Y_mean - b * X_mean

    The slope of the line (b) represents the change in Y for every unit change in X.

    The intercept (a) represents the value of Y when X equal to 0.
    So this is the value of Y when the line crosses the Y axis.
    """
    sum_x = 0
    sum_y = 0
    sum_xy = 0
    sum_x_squared = 0
    n = len(points)

    for point in points:
        sum_x += point.il
        sum_y += point.it
        sum_xy += point.il * point.it
        sum_x_squared += point.il * point.il

    x_mean = sum_x / n
    y_mean = sum_y / n
    slope = (sum_xy - n * x_mean * y_mean) / (sum_x_squared - n * x_mean * x_mean)
    intercept = y_mean - slope * x_mean
    return slope, intercept


def distance_to_line(point_x: float, point_y: float, line_slope: float, line_intercept: float) -> float:
    """
    Calculates the perpendicular distance from a point to a line.
    """
    numerator = abs(line_slope * point_x - point_y + line_intercept)
    denominator = math.sqrt(line_slope * line_slope + 1)
    return numerator / denominator


def distance_to_line_y(point_x: float, point_y: float, line_slope: float, line_intercept: float) -> float:
    """
    Calculates the absolute distance from a point to a line along the Y axis.
    """
    numerator = abs(line_slope * point_x - point_y + line_intercept)
    return numerator


def filter_points(points: List[PointI], slope: float, intercept: float, threshold: float) -> List[PointI]:
    filtered_points = []
    for point in points:
        distance = distance_to_line_y(point.il, point.it, slope, intercept)
        if distance <= threshold:
            filtered_points.append(point)
    return filtered_points
