import pipes
import subprocess

from typing import List, Optional, Tuple

from edge.types.exceptions import ExternalError, InternalError, ToutError


def run_silent(esc_cmd: str) -> int:
    """
    :param esc_cmd:
        Escaped shell-command. Use pipes.quote() to escape input form external sources
    """
    return subprocess.run(esc_cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT).returncode


def quote(data: str) -> str:
    return pipes.quote(data)


def exec_cli(cli_binary: str, cli_params: List[str] | Tuple[str],
             tout_sec: Optional[int] = 10,
             encoding: str = 'ascii') -> str:
    """
    :param cli_binary:
    :param cli_params:
    :param tout_sec:
    :param encoding: 'ascii' | 'utf-8'
    :returns: stdout
    """
    import subprocess
    from edge import eventlog

    eventlog.log_d(f'exec: <{cli_binary}>')

    try:

        cli_result = subprocess.run([cli_binary] + list(cli_params),
                                    capture_output=True, text=True, encoding=encoding,
                                    check=True, timeout=tout_sec)
        return cli_result.stdout

    except subprocess.TimeoutExpired as x:
        eventlog.log_e('fail: <{}> (timeout:{})'.format(x.cmd, x.timeout))
        raise ToutError

    except subprocess.CalledProcessError as x:
        if x.returncode == 64:  # EX_USAGE
            eventlog.log_e('fail: ExternalError <{}>'.format(x.cmd))
            raise ExternalError

        elif x.returncode == 70:  # EX_SOFTWARE
            eventlog.log_e('fail: InternalError <{}> ({})'.format(x.cmd, x.stderr))
            raise InternalError

        else:
            eventlog.log_e('fail: <{}> (returncode:{})'.format(x.cmd, x.returncode))
            raise InternalError
