import struct
import ujson

from typing import BinaryIO, Dict, Final, Literal, Optional, Tuple, TypeVar

from edge.services.types.tz import ITzService


class BlobMeta:
    def __str__(self) -> str:
        raise NotImplementedError

    def serialize(self) -> Dict:
        raise NotImplementedError


class JsonBlobMeta(BlobMeta):
    def __init__(self, jnode: Dict):
        BlobMeta.__init__(self)
        self._jnode = jnode

    def __str__(self) -> str:
        return ujson.dumps(self._jnode)

    def serialize(self) -> Dict:
        return self._jnode


class StructBlobMeta(BlobMeta):
    """
        struct meta_v0 {
        uint8_t    ver_;
        uint8_t    blob_name_size;
        uint16_t   blob_sign_size;
        uint16_t   blob_cert_size;
        uint8_t    reserved[26];
        uint8_t    blob_name_data[96];
        uint8_t    blob_sign_data[1024];
        uint8_t    blob_cert_data[8192];
    }

    struct meta_v1 {
        uint8_t    ver_;
        uint8_t    blob_name_size;
        uint16_t   blob_sign_size;
        uint16_t   blob_cert_size;
        uint8_t    reserved[26];
        uint8_t    blob_name_data[96];
        uint8_t    blob_sign_data[8192];
        uint8_t    blob_cert_data[8192];
    }
    """

    def __init__(self, sign_data: bytes, cert_data: bytes = b'') -> None:
        super().__init__()
        self.struct: Final[bytes] = struct.pack(
            '<BBHH26x96s8192s8192s',
            1, 0, len(sign_data), len(cert_data),
            b'', sign_data, cert_data)

    def __str__(self):
        raise NotImplementedError

    def serialize(self):
        raise NotImplementedError


BsChanRef = str
BlobProfile = Literal['args', 'plate', 'vehicle', 'thumb', 'scene', 'overview', 'measurement', 'live', 'trainer']
BlobProfileToBsChanMap = Dict[BlobProfile, BsChanRef]


class Blob:
    def __init__(self, profile: BlobProfile, file_ext: str):
        self._name: Optional[str] = None
        self._profile = profile
        self._file_ext = file_ext
        self._hash = None

    @property
    def name(self) -> str:
        return self._name or ''

    @name.setter
    def name(self, value: str):
        if len(value):
            self._name = value
        else:
            raise RuntimeError

    @property
    def profile(self) -> BlobProfile:
        return self._profile

    @property
    def min_ts(self) -> int:
        """
        :returns:
            value in usec
        """
        raise NotImplementedError

    @property
    def max_ts(self) -> int:
        """
        :return:
            value in usec
        """
        raise NotImplementedError

    @property
    def file_path(self) -> Optional[str]:
        raise NotImplementedError

    @property
    def file_ext(self) -> str:
        return self._file_ext

    @property
    def file_data(self) -> bytes:
        raise NotImplementedError

    @property
    def file_meta(self) -> Optional[bytes]:
        return None

    @property
    def hash(self) -> str:
        return self._hash

    @hash.setter
    def hash(self, value: Tuple[str, str]) -> None:
        assert len(value) == 2
        self._hash = value

    def serialize(self, tz_service: Optional[ITzService] = None) -> Dict:
        raise NotImplementedError


class MemBlob(Blob):
    def __init__(self, profile: BlobProfile, ts: int, data: bytes,
                 meta: Optional[BlobMeta] = None,
                 file_ext: str = 'jpeg'):

        super().__init__(profile, file_ext=file_ext)

        self._ts = ts
        self._data = data
        self._meta = meta

    @property
    def min_ts(self) -> int:
        return self._ts

    @property
    def max_ts(self) -> int:
        return self._ts

    @property
    def file_path(self) -> Optional[str]:
        return None

    @property
    def file_data(self) -> bytes:
        return self._data

    @property
    def meta(self) -> Optional[BlobMeta]:
        return self._meta

    @meta.setter
    def meta(self, value: Optional[BlobMeta]):
        self._meta = value

    def serialize(self, tz_service: Optional[ITzService] = None) -> Dict:
        # jnode = {
        #     "name": "0_150402739915.jpeg",
        #     "info": {
        #         "size": [1920, 1200, 100500],
        #         "trans": {
        #             "o": [0.0, 0.0],
        #             "i": [1.0, 0.0],
        #             "j": [0.0, 1.0]
        #         }
        #     },
        #     "hash": [
        #         "md5",
        #         "2ce2e89e20dd5ad3b611f12e56e6fa79f35be149d8b226c23666ced9c73acce2"
        #     ]
        # }

        jnode = {
            'name': self.name,
        }

        if tz_service:
            jnode['time'] = tz_service.map_ts_usec_to_isostr(self.min_ts)

        if self._meta:
            jnode['info'] = self._meta.serialize()

        if self.hash:
            jnode['hash'] = self.hash

        return jnode


class FileBlob(Blob):
    def __init__(self, min_ts: int, max_ts: int, data: BinaryIO,
                 meta: Optional[BlobMeta] = None,
                 profile: BlobProfile = 'measurement'):

        super().__init__(profile, file_ext='mp4')

        self._min_ts = min_ts
        self._max_ts = max_ts

        self._file = data
        self._meta = meta

    def __del__(self):
        self._file.close()

    @property
    def min_ts(self) -> int:
        return self._min_ts

    @property
    def max_ts(self) -> int:
        return self._max_ts

    @property
    def file_path(self) -> str:
        return self._file.name

    @property
    def file_data(self) -> bytes:
        return self._file.read()

    def serialize(self, tz_service: Optional[ITzService] = None) -> Dict:
        jnode = {
            'name': self.name,
        }

        if tz_service:
            jnode['time'] = [
                tz_service.map_ts_usec_to_isostr(self.min_ts),
                tz_service.map_ts_usec_to_isostr(self.max_ts)
            ]

        if self._meta:
            jnode['info'] = self._meta.serialize()

        if self.hash:
            jnode['hash'] = self.hash

        return jnode


TBlob = TypeVar('TBlob', bound=Blob)
