from typing import Self, Tuple

from edge.types.shape.rect import RectI, RectF


class QuadI:
    def __init__(self,
                 x0: int | float, y0: int | float,
                 x1: int | float, y1: int | float,
                 x2: int | float, y2: int | float,
                 x3: int | float, y3: int | float):

        self.x0 = int(x0)
        self.y0 = int(y0)

        self.x1 = int(x1)
        self.y1 = int(y1)

        self.x2 = int(x2)
        self.y2 = int(y2)

        self.x3 = int(x3)
        self.y3 = int(y3)

    @classmethod
    def create(cls, rect: RectI) -> Self:
        return cls(rect.il, rect.it,
                   rect.ir, rect.it,
                   rect.ir, rect.ib,
                   rect.il, rect.ib)

    def __contains__(self, point_xy: Tuple[int, int]) -> bool:
        min_x = min(self.x0, self.x3)
        max_x = max(self.x1, self.x2)
        min_y = min(self.y0, self.y1)
        max_y = max(self.y2, self.y3)

        same_x = min_x < point_xy[0] < max_x
        same_y = min_y < point_xy[1] < max_y
        return same_x and same_y

    def map(self,
            source_im_w: int,
            source_im_h: int,
            target_im_w: int,
            target_im_h: int,
            margin_tb: int = 0,
            margin_lr: int = 0
            ) -> Tuple[int, int, int, int, int, int, int, int]:

        x0 = int(float(self.x0) / source_im_w * target_im_w - margin_lr)
        y0 = int(float(self.y0) / source_im_h * target_im_h - margin_tb)

        x1 = int(float(self.x1) / source_im_w * target_im_w + margin_lr)
        y1 = int(float(self.y1) / source_im_h * target_im_h - margin_tb)

        x2 = int(float(self.x2) / source_im_w * target_im_w + margin_lr)
        y2 = int(float(self.y2) / source_im_h * target_im_h + margin_tb)

        x3 = int(float(self.x3) / source_im_w * target_im_w - margin_lr)
        y3 = int(float(self.y3) / source_im_h * target_im_h + margin_tb)

        return x0, y0, x1, y1, x2, y2, x3, y3

    @property
    def bbox_recti(self) -> RectI:
        il = min(self.x0, self.x3)
        ir = max(self.x1, self.x2)
        it = min(self.y0, self.y1)
        ib = max(self.y2, self.y3)

        return RectI(il, it, ir, ib)

    def get_rel_bbox_recf(self, source_im_w: int, source_im_h: int) -> RectF:
        fl = float(min(self.x0, self.x3)) / source_im_w
        fr = float(max(self.x1, self.x2)) / source_im_w
        ft = float(min(self.y0, self.y1)) / source_im_h
        fb = float(max(self.y2, self.y3)) / source_im_h

        return RectF(fl, ft, fr, fb)

    def map_bbox_wh(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        bbox: RectI = self.bbox_recti

        target_bbox_fw = float(bbox.ir - bbox.il) / source_im_w * target_im_w
        target_bbox_fh = float(bbox.ib - bbox.it) / source_im_h * target_im_h
        return int(target_bbox_fw), int(target_bbox_fh)

    def map_bbox_ref(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        source_bbox_rel: RectF = self.get_rel_bbox_recf(source_im_w, source_im_h)
        target_bbox_il = source_bbox_rel.fl * target_im_w
        target_bbox_ir = source_bbox_rel.fr * target_im_w
        target_bbox_fb = source_bbox_rel.fb * target_im_h

        return int(0.5 * (target_bbox_il + target_bbox_ir)), int(target_bbox_fb)

    def map_bbox_ref2(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        source_bbox_rel: RectF = self.get_rel_bbox_recf(source_im_w, source_im_h)
        target_bbox_il = source_bbox_rel.fl * target_im_w
        target_bbox_ir = source_bbox_rel.fr * target_im_w
        target_bbox_ft = float(max(self.y0, self.y1)) / source_im_h * target_im_h  # TODO: Should be min()?

        return int(0.5 * (target_bbox_il + target_bbox_ir)), int(target_bbox_ft)

    @property
    def center_l(self) -> int:
        cx: float = (self.x0 + self.x1 + self.x2 + self.x3) / 4.0
        return int(cx)

    @property
    def center_t(self) -> int:
        cy: float = (self.y0 + self.y1 + self.y2 + self.y3) / 4.0
        return int(cy)

    @property
    def center(self) -> Tuple[int, int]:
        return self.center_l, self.center_t

    @property
    def wh(self) -> Tuple[int, int]:
        bbox = self.bbox_recti
        return bbox.iw, bbox.ih


def encode__quadi(self: QuadI) -> Tuple[int, int, int, int, int, int, int, int]:
    return (self.x0, self.y0, self.x1, self.y1,
            self.x2, self.y2, self.x3, self.y3)


def decode__quadi(jnode: Tuple[int, int, int, int, int, int, int, int]) -> QuadI:
    if len(jnode) != 8:
        raise RuntimeError

    return QuadI(*jnode)
