from typing import Self, Tuple, Union


class RectF:
    def __init__(self, fl: float | int, ft: float | int, fr: float | int, fb: float | int):
        self.fl = float(fl)
        self.ft = float(ft)
        self.fr = float(fr)
        self.fb = float(fb)
        self.fw = float(fr - fl)
        self.fh = float(fb - ft)

    def __contains__(self, point_xy: Tuple[float, float]) -> bool:
        contains_x = self.fl < point_xy[0] < self.fr
        contains_y = self.ft < point_xy[1] < self.fb
        return contains_x and contains_y

    @property
    def ltrb(self) -> Tuple[float, float, float, float]:
        return self.fl, self.ft, self.fr, self.fb

    @classmethod
    def create(cls,
               center_l: Union[int, float],
               center_t: Union[int, float],
               width: Union[int, float],
               height: Union[int, float]) -> Self:

        center_fl, center_ft = float(center_l), float(center_t)
        half_fw, half_fh = width / 2.0, height / 2.0
        return cls(
            center_fl - half_fw, center_ft - half_fh,
            center_fl + half_fw, center_ft + half_fh)

    @property
    def center(self) -> Tuple[float, float]:
        return (self.fl + self.fr) / 2.0, (self.ft + self.fb) / 2.0

    @property
    def area(self) -> float:
        return self.fw * self.fh

    def get_visible_100(self,
                        image_w: Union[int, float] = 1.0,
                        image_h: Union[int, float] = 1.0) -> bool:

        return self.fl > 0.0 and self.fr < image_w and self.ft > 0.0 and self.fb < image_h

    def get_visible(self,
                    image_w: Union[int, float] = 1.0,
                    image_h: Union[int, float] = 1.0) -> float:
        """
        :returns:
            value from 0.0% to 100.0%
        """
        self_area = self.area
        if self_area > 0.0:
            self_visible = RectF(self.fl, self.ft, self.fr, self.fb)
            self_visible.strip_invisible(image_w, image_h)
            return 100.0 * self_visible.area / self_area

        else:
            return 0.0

    def strip_invisible(self,
                        image_w: Union[int, float] = 1.0,
                        image_h: Union[int, float] = 1.0) -> None:
        """
        :raise ValueError:
        """
        def norm(max_coord: Union[int, float], coord: float) -> float:
            if coord < 0.0:
                return 0.0

            elif coord > max_coord:
                return float(max_coord)

            else:
                return coord

        self.fl = norm(image_w, self.fl)
        self.ft = norm(image_h, self.ft)
        self.fr = norm(image_w, self.fr)
        self.fb = norm(image_h, self.fb)

        self.fw = self.fr - self.fl
        self.fh = self.fb - self.ft

    def make_visible(self,
                     image_w: Union[int, float] = 1.0,
                     image_h: Union[int, float] = 1.0) -> None:
        """
        :raise ValueError:
        """
        if self.fw > image_w or self.fh > image_h:
            raise ValueError

        # move self to a visible part of the image
        next_fl = self.fl
        if self.fl < 0.0:
            next_fl = 0.0

        elif self.fl + self.fw > image_w:
            next_fl = image_w - self.fw

        next_ft = self.ft
        if self.ft < 0.0:
            next_ft = 0.0

        elif self.ft + self.fh > image_h:
            next_ft = image_h - self.fh

        # fix rounding errors if any
        next_fw, next_fh = self.fw, self.fh

        next_fr = next_fl + next_fw
        if next_fr >= image_w:
            next_fw -= 1.0

            next_fr = next_fl + next_fw
            if next_fr >= image_w:
                raise ValueError

        next_fb = next_ft + next_fh
        if next_fb >= image_h:
            next_fh -= 1.0

            next_fb = next_ft + next_fh
            if next_fb >= image_h:
                raise ValueError

        # actualize self coordinates
        self.fl = next_fl
        self.ft = next_ft
        self.fr = next_fr
        self.fb = next_fb

        self.fw = self.fr - self.fl
        self.fh = self.fb - self.ft

    def scale_max(self,
                  min_w: Union[int, float] = 0.0,
                  min_h: Union[int, float] = 0.0,
                  max_w: Union[int, float] = 1.0,
                  max_h: Union[int, float] = 1.0) -> None:
        """
        :raise ValueError:
        """
        assert self.fh > 0.0
        wh_ratio = self.fw / self.fh

        def get_new_wh() -> Tuple[float, float]:
            """
            :raises ValueError:
            """
            if self.fw <= max_w and self.fh <= max_h:
                return self.fw, self.fh

            # strip extra h-pixels from self height (and keep aspect-ratio)
            if self.fh > max_h:
                next_fw = max_h * wh_ratio
                if min_w < next_fw <= max_w:
                    return next_fw, max_h

            # strip extra w-pixels from self width (and keep aspect-ratio)
            if self.fw > max_w:
                next_fh = max_w / wh_ratio
                if min_h < next_fh <= max_h:
                    return max_w, next_fh

            raise ValueError

        self.fw, self.fh = get_new_wh()

        half_fw, half_fh = self.fw / 2.0, self.fh / 2.0
        center_fl, center_ft = self.center

        self.fl = center_fl - half_fw
        self.ft = center_ft - half_fh
        self.fr = center_fl + half_fw
        self.fb = center_ft + half_fh


class RectI:
    def __init__(self, il: int | float, it: int | float, ir: int | float, ib: int | float):
        self.il = int(il)
        self.it = int(it)
        self.ir = int(ir)
        self.ib = int(ib)

        self.iw = int(ir - il)
        self.ih = int(ib - it)

    @classmethod
    def create(cls, rel_rect: RectF, image_w: int, image_h: int) -> Self:
        return cls(rel_rect.fl * image_w,
                   rel_rect.ft * image_h,
                   rel_rect.fr * image_w,
                   rel_rect.fb * image_h)

    def __contains__(self, point_xy: Tuple[int, int]) -> bool:
        contains_x = self.il < point_xy[0] < self.ir
        contains_y = self.it < point_xy[1] < self.ib
        return contains_x and contains_y

    @property
    def center(self) -> Tuple[int, int]:
        return (self.il + self.ir) // 2, (self.it + self.ib) // 2

    @property
    def ltrb(self) -> Tuple[int, int, int, int]:
        return self.il, self.it, self.ir, self.ib

    @property
    def wh(self) -> Tuple[int, int]:
        return self.iw, self.ih


def encode__recti(self: RectI) -> Tuple[int, int, int, int]:
    return self.il, self.it, self.ir, self.ib
