import os
import shutil
import time

from typing import Dict, Optional, Self, Union

from edge import eventlog
from edge.support import fs
from edge.webserver.cgi import CgiError500
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod.dg import DgMod


class UimodBackup:
    def __init__(self, source_filepath: str, backup_filepath: str):
        self._source_filepath = source_filepath
        self._backup_filepath = backup_filepath

    @classmethod
    def create(cls, backup_dir: str, uimod: Union[ph.UiMod, ph.UiMod2]) -> Optional[Self]:
        backup_uiconf_name = f'ui-{str(uimod)}-old.json'
        backup_uiconf_path = fs.join_path(backup_dir, backup_uiconf_name)
        try:
            eventlog.log_d(f'exec: backup {uimod.conf_path} to {backup_uiconf_path}')
            shutil.copy2(uimod.conf_path, backup_uiconf_path)
            eventlog.log_d('done')

            return cls(uimod.conf_path, backup_uiconf_path)

        except (IOError, os.error, shutil.Error) as x:
            eventlog.log_e(f'fail: cp {uimod.conf_path} to {backup_uiconf_path} ({x})')

    def restore(self) -> bool:
        eventlog.log_d(f'exec: restore {self._backup_filepath} to {self._source_filepath}')
        try:
            shutil.copy2(self._backup_filepath, self._source_filepath)
            eventlog.log_d('done')
            return True

        except (os.error, shutil.Error) as x:
            eventlog.log_e(f'fail: cp {self._backup_filepath} to {self._source_filepath} ({x})')

        return False

    def remove(self) -> None:
        eventlog.log_d(f'exec: delete {self._backup_filepath}')
        os.remove(self._backup_filepath)


class DgmodBackup:
    def __init__(self, source_dirpath: str, backup_dirpath: str):
        self._source_dirpath = source_dirpath
        self._backup_dirpath = backup_dirpath

    @classmethod
    def create(cls, dir_path: str, dgmod: DgMod) -> Optional[Self]:
        source_dirpath = dgmod.conf_dir
        backup_dirpath = fs.join_path(dir_path, f'edge-{dgmod.type}.{dgmod.id}')

        try:
            eventlog.log_d(f'exec: backup {source_dirpath} to {backup_dirpath}')
            shutil.copytree(source_dirpath, backup_dirpath)
            eventlog.log_d('done')

            return DgmodBackup(source_dirpath, backup_dirpath)

        except (IOError, os.error, shutil.Error) as x:
            eventlog.log_e(f'fail: backup {source_dirpath} to {backup_dirpath} ({x.message})')

    def restore(self) -> bool:
        def onerror_fn(error_func, error_path, exc_info):
            eventlog.log_w(f'fail: rm {error_path} ({exc_info[1].message})')

        try:
            eventlog.log_d(f'exec: restore {self._backup_dirpath} to {self._source_dirpath}')
            shutil.rmtree(self._source_dirpath, ignore_errors=False, onerror=onerror_fn)
            shutil.copytree(self._backup_dirpath, self._source_dirpath)
            eventlog.log_d('done')
            return True

        except shutil.Error as x:
            for srcname, dstname, why in x.args[0]:
                eventlog.log_e(f'fail: cp {srcname} to {dstname} ({why})')

        return False

    def remove(self) -> None:
        def onerror_fn(error_func, error_path, exc_info):
            eventlog.log_w(f'fail: rm {error_path} ({exc_info[1].message})')

        eventlog.log_d(f'exec: delete {self._backup_dirpath}')
        shutil.rmtree(self._backup_dirpath, ignore_errors=False, onerror=onerror_fn)


class NewFileBackup:
    def __init__(self, backup_filepath: str):
        self._backup_filepath = backup_filepath

    @classmethod
    def create(cls, backup_dir: str, uimod: Union[ph.UiMod, ph.UiMod2], new_uiconf: Dict) -> Optional[Self]:
        # backup configuration files
        new_name = f'ui-{str(uimod)}-new.json'
        new_filepath = fs.join_path(backup_dir, new_name)
        # noinspection PyBroadException
        try:
            eventlog.log_d(f'exec: create {new_filepath}')
            fs.write_json_file(new_uiconf, new_filepath, pretty=True)

        except Exception:
            eventlog.log_x()

        # handle backup succcess
        jnode_backup = cls(new_filepath)
        eventlog.log_d('done')
        return jnode_backup

    @staticmethod
    def restore() -> bool:
        return True

    def remove(self) -> None:
        os.remove(self._backup_filepath)


class CgiBackup(object):
    def __init__(self, bak_dir: str, bak_files: Dict[str, str]):
        self._bak_dir = bak_dir
        self._bak_files = bak_files

    def __call__(self, fn, *extra_args, **extra_kwargs):
        return lambda *fn_args, **fn_kwargs: self._exec(fn, *fn_args, **fn_kwargs)

    # noinspection PyBroadException
    def _exec(self, fn, *fn_args, **fn_kwargs):
        """
        :raise CgiError:
        """
        bak_dir = '{}.{}'.format(self._bak_dir, int(time.time() * 1000000))

        self._create_backup(bak_dir)
        try:
            answer = fn(*fn_args, **fn_kwargs)
            self._remove_backup(bak_dir)
            return answer

        except Exception:
            self._restore_backup(bak_dir)
            raise

    def _create_backup(self, bak_dir: str) -> None:
        """
        :raise CgiError:
        """
        try:
            os.makedirs(bak_dir)
            for (name, file_path) in self._bak_files.items():
                shutil.copy(file_path, os.path.join(bak_dir, name))

        except Exception:
            eventlog.log_e('fail: http500')
            eventlog.log_x()
            raise CgiError500

    def _restore_backup(self, bak_dir: str):
        """
        :raise CgiError:
        """
        # noinspection PyBroadException
        try:
            for (name, file_path) in self._bak_files.items():
                shutil.copy(os.path.join(bak_dir, name), file_path)

        except Exception:
            eventlog.log_x()

        self._remove_backup(bak_dir)

    @staticmethod
    def _remove_backup(bak_dir: str) -> None:
        # noinspection PyBroadException
        try:
            shutil.rmtree(bak_dir)
        except Exception:
            eventlog.log_x()
