__all__ = [
    'CgiError', 'CgiError400', 'CgiError404', 'CgiError405', 'CgiError409', 'CgiError500', 'CgiError503',
    'CgiLicError', 'CgiCamError',

    'CgiPathInfo', 'CgiPathInfoT',
    'CgiRequest', 'CgiResponse',

    'CgiHandler', 'cgi_handle',

    'CgiHandlerFactory',
    'CgiHandlerFactoryList',
    'CgiHandlerFactoryDict'
]

import fcntl
import os
import sys
import ujson

from typing import Dict, Callable, Generic, List, Optional, Self, TypeVar, Union

from edge import eventlog
from edge.types.exceptions import *


CGI_R200_SUCCESS = 200
CGI_R400_BAD_REQUEST = 400
CGI_R404_NOT_FOUND = 404
CGI_R405_METHOD_NOT_ALLOWED = 405
CGI_R409_CONFLICT = 409
CGI_R500_INTERNAL_SERVER_ERROR = 500
CGI_R503_SERVICE_UNAVAILABLE = 503


class CgiPathInfo(list):
    def __init__(self, raw_path_info: str):
        super().__init__(raw_path_info.split('/')[1:])


CgiPathInfoT = TypeVar('CgiPathInfoT')


class CgiRequest(Generic[CgiPathInfoT]):
    def __init__(self, method: str, path_info: CgiPathInfoT, query_string: Optional[str], body: Optional[str]):
        self.method = method
        self.path_info = path_info
        self.query_string = query_string
        self.body = body

        self._body_jnode: Optional[Dict] = None

    @property
    def query_params(self) -> List[str]:
        """
        :raise CgiError:
        """
        if self.query_string is None:
            return []

        self.assert_query_present()
        return self.query_string[1:].split('&')

    @property
    def body_jnode(self) -> Dict:
        """
        :raise CgiError:
        """
        if self.body is None:
            eventlog.log_e('fail: http400 (invalid body)')
            raise CgiError400

        elif self._body_jnode is None:
            try:
                self._body_jnode = ujson.loads(self.body)
                if self.body_jnode is None:
                    raise ValueError

            except ValueError:
                eventlog.log_e('fail: http400 (loads)')
                raise CgiError400

        return self._body_jnode

    @staticmethod
    def _read_stdin() -> Optional[str]:
        """
        :raises CgiError:
        """
        # make stdin a non-blocking file
        stdin_fd = sys.stdin.fileno()
        old_flags = fcntl.fcntl(stdin_fd, fcntl.F_GETFL)
        fcntl.fcntl(stdin_fd, fcntl.F_SETFL, old_flags | os.O_NONBLOCK)
        try:
            return sys.stdin.read() or None

        except Exception as x:
            eventlog.log_e(f'fail: http500 ({x})')
            raise CgiError500

        finally:
            fcntl.fcntl(stdin_fd, fcntl.F_SETFL, old_flags)

    @classmethod
    def unpack(cls, path_info_factory: Callable[[str], CgiPathInfoT] = CgiPathInfo) -> Self:
        """
        :raises CgiError:
        """
        try:
            # read request
            env_method = os.environ['REQUEST_METHOD']
            env_path_info = os.getenv('PATH_INFO') or None
            env_query_string = os.getenv('QUERY_STRING') or None
            body = cls._read_stdin() or None

            eventlog.log_i('%s|%s|%s|%s' % (env_method, env_path_info, env_query_string, body))

            # unpack request
            path_info = path_info_factory(env_path_info or '')
            return cls(env_method, path_info, env_query_string, body)

        except CgiError:
            eventlog.log_e('fail: unpack query-args')
            raise

        except Exception as x:
            eventlog.log_e(f'fail: {x}')
            raise CgiError500

    def assert_query_present(self):
        if self.query_string is None:
            eventlog.log_d('fail: http400')
            raise CgiError400

        elif not self.query_string.startswith('?'):
            eventlog.log_d('fail: http400')
            raise CgiError400

        elif len(self.query_string) <= 1:
            eventlog.log_d('fail: http400')
            raise CgiError400

    def assert_query_absent(self):
        if self.query_string:
            eventlog.log_d('fail: http400')
            raise CgiError400

    def assert_body_present(self):
        if not self.body:
            eventlog.log_d('fail: http400')
            raise CgiError400

    def assert_body_absent(self):
        if self.body:
            eventlog.log_d('fail: http400')
            raise CgiError400


class CgiResponse:
    def __init__(self, status: int = CGI_R200_SUCCESS, body: Optional[Union[Dict, List]] = None):
        self.status = str(status)
        self.body = body

    @classmethod
    def create(cls, reply_jnode: Dict) -> Self:
        reply_status: int = reply_jnode['status']
        if reply_jnode.get('envelope', True):
            if 'envelope' in reply_jnode:
                del reply_jnode['envelope']
            reply_body = reply_jnode

        elif reply_status == CGI_R200_SUCCESS:
            reply_body = reply_jnode['result']

        else:
            reply_body = None

        return cls(reply_status, reply_body)

    @classmethod
    def create200(cls,
                  result_jnode: Optional[Union[Dict, List]] = None,
                  envelope: bool = True
                  ) -> Self:

        return cls.create({
            'envelope': envelope,
            'status': CGI_R200_SUCCESS,
            'result': [] if result_jnode is None else result_jnode
        })

    def send(self):
        sys.stderr.write(self.status)
        if self.body:
            body_utf8 = ujson.dumps(self.body, ensure_ascii=False)
            sys.stdout.write(body_utf8)


class CgiError(GenericError):
    def __init__(self, reply: CgiResponse):
        self.reply = reply


class CgiError400(CgiError):
    def __init__(self, body_jnode: Optional[Union[Dict, List]] = None):
        super().__init__(CgiResponse(CGI_R400_BAD_REQUEST, body_jnode))


class CgiError404(CgiError):
    def __init__(self, body_jnode: Optional[Union[Dict, List]] = None):
        super().__init__(CgiResponse(CGI_R404_NOT_FOUND, body_jnode))


class CgiError405(CgiError):
    def __init__(self, body_jnode: Optional[Union[Dict, List]] = None):
        super().__init__(CgiResponse(CGI_R405_METHOD_NOT_ALLOWED, body_jnode))


class CgiError409(CgiError):
    def __init__(self, body_jnode: Optional[Union[Dict, List]] = None):
        super().__init__(CgiResponse(CGI_R409_CONFLICT, body_jnode))


class CgiError500(CgiError):
    def __init__(self, body_jnode: Optional[Union[Dict, List]] = None):
        super().__init__(CgiResponse(CGI_R500_INTERNAL_SERVER_ERROR, body_jnode))


class CgiError503(CgiError):
    def __init__(self, body_jnode: Optional[Union[Dict, List]] = None):
        super().__init__(CgiResponse(CGI_R503_SERVICE_UNAVAILABLE, body_jnode))


class CgiLicError(CgiError503):
    def __init__(self):
        super().__init__({
            'status': CGI_R503_SERVICE_UNAVAILABLE,
            'errors': [
                {
                    'code': 'lpr-licence-absent',
                    'message': 'free lpr licence is not available'
                }
            ]
        })


class CgiCamError(CgiError503):
    def __init__(self):
        super().__init__({
            'status': CGI_R503_SERVICE_UNAVAILABLE,
            'errors': [
                {
                    'code': 'videosource-error',
                    'message': 'something wrong with videosource'
                }
            ]
        })


CgiHandler = Callable[[], CgiResponse]
CgiHandlerFactory = Callable[[CgiRequest], CgiHandler]
CgiHandlerFactoryList = List[CgiHandlerFactory]
CgiHandlerFactoryDict = Dict[str, CgiHandlerFactoryList]


# noinspection PyBroadException
def cgi_handle(
        handler_factory: Callable[[CgiRequest], CgiHandler],
        pathinfo_factory: Callable[[str], CgiPathInfoT] = CgiPathInfo) -> None:

    response = CgiError500().reply
    try:
        request = CgiRequest.unpack(pathinfo_factory)
        handler = handler_factory(request)
        response = handler()

    except ParamsError:
        response = CgiError400().reply

    except NotFoundError:
        response = CgiError404().reply

    except BusyError:
        response = CgiError409().reply

    except CamServiceError:
        response = CgiCamError().reply

    except LicServiceError:
        response = CgiLicError().reply

    except ServiceError:
        response = CgiError503().reply

    except CgiError as x:
        response = x.reply

    except GenericError:
        pass

    except Exception:
        eventlog.log_x()

    finally:
        response.send()

    sys.exit(0)
