from gpsdclient import GPSDClient
import time
import os

from typing import Dict, Tuple, Union

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf import ph


def _geodata_rt_conf(new_jnode: Dict) -> Dict:
    az = new_jnode.get('azimuth', None)
    provider_jnode = new_jnode['providers'][0]

    if provider_jnode['provider'] == 'manual':
        jnode = {
            'azimuth': az,
            'source': ['static', provider_jnode['latitude'], provider_jnode['longitude']]
        }
    elif provider_jnode['provider'] == 'controller':
        jnode = {
            'azimuth': az,
            'source': ['gpsd', 'localhost', 2947]
        }
    else:
        jnode = dict()

    return jnode


def _test_uiconf_conf_changed(uiconf_conf: Dict, new_uiconf_conf: Dict) -> bool:
    if uiconf_conf.get('azimuth') != new_uiconf_conf.get('azimuth'):
        return True

    provider_info = uiconf_conf['providers'][0]
    new_provider_info = new_uiconf_conf['providers'][0]

    if provider_info['provider'] != new_provider_info['provider']:
        return True

    # old_mode == new_mode
    if provider_info['provider'] == 'manual':
        if provider_info['latitude'] != new_provider_info['latitude']:
            return True

        if provider_info['longitude'] != new_provider_info['longitude']:
            return True

    return False


def _test_gps_device_configured() -> bool:
    _GPSD_CONF_PATH = '/etc/default/gpsd'

    # gps_present = net.query_gps_count() > 0
    try:
        if os.path.isfile(_GPSD_CONF_PATH):
            with open(_GPSD_CONF_PATH, 'r') as gpsd_conf:
                gpsd_lines = gpsd_conf.readlines()

            for line in gpsd_lines:
                if line.startswith('DEVICES="/'):
                    return True

    except Exception as x:
        eventlog.log_e('fail: {}'.format(str(x)))

    return False


def _get_gps_device_lat_long2(
        host: str = 'localhost',
        port: str = '2947',
        tout_sec: float = 2.0) -> Union[Tuple[float, float], Tuple[None, None]]:

    call_ts = time.time()
    with GPSDClient(host=host, port=port, timeout=tout_sec) as client:
        try:
            for result in client.dict_stream(convert_datetime=False, filter=['DEVICES', 'TPV']):
                if (time.time() - call_ts) > tout_sec:
                    eventlog.log_d('fail: timeout')
                    break

                msg_class = result.get('class')
                if msg_class == 'DEVICES':
                    if len(result.get('devices', [])) == 0:
                        eventlog.log_w('fail: device-count:0')
                        return None, None

                elif msg_class == 'TPV':
                    lat_long = result.get('lat'), result.get('lon')
                    if None not in lat_long:
                        return lat_long

        except ToutError:
            pass

    return None, None


def _add_limits_stats(uiconf_conf: Dict) -> Dict:
    if _test_gps_device_configured():
        unit_ext_jnode = {
            'type': ph.UI_TYPE_GNSS,
            'conf': uiconf_conf,
            'limits': {
                'providers': ['manual', 'controller']
            }
        }

    else:
        unit_ext_jnode = {
            'type': ph.UI_TYPE_GNSS,
            'conf': uiconf_conf,
            'limits': {
                'providers': ['manual']
            }
        }

    uiconf_providers = uiconf_conf['providers']
    if len(uiconf_providers) != 1:
        raise NotImplementedError

    for provider in uiconf_providers:
        if provider['provider'] == 'manual':
            unit_ext_jnode['stats'] = {
                'provider': 'manual',
                'latitude': uiconf_providers[0]['latitude'],
                'longitude': uiconf_providers[0]['longitude']
            }

        elif provider['provider'] == 'controller':
            lat, lon = _get_gps_device_lat_long2()
            if None not in [lat, lon]:
                unit_ext_jnode['stats'] = {
                    'provider': 'controller',
                    'latitude': lat,
                    'longitude': lon
                }

    return unit_ext_jnode


# noinspection PyClassHasNoInit
class ApiGnss:
    @staticmethod
    def load_unit(uimod_ref: str) -> Dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_GNSS)
        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_ext_jnode = _add_limits_stats(uiconf['rw']['conf'])
        return unit_ext_jnode

    @staticmethod
    def save_unit(uimod_ref: str, unit_jnode: Dict) -> Dict:
        new_uiconf_conf = unit_jnode.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_GNSS)
        fs.validate_json(new_uiconf_conf, uimod.schema_path)

        uiconf = uimod.load_conf()
        uiconf_pub = uiconf['rw']
        if _test_uiconf_conf_changed(uiconf_pub['conf'], new_uiconf_conf):
            uiconf_pub['conf'] = new_uiconf_conf
            uiconf['rt'] = _geodata_rt_conf(new_uiconf_conf)
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        unit_ext_jnode = _add_limits_stats(new_uiconf_conf)
        return unit_ext_jnode
