import subprocess
import ujson

from typing import Dict, Optional

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf import ph


def get_stats_limits(host: str, port: int) -> Optional[Dict]:
    try:
        shell_params = f'env RPC_TARGET={host}:{port} /opt/edge/bin/ipcom-cli get-info'
        shell_result = subprocess.run(shell_params, capture_output=True, text=True, encoding='ascii',
                                      shell=True, check=True, timeout=10.0)

        return ujson.loads(shell_result.stdout)

    except subprocess.TimeoutExpired as x:
        eventlog.log_e(f'fail: <{x.cmd}> (timeout:{x.timeout})')

    except subprocess.CalledProcessError as x:
        eventlog.log_e(f'fail: <{x.cmd}> (returncode:{x.returncode})')


# noinspection PyClassHasNoInit
class IpcomWagaCameraApi:
    @staticmethod
    def load_unit(uimod_ref: str) -> Dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_IPCOM_WAGA)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        ext_uiconf_pub = uiconf['rw']
        server_jnode = ext_uiconf_pub['conf']['server']
        stats_jnode = get_stats_limits(server_jnode['host'], server_jnode['port'])

        reply = {
            'type': ph.UI_TYPE_IPCOM_WAGA,
            'conf': ext_uiconf_pub['conf']
            # "stats": {
            #     "uuid": "pidhorivka-uuid",
            #     "name": "WIM7",
            #     "vendor": "camea",
            #     "version": "0.0.2",
            #     "road": "M-22 Poltava-Oleksandria, km  74+810",
            #     "latitude": 49.1795,
            #     "longitude": 33.7992,
            #     "last-event": [
            #         1929233,
            #         "2021-11-08T15:40:00.075Z"
            #     ]
            # },
            # "limits": {
            #     "lanes": [
            #         {
            #             "name": "PI-KR-W1",
            #             "desc": "PP: Poltava - Oleksandria"
            #         },
            #         {
            #             "name": "PI-PL-W1",
            #             "desc": "ZP: Oleksandria - Polatava"
            #         }
            #     ]
            # }
        }
        if stats_jnode:
            reply.update(stats_jnode)

        return reply

    @staticmethod
    def save_unit(uimod_ref: str, new_uiconf: Dict) -> Dict:
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_IPCOM_WAGA)

        if uimod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        fs.validate_json(new_uiconf, uimod.schema_path)
        server_jnode = new_uiconf_conf['server']
        stats_jnode = get_stats_limits(server_jnode['host'], server_jnode['port'])

        uiconf = uimod.load_conf()
        if uiconf['rw']['conf'] != new_uiconf_conf:
            uiconf['rw']['conf'] = new_uiconf_conf
            if stats_jnode:
                uiconf['ro']['vendor'] = stats_jnode['stats']['vendor']

            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        reply = {
            'type': ph.UI_TYPE_IPCOM_WAGA,
            'conf': new_uiconf_conf,
            # "stats": {
            #     "uuid": "pidhorivka-uuid",
            #     "name": "WIM7",
            #     "vendor": "camea",
            #     "version": "0.0.2",
            #     "road": "M-22 Poltava-Oleksandria, km  74+810",
            #     "latitude": 49.1795,
            #     "longitude": 33.7992,
            #     "last-event": [
            #         1929233,
            #         "2021-11-08T15:40:00.075Z"
            #     ]
            # },
            # "limits": {
            #     "lanes": [
            #         {
            #             "name": "PI-KR-W1",
            #             "desc": "PP: Poltava - Oleksandria"
            #         },
            #         {
            #             "name": "PI-PL-W1",
            #             "desc": "ZP: Oleksandria - Polatava"
            #         }
            #     ]
            # }
        }

        if stats_jnode:
            reply.update(stats_jnode)

        return reply
