import math
import os
import subprocess
import ujson

from typing import Dict, List, Optional

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf import ph


def _get_live_image(ui_pid: str) -> Optional[str]:
    # get last overlay imagepath
    or_id = int(ui_pid) * 10
    or_store = f'/var/spool/edge/pipeline-{ui_pid}/or{or_id}-or{or_id}-store/'

    file_names: List[str] = os.listdir(or_store)

    if len(file_names):
        def decode_ts(file_name: str) -> int:
            return int(file_name.split('-')[-1])

        blob_name = max(file_names, key=decode_ts)
        blob_path = os.path.join(or_store, blob_name)
        return blob_path


def _landmark_cli(query_params: List[str], sensor_w_px: float, sensor_h_px: float,
                  focal_dist_px: float) -> Optional[Dict]:

    cam_fov_h: float = math.degrees(2 * math.atan2(sensor_w_px, 2 * focal_dist_px))
    cam_fov_v: float = math.degrees(2 * math.atan2(sensor_h_px, 2 * focal_dist_px))

    args: Dict = {
        '--fov-vertical': cam_fov_v,
        '--fov-horizontal': cam_fov_h,

        '--position-top': None,

        '--rotation-landmark-image': None,
        '--rotation-landmark-forward': None,
        '--rotation-landmark-left': None
    }

    for name_value in query_params:
        name, value = name_value.split('=', 1)

        if name == 'camera-up':
            args['--position-top'] = float(value)

        elif name == 'landmark-image':
            if value in ['bottom-left', 'bottom-right']:
                args['--rotation-landmark-image'] = value

        elif name == 'landmark-forward':
            args['--rotation-landmark-forward'] = float(value)

        elif name == 'landmark-left':
            args['--rotation-landmark-left'] = float(value)

    invalid_args = [arg_name_arg_value for arg_name_arg_value in iter(args.items()) if arg_name_arg_value[1] is None]
    if invalid_args:
        for invalid_arg_name, _ in invalid_args:
            eventlog.log_d(f'fail: param <{invalid_arg_name}> is absent')
        raise ParamsError

    else:
        cli_params = ['/opt/edge/bin/landmark_v2']
        for name, value in args.items():
            cli_params.append(f'{name}={value}')

        try:
            cli_result = subprocess.run(
                cli_params, capture_output=True, text=True, encoding='ascii',
                check=True, timeout=1.0)

            return ujson.loads(cli_result.stdout)

        except subprocess.TimeoutExpired as x:
            eventlog.log_e(f'fail: <{x.cmd}> (timeout:{x.timeout})')

        except subprocess.CalledProcessError as x:
            eventlog.log_e(f'fail: <{x.cmd}> (returncode:{x.returncode})')

        raise InternalError


# noinspection PyClassHasNoInit
class LocationApi:
    @staticmethod
    def load_unit(uimod_ref: str) -> Dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LOCATION)
        uiconf = uimod.load_conf(default=None)
        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_stats_jnode = cam_dgstats.get_stats(uimod.pid)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        uiconf_pub = uiconf['rw']
        reply = {
            'type': ph.UI_TYPE_LOCATION,
            'conf': uiconf_pub['args'],
            'stats': cam_stats_jnode
        }
        return reply

    @staticmethod
    def save_unit(uimod_ref: str, new_uiconf: Dict) -> Dict:
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LOCATION)
        fs.validate_json({'args': new_uiconf_conf}, uimod.schema_path)

        uiconf = uimod.load_conf()
        if uiconf['rw']['args'] != new_uiconf_conf:
            uiconf['rw']['args'] = new_uiconf_conf

            from edge.support import hg
            with hg.ConfRepo() as (hg_dir, hg_client):
                hgjpeg_path = fs.join_path(hg_dir, uimod.pid, '0.jpeg')
                hgconf_path = fs.join_path(hg_dir, uimod.pid, f'{ph.UI_TYPE_LOCATION}.0.json')

                add_files: bool = not fs.path_exists(hgconf_path)

                jpeg_path: Optional[str] = _get_live_image(uimod.pid)
                if jpeg_path is None:
                    eventlog.log_e(f'fail: get snapshot (term-ref:{uimod.pid})')
                    raise InternalError

                fs.silent_cp(jpeg_path, hgjpeg_path)
                fs.write_json_file(uiconf, hgconf_path, pretty=True)

                if add_files:
                    hg_client.add([hgconf_path.encode('ascii'), hgjpeg_path.encode('ascii')])

                hg_client.commit(message=f'edit({uimod.pid}/location.0)', user='admin')

                uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
                creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_stats_jnode = cam_dgstats.get_stats(uimod.pid)
        reply = {
            'type': ph.UI_TYPE_LOCATION,
            'conf': new_uiconf_conf,
            'stats': cam_stats_jnode
        }

        return reply

    @staticmethod
    def estimate_rotation(uimod_ref: str, query_params: List[str]) -> Dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LOCATION)

        whf = ph.CamUiStats.create(uimod.pid).get_sensorwpx_sensorhpx_focalpx()
        if whf is None:
            eventlog.log_e('fail: NotImplementedError')
            raise NotImplementedError

        landmark_jnode = _landmark_cli(query_params, *whf)

        landmark_rot = landmark_jnode['unit-rotation-auto']
        cam_rotation = {
            'camera': {
                'rotation': {
                    'elevation': landmark_rot['elevation'],
                    'azimuth': landmark_rot['azimuth']
                }
            }
        }

        return cam_rotation
