from typing import Dict, List, Tuple

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import ph
from edge.support.template import FormatTemplate

import edge.webserver.sysconf.edge_creator as creator


class SeatbeltApi:
    def __init__(self):
        pass

    @staticmethod
    def make_detector(term_pid: str, factory_conf: Dict) -> Tuple[str, Dict]:
        if not ph.RootInfoUiconf.validate_term(term_pid):
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        det_uimod_pid = term_pid
        det_uimod_type = factory_conf.get('type')
        det_name = factory_conf.get('name')

        if det_uimod_type != ph.UI_TYPE_SEATBELT or det_name is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        answer = ph.RootInfoUiconf.add_detector(det_uimod_pid, det_uimod_type, det_name)
        if answer is None:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        det_uimod_id, root_info = answer

        uimod = ph.UiMod(det_uimod_pid, det_uimod_type, det_uimod_id)
        uimod.install()
        uiconf = uimod.load_conf()
        uiconf['rw']['args']['name'] = det_name

        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})
        return uimod.ref, root_info

    @staticmethod
    def drop_detector(uimod_ref: str) -> Dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_SEATBELT)
        if uimod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        _, parent_info = ph.RootInfoUiconf.find_unit_parent_r(root_info, uimod_ref)
        parent_uimod = ph.UiMod.create(parent_info['ref'])
        parent_uiconf = parent_uimod.load_conf(default=True)

        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref, root_links)
        uirefs.remove(uimod.ref)

        root_info = ph.RootInfoUiconf.del_unit_r(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {parent_uimod: parent_uiconf})
        uimod.uninstall()

        ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        reply = {
            'info': [ext_root_info]
        }
        return reply

    @staticmethod
    def load_detector(uimod_ref: str) -> Dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_SEATBELT)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        uiconf_pub = uiconf['rw']
        reply = {
            'type': ph.UI_TYPE_SEATBELT,
            'conf': uiconf_pub['args']
        }
        reply.update(ph.get_detector_limits_stats(uimod.pid))
        return reply

    @staticmethod
    def _validate_format(jnode: List[Dict]) -> bool:
        key_to_true_map = {
            'speed': True,
            'datetime': True,
            'datetime-ms': True,
            'post-name': True,
            'geolocation': True,
            'plate': True,
            'lane': True,
            'vendor': True,
            'model': True,
            'serial': True
        }

        for row in jnode:
            for col in row['cols']:
                try:
                    FormatTemplate(template=col['format']).substitute(key_to_true_map)
                except KeyError as x:
                    eventlog.log_d(f'fail: invalid key ({str(x)})')
                    return False

        return True

    @staticmethod
    def save_detector(uimod_ref: str, new_uiconf: Dict) -> Dict:
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_SEATBELT)

        if uimod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        def validate_body() -> None:
            fs.validate_json({'args': new_uiconf_conf}, uimod.schema_path)

            # validate 'format' jnodes from json
            entry_images = new_uiconf_conf['images']['violation']

            entry_v_rows = entry_images['vehicle']['footer']['rows']
            if SeatbeltApi._validate_format(entry_v_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

            entry_s_rows = entry_images['scene']['footer']['rows']
            if SeatbeltApi._validate_format(entry_s_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

        validate_body()

        uiconf = uimod.load_conf()
        if uiconf['rw']['args'] != new_uiconf_conf:
            if new_uiconf_conf['name'] != uiconf['rw']['args']['name']:
                ph.RootInfoUiconf.set_unit_name(uimod.ref, new_uiconf_conf['name'])

            uiconf['rw']['args'] = new_uiconf_conf
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        reply = {
            'type': ph.UI_TYPE_SEATBELT,
            'conf': new_uiconf_conf
        }
        reply.update(ph.get_detector_limits_stats(uimod.pid))
        return reply
