from typing import Dict, List, Set, Union

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import InternalError


class UnityConfBuilder:
    def __init__(self):
        self._name_to_unit_map = dict()
        self._links = list()

    def has_units(self, unit_names: Union[List[str], Set[str]]) -> bool:
        for unit_name in unit_names:
            if unit_name not in self._name_to_unit_map:
                return False

        return True

    def unit(self, unit_name: str, unit_conf: Dict) -> None:
        if self.has_units({unit_name}):
            eventlog.log_e(f'fail: unit is present (name:{unit_name})')
            raise InternalError

        self._name_to_unit_map[unit_name] = unit_conf

    def link(self, source_unit: str, target_unit: str) -> None:
        """
        basic

        link('a', 'b')

        {
            "links":
            [
                ["a", "b"]
            ]
        }

        """
        if not self.has_units({source_unit, target_unit}):
            eventlog.log_e('fail')
            raise InternalError

        self._links.append((source_unit, target_unit))

    def link_seq(self, unit_list: List[str]) -> None:
        """
        sequential

        link_seq(['a', 'b', 'c', 'd'])

        {
            "links":
            [
                ["a", "b"],
                ["b", "c"],
                ["c", "d"]
            ]
        }
        """

        if not self.has_units(unit_list):
            eventlog.log_e('fail')
            raise InternalError

        for index in range(1, len(unit_list)):
            self.link(unit_list[index - 1], unit_list[index - 0])

    def link_dist(self,
                  source_units: Union[List[str], Set[str]],
                  target_units: Union[List[str], Set[str]]) -> None:
        """
        distributive

        link_dist(['s1', 's2', 's3'], ['t1', 't2'])

        {
            "links":
            [
                ["s1", "t1"],
                ["s2", "t1"],
                ["s3", "t1"],

                ["s1", "t2"],
                ["s2", "t2"],
                ["s3", "t2"]
            ]
        }

        """

        if not self.has_units(source_units):
            eventlog.log_e('fail')
            raise InternalError

        elif not self.has_units(target_units):
            eventlog.log_e('fail')
            raise InternalError

        for target_unit in target_units:
            for source_unit in source_units:
                self.link(source_unit, target_unit)

    def try_link_dist(self,
                      source_units: Union[List[str], Set[str]],
                      target_units: Union[List[str], Set[str]]) -> None:
        """
        distributive

        link_dist(['s1', 's2', 's3'], ['t1', 't2'])

        {
            "links":
            [
                ["s1", "t1"],
                ["s2", "t1"],
                ["s3", "t1"],

                ["s1", "t2"],
                ["s2", "t2"],
                ["s3", "t2"]
            ]
        }

        """

        known_units = list(self._name_to_unit_map.keys())
        source_units = [x for x in source_units if x in known_units]
        target_units = [x for x in target_units if x in known_units]

        for target_unit in target_units:
            for source_unit in source_units:
                self.link(source_unit, target_unit)

    def build(self, conf_dir: str) -> None:
        unity_conf_dir = fs.join_path(conf_dir, 'unit.conf.d')

        fs.silent_rm_r(unity_conf_dir)
        fs.silent_mkdir_r(unity_conf_dir)

        for conf_name, conf_data in self._name_to_unit_map.items():
            conf_path = '{}.conf'.format(fs.join_path(unity_conf_dir, conf_name))
            fs.write_json_file({'nodes': [conf_data]}, conf_path, pretty=True)

        conf_path = fs.join_path(unity_conf_dir, 'links.conf')
        fs.write_json_file({'links': self._links}, conf_path, pretty=True)
