from typing import Dict, List, Optional, Self, Set, Union

from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgconf.blobstore import BsDgconfEditor65535
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap


def bs65535_append_chan(es_uiconf: Dict) -> None:
    chanref = es_uiconf['ro']['channel']

    bs_dgconf = BsDgconfEditor65535.load()
    bs_dgconf.append_chan(chanref, es_uiconf)
    bs_dgconf.save()

    # sighup setupstore + restart blobstore
    dg.LifeCycle.sighup_ss_dgmod()


def bs65535_remove_chan(chanref: str) -> None:
    bs_dgconf = BsDgconfEditor65535.load()
    bs_dgconf.remove_chan(chanref)
    bs_dgconf.save()

    # sighup setupstore + restart blobstore
    dg.LifeCycle.sighup_ss_dgmod()


def bs65535_reappend_chan(es_uiconf: Dict) -> None:
    chanref = es_uiconf['ro']['channel']

    bs_dgconf = BsDgconfEditor65535.load()
    bs_dgconf.remove_chan(chanref)
    bs_dgconf.append_chan(chanref, es_uiconf)
    bs_dgconf.save()

    # restart blobstore
    dg.LifeCycle.sighup_dgmod(dg.EDGE_TYPE_BLOBSTORE, '65535')


# noinspection PyUnusedLocal
def bs_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_MEDIASTORE:
        return []

    try:
        ms_uimod = next(filter(lambda x: x.type == ph.UI_TYPE_MEDIASTORE, all_used_uimods))

        if forced:
            return [VideoBsCreator(ms_uimod)]

    except StopIteration:
        pass

    return []


class VideoBsCreator(BaseCreator):
    def __init__(self, ms_uimod: ph.UiMod):
        BaseCreator.__init__(self)
        self._ms_uimod = ms_uimod
        self._dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_BLOBSTORE, self._ms_uimod.pid, self._ms_uimod.id)
        self._main_conf = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self.dgmod.pid)
        cache_dir = fs.join_path(spool_dir, 'bs{0}-bs{0}-store'.format(self.dgmod.pid))
        archive_dir = fs.join_path(spool_dir, 'bs{0}-bs{0}-archive'.format(self.dgmod.pid))

        # create main.conf
        main_conf = {
            'monitoring': {
                'stat-path': fs.join_path(spool_dir, 'zabbix', 'bs0'),
                'stat-tout': 60
            },
            'storage':
            [
                {
                    'channel': '0',
                    'cache': {
                        'directory': cache_dir,
                        'limits': {
                            'max-abs-time': 2 * 60,
                            'max-du': 460
                        }
                    },
                    'archive': {
                        'directory': archive_dir,
                        'filesize': 4 * 1024,
                        'limits': {
                            'max-rel-time': 30,
                            'max-abs-time': 30,
                            'max-du': 4
                        }
                    }
                },
                {
                    'channel': '1',
                    'cache': {
                        'directory': cache_dir,
                        'limits': {
                            'max-abs-time': 2 * 60,
                            'max-du': 460
                        }
                    },
                    'archive': {
                        'directory': archive_dir,
                        'filesize': 4 * 1024,
                        'limits': {
                            'max-rel-time': 30,
                            'max-abs-time': 30,
                            'max-du': 120
                        }
                    }
                }
            ]
        }

        # create setupstore.conf
        self._main_conf = main_conf

    def complete(self) -> None:
        dgmod = self.dgmod
        dgmod.install()
        dgmod.save_main_conf(self._main_conf)
        dgmod.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.sighup_dgmod(self.dgmod.type, self.dgmod.id)


class VideoBsDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_MEDIASTORE:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_BLOBSTORE, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        spool_dir = ph.get_spool_dir(self._dgmod.pid)
        cache_dir = fs.join_path(spool_dir, 'bs{0}-bs{0}-store'.format(self._dgmod.pid))
        archive_dir = fs.join_path(spool_dir, 'bs{0}-bs{0}-archive'.format(self._dgmod.pid))

        fs.silent_rm_r(archive_dir)
        fs.silent_rm_r(cache_dir)
        self._dgmod.uninstall()
