import subprocess

from typing import Dict, Optional, Self

from edge import eventlog
from edge.support import fs, sh

EDGE_TYPE_SETUPSTORE = 'setupstore'
EDGE_TYPE_EVENTSTORE = 'eventstore'
EDGE_TYPE_LPR = 'lpr'
EDGE_TYPE_MEDIACODEC = 'mediacodec'
EDGE_TYPE_MEDIAGRABBER = 'mediagrabber'
EDGE_TYPE_GRABCONTROL = 'grabcontrol'
EDGE_TYPE_RADAR = 'radar'
EDGE_TYPE_RENDERER = 'overlayrender'
EDGE_TYPE_RTSP_GRABBER = 'rtspgrabber'
EDGE_TYPE_RTSP_SERVER = 'rtspserver'
EDGE_TYPE_SPEEDCONTROL = 'speedcontrol'
EDGE_TYPE_TRACKER = 'tracker'
EDGE_TYPE_WEBSERVER = 'webserver'
EDGE_TYPE_RAWCODEC = 'rawcodecgst'
EDGE_TYPE_BLOBSTORE = 'blobstore'
EDGE_TYPE_MEDIACODEC_GST = 'mediacodecgst'
EDGE_TYPE_TRAFFIC_LIGHT = 'traffic-light'
EDGE_TYPE_LBSAPP = 'lbsapp'
EDGE_TYPE_LEDDAR = 'leddar'


class DgMod:
    def __init__(self, dgmod_type: str, dgmod_id: str, uimod_pid: Optional[str] = None):
        self._uimod_pid = uimod_pid
        self.type = dgmod_type
        self.id = dgmod_id

    @classmethod
    def create(cls, dgmod_type: str, uimod_pid: str, uimod_id: str) -> Self:
        if dgmod_type in [EDGE_TYPE_WEBSERVER, EDGE_TYPE_SETUPSTORE]:
            dgmod_id = '0'

        elif dgmod_type == EDGE_TYPE_EVENTSTORE:
            dgmod_id = '65535'

        else:
            dgmod_id = uimod_id

        return cls(dgmod_type, dgmod_id, uimod_pid)

    def __hash__(self) -> int:
        return hash(self.type + self.id)

    def __str__(self) -> str:
        return f'<{self.type}|{self.id}>'

    def __eq__(self, other: Self) -> bool:
        return (isinstance(other, self.__class__) and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'id', None) == self.id)

    @property
    def pid(self) -> str:
        if self._uimod_pid is None:
            eventlog.log_e('fail: unsupported')
            raise NotImplementedError

        return self._uimod_pid

    @property
    def conf_dir(self) -> str:
        if self.type == EDGE_TYPE_SETUPSTORE:
            return f'/etc/edge/{self.type}'
        else:
            return f'/etc/edge/{self.type}.{self.id}'

    def get_conf_path(self, conf_name: str) -> str:
        return fs.join_path(self.conf_dir, conf_name)

    def load_json_conf(self, conf_name: str, **kwargs) -> Optional[Dict]:
        conf_path = self.get_conf_path(conf_name)
        return fs.load_json3(conf_path, **kwargs)

    def load_main_conf(self, **kwargs) -> Optional[Dict]:
        return self.load_json_conf('main.conf', **kwargs)

    def save_json_conf(self, conf_name: str, conf_jnode: Dict) -> None:
        conf_path = self.get_conf_path(conf_name)
        fs.write_json_file(conf_jnode, conf_path, pretty=True)

    def save_main_conf(self, conf_jnode: Dict) -> None:
        return self.save_json_conf('main.conf', conf_jnode)

    def save_misc_conf(self, conf_jnode: Dict) -> None:
        return self.save_json_conf('misc.conf', conf_jnode)

    def save_text_conf(self, conf_name: str, conf_text: str) -> None:
        conf_path = self.get_conf_path(conf_name)
        with open(conf_path, 'w') as text_file:
            text_file.write(conf_text)

    def save_haltaction_conf(self, conf_text: str) -> None:
        self.save_text_conf('haltaction.conf', conf_text)

    def save_setupstore_conf(self, active: bool) -> None:
        if active:
            conf_text = f'Active=true\nArgSet=--id={self.id} --log-level=info\n'
        else:
            conf_text = f'Active=false\nArgSet=--id={self.id} --log-level=info\n'

        self.save_text_conf('setupstore.conf', conf_text)

    def install(self) -> None:
        fs.silent_mkdir_r(self.conf_dir)

    def uninstall(self) -> None:
        fs.silent_rm_r(self.conf_dir)


CTRL_SS_DGMOD = DgMod(EDGE_TYPE_SETUPSTORE, dgmod_id='0')
CTRL_BS_DGMOD = DgMod(EDGE_TYPE_BLOBSTORE, dgmod_id='65535')
CTRL_ES_DGMOD = DgMod(EDGE_TYPE_EVENTSTORE, dgmod_id='65535')
CTRL_WS_DGMOD = DgMod(EDGE_TYPE_WEBSERVER, dgmod_id='0')


# noinspection PyClassHasNoInit
class LifeCycle:
    @staticmethod
    def restart_dgmod(dgmod_type: str, dgmod_id: Optional[str] = None) -> None:
        if dgmod_id is None:
            cli_params = ['pkill', '-9', dgmod_type]
            cli_result = subprocess.run(cli_params)
            eventlog.log_i(f'done: <{cli_result.args}> (exitcode:{cli_result.returncode})')

        else:
            eventlog.log_i(f'exec: {dgmod_type}.{dgmod_id}')
            sh.run_silent(
                "kill -9 `ps ax |"
                "grep \"{} --id.{}\" |"
                "grep -v grep |"
                "awk '{{print $1}}'`".format(dgmod_type, dgmod_id))

    @classmethod
    def restart_ws_dgmod(cls) -> None:
        eventlog.log_i('exec')
        sh.run_silent('/opt/edge/bin/setupstore-cli restart webserver')

    @staticmethod
    def sighup_dgmod(dgmod_type: str, dgmod_id: str) -> None:
        eventlog.log_i(f'exec: {dgmod_type}.{dgmod_id}')
        sh.run_silent(
            "kill -1 `ps ax |"
            "grep \"{} --id.{}\" |"
            "grep -v grep |"
            "awk '{{print $1}}'`".format(dgmod_type, dgmod_id))

    @staticmethod
    def sighup_dgmods(dgmod_type: str) -> None:
        cli_params = ['pkill', '-1', dgmod_type]
        cli_result = subprocess.run(cli_params)
        eventlog.log_i(f'done: <{cli_result.args}> (exitcode:{cli_result.returncode})')

    @staticmethod
    def sighup_ss_dgmod() -> None:
        eventlog.log_i('exec')
        sh.run_silent(f'pkill -1 {EDGE_TYPE_SETUPSTORE}')

    @staticmethod
    def sigint_dgmod(dgmod_type: str, dgmod_id: str) -> None:
        eventlog.log_i(f'exec: {dgmod_type}.{dgmod_id}')
        sh.run_silent(
            "kill -2 `ps ax |"
            "grep \"{} --id.{}\" |"
            "grep -v grep |"
            "awk '{{print $1}}'`".format(dgmod_type, dgmod_id))
