from typing import List, Optional, Self, Set, Union

from edge.support import net
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.uiconf.constroller import CtrlUiconf
from edge.webserver.sysconf.uiconf.db import DbUiconf


def _reconfigure_es(endpoints: List[str]) -> bool:
    for _ in [1, 2, 3]:
        try:
            net.query_es_config(endpoints, tout_sec=1)
            return True
        except net.NetError:
            pass

    return False


def _mk_react_path(rule_uimod: ph.UiMod) -> str:
    return f'/var/spool/edge/system/sc{rule_uimod.pid}-es65535-fact'


# noinspection PyUnusedLocal
def es_get_creators(
        entry_uimod: Union[ph.UiMod, ph.UiMod2],
        all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
        uimod_to_newconf_map: UiModToConfMap,
        forced: bool = False) -> List[BaseCreator]:

    if not forced:
        return []

    root_info = ph.RootInfoUiconf.get_info()
    tr_uirefs = ph.RootInfoUiconf.find_unit_refs_r(root_info, ph.UI_TYPE_TRAFFIC_RULES)
    rules_uimods = list(map(ph.UiMod.create, tr_uirefs))
    return [EsCreator(rules_uimods)]


class EsCreator(BaseCreator):
    def __init__(self, rules_uimods: List[ph.UiMod]):
        BaseCreator.__init__(self)
        self._rules_uimods = rules_uimods
        self._main_conf = None

    @property
    def dgmod(self) -> dg.DgMod:
        return dg.CTRL_ES_DGMOD

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        es_new_conf = {
            'api': net.get_es_uds_endpoint(),
            'db': DbUiconf.CONF_PATH,
            'monitoring': {
                'stat-path': '/var/spool/edge/system/zabbix/es65535',
                'stat-tout': 60
            },
            'fact-react': list(map(_mk_react_path, self._rules_uimods)),
            'systest': CtrlUiconf.get_systest()
        }

        self._main_conf = es_new_conf

    def complete(self) -> None:
        es_dgmod = dg.CTRL_ES_DGMOD
        es_dgmod.install()
        es_dgmod.save_main_conf(self._main_conf)
        es_dgmod.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        if not _reconfigure_es(self._main_conf['fact-react']):
            dg.LifeCycle.restart_dgmod(dg.EDGE_TYPE_EVENTSTORE)


class EsDeleter:
    def __init__(self, entry_uimod: ph.UiMod):
        self._entry_uimod = entry_uimod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_FACTORY:
            return cls(entry_uimod)

    def run(self) -> None:
        dgmod = dg.CTRL_ES_DGMOD
        main_conf = dgmod.load_main_conf()

        endpoint = _mk_react_path(self._entry_uimod)
        main_conf['fact-react'] = [x for x in main_conf['fact-react'] if x != endpoint]

        dgmod.save_main_conf(main_conf)

        if not _reconfigure_es(main_conf['fact-react']):
            dg.LifeCycle.restart_dgmod(dg.EDGE_TYPE_EVENTSTORE)
