from typing import List, Optional, Self, Set, Union

from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.dgmod.gc.gc_main import MainConfBuilder


# noinspection PyUnusedLocal
def gc_get_creators(uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if uimod.type != ph.UI_TYPE_MV_CAMERA:
        return []

    try:
        cam_uimod = next(filter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods))
        lpr_uimod = next(filter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod)
        ])
        if forced or activate:
            return [GcCreator(cam_uimod, lpr_uimod)]

    except StopIteration:
        pass

    return []


class GcCreator(BaseCreator):
    def __init__(self, cam_uimod: ph.UiMod, lpr_uimod: ph.UiMod):
        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._gc_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_GRABCONTROL, cam_uimod.pid, cam_uimod.id)
        self._main_active = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._gc_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        ui_pid = self._cam_uimod.pid
        spool_dir = ph.get_spool_dir(ui_pid)

        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        cam_uiconf_control = cam_uiconf['rw']['videostream']['control']

        cam_uiconf_pri = cam_uiconf['ro']
        #
        exp_time_conf = [0, 0, 0]
        gain_conf = [0, 0, 0]
        gain_step = 1.0
        exptime_step = 20
        if cam_uiconf_control['profile'] == 'auto-70-150':
            gain_conf = [cam_uiconf_control['gain-auto-min'], cam_uiconf_control['gain-auto-max'], gain_step]
            exp_time_conf = [
                cam_uiconf_control['exposure-auto-min'],
                cam_uiconf_control['exposure-auto-max'],
                exptime_step]
        #

        builder = MainConfBuilder()
        builder.setup_ir_control(cam_uiconf_pri.get('control-ir', True))
        builder.setup_gain_conf(gain_conf)
        builder.setup_exp_time_conf(exp_time_conf)
        builder.setup_add_scene_img_react(spool_dir, f'mg{self._cam_uimod.id}-gc{self._cam_uimod.id}-scene')
        builder.setup_add_scene_aoi_react(spool_dir, f'lr{self._lpr_uimod.id}-gc{self._cam_uimod.id}-scene')
        builder.setup_add_scene_control_event(f'/etc/edge/mediagrabber.{self._gc_dgmod.id}/dyn/')

        main_conf = builder.build()

        ss_active = cam_uiconf_control['profile'] == 'auto-70-150'
        self._main_active = main_conf, ss_active

    def complete(self) -> None:
        main_conf, ss_active = self._main_active

        edge_gc = self._gc_dgmod
        edge_gc.install()
        edge_gc.save_main_conf(main_conf)
        edge_gc.save_setupstore_conf(ss_active)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class GcDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_MV_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_GRABCONTROL, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
