from typing import List, Optional, Self, Set, Union

from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgconf.unity.bundle import (anpr, bus, essentials, speedcam, waga, mmr,
                                                        recipient, leddar, wiegand, sbr)
from edge.webserver.sysconf.dgconf.unity.builder import UnityConfBuilder
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap


# noinspection PyUnusedLocal
def lr_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_LPR:
        return []

    try:
        cam_uimod = next(filter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods))
        roadway_uimod = next(filter(lambda x: x.type == ph.UI_TYPE_ROADWAY, all_used_uimods))
        lpr_uimod = next(filter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods))
        fusion_uimod = next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods))
        tcp_recipient_uimods = [x for x in all_used_uimods if x.type == ph.UI_TYPE_TCP_RECIPIENT]
        ledar_uimods = [x for x in all_used_uimods if x.type == ph.UI_TYPE_LEDDAR]
        location_uimods = [x for x in all_used_uimods if x.type == ph.UI_TYPE_LOCATION]
        ipcom_waga_uimods = [x for x in all_used_uimods if x.type == ph.UI_TYPE_IPCOM_WAGA]

        activate = any([
            uimod_to_newconf_map.get(lpr_uimod),
            uimod_to_newconf_map.get(cam_uimod),
            uimod_to_newconf_map.get(roadway_uimod),

            any([uimod_to_newconf_map.get(x) for x in tcp_recipient_uimods]),
            any([uimod_to_newconf_map.get(x) for x in ledar_uimods]),
            any([uimod_to_newconf_map.get(x) for x in location_uimods]),
            any([uimod_to_newconf_map.get(x) for x in ipcom_waga_uimods])
        ])

        if forced or activate:
            return [LrCreator(lpr_uimod, cam_uimod, fusion_uimod, roadway_uimod, tcp_recipient_uimods, ledar_uimods,
                              location_uimods, ipcom_waga_uimods)]

    except StopIteration:
        pass

    return []


UNIT__ESSENTIALS0__API0 = 'essentials.0~api.0'
UNIT__ESSENTIALS0__COUNTERS0 = 'essentials.0~counters.0'
UNIT__MEDIA_SERVER0__MEDIA0 = 'media-server.0~media.0'
UNIT__EVENT_EMITTER0__TX_TRACK_BUS0 = 'event-emitter.0~tx-track-bus.0'
UNIT__EVENT_EMITTER0__TX_SCENE_BUS0 = 'event-emitter.0~tx-scene-bus.0'
UNIT__ANPR0__ANPR0 = 'anpr.0~anpr.0'
UNIT__MMR0__MMR0 = 'mmr.0~mmr.0'
UNIT__SBR0__SBR0 = 'sbr.0~sbr.0'
UNIT__SPEEDCAM0__SPEEDCAM0 = 'speedcam.0~bracket.0'
UNIT__IPCOM0_WAGA0 = 'ipcom.0~waga.0'
UNIT__LEDDAR0__LEDDAR0 = 'leddar.0~leddar.0'
UNIT__WIEGAND_PCB0__RECIPIENT0 = 'wiegand-pcb.0~recipient.0'
UNIT__WIEGAND_UDP0__RECIPIENT0 = 'wiegand-udp.0~recipient.0'
UNIT__RULE_PCB0__RULE_PCB0 = 'rule.pcb0~rule.pcb0'
UNIT__RULE_UDP0__RULE_UDP0 = 'rule.udp0~rule.udp0'
UNIT__RULE_TCPx__RULE_TCPx = 'rule.tcp{}~rule.tcp{}'


class LrCreator(BaseCreator):
    def __init__(self, lpr_uimod: ph.UiMod,
                 cam_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 roadway_uimod: ph.UiMod,
                 tcp_recipient_uimods: List[ph.UiMod],
                 ledar_uimods: List[ph.UiMod],
                 location_uimods: List[ph.UiMod],
                 ipcom_waga_uimods: List[ph.UiMod]):

        BaseCreator.__init__(self)
        self._lpr_uimod = lpr_uimod
        self._cam_uimod = cam_uimod
        self._fusion_uimod = fusion_uimod
        self._roadway_uimod = roadway_uimod
        self._tcp_recipient_uimods = tcp_recipient_uimods
        self._ledar_uimods = ledar_uimods
        self._location_uimods = location_uimods
        self._ipcom_waga_uimods = ipcom_waga_uimods

        self._lpr_dgmod = dg.DgMod.create(dg.EDGE_TYPE_LPR, lpr_uimod.pid, lpr_uimod.id)
        self._builder = UnityConfBuilder()

    @property
    def dgmod(self) -> dg.DgMod:
        return self._lpr_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        spool_path = ph.get_spool_dir(self._lpr_uimod.pid)

        lpr_uiconf = uimod_to_newconf_map.get(self._lpr_uimod) or self._lpr_uimod.load_conf()
        lpr_pri = lpr_uiconf['ro']
        lpr_pub = lpr_uiconf['rw']
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        cam_pri = cam_uiconf['ro']

        if self._cam_uimod.type == ph.UI_TYPE_MV_CAMERA:
            if cam_pri['raw-format'] == 'MONO8T':
                add_scene_react_dir = fs.join_path(
                    spool_path, f'mg{self._cam_uimod.id}-lr{self._lpr_uimod.id}-scene')
            else:
                add_scene_react_dir = fs.join_path(
                    spool_path, f'rc{self._cam_uimod.id}-lr{self._lpr_uimod.id}-scene')

        elif self._cam_uimod.type == ph.UI_TYPE_RTSP_CAMERA:
            add_scene_react_dir = fs.join_path(
                spool_path, f'rg{self._cam_uimod.id}-lr{self._lpr_uimod.id}-scene')

        else:
            raise NotImplementedError

        image_wh = ph.CamDgStats.create(self._lpr_uimod.pid).get_stream_wh() or (640, 480)
        scene_event_jnode = [
            fs.join_path(spool_path, f'lr{self._lpr_uimod.id}-tr{self._fusion_uimod.id}-scene')
        ]

        # enable gc
        if self._cam_uimod.type == ph.UI_TYPE_MV_CAMERA:
            if cam_uiconf['rw']['videostream']['control']['profile'] == 'auto-70-150':
                endpoint = fs.join_path(spool_path, f'lr{self._lpr_uimod.id}-gc{self._cam_uimod.id}-scene')
                scene_event_jnode.append(endpoint)

        self._builder.unit(*essentials.create_unit__essentials__api(
            UNIT__ESSENTIALS0__API0, self._lpr_uimod.pid))
        self._builder.unit(*essentials.create_unit__essentials__counters(
            UNIT__ESSENTIALS0__COUNTERS0, spool_path))
        self._builder.unit(*bus.create_unit__media_server__media(
            UNIT__MEDIA_SERVER0__MEDIA0, add_scene_react_dir))
        self._builder.unit(*bus.create_unit__event_emitter__tx_track_bus(
            UNIT__EVENT_EMITTER0__TX_TRACK_BUS0, lpr_pri.get('translate-plate') or True,
            fs.join_path(spool_path, f'lr{self._lpr_uimod.id}-lr{self._lpr_uimod.id}-store'),
            fs.join_path(spool_path, f'lr{self._lpr_uimod.id}-tr{self._fusion_uimod.id}-track')))
        self._builder.unit(*bus.create_unit__event_emitter__tx_scene_bus(
            UNIT__EVENT_EMITTER0__TX_SCENE_BUS0, scene_event_jnode))
        self._builder.unit(*anpr.create_unit__anpr__anpr(
            UNIT__ANPR0__ANPR0, lpr_pri, lpr_pub, image_wh))

        self._builder.link_dist({UNIT__MEDIA_SERVER0__MEDIA0, UNIT__ESSENTIALS0__API0}, {UNIT__ANPR0__ANPR0})
        self._builder.link_dist({UNIT__ANPR0__ANPR0}, {UNIT__ESSENTIALS0__API0, UNIT__EVENT_EMITTER0__TX_SCENE_BUS0})
        self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, UNIT__EVENT_EMITTER0__TX_SCENE_BUS0)

        units_seq = [UNIT__ANPR0__ANPR0]
        mmr_conf = lpr_pri.get(UNIT__MMR0__MMR0, None)
        if mmr_conf is not None:
            self._builder.unit(*mmr.create_unit__mmr___mmr(UNIT__MMR0__MMR0, mmr_conf))
            units_seq.append(UNIT__MMR0__MMR0)

        sbr_conf = lpr_pri.get(UNIT__SBR0__SBR0, None)
        if sbr_conf is not None:
            self._builder.unit(*sbr.create_unit__sbr___sbr(UNIT__SBR0__SBR0, sbr_conf))
            units_seq.append(UNIT__SBR0__SBR0)

        location_uimod = self._location_uimods[0] if len(self._location_uimods) > 0 else None

        if location_uimod:
            location_uiconf = uimod_to_newconf_map.get(location_uimod) or location_uimod.load_conf()
            speedcam_jnode = location_uiconf.get('ro', dict()).get('speedcam')
            if speedcam_jnode is not None:
                add_speedcam = speedcam_jnode.get('with-cam-n', False)
                if add_speedcam:
                    self._builder.unit(*speedcam.create_unit__speedcam___speedcam(
                        UNIT__SPEEDCAM0__SPEEDCAM0, location_uiconf, self._cam_uimod.conf_path, location_uimod.conf_path))
                    units_seq.append(UNIT__SPEEDCAM0__SPEEDCAM0)

        if len(self._ipcom_waga_uimods) == 1:
            waga_uimod = self._ipcom_waga_uimods[0]
            waga_conf = uimod_to_newconf_map.get(waga_uimod) or waga_uimod.load_conf()
            self._builder.unit(*waga.create_unit__ipcom___waga(UNIT__IPCOM0_WAGA0, waga_conf))
            units_seq.append(UNIT__IPCOM0_WAGA0)

        units_seq.append(UNIT__EVENT_EMITTER0__TX_TRACK_BUS0)
        self._builder.link_seq(units_seq)
        self._builder.link_dist([UNIT__ESSENTIALS0__COUNTERS0], units_seq)

        # create wiegand unit
        wiegand_conf = lpr_pri.get('wiegand-conf', None)
        if wiegand_conf:
            self._builder.unit(*wiegand.create_unit__wiegand___pcb(
                UNIT__WIEGAND_PCB0__RECIPIENT0, wiegand_conf))
            self._builder.unit(*recipient.create_unit__rule___rule2(
                UNIT__RULE_PCB0__RULE_PCB0, wiegand_conf['trigger']))

            self._builder.link_seq([UNIT__ANPR0__ANPR0, UNIT__RULE_PCB0__RULE_PCB0, UNIT__WIEGAND_PCB0__RECIPIENT0])
            self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, UNIT__WIEGAND_PCB0__RECIPIENT0)

        # create wiegand-udp unit
        wiegand_udp_conf = lpr_pri.get('wiegand-udp-conf', None)
        if wiegand_udp_conf:
            self._builder.unit(*wiegand.create_unit__wiegand___udp(UNIT__WIEGAND_UDP0__RECIPIENT0, wiegand_udp_conf))

            self._builder.unit(*recipient.create_unit__rule___rule2(
                UNIT__RULE_UDP0__RULE_UDP0, wiegand_udp_conf['trigger']))

            self._builder.link_seq([UNIT__ANPR0__ANPR0, UNIT__RULE_UDP0__RULE_UDP0, UNIT__WIEGAND_UDP0__RECIPIENT0])
            self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, UNIT__WIEGAND_UDP0__RECIPIENT0)

        dir_a = dir_r = dir_u = term_uuid = None
        if len(self._tcp_recipient_uimods):
            roadway_uiconf = uimod_to_newconf_map.get(self._roadway_uimod) or self._roadway_uimod.load_conf()
            ui_road_jnode = roadway_uiconf['rw']['args']
            ui_dir_jnode = ui_road_jnode.get('dir') or dict()
            dir_a = ui_dir_jnode.get('approaching')
            dir_r = ui_dir_jnode.get('receding')
            dir_u = ui_dir_jnode.get('unknown')
            term_uimod = ph.UiMod.create_term(self._lpr_uimod.pid)
            term_uuid = ph.TerminalUiconf(term_uimod.conf_path).get_uuid()

        tcp_index = 0
        for tcp_recipient_uimod in self._tcp_recipient_uimods:
            tcp_index += 1
            tcp_recipient_conf = uimod_to_newconf_map.get(tcp_recipient_uimod) or tcp_recipient_uimod.load_conf()

            self._builder.unit(*recipient.create_unit__recipient___vast2(tcp_recipient_uimod.ref, tcp_recipient_conf,
                                                                         dir_a, dir_r, dir_u, term_uuid))

            rule_ref = UNIT__RULE_TCPx__RULE_TCPx.format(tcp_index, tcp_index)
            self._builder.unit(*recipient.create_unit__rule___rule(rule_ref, tcp_recipient_conf))

            self._builder.link_seq([UNIT__ANPR0__ANPR0, rule_ref, tcp_recipient_uimod.ref])
            self._builder.link(tcp_recipient_uimod.ref, UNIT__ESSENTIALS0__COUNTERS0)

        if len(self._ledar_uimods) == 1:
            leddar_uimod = self._ledar_uimods[0]
            leddar_conf = uimod_to_newconf_map.get(leddar_uimod) or leddar_uimod.load_conf()
            leddar_endpoint = fs.join_path(spool_path, 'lg{0}-sc{0}-track'.format(self._lpr_uimod.id))

            self._builder.unit(*leddar.create_unit__leddar___leddar(
                UNIT__LEDDAR0__LEDDAR0, leddar_conf, leddar_endpoint))

            self._builder.link(UNIT__MEDIA_SERVER0__MEDIA0, UNIT__LEDDAR0__LEDDAR0)
            self._builder.link(UNIT__LEDDAR0__LEDDAR0, UNIT__ESSENTIALS0__COUNTERS0)

    def complete(self) -> None:
        lr_dgmod = self._lpr_dgmod
        self._builder.build(lr_dgmod.conf_dir)
        lr_dgmod.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class LrDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_LPR:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_LPR, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
