from typing import Dict, List, Optional, Self, Set, Tuple, Union

from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.dgmod.mc.mc_main import MainConfBuilder


# noinspection PyUnusedLocal
def mc_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type not in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA]:
        return []

    try:
        cam_uimod = next(filter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod)
        ])

        if forced or activate:
            return [McCreator(cam_uimod)]

    except StopIteration:
        pass

    return []


class McCreator(BaseCreator):
    def __init__(self, cam_uimod: ph.UiMod):
        super().__init__()

        self._cam_uimod = cam_uimod
        self._mc_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_MEDIACODEC_GST, cam_uimod.pid, cam_uimod.id)
        self._main_ss: Optional[Tuple[Dict, bool]] = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._mc_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        ui_pid = self._cam_uimod.pid
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        spool_dir = ph.get_spool_dir(ui_pid)

        quality = 30000
        if cam_uiconf:
            quality = cam_uiconf['ro'].get('h264Quality', 30000)

        # create main.conf
        builder = MainConfBuilder()
        builder.setup_blobstore(self._mc_dgmod.id, '0')
        builder.setup_react(spool_dir, 'mg{0}-mcgst{0}-scene'.format(self._cam_uimod.id))
        builder.setup_codec(quality)
        main_conf = builder.build()

        rgbs = cam_uiconf['ro'].get('rgbs', False)
        mcgst_scene = cam_uiconf['ro'].get('mcgst-scene', True)
        ss_active = mcgst_scene and not rgbs

        self._main_ss = main_conf, ss_active

    def complete(self) -> None:
        main_conf, ss_active = self._main_ss

        edge_mc = self._mc_dgmod
        edge_mc.install()
        edge_mc.save_main_conf(main_conf)
        edge_mc.save_setupstore_conf(ss_active)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class McDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA]:
            dgmod = dg.DgMod.create(
                dg.EDGE_TYPE_MEDIACODEC_GST, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
