from typing import Dict, Final, List, Optional

from edge import eventlog
from edge.support import fs


class MainConfBuilder:
    def __init__(self, uimod_id: str, spool_dir: str) -> None:
        self.uimod_id: Final[str] = uimod_id
        self._spool_dir: Final[str] = spool_dir
        self._raw_format_mono8: Optional[bool] = None

        self._jnode = {
            'FileApi': {
                'SceneEvent': {
                    'Dirs': list()
                },
                'HaltEvent': {
                    'SignalAbsentTout': 3000,
                    'DeviceAbsentTout': 60000
                }
            },
            'TxPipes': list(),
            'Driver': {
                'ParamSet': {
                    'VideoSource': dict(),
                    'IoBox': list()
                }
            }
        }
        self.setup_service_stats()

    def setup_strobe_event(self, fifo_name: str) -> None:
        self._jnode['TxPipes'].append(fs.join_path(self._spool_dir, fifo_name))

    def setup_scene_event_blobs(
            self, blob_dir: str,
            min_file_count: int,
            max_file_count: int,
            blob_format: str) -> None:

        self._jnode['FileApi'].update({
            'Store': fs.join_path(self._spool_dir, blob_dir),
            'FileCountLow': min_file_count,
            'FileCountHigh': max_file_count,
        })

        self._raw_format_mono8 = blob_format == 'MONO8T'
        self._jnode['FileApi']['SceneEvent'].update({
            'PixelFormat': blob_format
        })

    def setup_scene_event_args(self, args_dirs: List[str], require_mono8: bool = False) -> None:
        file_api_jnode: Dict = self._jnode['FileApi']
        scene_event_jnode: Dict = file_api_jnode['SceneEvent']
        dirs_jnode: List[str] = scene_event_jnode['Dirs']

        if not require_mono8:
            for args_dir in args_dirs:
                dirs_jnode.append(fs.join_path(self._spool_dir, args_dir))

        elif self._raw_format_mono8 is None:
            eventlog.log_e('fail')
            raise NotImplementedError()

        elif self._raw_format_mono8:
            for args_dir in args_dirs:
                dirs_jnode.append(fs.join_path(self._spool_dir, args_dir))

    def setup_service_stats(self) -> None:
        self._jnode.update({
            'Monitoring': {
                'StatPath': fs.join_path(self._spool_dir, 'zabbix', 'vs0'),
                'StatTout': 5
            }
        })

    def setup_cam_model(self, vendor: str, model: str, serial: str) -> None:
        if vendor == 'Basler':
            driver_file = '/opt/edge/libexec/mediagrabber/libmgbasler.so'

        elif vendor == 'Teledyne DALSA':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-dalsa.so'

        elif vendor == 'PointGrey':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-pointgrey.so'

        elif vendor == 'NPK ES-Experts':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-ese.so'

        elif vendor == 'JAI Corporation':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-jai.so'

        else:
            eventlog.log_e('fail')
            raise NotImplementedError()

        self._jnode['Driver'].update({
            'Vendor': vendor,
            'Model': model,
            'Id': serial,
            'File': driver_file
        })

    def setup_cam_ir(self, signal: str, pin: str, inverter: bool) -> None:
        self._jnode['Driver']['ParamSet']['IoBox'].append({
            'Source': signal,
            'Mode': 'Output',
            'Pin': pin,
            'Inverter': inverter
        })

    def setup_cam_sweep_auto(self, exp_time_limits: List[int], gain_limits: List[int]):
        self._jnode['Driver']['ParamSet']['VideoSource']['AutoSetup'] = {
            'ExpTimeLimits': exp_time_limits,
            'Target': {
                'Brightness': 0.50196
            },
            'GainLimits': gain_limits
        }

    def setup_cam_sweep_manual(self, exp_time: int, gain: int):
        self._jnode['Driver']['ParamSet']['VideoSource']['ManSetup'] = {
            'Lut12': '/opt/edge/share/setupstore/edge/mediagrabber/lut12.default.conf',
            'GainAll': gain,
            'ExpTime': exp_time
        }

    def setup_cam_sweep(self, mg_control_mode: str):
        self._jnode['Driver']['ParamSet']['VideoSource']['Sweep'] = {
            'Profile': 'MinGain',
            'ExpTimeMode': mg_control_mode,
            'GainMode': mg_control_mode,
            'LutMode': 'None'
        }

    def setup_cam_misc(self, data_size: int):
        self._jnode['Driver']['ParamSet']['Misc'] = {
            'TransferDataSize': data_size
        }

    def setup_cam_auqisition(self, w: int, h: int, x: int, y: int, fps: int, rotate180: bool):
        self._jnode['Driver']['ParamSet']['VideoSource']['Acquisition'] = {
            'Sensor': {
                'Width': w,
                'OffsetX': x,
                'OffsetY': y,
                'Height': h
            },
            'Gamma': -1,
            'Fps': fps,
            'Rotate180': rotate180
        }

    def build(self) -> Dict:
        return self._jnode
