from typing import List, Optional, Self, Set, Union

from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.dgmod.mg.mg_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.mg.mg_halt import HaltConfBuilder
from edge.webserver.sysconf.dgmod.mg.mg_misc import MiscConfBuilder
from edge.webserver.sysconf.dgmod.mg.uiconf import IUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.fusion import FusionUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.lpr import LprUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.mv_camera import MvCameraUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.radar import RadarUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.redlight import RedlightUiConfProvider, RailroadUiConfProvider


# noinspection PyUnusedLocal
def mg_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_MV_CAMERA:
        return []

    try:
        cam_uimod, fusion_uimod, lpr_uimod, ra_uimod, rl_uimods, rc_uimods = (
            next(filter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods), None),
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RED_LIGHT],
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RAILROAD_CROSSING])

        activate = any([
            uimod_to_newconf_map.get(cam_uimod),
            any([uimod_to_newconf_map.get(x) for x in rl_uimods]),
            any([uimod_to_newconf_map.get(x) for x in rc_uimods])
        ])
        return [MgCreator(cam_uimod, lpr_uimod, fusion_uimod, rl_uimods, rc_uimods, ra_uimod)]

    except StopIteration:
        pass

    return []


class MgCreator(BaseCreator):
    def __init__(self, cam_uimod: ph.UiMod,
                 lpr_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 rl_uimods: List[ph.UiMod],
                 rc_uimods: List[ph.UiMod],
                 ra_uimod: Optional[ph.UiMod]):

        super().__init__()

        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._fusion_uimod = fusion_uimod
        self._rl_uimods = rl_uimods
        self._rc_uimods = rc_uimods
        self._ra_uimod = ra_uimod
        self._edge_mg = dg.DgMod.create(
            dg.EDGE_TYPE_MEDIAGRABBER, cam_uimod.pid, cam_uimod.id)
        self._main_misc_halt = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._edge_mg

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        main_builder = MainConfBuilder(self._cam_uimod.id, ph.get_spool_dir(self._cam_uimod.pid))
        misc_builder = MiscConfBuilder()
        halt_builder = HaltConfBuilder()

        providers = list[IUiConfProvider]()

        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        providers.extend([
            MvCameraUiConfProvider(self._cam_uimod.id, cam_uiconf),
            FusionUiConfProvider(self._fusion_uimod.id),
            LprUiConfProvider(self._lpr_uimod.id)
        ])

        if self._ra_uimod:
            providers.extend([
                RadarUiConfProvider(self._ra_uimod.id)
            ])

        if len(self._rl_uimods) == 1:
            rl_uimod = self._rl_uimods[0]
            rl_uiconf = uimod_to_newconf_map.get(rl_uimod) or rl_uimod.load_conf()
            providers.extend([
                RedlightUiConfProvider(rl_uimod.id, rl_uiconf)
            ])

        elif len(self._rc_uimods) == 1:
            rl_uimod = self._rc_uimods[0]
            rl_uiconf = uimod_to_newconf_map.get(rl_uimod) or rl_uimod.load_conf()
            providers.extend([
                RailroadUiConfProvider(rl_uimod.id, rl_uiconf)
            ])

        for provider in providers:
            provider.setup_builders(main_builder, misc_builder, halt_builder)

        main_conf = main_builder.build()
        misc_conf = misc_builder.build()
        halt_conf = halt_builder.build()

        self._main_misc_halt = main_conf, misc_conf, halt_conf

    def complete(self) -> None:
        main_conf, misc_conf, halt_conf = self._main_misc_halt

        edge_mg = self._edge_mg
        edge_mg.install()
        edge_mg.save_main_conf(main_conf)
        edge_mg.save_misc_conf(misc_conf)
        edge_mg.save_haltaction_conf(halt_conf)
        edge_mg.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.sigint_dgmod(self.dgmod.type, self.dgmod.id)


class MgDeleter:
    def __init__(self, dgmod: dg.DgMod):

        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_MV_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_MEDIAGRABBER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
