from typing import List, Optional, Self, Set, Union

from edge import eventlog
from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.dgmod.ra.ra_halt import HaltConfBuilder
from edge.webserver.sysconf.dgmod.ra.ra_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.ra.uiconf import IUiConfProvider
from edge.webserver.sysconf.dgmod.ra.uiconf.fusion import FusionUiConfProvider
from edge.webserver.sysconf.dgmod.ra.uiconf.location import LocationUiConfProvider
from edge.webserver.sysconf.dgmod.ra.uiconf.mv_camera import MvCameraUiConfPorvider
from edge.webserver.sysconf.dgmod.ra.uiconf.rtsp_camera import RtspCameraUiConfPorvider


# noinspection PyUnusedLocal
def ra_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_RADAR:
        return []

    try:
        ra_uimod, cam_uimod, fusion_uimod, location_uimod = (
            next(filter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods)),
            next(filter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LOCATION, all_used_uimods)))

        activate = any([
            uimod_to_newconf_map.get(location_uimod),
            uimod_to_newconf_map.get(cam_uimod)
        ])

        if forced or activate:
            return [RaCreator(entry_uimod, ra_uimod, cam_uimod, fusion_uimod, location_uimod)]

    except StopIteration:
        pass

    return []


class RaCreator(BaseCreator):
    def __init__(self, entry_uimod: ph.UiMod,
                 ra_uimod: ph.UiMod,
                 cam_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 location_uimod: ph.UiMod):

        super().__init__()

        self._entry_uimod = entry_uimod
        self._ra_uimod = ra_uimod
        self._cam_uimod = cam_uimod
        self._fusion_uimod = fusion_uimod
        self._location_uimod = location_uimod
        self._edge_ra = dg.DgMod.create(dg.EDGE_TYPE_RADAR, ra_uimod.pid, ra_uimod.id)
        self._main_halt = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._edge_ra

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._ra_uimod.pid)

        api_builder = None
        main_builder = MainConfBuilder(self._ra_uimod.id, spool_dir)
        halt_builder = HaltConfBuilder()

        providers = list[IUiConfProvider]()

        loc_uiconf = uimod_to_newconf_map.get(self._location_uimod)
        loc_uiconf_changed = loc_uiconf is not None
        if loc_uiconf is None:
            loc_uiconf = self._location_uimod.load_conf()

        providers.extend([
            LocationUiConfProvider(self._location_uimod.id, loc_uiconf, loc_uiconf_changed)
        ])

        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()

        # send sms to module radar with new config
        if loc_uiconf_changed:
            eventlog.log_d(f'stat: newconf:<{loc_uiconf}>')

            ra_uds_filepath = fs.join_path(spool_dir, f'ra{self._ra_uimod.id}-ra{self._ra_uimod.id}-uds')
            ra_lock_dirpath = f'/run/edge/radar.{self._ra_uimod.pid}.api'

            from .ra_api import ApiConfBuilder
            api_builder = ApiConfBuilder(ra_uds_filepath, ra_lock_dirpath)

        if self._cam_uimod.type == ph.UI_TYPE_MV_CAMERA:
            loc_pri = loc_uiconf['ro']
            offset_frames = loc_pri.get('offset') or 0
            providers.extend([
                MvCameraUiConfPorvider(self._cam_uimod.id, cam_uiconf, offset_frames)
            ])

        elif self._cam_uimod.type == ph.UI_TYPE_RTSP_CAMERA:
            providers.extend([
                RtspCameraUiConfPorvider(self._cam_uimod.id, cam_uiconf)
            ])

        else:
            eventlog.log_e('fail')
            raise NotImplementedError()

        providers.extend([
            FusionUiConfProvider(self._fusion_uimod.id)
        ])

        for provider in providers:
            provider.setup_builders(api_builder, main_builder, halt_builder)

        if api_builder is not None:
            api_builder.build()

        main_conf = main_builder.build()
        halt_conf = halt_builder.build()

        self._main_halt = main_conf, halt_conf

    def complete(self) -> None:
        main_conf, halt_conf = self._main_halt

        edge_ra = self._edge_ra
        edge_ra.install()
        edge_ra.save_main_conf(main_conf)
        edge_ra.save_haltaction_conf(halt_conf)
        edge_ra.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        pass


class RaDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_RADAR:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RADAR, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
