from typing import List, Optional, Self, Set, Union

from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.dgmod.rc.rc_main import MainConfBuilder


# noinspection PyUnusedLocal
def rc_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_MV_CAMERA:
        return []

    try:
        cam_uimod, lpr_uimod, fusion_uimod = (
            next(filter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod)
        ])

        if forced or activate:
            return [RcCreator(cam_uimod, lpr_uimod, fusion_uimod)]

    except StopIteration:
        pass

    return []


class RcCreator(BaseCreator):
    def __init__(self, cam_uimod: ph.UiMod, lpr_uimod: ph.UiMod, fusion_uimod: ph.UiMod):
        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._fusion_uimod = fusion_uimod
        self._rc_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_RAWCODEC, cam_uimod.pid, cam_uimod.id)
        self._main_ss = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._rc_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        ui_pid = self._cam_uimod.pid
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        spool_dir = ph.get_spool_dir(ui_pid)

        builder = MainConfBuilder()
        builder.setup_react(spool_dir, 'mg{0}-rc{0}-scene'.format(self._cam_uimod.id))
        builder.setup_event_blobs(spool_dir, 'rc{0}-rc{0}-store'.format(self._cam_uimod.id))
        builder.setup_event_args(spool_dir, [
            f'rc{self._cam_uimod.id}-lr{self._lpr_uimod.id}-scene',
            f'rc{self._cam_uimod.id}-tr{self._fusion_uimod.id}-scene',
            'mg{0}-mcgst{0}-scene'.format(self._cam_uimod.id)
        ])

        ss_active = cam_uiconf['ro']['raw-format'] != 'MONO8T'
        if ss_active and cam_uiconf['rw']['videostream']['control']['profile'] == 'auto-70-150':
            builder.setup_event_args(spool_dir, ['mg{0}-gc{0}-scene'.format(self._cam_uimod.id)])

        main_conf = builder.build()
        self._main_ss = main_conf, ss_active

    def complete(self) -> None:
        main_conf, ss_active = self._main_ss

        edge_rc = self._rc_dgmod
        edge_rc.install()
        edge_rc.save_main_conf(main_conf)
        edge_rc.save_setupstore_conf(ss_active)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class RcDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_MV_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RAWCODEC, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
