from typing import Dict, Final, List

from edge import eventlog
from edge.support import fs


class _BusStreamerBuilder:
    def __init__(self, blobs_format: str) -> None:
        self._blobs_format: Final[str] = blobs_format
        self._blobs_dir: str | None = None
        self._args_dirs: List[str] = list()

    def setup_blobs(self, spool_dir: str, blobs_dir: str) -> None:
        self._blobs_dir = fs.join_path(spool_dir, blobs_dir)

    def setup_args(self, spool_dir: str, args_dirs: List[str]) -> None:
        for args_dir in args_dirs:
            self._args_dirs.append(fs.join_path(spool_dir, args_dir))

    def build(self) -> List[Dict]:
        if not self._args_dirs:
            return []

        if self._blobs_dir is None:
            eventlog.log_e('fail')
            raise NotImplementedError()

        return [{
            'type': 'bus-streamer',
            'conf': {
                'format': self._blobs_format,
                'blob-writer': {
                    'file-dir': self._blobs_dir,
                    'file-count': [40, 45]
                },
                'args-writer': self._args_dirs
            }
        }]


class MainConfBuilder:
    def __init__(self, uimod_id: str, spool_dir: str) -> None:
        self.uimod_id: Final[str] = uimod_id
        self._spool_dir: Final[str] = spool_dir

        self._bus_streamer_builder_raw = _BusStreamerBuilder('raw')
        self._bus_streamer_builder_mono8 = _BusStreamerBuilder('mono8')
        self._bus_streamer_grayscale: bool | None = None
        self._bus_streamer_dirs: List[str] = list()

        self._jnode = {
            'stats': {
                'filepath': fs.join_path(spool_dir, 'zabbix', 'vs0'),
                'interval': 5
            },
            'dec-source': {
                'conf': {
                    'target-w': 0,
                    'target-h': 0,
                    'with-deep-notify': False,
                    'wdt': 0
                }
            },
            'strobe-streamer': [],
            'dec-streamer': [],
            'enc-streamer': []
        }

    def setup_strobe_event(self, fifo_name: str) -> None:
        self._jnode['strobe-streamer'].extend([
            fs.join_path(self._spool_dir, fifo_name)
        ])

    def setup_enc_source_rtsp(self, stream_url: str, stream_format: str, allow_udp: bool) -> None:
        ui_to_edge_format_map = {
            'video/H264': 'h264',
            'video/JPEG': 'jpeg'
        }

        self._jnode.update({
            'enc-source': {
                'type': 'rtsp',
                'conf': {
                    'url': stream_url,
                    'video-format': ui_to_edge_format_map[stream_format],
                    'allow-udp-unicast': allow_udp,
                    'allow-udp-multicast': allow_udp,
                    'with-deep-notify': False,
                    'wdt': 0
                }
            }
        })

    def setup_enc_source_flexwath(self, conf: Dict) -> None:
        """
        :param conf:
            rtsp-camera#/ro/flexwatch-conf
        """
        self._jnode.update({
            'enc-source': {
                'type': 'flexwatch',
                'conf': conf
            }
        })

    def setup_dec_source(self, skip_rate: int, prefer_gpu: bool) -> None:
        self._jnode['dec-source']['conf'].update({
            'skiprate': skip_rate,
            'prefer-gpu': prefer_gpu
        })

    def setup_dec_streamer__bus_blobs_raw(self, blobs_dir: str) -> None:
        self._bus_streamer_builder_raw.setup_blobs(self._spool_dir, blobs_dir)

    def setup_dec_streamer__bus_blobs_mono8(self, blobs_dir: str) -> None:
        self._bus_streamer_builder_mono8.setup_blobs(self._spool_dir, blobs_dir)

    def setup_dec_streamer__bus_args_grayscale(self, grayscale: bool) -> None:
        self._bus_streamer_grayscale = grayscale

    def setup_dec_streamer__bus_args(self, args_dirs: List[str]) -> None:
        self._bus_streamer_dirs.extend(args_dirs)

    def setup_enc_streamer__bsv_mjpeg(self, blobstore_ident: int, blobstore_chan: str) -> None:
        self._jnode['enc-streamer'].append({
            'type': 'bsv-streamer',
            'conf': {
                'format': 'mjpeg',
                'host': blobstore_ident,
                'chan': blobstore_chan
            }
        })

    def build(self) -> Dict:
        if self._bus_streamer_grayscale is None:
            eventlog.log_e('fail')
            raise NotImplementedError()

        elif self._bus_streamer_grayscale:
            self._bus_streamer_builder_mono8.setup_args(self._spool_dir, self._bus_streamer_dirs)

        else:
            self._bus_streamer_builder_raw.setup_args(self._spool_dir, self._bus_streamer_dirs)

        self._jnode['dec-streamer'].extend(self._bus_streamer_builder_raw.build())
        self._jnode['dec-streamer'].extend(self._bus_streamer_builder_mono8.build())
        return self._jnode
