from typing import Dict, List, Optional, Self, Set, Union

from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.dgmod.rg.rg_halt import HaltConfBuilder
from edge.webserver.sysconf.dgmod.rg.rg_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.rg.uiconf import IUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.fusion import FusionUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.lpr import LprUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.radar import RadarUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.redlight import RedlightUiConfProvider, RailroadUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.rtsp_camera import RtspCameraUiConfProvider


# noinspection PyUnusedLocal
def rg_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_RTSP_CAMERA:
        return []

    try:
        cam_uimod, fusion_uimod, lpr_uimod, ra_uimod, rl_uimods, railroad_crossing_uimods = (
            next(filter(lambda x: x.type == ph.UI_TYPE_RTSP_CAMERA, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods), None),
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RED_LIGHT],
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RAILROAD_CROSSING])

        activate = any([
            uimod_to_newconf_map.get(cam_uimod),
            any([uimod_to_newconf_map.get(x) for x in rl_uimods]),
            any([uimod_to_newconf_map.get(x) for x in railroad_crossing_uimods])
        ])

        return [RgCreator(cam_uimod, lpr_uimod, fusion_uimod, rl_uimods, railroad_crossing_uimods, ra_uimod)]

    except StopIteration:
        pass

    return []


class RgCreator(BaseCreator):
    def __init__(self, cam_uimod: ph.UiMod,
                 lpr_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 rl_uimods: List[ph.UiMod],
                 railroad_crossing_uimods: List[ph.UiMod],
                 ra_uimod: Optional[ph.UiMod]):

        super().__init__()

        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._fusion_uimod = fusion_uimod
        self._rl_uimods = rl_uimods
        self._railroad_crossing_uimods = railroad_crossing_uimods
        self._ra_uimod = ra_uimod

        self._rg_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_RTSP_GRABBER, cam_uimod.pid, cam_uimod.id)
        self._main_halt = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._rg_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        main_builder = MainConfBuilder(self._cam_uimod.id, ph.get_spool_dir(self._cam_uimod.pid))
        halt_builder = HaltConfBuilder()

        providers = list[IUiConfProvider]()

        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        providers.extend([
            RtspCameraUiConfProvider(self._cam_uimod.id, cam_uiconf),
            FusionUiConfProvider(self._fusion_uimod.id),
            LprUiConfProvider(self._lpr_uimod.id)
        ])

        if self._ra_uimod:
            providers.extend([
                RadarUiConfProvider(self._ra_uimod.id)
            ])

        # add mg to rl path
        if len(self._rl_uimods) == 1:
            rl_uimod = self._rl_uimods[0]
            rl_uiconf: Dict = uimod_to_newconf_map.get(rl_uimod) or rl_uimod.load_conf()
            providers.extend([
                RedlightUiConfProvider(rl_uimod.id, rl_uiconf)
            ])

        elif len(self._railroad_crossing_uimods) == 1:
            rl_uimod = self._railroad_crossing_uimods[0]
            rl_uiconf: Dict = uimod_to_newconf_map.get(rl_uimod) or rl_uimod.load_conf()
            providers.extend([
                RailroadUiConfProvider(rl_uimod.id, rl_uiconf)
            ])

        for provider in providers:
            provider.setup_builders(main_builder, halt_builder)

        main_conf = main_builder.build()
        halt_conf = halt_builder.build()

        self._main_halt = main_conf, halt_conf

    def complete(self) -> None:
        main_conf, halt_conf = self._main_halt  # type: dict, str

        edge_rg = self._rg_dgmod
        edge_rg.install()
        edge_rg.save_main_conf(main_conf)
        edge_rg.save_haltaction_conf(halt_conf)
        edge_rg.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class RgDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_RTSP_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RTSP_GRABBER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
