from typing import Dict, List, Optional, Self, Set, Union

from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.dgmod.rn.rn_main import MainConfBuilder
from edge.webserver.sysconf.uiconf.time import TimeUiconf


# noinspection PyUnusedLocal
def or_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type not in [ph.UI_TYPE_HTTP_STREAMING, ph.UI_TYPE_RTSP_STREAMING]:
        return []

    try:
        (ra_uimod, rstreaming_uimod, hstreaming_uimod, fusion_uimod, ws_uimod, term_uimod,
         gnss_uimod, cam_uimod, roadway_uimod, location_uimod) = (
            next(filter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods), None),
            next(filter(lambda x: x.type == ph.UI_TYPE_RTSP_STREAMING, all_used_uimods), None),
            next(filter(lambda x: x.type == ph.UI_TYPE_HTTP_STREAMING, all_used_uimods), None),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_WEBSERVER, all_used_uimods), None),
            next(filter(lambda x: x.type == ph.UI_TYPE_TERMINAL, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_GNSS, all_used_uimods), None),
            next(filter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods), None),
            next(filter(lambda x: x.type == ph.UI_TYPE_ROADWAY, all_used_uimods), None),
            next(filter(lambda x: x.type == ph.UI_TYPE_LOCATION, all_used_uimods), None))

        spotspeed_uimods = [x for x in all_used_uimods if x.type == ph.UI_TYPE_SPOT_SPEED]
        spotspeed_uimod = None
        if len(spotspeed_uimods) == 1:
            spotspeed_uimod = spotspeed_uimods[0]

        if entry_uimod.type == ph.UI_TYPE_HTTP_STREAMING:
            streaming_uimod = hstreaming_uimod

        elif entry_uimod.type == ph.UI_TYPE_RTSP_STREAMING:
            streaming_uimod = rstreaming_uimod

        else:
            raise NotImplementedError

        activate = any([
            uimod_to_newconf_map.get(streaming_uimod),
            uimod_to_newconf_map.get(gnss_uimod),
            uimod_to_newconf_map.get(cam_uimod),
            uimod_to_newconf_map.get(roadway_uimod),
            uimod_to_newconf_map.get(term_uimod),
            uimod_to_newconf_map.get(spotspeed_uimod),
            uimod_to_newconf_map.get(location_uimod)
        ])

        if forced or activate:
            creator = _Creator(
                streaming_uimod, fusion_uimod, gnss_uimod, roadway_uimod, term_uimod, location_uimod, cam_uimod,
                webserver_uimod=ws_uimod, radar_uimod=ra_uimod, spotspeed_uimod=spotspeed_uimod)

            return [creator]

    except StopIteration:
        pass

    return []


class _Creator(BaseCreator):
    def __init__(self, streaming_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 gnss_uimod: ph.UiMod,
                 roadway_uimod: ph.UiMod,
                 terminal_uimod: ph.UiMod,
                 location_uimod: Optional[ph.UiMod],
                 cam_uimod: Optional[ph.UiMod] = None,
                 webserver_uimod: Optional[ph.UiMod] = None,
                 radar_uimod: Optional[ph.UiMod] = None,
                 spotspeed_uimod: Optional[ph.UiMod] = None):

        BaseCreator.__init__(self)
        self._streaming_uimod = streaming_uimod
        self._fusion_uimod = fusion_uimod
        self._gnss_uimod = gnss_uimod
        self._roadway_uimod = roadway_uimod
        self._terminal_uimod = terminal_uimod
        self._webserver_uimod = webserver_uimod
        self._radar_uimod = radar_uimod
        self._spotspeed_uimod = spotspeed_uimod
        self._location_uimod = location_uimod
        self._cam_uimod = cam_uimod
        self._main_ss = None

        self._rn_dgmod = dg.DgMod.create(dg.EDGE_TYPE_RENDERER, self._streaming_uimod.pid, self._streaming_uimod.id)

    @property
    def dgmod(self) -> dg.DgMod:
        return self._rn_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._streaming_uimod.pid)

        if self._streaming_uimod.type == ph.UI_TYPE_RTSP_STREAMING:
            codec = 'RGBA'
            scene_dir = f'or{self._streaming_uimod.id}-rs{self._streaming_uimod.id}-scene'

        elif self._streaming_uimod.type == ph.UI_TYPE_HTTP_STREAMING:
            codec = 'JPEG'
            scene_dir = f'or{self._streaming_uimod.id}-ws{self._webserver_uimod.id}-scene'

        else:
            raise NotImplementedError

        streaming_uiconf = uimod_to_newconf_map.get(self._streaming_uimod) or self._streaming_uimod.load_conf()
        streaming_pub: Dict = streaming_uiconf['rw']
        streaming_live: Dict = streaming_pub['live']
        streaming_vs: Dict = streaming_live['videostream']
        streaming_v: Dict = streaming_live['renderer']['vehicle']

        builder = MainConfBuilder()
        # builder.setup_trafficlight_react(spool_dir, 'tl{0}-or{0}-track/'.format(self._streaming_uimod.id))
        builder.setup_tracker_react(spool_dir, f'tr{self._fusion_uimod.id}-or{self._streaming_uimod.id}-scene')

        builder.setup_event_blobs(spool_dir, 'or{0}-or{0}-store/'.format(self._streaming_uimod.id), codec)
        builder.setup_event_args(spool_dir, [scene_dir])

        builder.setup_service_origin(ph.TermAboutUiconf.get_conf_path(self._streaming_uimod.pid))
        builder.setup_service_datetime(TimeUiconf.CONF_PATH)
        builder.setup_service_gnss(self._gnss_uimod.conf_path)
        builder.setup_service_stats(spool_dir, f'/zabbix/or{int(self._streaming_uimod.id) % 10}')

        builder.setup_image_footer(streaming_vs['frame-w'], streaming_live['footer'])
        builder.setup_image_anno_lr(streaming_v)

        if self._cam_uimod and self._location_uimod:
            builder.setup_service_cam(self._cam_uimod.conf_path, self._location_uimod.conf_path)

        if self._radar_uimod:
            builder.setup_image_anno_ra(streaming_v)

        if self._spotspeed_uimod is not None:
            spotspeed_uiconf = uimod_to_newconf_map.get(self._spotspeed_uimod) or self._spotspeed_uimod.load_conf()
            spotspeed_rules = spotspeed_uiconf['rw']['args']['rules']

            # do not render overspeed if there is > 1 spot-speed rules
            if len(spotspeed_rules) == 1 and spotspeed_rules[0]['enabled']:
                rule0_limits: Dict = spotspeed_rules[0]['limits']
                speed_limit = rule0_limits['speed-limit']
                max_overspeed = rule0_limits['max-overspeed']
                builder.setup_spotspeed(speed_limit, speed_limit + max_overspeed)

        roadway_uiconf = uimod_to_newconf_map.get(self._roadway_uimod) or self._roadway_uimod.load_conf()
        roadway_pub = roadway_uiconf['rw']
        roadway_args = roadway_pub['args']

        if len(roadway_args['lanes']) and streaming_v['video-lane'] != 'never':
            builder.setup_roadway(roadway_args)

        main_conf = builder.build()
        ss_active = bool(streaming_uiconf['rw']['enabled'])
        self._main_ss = main_conf, ss_active

    def complete(self) -> None:
        rn_dgmod = self.dgmod
        main_conf, ss_active = self._main_ss  # type: dict, bool
        rn_dgmod.install()
        rn_dgmod.save_main_conf(main_conf)
        rn_dgmod.save_setupstore_conf(ss_active)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class OrDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type in [ph.UI_TYPE_HTTP_STREAMING, ph.UI_TYPE_RTSP_STREAMING]:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RENDERER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
