from typing import List, Optional, Self, Set, Union

from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap


# noinspection PyUnusedLocal
def tl_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_TRAFFIC_LIGHT:
        return []

    try:
        cam_uimod, tl_uimod, rules_uimod, redlight_uimods, railroad_crossing_uimods = [
            next(filter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_TRAFFIC_LIGHT, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_TRAFFIC_RULES, all_used_uimods)),
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RED_LIGHT],
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RAILROAD_CROSSING]]

        activate = any([uimod_to_newconf_map.get(x) for x in redlight_uimods]) or len(redlight_uimods) == 0 or \
                   (any([uimod_to_newconf_map.get(x) for x in railroad_crossing_uimods]) or len(railroad_crossing_uimods) == 0)
        if forced or activate:
            return [TlCreator(cam_uimod, tl_uimod, rules_uimod, redlight_uimods, railroad_crossing_uimods)]

    except StopIteration:
        pass

    return []


class TlCreator(BaseCreator):
    def __init__(self, cam_uimod: ph.UiMod,
                 tl_uimod: ph.UiMod,
                 rules_uimod: ph.UiMod,
                 redlight_uimods: List[ph.UiMod],
                 railroad_crossing_uimods: List[ph.UiMod]):

        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._tl_uimod = tl_uimod
        self._rules_uimod = rules_uimod
        self._redlight_uimods = redlight_uimods
        self._railroad_crossing_uimods = railroad_crossing_uimods
        self._edge_tl = dg.DgMod.create(
            dg.EDGE_TYPE_TRAFFIC_LIGHT, tl_uimod.pid, tl_uimod.id)
        self._main_ss = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._edge_tl

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._tl_uimod.pid)

        # create main.conf
        main_conf = {
            # "device": {
            #     "vendor": "komkon",
            #     "params": {
            #         "endpoint": "/dev/ttyUSB3",
            #         "controller": [
            #             {
            #                 "id": 1,
            #                 "red-chan": 1,
            #                 "amber-chan": 2
            #             }
            #         ]
            #     }
            # },
            'changelog': {
                'entry-count': 1200,
                'entry-tout': 125,  # pass 600 for train-trafficlight
                'entry-age': 300
            },
            'track-event': [
                fs.join_path(spool_dir, f'tl0-sc{self._rules_uimod.id}-track')
            ],
            'monitoring': {
                'stat-file': fs.join_path(spool_dir, 'zabbix', f'tl{self._tl_uimod.id}'),
                'stat-tout': 30
            }
        }

        # if False:
        #     main_conf['track-event'].append(
        #         fs.join_path(spool_dir, f'tl{self._tl_uimod.id}-or{self._rules_uimod.id}-track'))

        # get device params
        if len(self._redlight_uimods) == 1:
            rl_uimod = self._redlight_uimods[0]
            rl_uiconf = uimod_to_newconf_map.get(rl_uimod) or rl_uimod.load_conf()
            tl_id = rl_uiconf['ro'].get('params', {}).get('controller', [{}])[0].get('id', 1)

            main_conf['delay-usec'] = rl_uiconf.get('delay-usec', 0)
            vendor = rl_uiconf['ro'].get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                tl_jnode = rl_uiconf['rw']['args']['scene']['area']['traffic-lights']
                params_jnode = {
                    'controller': [
                        {
                            'type': 'traffic-light',
                            'red-chan': 1,
                            'red-zone': tl_jnode['red'],
                            'amber-chan': 2,
                            'amber-zone': tl_jnode['amber'],
                            'green-chan': 3,
                            'green-zone': tl_jnode['green'],
                            'id': tl_id
                        }
                    ],
                    'endpoint': fs.join_path(spool_dir, f'mg{self._cam_uimod.id}-zc{rl_uimod.id}-scene')
                }

            else:
                params_jnode = rl_uiconf['ro'].get('params')

            if None not in [vendor, params_jnode]:
                main_conf['device'] = {
                    'vendor': vendor,
                    'params': params_jnode
                }

        elif len(self._railroad_crossing_uimods) == 1:
            rc_uimod = self._railroad_crossing_uimods[0]
            rc_uiconf = uimod_to_newconf_map.get(rc_uimod) or rc_uimod.load_conf()
            rc_uiconf_pri = rc_uiconf['ro']

            tl_id = rc_uiconf_pri.get('params', {}).get('controller', [{}])[0].get('id', 1)

            main_conf['delay-usec'] = rc_uiconf.get('delay-usec', 0)
            vendor = rc_uiconf_pri.get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                tl_jnode = rc_uiconf['rw']['args']['scene']['area']['traffic-lights']
                params_jnode = {
                    'controller': [
                        {
                            'type': 'train-light',
                            'red-chan': 1,
                            'red-zone1': tl_jnode['red1'],
                            'red-zone2': tl_jnode['red2'],
                            'id': tl_id,
                            'threshold': rc_uiconf_pri.get('threshold', 0.1),
                            'bin-threshold': rc_uiconf_pri.get('bin-threshold', 0.1),
                            'max-count': rc_uiconf_pri.get('max-count', 6)
                        }
                    ],
                    'endpoint': fs.join_path(spool_dir, f'mg{self._cam_uimod.id}-zc{rc_uimod.id}-scene')
                }

            else:
                params_jnode = rc_uiconf_pri.get('params')

            if None not in [vendor, params_jnode]:
                main_conf['device'] = {
                    'vendor': vendor,
                    'params': params_jnode
                }

        ss_active = len(self._redlight_uimods) == 1 or len(self._railroad_crossing_uimods) == 1

        self._main_ss = main_conf, ss_active

    def complete(self) -> None:
        main_conf, ss_active = self._main_ss  # type: dict, bool

        edge_tl = self._edge_tl
        edge_tl.install()
        edge_tl.save_main_conf(main_conf)
        edge_tl.save_setupstore_conf(ss_active)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.sighup_ss_dgmod()


class TlDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_TRAFFIC_LIGHT:
            dgmod = dg.DgMod.create(
                dg.EDGE_TYPE_TRAFFIC_LIGHT, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
