# coding=utf-8

from typing import Dict, List, Optional, Self, Set, Union

from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf


# noinspection PyUnusedLocal
def ws_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_WEBSERVER:
        return []

    try:
        terminal_uimods = [x for x in all_used_uimods if x.type == ph.UI_TYPE_TERMINAL]
        hstreaming_uimods = [x for x in all_used_uimods if x.type == ph.UI_TYPE_HTTP_STREAMING]
        assert len(hstreaming_uimods) == len(terminal_uimods)

        # TODO проверить что ref terminal коррелирует с ref http-streaming ?
        terminal_uimods.sort(key=lambda item: int(item.id))
        hstreaming_uimods.sort(key=lambda item: int(item.id))

        activate = any([
            any([uimod_to_newconf_map.get(x) for x in hstreaming_uimods])
        ])

        if forced or activate:
            return [WsCreator(entry_uimod, terminal_uimods, hstreaming_uimods)]

    except StopIteration:
        pass

    return []


class WsCreator(BaseCreator):
    def __init__(self, ui_entry: ph.UiMod, terminal_uimods: List[ph.UiMod], hstreaming_uimods: List[ph.UiMod]):
        BaseCreator.__init__(self)
        self._ui_entry_point = ui_entry
        self._main_conf = None
        self._hstreaming_uimods = hstreaming_uimods
        self._terminal_uimods = terminal_uimods
        self._edge_ws = dg.CTRL_WS_DGMOD

    @property
    def dgmod(self) -> dg.DgMod:
        return self._edge_ws

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        edge_ws = self._edge_ws
        main_conf = edge_ws.load_main_conf()

        media_jnode_list = []
        for index in range(0, len(self._terminal_uimods)):
            hstreaming_uimod = self._hstreaming_uimods[index]
            term_uimod = self._terminal_uimods[index]

            ui_pid = term_uimod.pid
            hstreaming_uiconf = uimod_to_newconf_map.get(hstreaming_uimod) or hstreaming_uimod.load_conf()
            spool_dir = ph.get_spool_dir(ui_pid)

            origin_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
            origin_jnode = {
                'origin': origin_uuid
            }

            if hstreaming_uiconf['rw']['enabled']:
                origin_jnode['LiveEndpoint'] = fs.join_path(spool_dir, f'or{hstreaming_uimod.id}-ws0-scene')

            media_jnode_list.append(origin_jnode)

        main_conf['Media'] = media_jnode_list
        self._main_conf = main_conf

    def complete(self) -> None:
        main_conf: Dict = self._main_conf

        edge_ws = self._edge_ws
        edge_ws.install()
        edge_ws.save_main_conf(main_conf)
        edge_ws.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_ws_dgmod()


class WsDeleter:
    def __init__(self, origin_uuid: str):
        self._term_uuid = origin_uuid

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_FACTORY:
            term_uimod = ph.UiMod.create_term(entry_uimod.pid)
            term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
            return cls(term_uuid)

    def run(self) -> None:
        dgmod = dg.CTRL_WS_DGMOD
        dgconf = dgmod.load_main_conf()
        dgconf['Media'] = [x for x in dgconf['Media'] if x['origin'] != self._term_uuid]
        dgmod.save_main_conf(dgconf)
