from collections import defaultdict
from functools import cmp_to_key
from typing import Dict, List, Optional, Tuple

from edge import eventlog
from edge.support import db as edge_db, net
from edge.support.db import DbArgs
from edge.services import db_value, tz
from edge.services.db_value import DbValueApi
from edge.services.types.jnode.tz import decode__tz_service_conf_file
from edge.webserver.sysconf import ph
from edge.webserver.systest.test import Test, TestError, TestLimitsError
from edge.webserver.systest.misc import *
from edge.webserver.sysconf.api_controller import ControllerApi
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf


def _get_section_jnode(db: DbValueApi) -> Dict:
    """
    :param db:
        should be under the db transaction context (operator 'with')
    """
    section_jnode = db.get_value(DB_SECTION)
    if section_jnode is None:
        section_jnode = {
            DB_KEY_TESTPACK: {
                'rw': {
                    'sessions': [],
                    'results': {}
                }
            }
        }

    return section_jnode


def _get_testpack_jnode(db: DbValueApi) -> Dict:
    """
    :param db:
        should be under the db transaction context (operator 'with')
    """
    testpack_jnode = db.get_value(DB_SECTION, DB_KEY_TESTPACK)
    if testpack_jnode is None:
        testpack_jnode = {
            'rw': {
                'sessions': [],
                'results': {}
            }
        }

    return testpack_jnode


def _get_session_jnode(testpack_jnode: Dict, session_ref: str) -> Dict:
    """
    :raise KeyError:
    """
    sessions_jnode = [jnode for jnode in testpack_jnode['rw']['sessions'] if jnode['session'] == session_ref]

    if len(sessions_jnode) == 1:
        return sessions_jnode[0]

    elif not sessions_jnode:
        raise KeyError

    raise NotImplementedError


def _get_session_limits(section_jnode: Dict, session_ref: str) -> Dict:
    def get_limits_map() -> Dict[str, List[str]]:
        """
        :returns:
            test-to-origins-map
        """
        test_to_origin = defaultdict(list)
        root_info = ph.RootInfoUiconf.get_info()

        term_infos = ph.InfoUiconf.map_units_r(root_info, unit_types=[ph.UI_TYPE_TERMINAL])
        for term_info in term_infos:
            term_ref = term_info['ref']
            term_uimod = ph.UiMod.create(term_ref)
            term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()

            ra_info = ph.InfoUiconf.find_unit_r(term_info, unit_type=ph.UI_TYPE_RADAR)

            test_to_origin['datetime'] += [term_uuid]
            if ra_info is not None:
                test_to_origin['position'] += [term_uuid]
                test_to_origin['spot-speed'] += [term_uuid]

        if net.query_gps_count() > 0:
            test_to_origin['geolocation'] += ['device']

        return test_to_origin

    def get_tests_map() -> Dict[str, List[str]]:
        """
        :returns:
            test-to-origins-map
        """
        session_jnode = _get_session_jnode(section_jnode[DB_KEY_TESTPACK], session_ref)
        tests_jnode = session_jnode['tests']

        test_to_origin = defaultdict(list)
        for test_jnode in tests_jnode:
            test_type, test_run = test_jnode['test']

            db_key = mk_test_db_key(session_jnode['session'], (test_type, test_run))
            ro_jnode = section_jnode[db_key].get('ro') or dict()
            origin = ro_jnode.get('origin', 'device')

            test_to_origin[test_type] += [origin]

        return test_to_origin

    limits_map = get_limits_map()
    tests_map = get_tests_map()

    def get_origin_to_test_map() -> Dict[str, List[str]]:
        """
        :returns:
            test-to-origins-map
        """
        origin_to_test = defaultdict(list)
        for test_type, src_origins in limits_map.items():
            dst_origins = tests_map.get(test_type, [])
            for origin in src_origins:
                if origin not in dst_origins:
                    origin_to_test[origin] += [test_type]

        return origin_to_test

    rest_map = get_origin_to_test_map()

    def mk_limits_jnode() -> Dict:
        limits_jnode = {
            'origins': list(),
            'device': {
                'tests': list()
            }
        }

        origins_jnode = limits_jnode['origins']
        device_jnode = limits_jnode['device']
        for origin, tests in rest_map.items():
            if origin != 'device':
                origins_jnode.append({
                    'tests': [dict(test=test_type) for test_type in tests],
                    'origin': {
                        'uuid': origin,
                    }
                })

            else:
                device_jnode['tests'] = [dict(test=test_type) for test_type in tests]

        return limits_jnode

    return mk_limits_jnode()


def _mk_test(services: Services, session_ref: str, test_ref: Tuple[str, str]) -> Test:
    test_type = test_ref[0]
    if test_type == 'geolocation':
        from .geolocation_test import GeolocationTest
        return GeolocationTest(services, session_ref, test_ref)

    elif test_type == 'datetime':
        from .datetime_test import DatetimeTest
        return DatetimeTest(services, session_ref, test_ref)

    elif test_type == 'spot-speed':
        from .spot_speed_test import SpotSpeedTest
        return SpotSpeedTest(services, session_ref, test_ref)

    elif test_type == 'position':
        from .position_test import PositionTest
        return PositionTest(services, session_ref, test_ref)

    raise NotImplementedError


class TestSession:
    def __init__(self, services: Services, session_ref: str):
        self.services = services
        self._ref = session_ref

    @property
    def ref(self) -> str:
        return self._ref

    @staticmethod
    def _add_test(db: DbValueApi, session_ref: str, session_jnode: Dict, test_jnode: Dict) -> Tuple[str, str]:
        """
        :param db:
            should be under the db transaction context (operator 'with')
        :param session_ref:
        :param session_jnode:
        :param test_jnode:
        """
        tests_jnode = session_jnode['tests']

        test_type = test_jnode['test']
        test_run = mk_ref([x['test'][1] for x in tests_jnode])
        test_ref = test_type, test_run

        db.set_value(DB_SECTION, mk_test_db_key(session_ref, test_ref), value_jnode={
            'ro': test_jnode.get('args'),
            'rw': {
                'log': []
            }
        })

        tests_jnode.append({'test': test_ref})
        return test_ref

    def add_test(self, args: Dict) -> Tuple[str, str]:  # TODO: validate args againts limits
        """
        :returns:
            test-type, test-run
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)

            session_jnode = _get_session_jnode(testpack_jnode, self.ref)
            test_ref = self._add_test(db_service, self.ref, session_jnode, args)

            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
            return test_ref

    def add_tests(self) -> None:
        with self.services.db as db_service:
            section_jnode = _get_section_jnode(db_service)
            testpack_jnode = section_jnode[DB_KEY_TESTPACK]

            session_jnode = _get_session_jnode(testpack_jnode, self.ref)
            limits_jnode = _get_session_limits(section_jnode, self.ref)

            # make device tests
            for test_jnode in limits_jnode['device']['tests']:
                self._add_test(db_service, self.ref, session_jnode, {
                    'test': test_jnode['test']
                })

            # make origins tests
            for origin_jnode in limits_jnode['origins']:
                origin = origin_jnode['origin']['uuid']
                for test_jnode in origin_jnode['tests']:
                    self._add_test(db_service, self.ref, session_jnode, {
                        'test': test_jnode['test'],
                        'args': {
                            'origin': origin
                        }
                    })

            # sort tests
            def _cmp_test(left: Dict, right: Dict) -> int:
                l_type = left['test'][0]
                r_type = right['test'][0]
                _type_to_order_map = {
                    'datetime': 0,
                    'geolocation': 1,
                    'spot-speed': 2,
                    'position': 3
                }
                return _type_to_order_map[l_type] - _type_to_order_map[r_type]

            session_jnode['tests'].sort(key=cmp_to_key(_cmp_test))

            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

    def get_test(self, ref: Tuple[str, str]) -> Optional[Test]:
        """
        :param ref:
            test-type, test-run
        """
        with self.services.db as db_service:
            test_jnode = db_service.get_value(DB_SECTION, mk_test_db_key(self.ref, ref))
            if not test_jnode:
                return None

            return _mk_test(self.services, self.ref, ref)

    def del_test(self, ref: Tuple[str, str]) -> bool:
        """
        :param ref:
            test-type, test-run
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = _get_session_jnode(testpack_jnode, self.ref)

            last_tests = sessions_jnode['tests']
            next_tests = [test_jnode for test_jnode in last_tests if test_jnode['test'] != ref]

            if len(next_tests) == len(last_tests):
                return False

            sessions_jnode['tests'] = next_tests
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
            db_service.drop_value(DB_SECTION, key=mk_test_db_key(self.ref, ref))
            return True

    def set_summary(self, jnode) -> None:
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = _get_session_jnode(testpack_jnode, self.ref)

            sessions_jnode['summary'] = jnode.get('summary')
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)


class TestPack:
    def __init__(self, services: Services):
        self.services = services

    def add_results(self, args: Optional[Dict]) -> None:
        raw_crt_jnode = args.get('certificate') if args else None

        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            crt_jnode = dict()
            if raw_crt_jnode:
                crt_jnode['ref-number'] = raw_crt_jnode['ref-number']
                crt_jnode['max-time'] = raw_crt_jnode['max-time']

                min_time = raw_crt_jnode.get('min-time')
                if min_time:
                    crt_jnode['min-time'] = min_time

                authority = raw_crt_jnode.get('authority')
                if authority:
                    crt_jnode['authority'] = authority

            if crt_jnode:
                testpack_jnode['rw']['results']['certificate'] = crt_jnode
            else:
                try:
                    del testpack_jnode['rw']['results']['certificate']
                except KeyError:
                    pass

            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

        if crt_jnode:
            ControllerApi.install_systest_crt(crt_jnode)
        else:
            ControllerApi.uninstall_systest_crt()

    def add_report(self, tmp_blob_filepath: str) -> None:
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            old_report_ref = testpack_jnode['rw']['results'].get('report')
            if old_report_ref:
                ControllerApi.uninstall_systest_report(old_report_ref)

            report_ref = f'{mk_ref()}.pdf'
            testpack_jnode['rw']['results']['report'] = report_ref
            ControllerApi.install_systest_report(report_ref, tmp_blob_filepath)
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

    def delete_report(self, report_ref: str) -> None:
        """
        :raise TestError:
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)

            if testpack_jnode['rw']['results'].get('report') != report_ref:
                raise TestError

            del testpack_jnode['rw']['results']['report']
            ControllerApi.uninstall_systest_report(report_ref)
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

    def add_session(self) -> TestSession:
        """
        :raise TestLimitsError:
        :raise TestError:
        """
        # reconfig radar
        try:
            ra_origin_pid = ph.RootInfoUiconf.get_raterm_pid()
            if ra_origin_pid:
                min_rcs = get_rcs(ra_origin_pid, systest_active=True)
                if min_rcs is not None:
                    net.query_ra_alignment(ra_origin_pid, min_rcs=min_rcs, alignment=True)
        except net.NetError:
            eventlog.log_e('fail: setup radar')
            raise TestError

        ControllerApi.start_systest()

        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = testpack_jnode['rw']['sessions']
            if len(sessions_jnode) > 0:
                raise TestLimitsError

            session_ref = mk_ref(known_refs=list())
            session_jnode = {
                'session': session_ref,
                'tests': []
            }

            testpack_jnode['rw']['sessions'] = [session_jnode]
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
            return TestSession(self.services, session_ref)

    def get_session(self, ref: str) -> Optional[TestSession]:
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)

        try:
            _get_session_jnode(testpack_jnode, ref)
            return TestSession(self.services, ref)
        except KeyError:
            pass

    def del_session(self, ref: str) -> bool:
        """
        :raise TestError:
        """
        try:
            ra_origin_pid = ph.RootInfoUiconf.get_raterm_pid()
            if ra_origin_pid:
                min_rcs = get_rcs(ra_origin_pid, systest_active=False)
                if min_rcs is not None:
                    net.query_ra_alignment(ra_origin_pid, min_rcs=min_rcs, alignment=False)
        except net.NetError:
            eventlog.log_e('fail: setup radar')
            raise TestError

        ControllerApi.finish_systest()

        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = testpack_jnode['rw']['sessions']
            session_jnode = None
            for jnode in sessions_jnode:
                if jnode['session'] == ref:
                    session_jnode = jnode

            if session_jnode:
                sessions_jnode.remove(session_jnode)
                db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
                db_service.drop_value(DB_SECTION, mk_test_db_glob(ref))
                return True

        return False

    def pack(self) -> Dict:
        with self.services.db as db_service:
            section_jnode = _get_section_jnode(db_service)
            testpack_rw = section_jnode[DB_KEY_TESTPACK]['rw']

        for session_jnode in testpack_rw['sessions']:
            session_jnode['limits'] = _get_session_limits(
                section_jnode, session_jnode['session'])

        return testpack_rw


def init(db_args: Optional[DbArgs] = None) -> TestPack:
    if db_args is None:
        db_args = edge_db.DbArgs.deserialize2()

    db_service = db_value.init_service(db_args)

    tz_args = decode__tz_service_conf_file()
    tz_service = tz.init_service(tz_args)

    services = Services(tz_service, db_service)
    return TestPack(services)


# noinspection PyUnusedLocal
def fini(testpack: TestPack):
    testpack.services.db.destroy()
    testpack.services.tz.destroy()
