import os
import ujson

from twisted.internet.error import ConnectionDone
from twisted.internet.protocol import connectionDone
from twisted.protocols.basic import LineOnlyReceiver
from twisted.protocols.policies import LimitTotalConnectionsFactory
from twisted.python.failure import Failure

from typing import cast, Callable, Dict, Optional

from edge import eventlog

from eventstore.services.types import ApiArgs


class IApiService:

    def set_handler(self, handler: Callable[[Optional[Dict]], Dict]) -> None:
        raise NotImplementedError

    def destroy(self) -> None:
        pass


class _ApiClientProtocol(LineOnlyReceiver):
    delimiter = b'\0'

    def __init__(self):
        self.timeout_call = None

    def connectionMade(self) -> None:
        eventlog.log_d('exec')

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        if reason.check([ConnectionDone]):
            eventlog.log_e(f'done: {reason}')
        else:
            eventlog.log_d('done')

    def lineReceived(self, query_bytes: bytes) -> None:
        query_utf8 = str(query_bytes, 'utf-8')
        eventlog.log_d(f'query:<{query_utf8}>')

        # noinspection PyBroadException
        try:
            query_jnode = ujson.loads(query_utf8)
            reply_jnode = self.factory.exec_query(query_jnode)
        except Exception as x:
            eventlog.log_x()
            reply_jnode = {
                'code': 500,
                'text': str(x)
            }

        # noinspection PyBroadException
        try:
            reply_utf8 = ujson.dumps(reply_jnode, ensure_ascii=False)
            self.sendLine(bytes(reply_utf8, 'utf-8'))
            eventlog.log_d(f'reply:<{query_utf8}>')
        except Exception:
            eventlog.log_x()


class _ApiServer(LimitTotalConnectionsFactory, IApiService):
    connectionLimit = 1

    def __init__(self, twisted_reactor, endpoint: str):
        """
        :param twisted_reactor:
            twisted.internet.interfaces.IReactorUNIX
        :param endpoint
        """
        IApiService.__init__(self)

        self._handler = None
        self.protocol = _ApiClientProtocol  # override Factory::protocol
        twisted_reactor.listenUNIX(endpoint, self)

    def set_handler(self, handler: Callable[[Optional[Dict]], None]) -> None:
        self._handler = handler

    def exec_query(self, query_args: Dict) -> Dict:
        return self._handler(query_args)


def init_service(twisted_reactor, args: ApiArgs) -> IApiService:
    if os.path.exists(args.endpoint):
        os.unlink(args.endpoint)

    return _ApiServer(twisted_reactor, args.endpoint)


def fini_service(service: IApiService) -> None:
    self = cast(IApiService, service)
    self.destroy()
