import os

from typing import Dict, List, Optional, Self, Tuple

from edge import eventlog
from edge.image.jnode.types import decode__im_footer_conf
from edge.image.types.footer import ImFooterConf
from edge.services.types.jnode.cam import decode__cam_service_conf
from edge.services.types.cam import CamServiceConf
from edge.services.types.jnode.gnss import decode__gnss_service_conf_file
from edge.services.types.jnode.tz import decode__tz_service_conf_file
from edge.services.types.lane import LaneServiceConf, LaneConf
from edge.services.types.about import AboutServiceConf
from edge.services.types.gnss import GnssServiceConf
from edge.services.types.tz import TzServiceConf
from edge.types.exceptions import JnodeError

from renderer.services.jnode.tl import JsonTlServiceArgs


class StatsServiceArgs:
    def __init__(self, stat_filepath: str, stat_tout_sec: int):
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            main.conf#/services/monitoring

        :raise JnodeError:
        """
        stat_path = jnode.get('path')
        if not (type(stat_path) in [str] and len(stat_path) < 64 and os.path.isabs(stat_path)):
            eventlog.log_e_param('#/path')
            raise JnodeError

        stat_tout_sec = jnode.get('tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_e_param('#/tout')
            raise JnodeError

        return cls(stat_path, stat_tout_sec)


class ReactArgs:
    def __init__(self, endpoint: str):
        self.endpoints = [endpoint]

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        return cls(jnode['Url'])


class ImLoaderArgs:
    def __init__(self, dh: int, image_wh: Tuple[int, int]):
        self.dh = dh
        self.image_wh = image_wh


class ImQueueArgs:
    def __init__(self, endpoint: str, min_count: int, max_count: int):
        self.endpoint = endpoint
        self.min_file_count = min_count
        self.max_file_count = max_count


class ImStoreArgs:
    def __init__(self, im_format: str, im_q: ImQueueArgs):
        self.codec = im_format
        self.image_queue_args = im_q

    @classmethod
    def unpack(cls,  jnode: Dict) -> Self:
        """
        :param jnode:
            main.conf#/image-store

        :raise JnodeError:
        """
        im_format = jnode.get('format')
        if im_format not in ['JPEG', 'RGBA']:
            eventlog.log_e_param('#/image-store')
            raise JnodeError

        im_max_count = jnode.get('max-file-count', 4)
        if not (int == type(im_max_count) and 4 <= im_max_count <= 8):
            eventlog.log_e_param('#/max-file-count')
            raise JnodeError

        endpoint = jnode.get('endpoint')
        if not (type(endpoint) in [str] and len(endpoint) < 64 and os.path.isabs(endpoint)):
            eventlog.log_e_param('#/endpoint')
            raise JnodeError

        im_q = ImQueueArgs(endpoint, im_max_count - 2, im_max_count)
        return cls(im_format, im_q)


class JsonQueueArgs:
    def __init__(self, endpoint: str, min_count: int, max_count: int):
        self.endpoint = endpoint
        self.min_file_count = min_count
        self.max_file_count = max_count


class EventArgs:
    def __init__(self, q_args: List[JsonQueueArgs], log_level: str = 'info'):
        self.json_queue_args_list = q_args
        self.log_level = log_level

    @classmethod
    def unpack(cls, jnode: Dict, log_level: str) -> Self:
        """
        :param jnode:
            main.conf#/event-store
        :param log_level:

        :raise JnodeError:
        """
        max_file_count = jnode.get('max-file-count', 4)
        if not (int == type(max_file_count) and 4 <= max_file_count <= 8):
            eventlog.log_e_param('#/max-file-count')
            raise JnodeError

        q_args = [JsonQueueArgs(endpoint, max_file_count - 5, max_file_count)
                  for endpoint in jnode.get('endpoints') or list()]

        return cls(q_args, log_level)


class LrAnnoArgs:
    _MODE_NEVER = 'never'
    _MODE_ALWAYS = 'always'

    def __init__(self,
                 dir_mode: str = 'never',
                 lane_mode: str = 'never',
                 text_mode: str = 'never',
                 trace_mode: str = 'never',
                 center_mode: str = 'never'):

        self.show_dir = dir_mode == 'always'
        self.show_lane = lane_mode == 'always'
        self.show_text = text_mode == 'always'
        self.show_trace = trace_mode == 'always'
        self.show_center = center_mode == 'always'

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            main.conf#/Renderer/profile/vehicle/lr

        :raise JnodeError:
        """
        modes = [cls._MODE_NEVER, cls._MODE_ALWAYS]

        # unpack #/plate
        plate_mode = jnode.get('plate', cls._MODE_NEVER)
        if plate_mode not in modes:
            eventlog.log_e_param('#/plate')
            raise JnodeError

        # unpack #/trace
        trace_mode = jnode.get('trace', cls._MODE_NEVER)
        if trace_mode not in modes:
            eventlog.log_e_param('#/trace')
            raise JnodeError

        # unpack #/dir
        dir_mode = jnode.get('dir', cls._MODE_NEVER)
        if dir_mode not in modes:
            eventlog.log_e_param('#/dir')
            raise JnodeError

        # unpack #/lane
        lane_mode = jnode.get('lane', cls._MODE_NEVER)
        if lane_mode not in modes:
            eventlog.log_e_param('#/lane')
            raise JnodeError

        # unpack #/center
        center_mode = jnode.get('center', cls._MODE_NEVER)
        if center_mode not in modes:
            eventlog.log_e_param('#/center')
            raise JnodeError

        return LrAnnoArgs(dir_mode, lane_mode, plate_mode, trace_mode, center_mode)


class RaAnnoArgs:
    _MODE_IF_PLATE = 'if-plate'
    _MODE_ALWAYS = 'always'
    _MODE_NEVER = 'never'

    def __init__(self,
                 yellow_speed: int,
                 red_speed: int,
                 len_mode: str = 'never',
                 dir_mode: str = 'never',
                 target_mode: str = 'never',
                 speed_mode: str = 'never',
                 vehicle_mode: str = 'never',
                 radar_present: bool = False):

        self._speed_mode = speed_mode
        self._len_mode = len_mode
        self._dir_mode = dir_mode
        self._target_mode = target_mode
        self._vehicle_mode = vehicle_mode

        self.yellow_speed = yellow_speed
        self.red_speed = red_speed
        self.radar_present = radar_present

    @classmethod
    def _show(cls, mode: str, lr_ra_match: bool) -> bool:
        if mode == cls._MODE_ALWAYS:
            return True

        elif mode == cls._MODE_IF_PLATE:
            return lr_ra_match

        else:
            return False

    def show_len(self, lr_ra_match: bool = False) -> bool:
        return self._show(self._len_mode, lr_ra_match)

    def show_dir(self, lr_ra_match: bool = False) -> bool:
        return self._show(self._dir_mode, lr_ra_match)

    def show_target(self, lr_ra_match: bool = False) -> bool:
        return self._show(self._target_mode, lr_ra_match)

    def show_vehicle(self, lr_ra_match: bool = False) -> bool:
        return self._show(self._vehicle_mode, lr_ra_match)

    def show_speed(self, lr_ra_match: bool = False) -> bool:
        return self._show(self._speed_mode, lr_ra_match)

    @classmethod
    def unpack(cls, jnode: Dict, yellow_speed: int, red_speed: int) -> Self:
        """
        :param jnode:
            main.conf#/Renderer/profile/vehicle/ra
        :param yellow_speed:
        :param red_speed:

        :raise JnodeError:
        """
        modes = [cls._MODE_NEVER, cls._MODE_IF_PLATE, cls._MODE_ALWAYS]

        # unpack #/target
        target_mode = jnode.get('target', cls._MODE_NEVER)
        if target_mode not in modes:
            eventlog.log_e_param('#/target')
            raise JnodeError

        # unpack #/vehicle
        vehicle_mode = jnode.get('vehicle', cls._MODE_NEVER)
        if vehicle_mode not in modes:
            eventlog.log_e_param('#/vehicle')
            raise JnodeError

        # unpack #/speed
        speed_mode = jnode.get('speed', cls._MODE_NEVER)
        if speed_mode not in modes:
            eventlog.log_e_param('#/speed')
            raise JnodeError

        # unpack #/plate
        dir_mode = jnode.get('dir', cls._MODE_NEVER)
        if dir_mode not in modes:
            eventlog.log_e_param('#/dir')
            raise JnodeError

        # unpack #/len
        len_mode = jnode.get('len', cls._MODE_NEVER)
        if len_mode not in modes:
            eventlog.log_e_param('#/len')
            raise JnodeError

        return RaAnnoArgs(yellow_speed, red_speed, len_mode, dir_mode, target_mode, speed_mode, vehicle_mode,
                          radar_present=True)


class RendererArgs:
    def __init__(self, lr_anno: LrAnnoArgs, ra_anno: RaAnnoArgs):
        self.lr_anno = lr_anno
        self.ra_anno = ra_anno

    @classmethod
    def unpack(cls, jnode: Dict, yellow_speed: int, red_speed: int) -> Self:
        """
        :param jnode: main.conf#/Renderer/profile/vehicle
        :param yellow_speed:
        :param red_speed:
        """
        lr_jnode = jnode.get('lr')
        lr_anno = LrAnnoArgs.unpack(lr_jnode) if lr_jnode else LrAnnoArgs()

        ra_jnode = jnode.get('ra')
        if ra_jnode:
            ra_anno = RaAnnoArgs.unpack(ra_jnode, yellow_speed, red_speed)
        else:
            ra_anno = RaAnnoArgs(yellow_speed, red_speed, radar_present=False)

        return RendererArgs(lr_anno, ra_anno)


def decode__lane_service_conf(jnode: Dict) -> LaneServiceConf:
    """
    :param jnode:
        main.conf#/road-mask
    """
    lanes_jnode: List[Dict] = jnode.get('lanes', [])
    lane_to_mask_list: List[LaneConf] = [(x['name'], x['mask']) for x in lanes_jnode]
    mask_w, mask_h = jnode.get('scene-wh', [0, 0])

    return LaneServiceConf(lane_to_mask_list, mask_w, mask_h)


class KernelConf:
    def __init__(self,
                 react: ReactArgs,
                 event: EventArgs,
                 image_store: ImStoreArgs,
                 renderer: RendererArgs,
                 lane_service: LaneServiceConf,
                 lr_store_tout: int,
                 tz: TzServiceConf,
                 gpsd: GnssServiceConf,
                 footer: ImFooterConf,
                 monitoring: StatsServiceArgs,
                 im_loader: ImLoaderArgs,
                 about: AboutServiceConf,
                 tl_service_args: Optional[JsonTlServiceArgs],
                 landmark_args: Optional[CamServiceConf] = None):

        self.lane = lane_service
        self.react = react
        self.scene_event = event
        self.image_store = image_store
        self.renderer = renderer
        self.lr_store_tout = lr_store_tout
        self.tz = tz
        self.gpsd = gpsd
        self.footer = footer
        self.monitoring = monitoring
        self.im_loader = im_loader
        self.about = about
        self.tl = tl_service_args
        self.landmark_args = landmark_args


def decode__kernel_conf(jnode: Dict, log_level: str) -> KernelConf:
    """
    :param jnode:
        main.conf#/
    :param log_level:

    :raise JnodeArgs:
    """
    lane = decode__lane_service_conf(jnode.get('road-mask', {}))

    rn_jnode = jnode['Renderer']
    profile_jnode = rn_jnode['profile']
    footer = decode__im_footer_conf(profile_jnode['footer'])

    react = ReactArgs.unpack(jnode['Source'])
    event = EventArgs.unpack(jnode['scene-event'], log_level)

    im_loader = ImLoaderArgs(footer.height, (profile_jnode['width'], 65535))
    im_store = ImStoreArgs.unpack(jnode['image-store'])

    vehicle_jnode = profile_jnode.get('annotation', {}).get('vehicle', {})

    tl_service_args = JsonTlServiceArgs.deserialize(jnode.get('tl-args'))

    yellow = rn_jnode.get('YellowSpeedThreshold', 65535)
    red = rn_jnode.get('RedSpeedThreshold', 65535)
    renderer = RendererArgs.unpack(vehicle_jnode, yellow, red)

    # define LrStoreTout
    def_lr_store_tout = 1
    lr_store_tout = rn_jnode.get('LrStoreTout', def_lr_store_tout)
    if not (1 <= lr_store_tout <= 60):

        eventlog.log_w('The param LrStoreTout has invalid value %d. The default value %d is used.'
                       % (lr_store_tout, def_lr_store_tout))
        lr_store_tout = def_lr_store_tout

    # deserialize services_jnode
    services_jnode = jnode['services']

    conf_path = services_jnode['origin']
    about = AboutServiceConf(conf_path)

    # datetime service
    conf_path = services_jnode.get('datetime')
    if not conf_path:
        eventlog.log_e_param('#/services/datetime')
        raise JnodeError

    tz = decode__tz_service_conf_file(conf_path)

    # geolocation service
    conf_path = services_jnode.get('geolocation')
    if not conf_path:
        eventlog.log_e_param('#/services/geolocation')
        raise JnodeError

    gpsd = decode__gnss_service_conf_file(conf_path)

    # stat service
    conf_jnode = services_jnode['monitoring']
    stats = StatsServiceArgs.deserialize(conf_jnode)

    landmark_args = None
    if renderer.ra_anno.radar_present:
        cam_jnode = services_jnode.get('cam')
        if cam_jnode is None:
            eventlog.log_e_param('#/services/cam')
            raise JnodeError

        # radar terminal
        landmark_args = decode__cam_service_conf(cam_jnode)

    return KernelConf(react, event, im_store, renderer, lane, lr_store_tout, tz, gpsd, footer,
                      stats, im_loader, about, tl_service_args, landmark_args)
