from typing import Dict

from edge import eventlog
from edge.image.builder import LiveImBuilder, LiveImConf
from edge.support.imloader import ImrefLoader
from edge.image.types.footer import ImFooterVars
from edge.types.image import ImRgba
from edge.services import lane, cam, tz, gnss, react, about

from renderer._args import KernelConf, ImLoaderArgs
from renderer._color_store import LrColorToutStore
from renderer._event import SceneEventConsumer
from renderer._models import TrScene
from renderer._renderer import Renderer
from renderer._stats import init_service as init_monitoring, fini_service as fini_monitoring, IStatService
from renderer.services.tl import init_service
from renderer.services.types.tl import ITlService


class _Services:
    def __init__(self,
                 tz_service: tz.ITzService,
                 gps_service: gnss.IGnssService,
                 stat_service: IStatService,
                 react_service: react.IReactService,
                 about_service: about.IAboutService,
                 tl_service: ITlService):

        self.tz = tz_service
        self.gps = gps_service
        self.stat = stat_service
        self.react = react_service
        self.about = about_service
        self.tl = tl_service


class BaseKernel:
    def __init__(self,
                 event_consumer: SceneEventConsumer,
                 image_renderer: Renderer,
                 im_engine: LiveImBuilder,
                 services: _Services,
                 im_loader_args: ImLoaderArgs,
                 twisted_reactor):

        self._event_consumer = event_consumer
        self._image_renderer = image_renderer

        self._im_engine = im_engine
        self._services = services
        self._services.react.set_handler(self.handle_react)
        self._footer_height = im_loader_args.dh

        self._im_loader = ImrefLoader(im_loader_args.image_wh[0], im_loader_args.image_wh[1])
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._im_loader.destroy)
        self._rgba_image = ImRgba()

        #
        self._react_ts = 0

    # noinspection PyUnusedLocal
    def handle_react(self, react_name: str, react_jnode: Dict) -> None:
        # noinspection PyBroadException
        tr_scene = TrScene.unpack(react_jnode)
        try:
            self._im_loader.load_rgba(self._rgba_image, tr_scene.im_ref, self._footer_height)
        except ValueError as x:
            eventlog.log_e(f'fail: load image {tr_scene.im_ref.url} <{str(x)}> ')
            return

        react_ts = tr_scene.im_ref.ts
        if react_ts > self._react_ts:
            self._react_ts = react_ts

            origin = self._services.about.get_origin(self._rgba_image.ts)
            im_vars = ImFooterVars(
                self._services.tz, self._services.about, self._rgba_image.ts, defvars={},
                geolocation=origin.geolocation, location=origin.location)
            self._image_renderer.render_tr_scene(tr_scene, self._rgba_image, self._services.tl)
            self._im_engine.build_scene(self._rgba_image, im_vars)
            self._event_consumer.consume_scene(react_ts, self._rgba_image)
            self._services.stat.wh.im_size(self._rgba_image.w, self._rgba_image.h, self._rgba_image.dh)


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, conf: KernelConf):
        lane_service = lane.init_service(conf.lane)

        react_service = react.init_service(twisted_reactor, set(conf.react.endpoints))

        tz_service = tz.init_service(conf.tz)
        gpsd_service = gnss.init_service(twisted_reactor, conf.gpsd, tz_service)
        stat_service = init_monitoring(twisted_reactor, conf.monitoring)

        about_service = about.init_service(conf.about, gpsd_service)

        tl_service = init_service(twisted_reactor, conf.tl)

        cam_service = None
        if conf.landmark_args is not None:
            cam_service = cam.init_service(conf.landmark_args)

        services = _Services(tz_service, gpsd_service, stat_service, react_service, about_service, tl_service)

        event_consumer = SceneEventConsumer.create(twisted_reactor, conf.scene_event, conf.image_store)

        color_tout_store = LrColorToutStore(twisted_reactor, conf.lr_store_tout)

        renderer = Renderer(conf.renderer, color_tout_store, lane_service, cam_service)

        live_args = LiveImConf(conf.footer)
        im_engine = LiveImBuilder(live_args)

        super().__init__(event_consumer, renderer, im_engine, services, conf.im_loader, twisted_reactor)

        self._twisted_reactor = twisted_reactor

        def shutdown() -> None:
            fini_monitoring(stat_service)

            gpsd_service.destroy()
            tz_service.destroy()
            react_service.destroy()
            about_service.destroy()
            tl_service.destroy()
            if cam_service:
                cam_service.destroy()
            lane_service.destroy()

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', shutdown)

    def run(self) -> None:
        self._twisted_reactor.run()
