from typing import Dict, List, Self, Tuple, Union

from edge.types.jnode.imageref import ImageRef, decode__image_ref
from edge.types.shape.quad import QuadI, decode__quadi


class RaMargin:
    def __init__(self,
                 left: Union[float, int],
                 top: Union[float, int],
                 right: Union[float, int],
                 bottom: Union[float, int]):

        self.left = float(left)
        self.right = float(right)
        self.top = float(top)
        self.bottom = float(bottom)

    @classmethod
    def unpack(cls, jnode: List[float]) -> Self:
        """
        :param jnode:
            TrScene#/Ra-ltrb
        """
        if len(jnode) != 4:
            raise ValueError

        return cls(*jnode)


class FaceBbox:
    def __init__(self, x0: int, y0: int, x1: int, y1: int):
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1

    @classmethod
    def unpack(cls, jnode: List[float]) -> Self:
        if len(jnode) != 4:
            raise RuntimeError

        return cls(*jnode)


class LrSceneItem:
    def __init__(self, lr_oid: int, lr_dir: int, lr_text: str, lr_quad: QuadI, lr_trace: List):
        self.lr_oid = lr_oid
        self.lr_dir = lr_dir
        self.lr_text = lr_text
        self.lr_quad = lr_quad
        self.lr_trace = lr_trace

    def map_bbox_ref(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        return self.lr_quad.map_bbox_ref(source_im_w, source_im_h, target_im_w, target_im_h)

    def map_bbox_ref2(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        return self.lr_quad.map_bbox_ref2(source_im_w, source_im_h, target_im_w, target_im_h)

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/Lr/:index
            TrScene#/LrRa/:index/Lr
        """
        lr_plate_jnode = jnode['Plate']

        lr_oid = jnode['Id']
        lr_dir = jnode['Dir']
        lr_text = lr_plate_jnode['P']
        lr_quad = decode__quadi(lr_plate_jnode['Q'])
        lr_trace = jnode['Trace']

        return cls(lr_oid, lr_dir, lr_text, lr_quad, lr_trace)


class FaceSceneItem:
    def __init__(self, face_id: int, face_conf: float, face_bbox: FaceBbox):
        self.face_id = face_id
        self.face_conf = face_conf
        self.face_bbox = face_bbox

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/Lr/:index
            TrScene#/LrRa/:index/Lr
        """
        face_id = jnode['Id']
        face_conf = jnode['Conf']
        face_bbox = FaceBbox.unpack(jnode['R'])

        return cls(face_id, face_conf, face_bbox)


class RaSceneItem:
    def __init__(self, rel_im_x: float, rel_im_y: float, ra_speed: int, ra_len: int, ra_ltrb: RaMargin):
        self.ra_rel_im_x = rel_im_x
        self.ra_rel_im_y = rel_im_y

        self.ra_speed = ra_speed
        self.ra_len = ra_len
        self.ra_ltrb = ra_ltrb

    @property
    def ra_dir(self) -> int:
        if self.ra_speed > 0:
            return 12
        elif self.ra_speed < 0:
            return 11
        else:
            return 1

    def map_bbox_ref(self, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        abs_x = self.ra_rel_im_x * target_im_w
        abs_y = self.ra_rel_im_y * target_im_h
        return int(abs_x), int(abs_y)

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/Ra/:index
            TrScene#/LrRa/:index/Ra
        """
        rel_im_x, rel_im_y = jnode['xy']
        ra_speed = int(jnode['speed'])
        ra_len = int(jnode['len'])
        ra_ltrb = RaMargin.unpack(jnode['r'])

        return cls(float(rel_im_x), float(rel_im_y), ra_speed, ra_len, ra_ltrb)


class TrSceneItem:
    def __init__(self, lr_item: LrSceneItem, ra_item: RaSceneItem):
        self.lr_item = lr_item
        self.ra_item = ra_item

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/LrRa/:index
        """
        lr_item = LrSceneItem.unpack(jnode['Lr'])
        ra_item = RaSceneItem.unpack(jnode['Ra'])
        return cls(lr_item, ra_item)


class TrScene:
    def __init__(self,
                 im_ref: ImageRef,
                 tr_items: List[TrSceneItem],
                 lr_items: List[LrSceneItem],
                 ra_items: List[RaSceneItem],
                 face_items: List[FaceSceneItem]):

        self.im_ref = im_ref

        self.tr_items = tr_items
        self.lr_items = lr_items
        self.ra_items = ra_items
        self.face_items = face_items

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/
        """
        im_ref = decode__image_ref(jnode['Image'])

        tr_items = [TrSceneItem.unpack(item_jnode) for item_jnode in jnode.get('LrRa', [])]
        lr_items = [LrSceneItem.unpack(item_jnode) for item_jnode in jnode.get('Lr', [])]
        ra_items = [RaSceneItem.unpack(item_jnode) for item_jnode in jnode.get('Ra', [])]
        face_items = [FaceSceneItem.unpack(item_jnode) for item_jnode in jnode.get('Face', [])]

        return cls(im_ref, tr_items, lr_items, ra_items, face_items)
