from twisted.internet import inotify
from twisted.python.filepath import FilePath

from typing import Dict, List, Optional, Self, Tuple

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import JnodeError

from edge.types.timespan import TimeSpan, decode__timespans
from renderer.services.jnode.tl import JsonTlServiceArgs
from renderer.services.types.tl import ITlService


class _StubService(ITlService):
    def __init__(self):
        ITlService.__init__(self)

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        return []

    def get_times(self, ts: int) -> Tuple[Optional[int], Optional[int], Optional[int]]:
        return None, None, None


def _print_ti(string: str, ti_l: List[TimeSpan]):
    print(string)
    for ti in ti_l:
        print((ti.min_ts, '-', ti.max_ts))


class _TrafficLight:
    def __init__(self, ch_id: int,
                 red: List[TimeSpan],
                 amber: List[TimeSpan],
                 green: Optional[List[TimeSpan]] = None):

        self.ch_id = ch_id
        self.red = red
        self.amber = amber
        self.green = green or list()

        self.prohibiting = self.red

    def get_red2(self, ts: int) -> Optional[int]:
        red_spans = [time_span for time_span in self.red if ts in time_span]

        if not len(red_spans):
            # current collor is green
            return None

        # red started before ts
        red_spans: List[TimeSpan] = [span for span in red_spans if span.min_ts < ts]
        if not len(red_spans):
            return None

        # check "just red" instead of "any red"
        # just_red = any_red - any_amber
        any_red_time = ts - red_spans[0].min_ts
        return any_red_time

    def get_green2(self, ts: int) -> Optional[int]:
        green_spans = [time_span for time_span in self.green if ts in time_span]

        if not len(green_spans):
            # current collor is green
            return None

        # green started before ts
        green_spans: List[TimeSpan] = [span for span in green_spans if span.min_ts < ts]
        if not len(green_spans):
            return None

        any_green_time = ts - green_spans[0].min_ts
        return any_green_time

    def get_amber2(self, ts: int) -> Optional[int]:
        amber_spans = [time_span for time_span in self.amber if ts in time_span]

        if not len(amber_spans):
            # current collor is green
            return None

        # green started before ts
        amber_spans: List[TimeSpan] = [span for span in amber_spans if span.min_ts < ts]
        if not len(amber_spans):
            return None

        any_green_time = ts - amber_spans[0].min_ts
        return any_green_time


class TlLog:
    def __init__(self, time: TimeSpan, channels: List[_TrafficLight]):
        self.channels = channels
        self.time = time

    @classmethod
    def deserialize(cls, tl_ident: int, jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        ts_jnode = jnode.get('ts')

        if type(ts_jnode) != list or len(ts_jnode) != 2:
            eventlog.log_e_param('#/ts')
            raise JnodeError

        time_interval = TimeSpan(*ts_jnode)

        traffic_lights_jnode = jnode.get('tl')
        if type(traffic_lights_jnode) != list:
            eventlog.log_e_param('#/tl')

        traffic_lights = []
        for tl_jnode in traffic_lights_jnode:
            if tl_ident != tl_jnode['id']:
                eventlog.log_i('tl_ident != tl_jnode[id]')
                continue

            red = tl_jnode.get('red', [])
            amber = tl_jnode.get('amber', [])
            green = tl_jnode.get('green', [])

            # eventlog.log_i('{} {} {}'.format(len(red), len(amber), len(green)))

            red_intervals = decode__timespans(red)
            amber_intervals = decode__timespans(amber)
            green_intervals = decode__timespans(green)

            traffic_light = _TrafficLight(tl_ident, red_intervals, amber_intervals, green_intervals)
            traffic_lights.append(traffic_light)

        if not traffic_lights:
            eventlog.log_d('fail: duplicate idents found')
            raise JnodeError

        return TlLog(time_interval, traffic_lights)


class _TlService(ITlService):
    def __init__(self, twisted_reactor, args: JsonTlServiceArgs):
        """
        :param twisted_reactor:
            twisted.internet.interfaces.IReactorUNIX
        :param args: JsonTlServiceArgs
        """
        ITlService.__init__(self)

        self._twisted_reactor = twisted_reactor

        self._tl_log: Optional[TlLog] = None
        self._tl_ident = args.tl_ident

        self._endpoint = args.endpoint
        self._inotify = inotify.INotify()
        self._inotify.startReading()
        self._reset_inotify_watch()

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        return []

    def destroy(self) -> None:
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

    def _reset_inotify_watch(self) -> None:
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

        fs.silent_mkdir_r(self._endpoint)

        try:
            self._inotify.watch(
                FilePath(self._endpoint),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_react])
        except inotify.INotifyError as x:
            eventlog.log_d(f'fail: reset inotify watch ({x})')

    # noinspection PyUnusedLocal
    def _handle_react(self, ignored, filepath, mask) -> None:
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_w(f'endpoint was changed to {self._endpoint}')
            return

        try:
            react_jnode = fs.load_json3(filepath.path)
            if react_jnode:
                self._tl_log = TlLog.deserialize(self._tl_ident, react_jnode)

        except JnodeError as x:
            return

        finally:
            fs.silent_rm_r(filepath.path)

    def get_times(self, ts: int) -> Tuple[Optional[int], Optional[int], Optional[int]]:
        """
        :param ts:
            value in usec

        :returns:
            (green-amber-time, red-time, red-amber-time).
            Values in usec
        """
        if self._tl_log is None:
            eventlog.log_d('fail: tl log is absent')
            return None, None, None

        traffic_lights = [ch for ch in self._tl_log.channels if ch.ch_id == self._tl_ident]
        assert 1 == len(traffic_lights)

        tl = traffic_lights[0]  # type: _TrafficLight
        # eventlog.log_d('ts {} gr {} red {} amb {} ( {} {} {})'
        #                .format(ts,
        #                        tl.green[-1].max_ts,
        #                        tl.red[-1].max_ts,
        #                        tl.amber[-1].max_ts,
        #                        (ts - tl.green[-1].min_ts) / 1e6,
        #                        (ts - tl.red[-1].min_ts) / 1e6,
        #                        (ts - tl.amber[-1].min_ts) / 1e6))

        return tl.get_green2(ts), tl.get_red2(ts), tl.get_amber2(ts)


def init_service(twisted_reactor, args: Optional[JsonTlServiceArgs]) -> ITlService:
    """
    :param twisted_reactor:
        twisted.internet.interfaces.IReactorUNIX
    :param args:
    """
    if args is None:
        rl = _StubService()
    else:
        rl = _TlService(twisted_reactor, args)
    return rl


def fini_service(service: ITlService) -> None:
    service.destroy()
