import ujson

from typing import Any, Dict, Final, Optional
from os.path import join as join_path

from edge import eventlog
from edge.services.types.jnode.cam import decode__cam_service_conf
from edge.services.types.cam import CamServiceConf
from edge.services.types.jnode.gnss import decode__gnss_service_conf_file
from edge.services.types.jnode.tz import decode__tz_service_conf
from edge.services.types.about import AboutServiceConf
from edge.services.types.gnss import GnssServiceConf
from edge.services.types.tz import TzServiceConf
from edge.types.exceptions import JnodeError

from speedcontrol.conf.jnode import detectors
from speedcontrol.services.types.jnode import db, road
from speedcontrol.services.types.jnode import trainer, leddar, tl, event, react_tr
from speedcontrol.services.types.jnode.plate import JsonPlateInfoServiceConf
from speedcontrol.services.types.jnode.raas import JsonRaasServiceArgs
from speedcontrol.services.types.jnode.sign import JsonSignServiceArgs
from speedcontrol.services.types.jnode.stats import JsonStatServiceArgs
from speedcontrol.services.types.jnode.systest import JsonSystestServiceConf
from speedcontrol.services.types.bs import BsServiceConf
from speedcontrol.services.types.raas import RaasServiceConf
from speedcontrol.services.types.sign import SignServiceConf
from speedcontrol.services.types.stats import StatsServiceConf
from speedcontrol.services.types.tl import TlServiceConf
from edge.services.types.vs import VsServiceConf


class KernelConf:
    def __init__(self,
                 road_conf: road.RoadConf,
                 trainer_conf: trainer.TrainerServiceConf,
                 tr_react_conf: react_tr.TrReactServiceConf,
                 event_conf: event.EventServiceConf,
                 stats_conf: StatsServiceConf,
                 tz_conf: TzServiceConf,
                 raas_conf: Optional[RaasServiceConf],
                 tl_conf: Optional[TlServiceConf],
                 gnss_conf: GnssServiceConf,
                 about_conf: AboutServiceConf,
                 detectors_conf: detectors.DetectorsConf,
                 bs_conf: BsServiceConf,
                 vs_conf: VsServiceConf | None,
                 sign_conf: Optional[SignServiceConf],
                 cam_conf: Optional[CamServiceConf],
                 db_conf: db.DbServiceConf,
                 systest_conf: JsonSystestServiceConf,
                 leddar_conf: Optional[leddar.LeddarServiceConf],
                 plateinfo_conf: JsonPlateInfoServiceConf):

        self.road = road_conf
        self.trainer = trainer_conf

        self.tr_react = tr_react_conf
        self.event = event_conf
        self.db = db_conf

        self.counters = stats_conf

        self.tz = tz_conf
        self.raas = raas_conf
        self.tl = tl_conf
        self.leddar = leddar_conf
        self.gnss = gnss_conf
        self.about = about_conf

        self.detectors = detectors_conf

        self.bs = bs_conf
        self.vs = vs_conf
        self.sign = sign_conf
        self.systest = systest_conf
        self.plateinfo = plateinfo_conf
        self.cam: Final[CamServiceConf | None] = cam_conf


def decode__vs_service_conf(with_videostore: bool, module_id: int, log_level: str) -> VsServiceConf | None:
    if with_videostore:
        tmp_path = join_path('/run', 'edge', 'sc', str(module_id))
        return VsServiceConf(module_id, tmp_path, log_level=log_level)


def decode__kernel_conf(main_conf_jnode: Dict, log_level: str, module_id: int) -> KernelConf:
    # define services
    services_jnode = main_conf_jnode.get('services')
    if type(services_jnode) != dict:
        eventlog.log_e_param('#/services')
        raise JnodeError

    # define stat-counters-args
    monitoring_jnode = services_jnode.get('stat')
    if not monitoring_jnode:
        raise ValueError('#/services/stat')

    # define blobstore args
    video_conf = decode__vs_service_conf(services_jnode.get('vs', True), module_id, log_level)

    stats = JsonStatServiceArgs.deserialize(monitoring_jnode)
    systest = JsonSystestServiceConf.deserialize(main_conf_jnode)
    template = JsonPlateInfoServiceConf.deserialize(main_conf_jnode)

    # define def-footer-mapping
    tz_conf_filepath = services_jnode['tz.conf']
    with open(tz_conf_filepath) as tz_conf_jnode:
        tz_conf_jnode = ujson.load(tz_conf_jnode)

    tz = decode__tz_service_conf(tz_conf_jnode)

    # define road args
    road_conf_jnode = {}
    road_conf_filepath = services_jnode.get('road.conf')
    if road_conf_filepath:
        with open(road_conf_filepath) as jnode:
            road_conf_jnode = ujson.load(jnode)

    road_conf = road.decode__road_conf(road_conf_jnode)

    # define origin args
    conf_path = services_jnode.get('origin')
    about = AboutServiceConf(conf_path)

    gpsd = decode__gnss_service_conf_file(services_jnode['gpsd'])

    def_footer_mapping: Dict[str, Any] = {
        'tz': tz.tz_info,
        'post-name': '',
        'geolocation': '',
        'plate': '',
        'speed': '-',
        'approaching': road_conf.dir_alias_a,
        'receding': road_conf.dir_alias_r,
        'unknown': road_conf.dir_alias_u,
        'video-speed': '-',
        'video-speed2': '-',
        'radar-speed': '-',
        'radar-speed2': '',
        'max-speed': '-',
        'speed-limit': '-',
        'radar-overspeed': '-'
    }

    # define db args
    db_conf_abspath = services_jnode.get('db.conf')
    if not db_conf_abspath:
        eventlog.log_e('fail: param #/services/db.conf has invalid value')
        raise JnodeError

    try:
        with open(db_conf_abspath) as db_conf:
            db_conf_jnode = ujson.load(db_conf)

    except Exception as x:
        eventlog.log_e(f'fail: load {db_conf_abspath} ({x})')
        raise JnodeError

    db_conf = db.decode__db_service_conf(module_id, db_conf_jnode)

    # define detectors args
    detectors_jnode = main_conf_jnode.get('detectors', {})
    detectors_conf = detectors.decode__detectors_conf(detectors_jnode, def_footer_mapping)

    # define blob-store-args
    blob_store = BsServiceConf(module_id, log_level)

    # define others
    trainer_jnode = services_jnode.get('trainer')
    trainer_conf = trainer.decode__trainer_service_conf(trainer_jnode) if trainer_jnode else None

    event_conf = event.decode__event_service_conf(services_jnode['event'])
    leddar_conf = leddar.decode__leddar_service_conf(services_jnode.get('leddar', None))
    raas_conf = JsonRaasServiceArgs.deserialize(services_jnode.get('raas', None))
    react_conf = react_tr.decode__tr_react_service_conf(main_conf_jnode)
    tl_conf = tl.decode__tl_service_conf(services_jnode.get('tl', None))

    # define sign service args
    sign_args = None
    sign_jnode = services_jnode.get('sign')
    if sign_jnode:
        sign_args = JsonSignServiceArgs.deserialize(sign_jnode, log_level)

    cam: Optional[CamServiceConf] = decode__cam_service_conf(services_jnode.get('cam'))

    kernel_args = KernelConf(
        road_conf, trainer_conf, react_conf, event_conf,
        stats, tz, raas_conf, tl_conf, gpsd, about, detectors_conf, blob_store, video_conf,
        sign_args, cam, db_conf, systest, leddar_conf, template)

    return kernel_args
