import functools
import math

from typing import Callable, Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.image.scaler import WhScaler
from edge.support.mask import Mask
from edge.support.service import ServicePack
from edge.types.blob import FileBlob, BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.shape.quad import QuadI
from edge.types.timespan import TimeSpan
from edge.types.track import TrTrackItem

from speedcontrol.detectors.base.best import check_quad_dist_from_img_borders
from speedcontrol.detectors.buslane.rule import Rule
from speedcontrol.detectors.buslane.types.deps import DetectorDeps
from speedcontrol.detectors.buslane.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.buslane import Evidence
from edge.services.types.vs import VsSessionAbsentError, VsError
from speedcontrol.trackset.trackset import Trackset


def _test_points_far_enought(quad1: QuadI, quad2: QuadI) -> bool:
    _, plate_h = quad1.wh
    max_distance = plate_h * 0.7  # % of plate height

    rect1_cx, rect1_cy = quad1.center
    rect2_cx, rect2_cy = quad2.center

    if math.fabs(rect1_cx - rect2_cx) < max_distance:
        if math.fabs(rect1_cy - rect2_cy) < max_distance:
            # points are in same rect
            return False

    # points are far enought
    return True


def _filter_tr_items(tr_items: List[TrTrackItem], lr_scene_w: int, lr_scene_h: int) -> List[TrTrackItem]:
    result = []
    # todo: lr_item can be null ?
    init_point: TrTrackItem = tr_items[0]
    result.append(init_point)

    for index in range(1, len(tr_items), 1):
        fini_point: TrTrackItem = tr_items[index]
        if fini_point.lr_item is None:
            continue

        if init_point.lr_item is None:
            continue

        if not check_quad_dist_from_img_borders(fini_point.lr_item.plate_quad, lr_scene_w, lr_scene_h):
            continue

        if _test_points_far_enought(init_point.lr_item.plate_quad, fini_point.lr_item.plate_quad):
            result.append(fini_point)
            init_point = fini_point

    return result


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf

        self._handler: Callable[[Event], None] | None = None

        imbuilder = EventMeasCamImBuilder(conf.im_engine)
        self._ime_init = ImBuilderEnv(imbuilder)
        self._ime_mid = ImBuilderEnv(imbuilder)
        self._ime_fini = ImBuilderEnv(imbuilder)
        self._ime_best = ImBuilderEnv(imbuilder)

        if self._conf.video:
            self._im_fitter = WhScaler(self._conf.video.max_wh)

        self._rules: List[Rule] = []
        for rule_conf in conf.rules:
            if rule_conf.enabled:
                rule = Rule(deps, rule_conf, conf.name,
                            self._ime_init, self._ime_mid, self._ime_fini, self._ime_best,
                            conf.min_ts_dist_usec)
                self._rules.append(rule)

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track: Trackset, base_event: BaseEvent) -> None:
        if '*' in base_event.plate_ucode:
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            return

        if (track.lr_track.max_ts - track.lr_track.min_ts) / 1e6 < 2:
            eventlog.log_i('fail: to short track {} {:6.3} s'.format(
                base_event.plate_utf8, (track.lr_track.max_ts - track.lr_track.min_ts) / 1e6))
            # too short trek
            return None

        sc_detector = Mask(self._conf.scene)

        def get_parts() -> List[TrTrackItem]:
            violation_tritem = []

            lr_scene_img = track.lr_track.imref
            # TODO: rewrite filter
            for tr_item in _filter_tr_items([x for x in track.tr_track.values()], lr_scene_img.width, lr_scene_img.height):
                # for tr_item in track.tr_track:
                if tr_item.lr_item:
                    if not sc_detector.check_quad_abs(tr_item.lr_item.plate_quad):
                        continue

                    violation_tritem.append(tr_item)

            return violation_tritem

        violation_tr_items = get_parts()

        if len(violation_tr_items) < 3:
            return None

        c_v_ts = violation_tr_items[0].ts
        l_v_ts = violation_tr_items[-1].ts

        if (l_v_ts - c_v_ts) / 1e6 < 2:
            eventlog.log_i('fail: to short track {} {:6.3} s'.format(
                base_event.plate_utf8, (l_v_ts - c_v_ts) // 1e6))
            return None

        for rule in self._rules:
            try:
                reply: Optional[Tuple[Evidence, TimeSpan]] = rule.check(track, base_event, violation_tr_items)
                if reply is None:
                    continue
            except GenericError:
                eventlog.log_w('fail')
                continue

            evidence, time_span = reply
            handler = functools.partial(self.handle_evidence, base_event, evidence)
            if self._conf.video:
                scaled_w, scaled_h = self._im_fitter.scaledown(self._ime_init.scene_rgba.wh)

                # noinspection PyBroadException
                try:
                    return self._deps.vs.export_video(
                        time_span.min_ts // 1000 - self._conf.video.pre_record,
                        time_span.max_ts // 1000 + self._conf.video.post_record,
                        scaled_w, scaled_h, self._conf.video.quality,
                        handler)

                except VsSessionAbsentError:
                    eventlog.log_w('fail: VsSessionAbsentError')

                except VsError:
                    eventlog.log_w('fail: VsError')

                except GenericError:
                    eventlog.log_w('fail')

                except Exception:
                    eventlog.log_x()

            handler(None)
            break

    def handle_evidence(self, base: BaseEvent, evidence: Evidence, video_blob: Optional[FileBlob]) -> None:
        if video_blob is not None:
            evidence.add_videos([video_blob])

        self._handler(Event(self._conf.identity, base, evidence))


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detectors.plate_lost.add(_Detector(DetectorDeps.create(services), conf))
        except GenericError:
            eventlog.log_e('fail')
            raise
