from typing import Final, List, Optional, Union, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.image.types.footer import ImFooterVars
from edge.support.schedule import Schedule
from edge.types.dir import Dir
from edge.types.jnode.dir import encode_dir_text
from edge.types.timespan import TimeSpan
from edge.types.track import TrTrackItem

from speedcontrol.detectors.buslane.types.conf import RuleConf
from speedcontrol.detectors.buslane.types.deps import RuleDeps
from speedcontrol.events.detector import RuleInfo, DetectorInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.buslane import Evidence
from speedcontrol.events.schedule import ScheduleInfo
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import Vehicle, ViolationVehicle
from speedcontrol.events.waypoint import Waypoint
from speedcontrol.events.waypoint.measurement import BestWaypoint
from speedcontrol.events.waypoint.violation import ViolationWaypoint
from edge.services.types.vs import VsImageInfo
from speedcontrol.trackset.trackset import Trackset


class _SceneMsRef(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem):
        super().__init__(tr_item.lr_item.ts)
        self.tr_item = tr_item


def _create_detector_info(detector_name: str, conf: RuleConf) -> DetectorInfo:
    schedule_info = ScheduleInfo(
        conf.schedule.min_time,
        conf.schedule.max_time,
        conf.schedule.days_to_bool)

    ticket_info = TicketInfo(conf.ticket.name)

    rule_info = RuleInfo(conf.name, schedule_info, ticket_info, {
        'types': conf.limits.vehicle_types
    })

    return DetectorInfo(detector_name, rule_info)


class Rule:
    def __init__(self,
                 deps: RuleDeps,
                 conf: RuleConf,
                 detector_name: str,
                 ime_init: ImBuilderEnv,
                 ime_mid: ImBuilderEnv,
                 ime_fini: ImBuilderEnv,
                 ime_best: ImBuilderEnv,
                 min_ts_dist_usec: Union[int, float] = 1e6):

        self._deps = deps
        self._conf = conf
        self._info = _create_detector_info(detector_name, conf)

        self._ime_init: Final[ImBuilderEnv] = ime_init
        self._ime_mid: Final[ImBuilderEnv] = ime_mid
        self._ime_fini: Final[ImBuilderEnv] = ime_fini
        self._ime_best: Final[ImBuilderEnv] = ime_best

        self._schedule = Schedule(conf.schedule, self._deps.tz)
        self._min_ts_dist_usec = min_ts_dist_usec

    def check(self,
              track: Trackset,
              base: BaseEvent,
              violation_tr_items: List[TrTrackItem]) -> Optional[Tuple[Evidence, TimeSpan]]:

        if base.best_wp.vehicle.mmr_class not in self._conf.limits.vehicle_types:
            return None

        # print 'rule %s' % self._args.name
        c_v_ts = violation_tr_items[0].ts
        l_v_ts = violation_tr_items[-1].ts

        if not (self._schedule.check_ts_usec(l_v_ts) or self._schedule.check_ts_usec(c_v_ts)):
            return None

        scene_ms_refs: List[_SceneMsRef] = list(map(_SceneMsRef, violation_tr_items))
        init_scene_ms_refs = scene_ms_refs[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)
        fini_scene_ms_refs = scene_ms_refs[:-6:-1]

        if track.lr_track.dir == Dir.CODE_A:
            best_scene_ms_refs = fini_scene_ms_refs
        elif track.lr_track.dir == Dir.CODE_R:
            best_scene_ms_refs = init_scene_ms_refs
        else:
            best_scene_ms_refs = []

        center_ts = (c_v_ts + l_v_ts) * 0.5

        mid_scene_ms_refs = [it for it in scene_ms_refs if center_ts < it.ts < l_v_ts - self._min_ts_dist_usec]

        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)

        init_winner_img = self._deps.vs.export_image_rgba(
            self._ime_init.scene_rgba, init_scene_ms_refs, self._ime_init.scene_dh)

        mid_winner_img = self._deps.vs.export_image_rgba(
            self._ime_mid.scene_rgba, mid_scene_ms_refs, self._ime_mid.scene_dh)

        fini_winner_img = self._deps.vs.export_image_rgba(
            self._ime_fini.scene_rgba, fini_scene_ms_refs, self._ime_fini.scene_dh)

        best_winner_img = self._deps.vs.export_image_rgba(
            self._ime_best.scene_rgba, best_scene_ms_refs, self._ime_best.scene_dh)

        if None in [init_winner_img, fini_winner_img, mid_winner_img, best_winner_img]:
            eventlog.log_i('fail: get init/fini/mid/best image from vs')
            return None

        if mid_winner_img.ts - init_winner_img.ts < self._min_ts_dist_usec:
            eventlog.log_i('fail: too short p-p interval (plate:{}, mid-to-inif:{:6.3}sec'.format(
                base.plate_utf8, (mid_winner_img.ts - init_winner_img.ts) / 1e6))
            return None

        if fini_winner_img.ts - mid_winner_img.ts < self._min_ts_dist_usec:
            eventlog.log_i('fail: too short p-p interval (plate:{}, fini-to-mid:{:6.3}sec'.format(
                base.plate_utf8, (fini_winner_img.ts - mid_winner_img.ts) / 1e6))
            return None

        return self._mk_evidence(base, track,
                                 init_winner_img.tr_item,
                                 mid_winner_img.tr_item,
                                 fini_winner_img.tr_item,
                                 best_winner_img.tr_item)

    def _mk_evidence(self,
                     base_event: BaseEvent,
                     trackset: Trackset,
                     init_tr_item: TrTrackItem,
                     mid_tr_item: TrTrackItem,
                     fini_tr_item: TrTrackItem,
                     best_tr_item: TrTrackItem) -> Optional[Tuple[Evidence, TimeSpan]]:

        best_waypoint: BestWaypoint = base_event.best_wp
        best_vehicle: Vehicle = best_waypoint.vehicle

        def mk_waypoint(
                im_env: ImBuilderEnv,
                tr_item: TrTrackItem,
                w_type: str = Waypoint.NAME_VIOLATION) -> ViolationWaypoint:
            """
            :raise RuntimeError:
            """
            waypoint_origin = self._deps.about.get_origin(im_env.scene_rgba.ts)

            if tr_item.lane_item is not None:
                waypoint_origin.location.set_lane(tr_item.lane_item.lane)

            waypoint_ts = im_env.scene_rgba.ts
            wp_direction = trackset.tr_track.get_item_dir(tr_item.ts)
            wp_dir_alias = self._conf.footer_defvars.get(encode_dir_text(wp_direction))

            footer_vars = ImFooterVars(
                self._deps.tz,
                self._deps.about,
                waypoint_ts,
                defvars=self._conf.footer_defvars,
                location=waypoint_origin.location,
                geolocation=waypoint_origin.geolocation,
                plate_text=base_event.plate_ucode,
                spot_speed=tr_item.ra_item and tr_item.ra_item.speed_kmh,
                direction=wp_direction,
                spot_speed_ra_n=best_vehicle.spotspeeds.ra_n,
                spot_speed_cam_2=best_vehicle.spotspeeds.cam_2,
                spot_speed_cam_n=best_vehicle.spotspeeds.cam_n,
                spot_speed_wim=best_vehicle.spotspeeds.wim,
                spot_speed_main=best_vehicle.spotspeeds.main,
                spot_speed_control=best_vehicle.spotspeeds.control)

            if w_type == Waypoint.NAME_VIOLATION:
                scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(waypoint_ts)
                im_blobs = im_env.builder.build(
                    im_env.scene_rgba, scene_dt_utc, waypoint_origin.geolocation,
                    footer_vars, trackset.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True)

                if len(im_blobs) != 3:
                    eventlog.log_i('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint_v = ViolationVehicle(
                    direction=wp_direction,
                    direction_alias=wp_dir_alias,
                    plate_quads=[tr_item.lr_item.plate_quad])

                waypoint = ViolationWaypoint(
                    Waypoint.NAME_VIOLATION,
                    waypoint_ts, None, waypoint_origin.location, im_blobs,
                    waypoint_v)

            elif w_type == Waypoint.NAME_VIOLATION_BEST:
                scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(waypoint_ts)
                im_blobs = im_env.builder.build(
                    im_env.scene_rgba, scene_dt_utc, waypoint_origin.geolocation,
                    footer_vars, trackset.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True, thumb=True)

                if len(im_blobs) != 4:
                    eventlog.log_i('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint_v = ViolationVehicle(
                    direction=wp_direction,
                    direction_alias=wp_dir_alias,
                    plate_quads=[tr_item.lr_item.plate_quad])

                waypoint = ViolationWaypoint(
                    Waypoint.NAME_VIOLATION_BEST,
                    waypoint_ts, None, waypoint_origin.location, im_blobs,
                    waypoint_v)

            else:
                eventlog.log_e('fail: wp not supported')
                raise RuntimeError

            return waypoint

        try:
            init_wp = mk_waypoint(self._ime_init, init_tr_item)
            mid_wp = mk_waypoint(self._ime_mid, mid_tr_item)
            fini_wp = mk_waypoint(self._ime_fini, fini_tr_item)
            best_wp = mk_waypoint(self._ime_best, best_tr_item, Waypoint.NAME_VIOLATION_BEST)
        except RuntimeError:
            return None

        init_ts = init_wp.main_ts
        fini_ts = fini_wp.main_ts
        event_ts = fini_ts
        duration = int((fini_ts - init_ts) // 1000000)

        evidence = Evidence(event_ts, self._deps.about.get_origin(event_ts), self._info, duration)
        evidence.add_waypoints([
            init_wp,
            mid_wp,
            fini_wp,
            best_wp
        ])

        return evidence, TimeSpan(init_ts, fini_ts)
