import math
import time

from datetime import datetime, timedelta
from typing import Callable, Final, List

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.support.mask import Mask
from edge.support.service import ServicePack
from edge.types.blob import BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.shape.rect import RectF
from edge.types.track import LrTrackItem
from edge.types.trend import LrTrend
from edge.types.jnode.trend import encode__lr_trend, decode__lr_trend

from speedcontrol.detectors.parking.rule import Rule
from speedcontrol.detectors.parking.types.conf import DetectorConf
from speedcontrol.detectors.parking.types.deps import DetectorDeps
from speedcontrol.detectors.parking.types.jnode.conf import decode__detector_conf
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.trackset.trackset import Trackset


def get_trace_duration(trace: List[LrTrackItem]) -> int:
    return trace[-1].ts - trace[0].ts


def check_rects(rect1: RectF, rect2: RectF) -> bool:
    podgonian = 2  # TODO
    rect1_c_y = (rect1.ft + rect1.fb) / 2
    rect1_c_x = (rect1.fr + rect1.fl) / 2

    rect2_c_y = (rect2.ft + rect2.fb) / 2
    rect2_c_x = (rect2.fr + rect2.fl) / 2

    dist = math.sqrt(math.pow(rect1_c_x - rect2_c_x, 2) + math.pow(rect1_c_y - rect2_c_y, 2))
    rect_1_h = rect1.fb - rect1.ft
    return dist < podgonian * rect_1_h


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf

        self._scene_mask: Final[Mask] = Mask(conf.scene)

        imbuilder = EventMeasCamImBuilder(conf.init_imbuilder)
        self._ime_init = ImBuilderEnv(imbuilder)
        self._ime_fini = ImBuilderEnv(imbuilder)

        self._rules: List[Rule] = []
        for rule_conf in conf.rules:
            rule = Rule(conf.name, deps, rule_conf, self._ime_init, self._ime_fini)
            self._rules.append(rule)

        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], None]):
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track: Trackset, base_event: BaseEvent) -> None:
        if '*' in base_event.plate_ucode:
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            return

        area_items = [lr_item for lr_item in track.lr_track if self._scene_mask.check_quad_abs(lr_item.plate_quad)]

        if not area_items:
            return

        # 1 load logs with plate-dist = 1
        # 1.1 if the is an result-entry with dist == 0, prolong it
        # 1.2 else create the new one
        # 2 try to merge result-entries

        answer = self._deps.db.get_log(int(self._conf.identity.chan), base_event.plate_ucode, plate_dist=0)
        # print 'db.get-log: %s, log-count=%s' % (base.get_plate_text(), len(log_jnodes))
        trend = decode__lr_trend(answer[0][2]) if answer else LrTrend()
        trend.prolong(area_items, self._conf.log_range_tout, self._conf.log_lost_tout)
        trend_copy = trend.retrend()

        for rule in self._rules:
            evidence = rule.check(track, base_event, trend_copy)
            if evidence:
                trend.set_event(evidence.main_ts)
                self._handler(Event(self._conf.identity, base_event, evidence))
                break

        self._deps.db.set_log(
            int(self._conf.identity.chan), base_event.plate_ucode,
            trend.min_ts, trend.max_ts, encode__lr_trend(trend))

        # ts = (time.time() + 60) * 1e6  # (now + 1 min) in microseconds
        # max_ts = int(ts - self._log_lost_tout)
        max_datetime = datetime.now() - timedelta(days=1)
        max_ts = int(time.mktime(max_datetime.timetuple()) * 1e6)

        self._deps.db.clear_log(int(self._conf.identity.chan), max_ts)


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detectors.plate_lost.add(_Detector(DetectorDeps.create(services), conf))
        except GenericError:
            eventlog.log_e('fail')
            raise
