from typing import Final, Optional, List

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.image.types.footer import ImFooterVars
from edge.services.types.vs import VsImageInfo
from edge.types.blob import MemBlob
from edge.types.origin import Origin
from edge.types.shape.quad import QuadI
from edge.types.trend import LrTrend, LrTrendItem

from edge.support.schedule import Schedule
from speedcontrol.detectors.parking.types.conf import RuleConf
from speedcontrol.detectors.parking.types.deps import RuleDeps
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.parking import Evidence
from speedcontrol.events.schedule import ScheduleInfo
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import ViolationVehicle
from speedcontrol.events.waypoint.violation import ViolationWaypoint
from speedcontrol.trackset.trackset import Trackset


def _create_detector_info(detector_name: str, conf: RuleConf) -> DetectorInfo:
    schedule_info = ScheduleInfo(
        conf.schedule.min_time,
        conf.schedule.max_time,
        conf.schedule.days_to_bool)

    ticket_info = TicketInfo(conf.ticket.name)

    rule_info = RuleInfo(conf.name, schedule_info, ticket_info, {
        'max-standing-time': conf.limits.max_standing_time
    })

    return DetectorInfo(detector_name, rule_info)


class Rule:
    def __init__(self,
                 detector_name: str,
                 deps: RuleDeps,
                 conf: RuleConf,
                 ime_init: ImBuilderEnv,
                 ime_fini: ImBuilderEnv):

        self._deps: Final[RuleDeps] = deps
        self._conf: Final[RuleConf] = conf
        self._info: Final[DetectorInfo] = _create_detector_info(detector_name, conf)
        self._schedule: Final[Schedule] = Schedule(conf.schedule, deps.tz)

        self._ime_init = ime_init
        self._ime_fini = ime_fini

    def check(self, trackset: Trackset, base_event: BaseEvent, trend: LrTrend) -> Optional[Evidence]:
        for item in trend:
            # check constraints
            if item.event:
                continue

            elif item.duration_sec < self._conf.limits.max_standing_time:
                continue

            elif not self._schedule.check_ts_usec(item.min_ts):
                continue

            elif not self._schedule.check_ts_usec(item.max_ts):
                continue

            # load init-image
            im_info = self._deps.vs.export_image_rgba(self._ime_init.scene_rgba, [VsImageInfo(item.min_ts)],
                                                      self._ime_init.scene_dh)
            if im_info is None:
                eventlog.log_i('fail: load init-image')
                continue

            # load fini-image
            im_info = self._deps.vs.export_image_rgba(self._ime_fini.scene_rgba, [VsImageInfo(item.max_ts)],
                                                      self._ime_fini.scene_dh)
            if im_info is None:
                eventlog.log_i('fail: load fini-image')
                continue

            plate_bbox = trend.get_vrp_bbox(item)
            plate_quad = QuadI.create(plate_bbox)

            evidence = self._mk_evidence(trackset, base_event, item, plate_quad)
            if evidence:
                return evidence

    def _mk_evidence(self,
                     trackset: Trackset,
                     base_event: BaseEvent,
                     item: LrTrendItem,
                     item_plate_quad: QuadI) -> Optional[Evidence]:

        assert item.min_ts == self._ime_init.scene_rgba.ts
        assert item.max_ts == self._ime_fini.scene_rgba.ts

        def mk_waypoint_blobs(env: ImBuilderEnv, origin: Origin) -> List[MemBlob]:
            footer_vars = ImFooterVars(
                self._deps.tz, self._deps.about, env.scene_rgba.ts,
                defvars=self._conf.footer_defvars,
                location=origin.location,
                geolocation=origin.geolocation,
                plate_text=base_event.plate_ucode)

            scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(env.scene_rgba.ts)
            blobs = env.builder.build(
                env.scene_rgba, scene_dt_utc, origin.geolocation,
                footer_vars, trackset.lr_track, item_plate_quad,
                vehicle=True, scene=True, plate=True)

            return blobs

        lane = trackset.tr_track.get_item_lane(item.max_ts) or ''

        init_origin = self._deps.about.get_origin(item.min_ts, lane)
        init_blobs = mk_waypoint_blobs(self._ime_init, init_origin)
        if len(init_blobs) != 3:
            eventlog.log_i('fail: mk-waypoint-blobs <violation>')
            return

        fini_origin = self._deps.about.get_origin(item.max_ts, lane)
        fini_blobs = mk_waypoint_blobs(self._ime_fini, fini_origin)
        if len(fini_blobs) != 3:
            eventlog.log_i('fail: mk-waypoint-blobs <violation>')
            return

        evidence = Evidence(fini_origin, item.max_ts, self._info, item.duration_sec)

        init_waipoint = ViolationWaypoint(
            ViolationWaypoint.NAME_VIOLATION, item.min_ts, None, init_origin.location, init_blobs,
            ViolationVehicle(plate_rects=[item_plate_quad.bbox_recti]))

        fini_waypoint = ViolationWaypoint(
            ViolationWaypoint.NAME_VIOLATION, item.max_ts, None, fini_origin.location, fini_blobs,
            ViolationVehicle(plate_rects=[item_plate_quad.bbox_recti]))

        evidence.add_waypoints([init_waipoint, fini_waypoint])
        return evidence
