import functools

from typing import Callable, Dict, Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.image.scaler import WhScaler
from edge.services.types.vs import VsSessionAbsentError, VsError
from edge.support.mask import IntersectionMask
from edge.support.service import ServicePack
from edge.types.blob import FileBlob, BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.timespan import TimeSpan
from edge.types.track import TrTrackItem

from speedcontrol.detectors.base.best import check_quad_dist_from_img_borders
from speedcontrol.detectors.redlight.rule import Rule
from speedcontrol.detectors.redlight.types.deps import DetectorDeps
from speedcontrol.detectors.redlight.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.detectors.redlight.types.stats import DetectorStats
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.buslane import Evidence
from speedcontrol.trackset.trackset import Trackset


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf
        self._stats: Final[DetectorStats] = DetectorStats()
        self._scene_mask: Final[IntersectionMask] = IntersectionMask(conf.scene)

        imbuilder = EventMeasCamImBuilder(conf.imbuilder)

        self._ime_prebar = ImBuilderEnv(imbuilder)
        self._ime_postbar_init = ImBuilderEnv(imbuilder)
        self._ime_postbar_fini = ImBuilderEnv(imbuilder)

        if self._conf.video:
            self._im_fitter = WhScaler((self._conf.video.max_w, self._conf.video.max_h))

        self._rules: List[Rule] = []

        rule_index = 0
        for rule_conf in conf.rules:
            rule_name = conf.identity.child_ref(IDetector.TYPE_REDLIGHT_RULE, rule_index)
            rule = Rule(deps, rule_conf, conf.name, rule_name,
                        self._ime_prebar, self._ime_postbar_init, self._ime_postbar_fini)

            self._rules.append(rule)
            rule_index += 1

        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = handler

    def analyze(self, trackset: Trackset, base_event: BaseEvent) -> None:
        def split_track(
                sc_detector: IntersectionMask
        ) -> Optional[Tuple[List[TrTrackItem], List[TrTrackItem], List[TrTrackItem]]]:
            lr_scene_img = trackset.lr_track.imref
            tr_items = [item for item in trackset.tr_track.values() if item.lr_item and check_quad_dist_from_img_borders(
                    item.lr_item.plate_quad, lr_scene_img.width, lr_scene_img.height)]

            pre_stopbar, on_stopbar, post_stop_bar = [], [], []
            for tr_item in tr_items:
                check_answer = sc_detector.check_point_abs(tr_item.lr_item.plate_quad.center, self._conf.dir_r)

                if check_answer == sc_detector.STOPBAR_PRE:
                    pre_stopbar.append(tr_item)

                elif check_answer == sc_detector.STOPBAR_ON:
                    if len(pre_stopbar):
                        on_stopbar.append(tr_item)

                    else:
                        # we didn't have pre-stopbar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err in'
                        return None

                elif check_answer == sc_detector.STOPBAR_POST:
                    if len(pre_stopbar) == 0 or len(on_stopbar) == 0:
                        # we didn't have pre-stop-bar or inside-stop-bar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err post'
                        return None

                    post_stop_bar.append(tr_item)

            return pre_stopbar, on_stopbar, post_stop_bar

        if '*' in base_event.plate_ucode:
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            self._stats.lr_invalid += 1
            return

        answer = split_track(self._scene_mask)
        if answer is None:
            # controlling direction of movement is not defined
            eventlog.log_e('fail: skip event due to lack of points')
            self._stats.zone_invalid += 1
            return

        elif 0 in list(map(len, answer)):
            # stopbar crossing is not detected
            eventlog.log_e('fail: skip event due to lack of points')
            self._stats.zone_invalid += 1
            return

        for rule in self._rules:
            try:
                reply: Optional[Tuple[Evidence, TimeSpan]] = rule.check(trackset, base_event, *answer)
                if reply is None:
                    continue
            except GenericError:
                eventlog.log_w('fail')
                continue

            evidence, time_span = reply
            handler = functools.partial(self.handle_evidence, base_event, evidence)
            if self._conf.video:
                scaled_w, scaled_h = self._im_fitter.scaledown(self._ime_prebar.scene_rgba.wh)

                # noinspection PyBroadException
                try:
                    return self._deps.vs.export_video(
                        time_span.min_ts // 1000 - self._conf.video.pre_record,
                        time_span.max_ts // 1000 + self._conf.video.post_record,
                        scaled_w, scaled_h, self._conf.video.quality,
                        handler)

                except VsSessionAbsentError:
                    eventlog.log_w('fail: VsSessionAbsentError')

                except VsError:
                    eventlog.log_w('fail: VsError')

                except GenericError:
                    eventlog.log_w('fail')

                except Exception:
                    eventlog.log_x()

            handler(None)
            break

    def handle_evidence(self, base: BaseEvent, evidence: Evidence, video_blob: Optional[FileBlob]) -> None:
        if video_blob:
            evidence.add_videos([video_blob])

        self._stats.emit_total += 1

        if not self._handler(Event(self._conf.identity, base, evidence)):
            self._stats.emit_fail += 1

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        total_fails = self._stats.lr_invalid + self._stats.zone_invalid
        jnode = {
            'type': 'red-light',
            'name': f'{self._conf.identity.ref}',
            'stat': {
                'assert-track': {
                    'count': [self._stats.total, self._stats.total - total_fails, total_fails],
                    'zone-unmatch': self._stats.zone_invalid,
                    'lr-invalid': self._stats.lr_invalid
                },
                'emit-event': {
                    'count': [
                        self._stats.emit_total,
                        self._stats.emit_total - self._stats.emit_fail,
                        self._stats.emit_fail
                    ]
                }
            }
        }

        json_handler(jnode)
        list(map(lambda x: x.collect_stats(json_handler), self._rules))


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detectors.plate_lost.add(_Detector(DetectorDeps.create(services), conf))
        except GenericError:
            eventlog.log_e('fail')
            raise
