from typing import Callable, Dict, Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.image.types.footer import ImFooterVars
from edge.types.timespan import TimeSpan

from speedcontrol.detectors.redlight.types.conf import RuleConf
from speedcontrol.detectors.redlight.types.deps import RuleDeps
from speedcontrol.detectors.redlight.types.stats import RuleStats
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence import BaseEvidence
from speedcontrol.events.evidence.redlight import Evidence
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import Vehicle, ViolationVehicle
from speedcontrol.events.waypoint import Waypoint
from speedcontrol.events.waypoint.redlight import RedLightViolationWaypoint
from edge.services.types.vs import VsImageInfo
from edge.types.track import TrTrackItem
from speedcontrol.trackset.trackset import Trackset


def _create_detector_info(detector_name: str, conf: RuleConf) -> DetectorInfo:
    ticket_info = TicketInfo(conf.ticket.name)
    rule_info = RuleInfo(conf.name, None, ticket_info)
    return DetectorInfo(detector_name, rule_info)


class _SceneImageInfo(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem):
        super().__init__(tr_item.lr_item.ts)
        self.tr_item = tr_item


class Rule:
    def __init__(self,
                 deps: RuleDeps,
                 conf: RuleConf,
                 detector_name: str,
                 stats_name: str,
                 ime_prebar: ImBuilderEnv,
                 ime_postbar_init: ImBuilderEnv,
                 ime_postbar_fini: ImBuilderEnv):

        self._deps: Final[RuleDeps] = deps
        self._conf: Final[RuleConf] = conf
        self._info: Final[DetectorInfo] = _create_detector_info(detector_name, conf)
        self._stats: Final[RuleStats] = RuleStats()

        self._stats_name = stats_name
        self._detector_name = detector_name

        self._ime_prebar = ime_prebar
        self._ime_postbar_init = ime_postbar_init
        self._ime_postbar_fini = ime_postbar_fini

    def check(self, trackset: Trackset, base_event: BaseEvent,
              pre_stopbar: List[TrTrackItem],
              on_stopbar: List[TrTrackItem],
              post_stopbar: List[TrTrackItem]
              ) -> Optional[Tuple[BaseEvidence, TimeSpan]]:

        self._stats.total += 1

        tl_result: List[TimeSpan] = self._deps.tl.check(trackset.lr_track.min_ts, trackset.lr_track.max_ts)

        if not len(tl_result):
            self._stats.valid_light += 1
            return None

        for violation in tl_result:
            self._stats.violation_total += 1

            # check pre-stopbar conditions
            pre_ts_range = violation.intersect(
                TimeSpan(pre_stopbar[0].ts, pre_stopbar[-1].ts))
            if pre_ts_range is None:
                self._stats.pre_items_invalid += 1
                continue

            pre_tr_items = [tr_item for tr_item in pre_stopbar if tr_item.ts in pre_ts_range]
            if not pre_tr_items:
                self._stats.pre_items_invalid += 1
                continue

            # check on-stopbar conditions
            on_ts_range = violation.intersect(
                TimeSpan(on_stopbar[0].ts, on_stopbar[-1].ts))
            if on_ts_range is None:
                self._stats.on_items_invalid += 1
                continue

            on_tr_items = [tr_item for tr_item in on_stopbar if tr_item.ts in on_ts_range]
            if not on_tr_items:
                self._stats.on_items_invalid += 1
                continue

            # check post-stopbar conditions
            post_ts_range = violation.intersect(
                TimeSpan(post_stopbar[0].ts, post_stopbar[-1].ts))
            if post_ts_range is None:
                self._stats.post_items_invalid += 1
                continue

            post_tr_items = [tr_item for tr_item in post_stopbar if tr_item.ts in post_ts_range]
            if not post_tr_items:
                self._stats.post_items_invalid += 1
                continue

            # prepare im-environments
            pre_im_infos = list(map(_SceneImageInfo, pre_tr_items))
            post_im_infos = list(map(_SceneImageInfo, post_tr_items))
            self._stats.make_evidence_total += 1

            pre_im_info = self._deps.vs.export_image_rgba(self._ime_prebar.scene_rgba, pre_im_infos[:-6:-1],
                                                          self._ime_prebar.scene_dh)
            if pre_im_info is None:
                eventlog.log_i('fail: load pre-stopbar image')
                self._stats.load_images += 1
                continue

            post_im_info1 = self._deps.vs.export_image_rgba(self._ime_postbar_init.scene_rgba, post_im_infos[0:5],
                                                            self._ime_postbar_init.scene_dh)
            if post_im_info1 is None:
                eventlog.log_i('fail: load post-stopbar image-1')
                self._stats.load_images += 1
                continue

            post_im_info2 = self._deps.vs.export_image_rgba(self._ime_postbar_fini.scene_rgba, post_im_infos[:-6:-1],
                                                            self._ime_postbar_fini.scene_dh)
            if post_im_info2 is None:
                eventlog.log_i('fail: load post-stopbar image-2')
                self._stats.load_images += 1
                continue

            # prepare evidence
            evidence = self._mk_evidence(
                base_event, trackset,
                pre_im_info.tr_item,
                post_im_info1.tr_item,
                post_im_info2.tr_item)

            if evidence:
                ref_ts = post_tr_items[0].lr_item.ts
                return evidence, TimeSpan(ref_ts, ref_ts)

    def _mk_evidence(self,
                     base_event: BaseEvent,
                     track: Trackset,
                     pre_tr_item: TrTrackItem,
                     post_tr_item1: TrTrackItem,
                     post_tr_item2: TrTrackItem
                     ) -> Optional[Evidence]:

        # get best vehicle
        best_waypoint = base_event.best_wp
        best_vehicle: Vehicle = best_waypoint.vehicle

        def mk_waypoint(deps: RuleDeps, im_env: ImBuilderEnv, tr_item: TrTrackItem, waypoint_name: str) -> Waypoint:
            origin = deps.about.get_origin(im_env.scene_rgba.ts)
            wp_direction = track.tr_track.get_item_dir(tr_item.ts)

            if tr_item.lane_item is not None:
                origin.location.set_lane(tr_item.lane_item.lane)

            green_amber_time, red_time, red_amber_time, halt_time = deps.tl.get_times(im_env.scene_rgba.ts)
            footer_vars = ImFooterVars(deps.tz, self._deps.about, im_env.scene_rgba.ts,
                                       defvars=self._conf.footer_defvars,
                                       location=origin.location,
                                       geolocation=origin.geolocation,
                                       plate_text=base_event.plate_ucode,
                                       spot_speed=tr_item.ra_item and tr_item.ra_item.speed_kmh,
                                       direction=wp_direction,
                                       spot_speed_ra_n=best_vehicle.spotspeeds.ra_n,
                                       spot_speed_cam_2=best_vehicle.spotspeeds.cam_2,
                                       spot_speed_cam_n=best_vehicle.spotspeeds.cam_n,
                                       spot_speed_wim=best_vehicle.spotspeeds.wim,
                                       spot_speed_main=best_vehicle.spotspeeds.main,
                                       spot_speed_control=best_vehicle.spotspeeds.control,
                                       halt_time=halt_time)

            scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(im_env.scene_rgba.ts)
            im_blobs = im_env.builder.build(
                im_env.scene_rgba, scene_dt_utc, origin.geolocation,
                footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                vehicle=False, scene=True, plate=True)

            if len(im_blobs) != 2:
                eventlog.log_i('fail: mk image-blobs')
                raise RuntimeError

            waypoint = RedLightViolationWaypoint(
                waypoint_name, im_env.scene_rgba.ts, origin.location, im_blobs,
                ViolationVehicle(plate_quads=[tr_item.lr_item.plate_quad]),
                green_amber_time, red_time, red_amber_time, halt_time)

            return waypoint

        try:
            pre_waypoint = mk_waypoint(
                self._deps, self._ime_prebar, pre_tr_item,
                Waypoint.NAME_PRE_STOPBAR)

            post_waypoint1 = mk_waypoint(
                self._deps, self._ime_postbar_init, post_tr_item1,
                Waypoint.NAME_POST_STOPBAR)

            post_waypoint2 = mk_waypoint(
                self._deps, self._ime_postbar_fini, post_tr_item2,
                Waypoint.NAME_POST_STOPBAR)

        except RuntimeError:
            eventlog.log_e('fail: make images')
            self._stats.make_evidence_make_images += 1
            return None

        event_ts = self._ime_postbar_init.scene_rgba.ts

        evidence = Evidence(
            self._deps.about.get_origin(event_ts),
            event_ts, self._info)
        evidence.add_waypoints([pre_waypoint, post_waypoint1, post_waypoint2])
        return evidence

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        violation_total_fail = self._stats.pre_items_invalid + self._stats.on_items_invalid + \
                               self._stats.post_items_invalid + self._stats.load_images
        jnode = {
            'type': 'red-light~rule',
            'name': self._stats_name,
            'stat': {
                'assert-track': {
                    'count': [
                        self._stats.total,
                        self._stats.total - self._stats.valid_light,
                        self._stats.valid_light
                    ],
                    'green-light': self._stats.valid_light
                },
                'assert-violation': {
                    'count': [
                        self._stats.violation_total,
                        self._stats.violation_total - violation_total_fail,
                        violation_total_fail
                    ],
                    'pre-zone-unmatch': self._stats.pre_items_invalid,
                    'on-zone-unmatch': self._stats.on_items_invalid,
                    'post-zone-unmatch': self._stats.post_items_invalid,
                    'load-images': self._stats.load_images,
                },
                'make-evidence': {
                    'count': [
                        self._stats.make_evidence_total,
                        self._stats.make_evidence_total - self._stats.make_evidence_make_images,
                        self._stats.make_evidence_make_images
                    ],
                    'make-images': self._stats.make_evidence_make_images
                }
            }
        }
        json_handler(jnode)
