import functools

from typing import Callable, Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.image.scaler import WhScaler
from edge.support.service import ServicePack
from edge.types.blob import FileBlob, BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.timespan import TimeSpan

from speedcontrol.detectors.redlight_leddar.rule import Rule
from speedcontrol.detectors.redlight_leddar.types.deps import DetectorDeps
from speedcontrol.detectors.redlight_leddar.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.redlight import Evidence
from edge.services.types.vs import VsSessionAbsentError, VsError
from speedcontrol.trackset.trackset import Trackset


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf

        if self._conf.video:
            self._im_fitter = WhScaler((self._conf.video.max_w, self._conf.video.max_h))

        self._ime_postbar = ImBuilderEnv(EventMeasCamImBuilder(conf.imbuilder))

        self._rules: List[Rule] = []
        for rule_conf in conf.rules:
            rule = Rule(deps, rule_conf, conf.name, self._ime_postbar)
            self._rules.append(rule)

        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, trackset: Trackset, base_event: BaseEvent) -> None:
        if trackset.le_track is None:
            return

        elif trackset.le_track.ref != self._deps.leddar.ref():
            return

        for rule in self._rules:
            try:
                reply: Optional[Tuple[Evidence, TimeSpan]] = rule.check(trackset, base_event)
                if reply is None:
                    continue
            except GenericError:
                eventlog.log_w('fail')
                continue

            evidence, time_span = reply
            handler = functools.partial(self.handle_evidence, base_event, evidence)
            if self._conf.video:
                scaled_w, scaled_h = self._im_fitter.scaledown(self._ime_postbar.scene_rgba.wh)

                # noinspection PyBroadException
                try:
                    return self._deps.vs.export_video(
                        time_span.min_ts // 1000 - self._conf.video.pre_record,
                        time_span.max_ts // 1000 + self._conf.video.post_record,
                        scaled_w, scaled_h, self._conf.video.quality,
                        handler)

                except VsSessionAbsentError:
                    eventlog.log_w('fail: VsSessionAbsentError')

                except VsError:
                    eventlog.log_w('fail: VsError')

                except GenericError:
                    eventlog.log_w('fail')

                except Exception:
                    eventlog.log_x()

            handler(None)
            break

    def handle_evidence(self, base: BaseEvent, evidence: Evidence, video_blob: FileBlob | None) -> None:
        if video_blob:
            evidence.add_videos([video_blob])

        self._handler(Event(self._conf.identity, base, evidence))


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detectors.plate_lost.add(_Detector(DetectorDeps.create(services), conf))
        except GenericError:
            eventlog.log_e('fail')
            raise
