from typing import Callable, Final, List, Optional, Dict

from edge import eventlog
from edge.image.builder import EventMeasCamImBuilder
from edge.image.types.footer import ImFooterVars
from edge.services.types.vs import VsImageInfo
from edge.support.imloader import RgbaImLoader, ImStats
from edge.support.service import ServicePack
from edge.types.blob import BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError, InternalError
from edge.types.identity import Identity
from edge.types.image import ImRgba
from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectF

from speedcontrol.detectors.seatbelt.types.conf import DetectorConf
from speedcontrol.detectors.seatbelt.types.deps import DetectorDeps
from speedcontrol.detectors.seatbelt.types.jnode.conf import decode__detector_conf
from speedcontrol.detectors.types.detector import DetectorGroups, DetectorUnitInfo, IDetector
from speedcontrol.events.detector import RuleInfo, DetectorInfo
from speedcontrol.events.event import Event, BaseEvent
from speedcontrol.events.evidence.seatbelt import Evidence
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import ViolationVehicle
from speedcontrol.events.waypoint.violation import ViolationWaypoint
from speedcontrol.trackset.trackset import Trackset


def _create_detector_info(detector_name: str, ticket: TicketInfo) -> DetectorInfo:
    rule_info = RuleInfo('default', None, ticket)
    return DetectorInfo(detector_name, rule_info)


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._conf: Final[DetectorConf] = conf
        self._deps: Final[DetectorDeps] = deps
        self._info = _create_detector_info(conf.name, conf.ticket)
        self._stats_meas_im = ImStats()

        self._best_im_loader = RgbaImLoader(self._stats_meas_im, self._deps.vs, 14)
        self._best_im_builder = EventMeasCamImBuilder(conf.imbuilder)
        self._best_im_rgba: ImRgba = ImRgba()

        self._handler: Callable[[Event], None] | None = None

    def destroy(self) -> None:
        self._best_im_loader.destroy()

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def mk_waypoint(self, track: Trackset, plate_text: str, plate_quad: QuadI,
                    pixelate_bboxes: Optional[List[RectF]],
                    draw_bboxes: Optional[List[RectF]]) -> ViolationWaypoint:

        waypoint_ts = self._best_im_rgba.ts

        waypoint_origin = self._deps.about.get_origin(waypoint_ts)

        footer_vars = ImFooterVars(
            self._deps.tz, self._deps.about, waypoint_ts, defvars=self._conf.def_footer_mapping,
            location=waypoint_origin.location,
            geolocation=waypoint_origin.geolocation,
            plate_text=plate_text)

        scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(waypoint_ts)
        im_blobs = self._best_im_builder.build(
            self._best_im_rgba, scene_dt_utc, waypoint_origin.geolocation,
            footer_vars, track.lr_track, plate_quad,
            vehicle=True, scene=True, plate=True, thumb=True, pixelate_bboxes=pixelate_bboxes,
            draw_bboxes=draw_bboxes)

        if len(im_blobs) != 4:
            eventlog.log_w('fail: get all blobs for evidence')
            raise InternalError

        waypoint_v = ViolationVehicle(
            plate_quads=[plate_quad])

        waypoint = ViolationWaypoint(
            ViolationWaypoint.NAME_VIOLATION_BEST,
            waypoint_ts, None, waypoint_origin.location, im_blobs,
            waypoint_v)

        return waypoint

    def check(self, track: Trackset, base_event: BaseEvent) -> Optional[Evidence]:
        best_point = base_event.best_wp
        if best_point is None:
            eventlog.log_e('fail: find best-waypoint')
            return

        seats_jnode = best_point.vehicle.seats_jnode
        if seats_jnode is None:
            eventlog.log_d('done: no seats info avaliable')
            return

        if len(self._conf.seats) != len(seats_jnode):
            eventlog.log_w('fail')
            return

        seats_to_blur: List[RectF] = []
        seats_to_draw: List[RectF] = []
        evidence_seats_jnode: List[Dict] = []

        for index in range(0, len(self._conf.seats)):
            i_conf = self._conf.seats[index]
            i_seat = seats_jnode[index]
            i_seat_bbox_jnode = i_seat['bbox']
            i_seat_bbox = RectF(*i_seat_bbox_jnode)

            i_seat_status = i_seat['belt']
            i_seat_w = i_seat['belt-weight']

            if i_seat_w < i_conf.min_weight:
                seats_to_blur.append(i_seat_bbox)

            elif i_seat_status != 'present':
                seats_to_blur.append(i_seat_bbox)

            else:
                seats_to_draw.append(i_seat_bbox)
                evidence_seats_jnode.append({
                    'bbox': i_seat_bbox_jnode,
                    'belt': 'absent',
                    'belt-weight': i_seat_w
                })

        if len(seats_to_draw) == 0:
            return

        im_infos: List[VsImageInfo] = [
            VsImageInfo(track.lr_track.imref.ts)
        ]

        im_info: Optional[VsImageInfo] = self._best_im_loader.load(self._best_im_rgba, im_infos, track.lr_track.imref)
        im_ts = track.lr_track.imref.ts if im_info is None else im_info.ts

        best_lr_item = track.lr_track.get_item(im_ts)
        best_quad = best_lr_item.plate_quad if best_lr_item else None
        if best_quad is None:
            eventlog.log_i(f'fail: reject track (plate:{base_event.plate_utf8}, reason:plate-quad)')
            return None

        try:
            best_wp = self.mk_waypoint(track, base_event.plate_ucode, best_quad, seats_to_blur, seats_to_draw)
        except GenericError:
            eventlog.log_i(f'fail: reject track (plate:{base_event.plate_utf8}, reason:make-images)')
            return None

        origin = self._deps.about.get_origin(best_point.main_ts)
        evidence = Evidence(origin, best_point.main_ts, self._info, evidence_seats_jnode)
        evidence.add_waypoints([best_wp])
        return evidence

    def analyze(self, track: Trackset, base: BaseEvent) -> None:
        evidence = self.check(track, base)
        if evidence:
            self._handler(Event(self._conf.identity, base, evidence))


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            pass
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detectors.plate_lost.add(_Detector(DetectorDeps.create(services), conf))
        except GenericError:
            eventlog.log_e('fail')
            raise
