from typing import Callable, Dict, Final, List

from twisted.internet.defer import DeferredList

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.services.types.tz import ITzService
from edge.support.service import ServicePack
from edge.types.blob import BsChanRef, BlobProfileToBsChanMap
from edge.types.dir import Dir
from edge.types.exceptions import GenericError
from edge.types.identity import Identity

from speedcontrol.detectors.sectionspeed.rule import Rule
from speedcontrol.detectors.sectionspeed.types.deps import DetectorDeps
from speedcontrol.detectors.sectionspeed.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.jnode.event import encode__base_event
from speedcontrol.trackset.trackset import Trackset


class BaseEvent2(BaseEvent):
    def __init__(self, base_event: BaseEvent, remote_m0: Dict):
        """
        :param base_event:
        :param remote_m0:
            remote-args.json#/measurements/0
        """
        super().__init__(base_event.measurements, base_event.alarms_jnode)
        self._measurement_jnode = remote_m0

    def serialize(self, tz_service: ITzService) -> Dict:
        base_jnode = encode__base_event(self, tz_service)
        base_jnode['measurements'].append(self._measurement_jnode)
        return base_jnode


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._conf: Final[DetectorConf] = conf
        self._ime_fini = ImBuilderEnv(EventMeasCamImBuilder(conf.im_engine))

        self._rules: List[Rule] = []
        for rule_conf in conf.rules:
            if rule_conf.enabled:
                rule = Rule(deps, rule_conf, conf.name, conf.section, conf.ignore_ra, self._ime_fini)
                self._rules.append(rule)

        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, trackset: Trackset, base_event: BaseEvent):
        def log_exc(e):
            eventlog.log_w(str(e))

        def handle_results(d_results):
            """
            :param d_results: list of pairs (deferred success, deferred result)
            :type d_results: list (bool, (SectionSpeedEvidence, dict) | None)
            """
            for success, reply in d_results:
                if success and reply:
                    (evidence, jnode) = reply
                    self._handler(Event(self._conf.identity, BaseEvent2(base_event, jnode), evidence))

        if trackset.lr_track.dir != Dir(self._conf.section.direction):
            return

        if '*' in base_event.plate_ucode:
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            return

        deferreds = []
        for rule in self._rules:
            deferreds.append(rule.check(trackset, base_event))

        d_list = DeferredList(deferreds)
        d_list.addCallbacks(handle_results, log_exc)


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detectors.plate_lost.add(_Detector(DetectorDeps.create(services), conf))
        except GenericError:
            eventlog.log_e('fail')
            raise
