import base64
import ujson
import urllib.parse

from typing import Dict, Final, List, Optional, Tuple

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.web.client import Agent, readBody
from twisted.web.http_headers import Headers

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.image.types.footer import ImFooterVars

from edge.support.isotime import isostr_to_timestamp_usec
from edge.support.schedule import Schedule
from edge.support.speed import map_speed_int
from edge.types.blob import MemBlob, JsonBlobMeta
from edge.types.shape.quad import QuadI, decode__quadi
from edge.types.track import TrTrackItem

from speedcontrol.detectors.sectionspeed.types.conf import RuleConf, SectionConf
from speedcontrol.detectors.sectionspeed.types.deps import RuleDeps
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.sectionspeed import Evidence
from speedcontrol.events.position import Position
from speedcontrol.events.schedule import ScheduleInfo
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import ViolationVehicle
from speedcontrol.events.waypoint import Waypoint
from speedcontrol.events.waypoint.violation import ViolationWaypoint
from edge.services.types.vs import VsImageInfo
from speedcontrol.trackset.trackset import Trackset


def _create_detector_info(detector_name: str, conf: RuleConf) -> DetectorInfo:
    schedule_info = ScheduleInfo(
        conf.schedule.min_time,
        conf.schedule.max_time,
        conf.schedule.days_to_bool)

    ticket_info = TicketInfo(conf.ticket.name)

    rule_info = RuleInfo(conf.name, schedule_info, ticket_info, {
        'speed-limit': conf.limits.speed_limit,
        'max-overspeed': conf.limits.over_speed
    })

    return DetectorInfo(detector_name, rule_info)


class _SceneMsRef(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem):
        super().__init__(tr_item.ts)
        self.tr_item = tr_item


class Rule:
    def __init__(self,
                 deps: RuleDeps,
                 conf: RuleConf,
                 detector_name: str,
                 section_args: SectionConf,
                 ignore_ra: bool,
                 fini_im_env: ImBuilderEnv):

        self._deps: Final[RuleDeps] = deps
        self._conf: Final[RuleConf] = conf
        self._info: Final[DetectorInfo] = _create_detector_info(detector_name, conf)

        self._schedule: Final[Schedule] = Schedule(conf.schedule, deps.tz)

        self._det_conf = section_args
        self._fini_im_env = fini_im_env
        self._ignore_ra = ignore_ra

        self._agent = Agent(reactor, connectTimeout=60)

    @inlineCallbacks
    def _get_remote_event_file(self, event_id: str, img_name: str) -> bytes:
        det_conf = self._det_conf
        user, password = det_conf.init_area.auth
        base_url = det_conf.init_area.controller

        url = '{url}/api/v2/origins/event/dist/{event_id}/by-ref/{img_name}'.format(
            url=base_url, event_id=event_id, img_name=img_name)
        b64 = base64.b64encode('{}:{}'.format(user, password))
        resp = yield self._agent.request(b'GET', url, Headers({
            'Authorization': ['Basic {}'.format(b64)]
        }), None)

        if resp.code != 200:
            eventlog.log_e('fail: get file: <{0}>, <{1}>'.format(url, resp.code))
            returnValue(None)

        file_contents = yield readBody(resp)
        returnValue(file_contents)

    @inlineCallbacks
    def _get_remote_feed_record(self, plate_text: str) -> Optional[Dict]:
        det_conf = self._det_conf
        user, password = det_conf.init_area.auth
        base_url = det_conf.init_area.controller
        origin = det_conf.init_area.origin
        direction = det_conf.direction

        url = ('{url}/api/v2/origins/event/feed?origin={origin}&type=transit&'
               'dir={dir}&position=required&plate-method=exact&plate-pattern={plate}&limit=1'
               ).format(url=base_url, origin=origin, dir=direction, plate=plate_text)

        url = urllib.parse.quote(url.encode('utf8'), ':/&?=')
        b64 = base64.b64encode('{}:{}'.format(user, password))
        resp = yield self._agent.request(b'GET', url, Headers({
            "Authorization": [f"Basic {b64}"]
        }), None)

        if resp.code != 200:
            eventlog.log_e('fail: get feed record: <{0}>, <{1}>'.format(url, resp.code))
            returnValue(None)

        feed_str = yield readBody(resp)
        feed_records = ujson.loads(feed_str)  # list of feed records
        if feed_records:
            returnValue(feed_records[0])

        else:
            returnValue(None)

    def _estimate_speed_dist(self,
                             remote_ts: int,
                             remote_dist: int,
                             local_ts: int,
                             local_dist: int) -> Tuple[int, int]:
        """
        :param remote_ts:
            time in usec
        :param remote_dist:
            init distance in cm
        :param local_ts:
            time in usec
        :param local_dist:
            fini distance in cm

        :returns:
            (speed in km/h, distance in m)
        """
        conf_dir, conf_dist = self._det_conf.direction, self._det_conf.dist
        if conf_dir == conf_dir.CODE_R:
            dist_m = int(conf_dist - remote_dist / 100.0 + local_dist / 100.0)
        else:
            dist_m = int(conf_dist + remote_dist / 100.0 - local_dist / 100.0)

        time_hours = (local_ts - remote_ts) / 1e6 / 3600
        dist_km = dist_m / 1000.0

        section_speed: int = map_speed_int(dist_km / time_hours)
        return section_speed, dist_m

    def _estimate_speed(self, remote_record: Dict, local_track:  Trackset, base: BaseEvent) -> int:
        remote_e = remote_record['event']
        remote_v = remote_e['vehicle']
        remote_p = remote_v.get('position') or dict()

        remote_ts = isostr_to_timestamp_usec(remote_e['time'])  # usec
        remote_dist = remote_p.get('forward', 0)  # cm

        conf_dir, conf_dist = self._det_conf.direction, self._det_conf.dist

        local_best_wp = base.measurements[0].best_waypoint
        local_best_position = local_best_wp.vehicle.position

        if conf_dist < self._det_conf.asssert_dist and not self._ignore_ra:
            assert local_track.ra_track
            local_ts = local_track.ra_track[0].frame_ts
            local_dist = local_track.ra_track[0].scene_f_cm

        elif local_track.ra_track and not self._ignore_ra:
            # TODO: почему-бы не брать best ?
            local_ts = local_track.ra_track[0].frame_ts
            local_dist = local_track.ra_track[0].scene_f_cm

        elif local_best_position is not None:
            local_ts = local_best_wp.main_ts
            local_dist = local_best_position.forward

        elif conf_dir == conf_dir.CODE_A:
            local_ts = local_track.lr_track.max_ts
            local_dist = 0

        elif conf_dir == conf_dir.CODE_R:
            local_ts = local_track.lr_track.min_ts
            local_dist = 0

        else:
            local_ts = local_best_wp.main_ts
            local_dist = 0

        speed, _ = self._estimate_speed_dist(remote_ts, remote_dist, local_ts, local_dist)
        return speed

    @inlineCallbacks
    def check(self, local_track: Trackset, base: BaseEvent) -> Optional[Evidence]:
        if self._det_conf.dist < self._det_conf.asssert_dist and local_track.ra_track is None:
            eventlog.log_w('fail: radar measurements are required')
            returnValue(None)
            return

        min_ts = local_track.lr_track.min_ts
        max_ts = local_track.lr_track.max_ts
        if not(self._schedule.check_ts_usec(max_ts) or self._schedule.check_ts_usec(min_ts)):
            returnValue(None)

        plate_ucode = base.plate_ucode
        try:
            feed_record = yield self._get_remote_feed_record(plate_ucode)
        except Exception:
            eventlog.log_e('fail: get remote feed record')
            returnValue(None)
            return

        if feed_record is None:  # !200 or empty list
            returnValue(None)
            return

        try:
            est_section_speed: int = self._estimate_speed(feed_record, local_track, base)
        except KeyError:
            returnValue(None)
            return

        conf_speed_limit = self._conf.limits.speed_limit
        conf_max_overspeed = self._conf.limits.over_speed
        conf_max_section_speed = conf_speed_limit + conf_max_overspeed
        eventlog.log_d('plate:{}, speed:{}, max-speed:{}'.format(
            base.plate_utf8, est_section_speed, conf_max_section_speed))
        if est_section_speed <= conf_max_section_speed:
            returnValue(None)
            return

        # download args.json and images
        remote_e = feed_record['event']
        try:
            remote_uuid = remote_e['uuid']
            remote_args_name = feed_record['dist']['args']
            remote_args_bytes = yield self._get_remote_event_file(remote_uuid, remote_args_name)
            remote_args = ujson.loads(remote_args_bytes)

            remote_m0 = remote_args['measurements'][0]
            remote_best = remote_m0['waypoints']['best']
            remote_v = remote_best['vehicle']
            remote_plate_quad = decode__quadi(remote_v['license-plates'][0]['bbox-4'])

            remote_i = remote_best['images']
            remote_scene_imname = remote_i['scene']['name']
            remote_plate_imname = remote_i['license-plates'][0]['name']

            remote_scene_iminfo = remote_i['scene']['info']

            remote_scene_imbytes = yield self._get_remote_event_file(remote_uuid, remote_scene_imname)
            remote_plate_imbytes = yield self._get_remote_event_file(remote_uuid, remote_plate_imname)

        except Exception:
            returnValue(None)
            return

        if local_track.ra_track is not None:
            candidate_items: List[TrTrackItem] = [x for x in local_track.tr_track.values() if x.lr_item and x.ra_item]
        else:
            candidate_items = [x for x in local_track.tr_track.values()]

        scene_ms_refs = list(map(_SceneMsRef, candidate_items))

        local_scene_iminfo: Optional[_SceneMsRef] = self._deps.vs.export_image_rgba(
            self._fini_im_env.scene_rgba, scene_ms_refs[:-6:-1], self._fini_im_env.scene_dh)
        if not local_scene_iminfo:
            eventlog.log_w('fail: get local_scene_im from video storage')
            returnValue(None)
            return

        local_ra_item = local_scene_iminfo.tr_item.ra_item
        local_cam_f = 0
        local_cam_l = 0
        # если нет радара и камера MV, пересчитать forward через image-mapper
        if self._deps.cam_service:
            plate_cx, plate_cy = local_scene_iminfo.tr_item.lr_item.plate_quad.center
            world_f, world_l = self._deps.cam_service.mk_world_point(plate_cx, plate_cy, world_t=0.5)
            # m to cm transform

            local_cam_l = int(world_l * 100)
            local_cam_f = - int(world_f * 100)

        if local_ra_item and not self._ignore_ra:
            local_f = local_ra_item.scene_f_cm
            local_l = local_ra_item.scene_l_cm
        else:
            local_l = local_cam_l
            local_f = local_cam_f

        local_pos = None if local_f == local_l == 0 else Position(local_f, local_l)
        local_ts = local_scene_iminfo.tr_item.ts

        remote_p = remote_e['vehicle'].get('position') or dict()
        remote_f = remote_p.get('forward', 0)  # cm
        remote_l = remote_p.get('left', 0)  # cm
        remote_pos = Position(remote_f, remote_l) if remote_p else None
        remote_ts = isostr_to_timestamp_usec(remote_e['time'])  # usec

        section_speed, section_dist = self._estimate_speed_dist(remote_ts, remote_f, local_ts, local_f)

        if local_ra_item:
            eventlog.log_d(
                'plate:{}, '
                'remote_f:{:6}, remote_ts: {}, '
                'local_f:{:6}, local_ra_f:{:6}, local_ts: {}, '
                'speed:{:6}, max-speed:{:6}'.format(
                    base.plate_utf8,
                    remote_f, remote_ts,
                    local_f, local_ra_item.scene_f_cm, local_ts,
                    section_speed, conf_max_section_speed))
        else:
            eventlog.log_d(
                'plate:{}, '
                'remote_f:{:6}, remote_ts:{}, '
                'local_f:{:6}, local_ts:{}, '
                'speed:{:6}, max-speed:{:6}'.format(
                    base.plate_utf8,
                    remote_f, remote_ts,
                    local_f, local_ts,
                    section_speed, conf_max_section_speed))

        if section_speed <= conf_max_section_speed:
            returnValue(None)
            return

        evidence = self._mk_evidence(plate_ucode, section_dist, section_speed, remote_best, remote_scene_imbytes,
                                     remote_scene_iminfo, remote_plate_quad, remote_plate_imbytes, remote_pos,
                                     local_track, local_scene_iminfo.tr_item, local_pos)

        del remote_best['images']
        remote_m0['origin'] = remote_args['origin']
        remote_args_crt = remote_args.get('certificate')
        if remote_args_crt:
            remote_m0['certificate'] = remote_args_crt

        # remote_m0['transit-uuid'] = args_json['uuid']
        returnValue((evidence, remote_m0))

    def _mk_evidence(self,
                     plate_text: str,
                     section_dist_m: int,
                     section_speed_kmh: int,
                     remote_best: Dict,
                     remote_scene_imbytes: bytes,
                     remote_scene_iminfo: Dict,
                     remote_plate_quad: QuadI,
                     remote_plate_imbytes: bytes,
                     remote_pos: Optional[Position],
                     local_track: Trackset,
                     local_tr_item: TrTrackItem,
                     local_pos: Optional[Position]) -> Evidence:

        def create_remote_wp() -> Waypoint:
            remote_ts = isostr_to_timestamp_usec(remote_best['time'])
            remote_control_ts_str = remote_best.get('control-time')
            remote_control_ts = isostr_to_timestamp_usec(remote_control_ts_str) if remote_control_ts_str else None

            scene = MemBlob('scene', remote_ts, bytes(remote_scene_imbytes), meta=JsonBlobMeta(remote_scene_iminfo))
            plate = MemBlob('plate', remote_ts, bytes(remote_plate_imbytes))

            answer = ViolationWaypoint(
                Waypoint.NAME_INIT, remote_ts, remote_control_ts, None, [scene, plate],
                ViolationVehicle(position=remote_pos, plate_quads=[remote_plate_quad]))

            return answer

        remote_wp = create_remote_wp()

        local_plate_quad = local_tr_item.lr_item.plate_quad

        main_ts: int = self._fini_im_env.scene_rgba.ts
        control_ts: int = self._deps.clock.get_control_ts(main_ts)
        overspeed_kmh: int = section_speed_kmh - self._conf.limits.speed_limit

        waypoint_origin = self._deps.about.get_origin(main_ts)
        footer_vars = ImFooterVars(self._deps.tz, self._deps.about,
                                   main_ts,
                                   control_ts=control_ts,
                                   defvars=self._conf.footer_defvars,
                                   location=waypoint_origin.location,
                                   geolocation=waypoint_origin.geolocation,
                                   plate_text=plate_text,
                                   section_speed=section_speed_kmh,
                                   section_overspeed=overspeed_kmh,
                                   direction=self._det_conf.direction,
                                   speed_limit=self._conf.limits.speed_limit)

        scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(main_ts)
        im_blobs = self._fini_im_env.builder.build(
            self._fini_im_env.scene_rgba, scene_dt_utc, waypoint_origin.geolocation,
            footer_vars, local_track.lr_track,
            local_plate_quad, scene=True, plate=True)

        if len(im_blobs) != 2:
            eventlog.log_i('fail: get all blobs for evidence')
            raise RuntimeError

        local_wp = ViolationWaypoint(
            Waypoint.NAME_FINI, main_ts, control_ts, None, im_blobs,
            ViolationVehicle(position=local_pos, plate_quads=[local_plate_quad]))

        evidence = Evidence(
            self._deps.about.get_origin(main_ts), main_ts,
            section_speed_kmh, overspeed_kmh, self._info, self._det_conf.direction,
            self._det_conf.dist, section_dist_m, control_ts=control_ts)
        evidence.add_waypoints([remote_wp, local_wp])
        return evidence
