from typing import Dict
from uuid import UUID

from edge import eventlog
from edge.image.jnode.types import decode__im_builder_conf
from edge.types.jnode.dir import decode__dir_text
from edge.types.exceptions import JnodeError

from speedcontrol.detectors.types.jnode.conf import decode__base_rule_conf
from speedcontrol.detectors.sectionspeed.types.conf import (
    DetectorConf, RuleConf, RuleLimits,
    SectionConf, InitAreaConf, FiniAreaConf)
from edge.support.mask import decode__mask_region2
from edge.types.identity import Identity


def decode__rule_limits(jnode: Dict) -> RuleLimits:
    """
    :rtype: JsonLimitsArgs
    :raise JnodeError:
    """
    # deserialize #/speed-limit
    speed_limit = jnode.get('speed-limit')
    if not (int == type(speed_limit) and 0 <= speed_limit <= 250):
        eventlog.log_e_param('#/speed-limit')
        raise JnodeError

    # deserialize #/max-overspeed
    max_overspeed = jnode.get('max-overspeed')
    if not (int == type(speed_limit) and 0 <= max_overspeed <= 250):
        eventlog.log_e_param('#/max-overspeed')
        raise JnodeError

    return RuleLimits(speed_limit, max_overspeed)


def decode__rule_conf(rule_jnode: Dict, footer_defvars: Dict) -> RuleConf:
    """
    :rtype: JsonRuleArgs
    :raise JnodeError:
    """
    if type(rule_jnode) != dict:
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize #/name
    # deserialize #/enabled
    # deserialize #/schedule
    # deserialize #/ticket
    rule = decode__base_rule_conf(rule_jnode)

    # deserialize #/limits
    limits_jnode = rule_jnode.get('limits')
    if type(rule_jnode) != dict:
        eventlog.log_e_param('#/limits')
        raise JnodeError

    limits = decode__rule_limits(limits_jnode)
    return RuleConf(rule.name, rule.enabled, rule.schedule, rule.ticket, limits, footer_defvars)


class JsonInitAreaArgs(InitAreaConf):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def deserialize(init_area_jnode: Dict):
        """
        :rtype: JsonInitAreaArgs
        :raise JnodeError:
        """
        if type(init_area_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/origin
        origin = init_area_jnode.get('origin')
        try:
            _ = UUID(origin)
        except ValueError:
            eventlog.log_e_param('#/origin')
            raise JnodeError

        # deserialize #/auth
        auth_jnode = init_area_jnode.get('auth')
        if not isinstance(auth_jnode, (list, tuple)) or len(auth_jnode) != 2:
            eventlog.log_e_param('#/auth')
            raise JnodeError

        user, pwd = auth_jnode
        if not (isinstance(user, str) and len(user) >= 1):
            eventlog.log_e_param('#/auth/0')
            raise JnodeError

        if not (isinstance(pwd, str) and len(pwd) >= 6):
            eventlog.log_e_param('#/auth/1')
            raise JnodeError

        # deserialize #/controller
        controller = init_area_jnode.get('controller')
        if not (controller and isinstance(controller, str)):
            eventlog.log_e_param('#/controller')
            raise JnodeError

        return JsonInitAreaArgs(controller, (user, pwd), origin)


class JsonFiniAreaArgs(FiniAreaConf):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def deserialize(fini_area_jnode: Dict):
        """
        :rtype: JsonFiniAreaArgs
        :raise JnodeError:
        """
        if type(fini_area_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/mask
        mask_jnode = fini_area_jnode.get('mask')
        mask = decode__mask_region2(mask_jnode)

        # deserialize #/scene-wh
        scene_wh_jnode = fini_area_jnode.get('scene-wh')
        if not isinstance(scene_wh_jnode, (list, tuple)) or len(scene_wh_jnode) != 2:
            eventlog.log_e_param('#/scene-wh')
            raise JnodeError

        w, h = scene_wh_jnode
        if not (int == type(w) and 640 <= w <= 65535):
            eventlog.log_e_param('#/scene-wh/0')
            raise JnodeError

        if not (int == type(h) and 480 <= h <= 65535):
            eventlog.log_e_param('#/scene-wh/1')
            raise JnodeError

        return JsonFiniAreaArgs(mask, (w, h))


class JsonSectionArgs(SectionConf):
    DIRECTIONS = ['approaching', 'receding']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def deserialize(section_jnode: Dict):
        """
        :rtype: JsonSectionArgs
        :raise JnodeError:
        """
        if type(section_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/dir
        direction_jnode = section_jnode.get('dir')
        try:
            direction = decode__dir_text(direction_jnode)
        except ValueError:
            direction = None

        if not direction or direction_jnode not in JsonSectionArgs.DIRECTIONS:
            eventlog.log_e_param('#/dir')
            raise JnodeError

        # deserialize #/dist
        dist = section_jnode.get('dist')
        if not (isinstance(dist, (int, float)) and dist >= 300):
            eventlog.log_e_param('#/dist')
            raise JnodeError

        # deserialize #/assert-dist
        assert_dist = section_jnode.get('assert-dist')
        if not (isinstance(assert_dist, (int, float)) and assert_dist >= 300):
            eventlog.log_e_param('#/assert_dist')
            raise JnodeError

        # deserialize #/init-area
        init_area_jnode = section_jnode.get('init-area')
        if not isinstance(init_area_jnode, dict):
            eventlog.log_e_param('#/init-area')
            raise JnodeError

        init_area = JsonInitAreaArgs.deserialize(init_area_jnode)

        # deserialize #/fini-area
        fini_area_jnode = section_jnode.get('fini-area')
        if not isinstance(fini_area_jnode, dict):
            eventlog.log_e_param('#/fini-area')
            raise JnodeError

        fini_area = JsonFiniAreaArgs.deserialize(fini_area_jnode)

        return JsonSectionArgs(init_area, fini_area, direction, dist, assert_dist)


def decode__detector_conf(identity: Identity, jnode: Dict, footer_defvars: Dict) -> DetectorConf:
    """
    :raise JnodeError:
    """
    # deserialize #/name
    name = jnode.get('name')
    if not type(name) in [str] or len(name) > 32:
        eventlog.log_e_param('#/name')
        raise JnodeError

    # deserialize #/store
    store = jnode.get('store')
    if not (isinstance(store, str) and 1 <= len(store) <= 2):
        eventlog.log_e_param('#/store')
        raise JnodeError

    # deserialize #/rules
    rules_jnode = jnode.get('rules')
    if type(rules_jnode) != list:
        eventlog.log_e_param('#/rules')
        raise JnodeError

    rules = []
    for rule_jnode in rules_jnode:
        rule = decode__rule_conf(rule_jnode, footer_defvars)
        if rule.enabled:
            rules.append(rule)

    # deserialize #/images
    images_jnode = jnode.get('images')
    if type(images_jnode) != dict:
        eventlog.log_e_param('#/images')
        raise JnodeError

    im_engine = decode__im_builder_conf(images_jnode['fini'])

    # deserialize #/section
    section_jnode = jnode.get('section')
    if type(section_jnode) != dict:
        eventlog.log_e_param('#/section')
        raise JnodeError

    section = JsonSectionArgs.deserialize(section_jnode)

    # deserialize igrone-ra
    ignore_ra = jnode.get('igrone-ra')
    return DetectorConf(identity, name, store, rules, im_engine, section, ignore_ra)
