import functools

from typing import Callable, Dict, Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.image.scaler import WhScaler
from edge.services.types.vs import VsError, VsSessionAbsentError
from edge.support.service import ServicePack
from edge.types.blob import FileBlob, BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.timespan import TimeSpan

from speedcontrol.detectors.spotspeed.rule import Rule
from speedcontrol.detectors.spotspeed.types.conf import DetectorConf
from speedcontrol.detectors.spotspeed.types.deps import DetectorDeps
from speedcontrol.detectors.spotspeed.types.jnode.conf import decode__detector_conf
from speedcontrol.detectors.spotspeed.types.jnode.stats import encode__detector_stats
from speedcontrol.detectors.spotspeed.types.stats import DetectorStats
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.spotspeed import Evidence
from speedcontrol.trackset.trackset import Trackset


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf
        self._stats: Final[DetectorStats] = DetectorStats()

        if self._conf.video:
            self._im_fitter = WhScaler(self._conf.video.max_wh)

        imbuilder = EventMeasCamImBuilder(conf.im_engine)
        self._ime_init = ImBuilderEnv(imbuilder)
        self._ime_best = ImBuilderEnv(imbuilder)
        self._ime_fini = ImBuilderEnv(imbuilder)

        self._rules: List[Rule] = []
        for rule_conf in conf.rules:
            rule = Rule(deps, rule_conf, self._ime_init, self._ime_fini, self._ime_best)
            self._rules.append(rule)

        self._handler: Callable[[Event], bool] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], bool]) -> None:
        assert self._handler is None
        self._handler = event_handler

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        json_handler(encode__detector_stats(self._conf.identity, self._stats))
        list(map(lambda x: x.collect_stats(json_handler), self._rules))

    def _test_suitable(self, trackset: Trackset, base_event: BaseEvent) -> bool:
        self._stats.total += 1

        if trackset.lr_track is None:
            self._stats.vrp_absent += 1
            eventlog.log_i('fail: reject track (plate:-, reason:lr-absent)')
            return False

        if '*' in base_event.plate_ucode:
            self._stats.vrp_text_invalid += 1
            eventlog.log_i(f'fail: reject track (plate:{base_event.plate_utf8}, reason:lr-invalid)')
            return False

        if trackset.ra_track is None:
            self._stats.ra_track_absent += 1
            eventlog.log_i(f'fail: reject track (plate:{base_event.plate_utf8}, reason:ra-absent)')
            return False

        if base_event.best_wp.vehicle.spotspeeds.main is None:
            self._stats.main_speed_absent += 1
            eventlog.log_i(f'fail: reject track (plate:{base_event.plate_utf8}, reason:main-speed-absent)')
            return False

        if base_event.best_wp.vehicle.spotspeeds.control is None:
            self._stats.control_speed_absent += 1
            eventlog.log_i(f'fail: reject track (plate:{base_event.plate_utf8}, reason:control-speed-absent)')
            return False

        eventlog.log_i(f'done: accept track (plate:{base_event.plate_utf8})')
        return True

    def analyze(self, trackset: Trackset, base_event: BaseEvent) -> None:
        if not self._test_suitable(trackset, base_event):
            return

        for rule in self._rules:
            try:
                reply: Optional[Tuple[Evidence, TimeSpan]] = rule.check(trackset, base_event)
                if reply is None:
                    continue
            except GenericError:
                eventlog.log_w(f'fail: plate:{base_event.plate_utf8}')
                continue

            evidence, time_span = reply
            handler = functools.partial(self._handle_evidence, base_event, evidence)
            if self._conf.video:
                scaled_w, scaled_h = self._im_fitter.scaledown(self._ime_init.scene_rgba.wh)

                # noinspection PyBroadException
                try:
                    return self._deps.vs.export_video(
                        time_span.min_ts // 1000 - self._conf.video.pre_record,
                        time_span.max_ts // 1000 + self._conf.video.post_record,
                        scaled_w, scaled_h, self._conf.video.quality,
                        handler)

                except VsSessionAbsentError:
                    eventlog.log_w(f'fail: VsSessionAbsentError (plate:{base_event.plate_utf8})')

                except VsError:
                    eventlog.log_w(f'fail: VsError (plate:{base_event.plate_utf8})')

                except GenericError:
                    eventlog.log_w(f'fail: plate:{base_event.plate_utf8}')

                except Exception:
                    eventlog.log_x()

            handler(None)
            break

    def _handle_evidence(self, base_event: BaseEvent, evidence: Evidence, video_blob: Optional[FileBlob]) -> None:
        if video_blob is not None:
            evidence.add_videos([video_blob])

        self._stats.emit_total += 1

        if not self._handler(Event(self._conf.identity, base_event, evidence)):
            self._stats.emit_fail += 1

        eventlog.log_i(f'done: plate:{base_event.plate_utf8}, with-video:{video_blob is not None}')


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detector = _Detector(DetectorDeps.create(services), conf)
            detectors.plate_lost.add(detector)
        except GenericError:
            eventlog.log_e(f'fail: {info.identity.ref}')
            raise
