from typing import Callable, List, Optional, Tuple, Set

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.image.types.footer import ImFooterVars
from edge.services.types.vs import VsImageInfo
from edge.support.schedule import Schedule
from edge.types.jnode.dir import encode_dir_text
from edge.types.timespan import TimeSpan
from edge.types.track import TrTrackItem
from edge.webserver.sysconf.uiconf.constroller import CtrlUiconf

from speedcontrol.detectors.base.best import check_quad_dist_from_img_borders
from speedcontrol.detectors.spotspeed_waga.types.conf import RuleConf
from speedcontrol.detectors.spotspeed_waga.types.deps import RuleDeps
from speedcontrol.detectors.spotspeed_waga.types.jnode.stats import encode__rule_stats
from speedcontrol.detectors.spotspeed_waga.types.stats import RuleStats
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.spotspeed_waga import SpotSpeedWagaEvidence
from speedcontrol.events.position import Position
from speedcontrol.events.schedule import ScheduleInfo
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import Vehicle, ViolationVehicle
from speedcontrol.events.waypoint import Waypoint
from speedcontrol.events.waypoint.violation import ViolationWaypoint
from speedcontrol.trackset.tr import TrTrack
from speedcontrol.trackset.trackset import Trackset


class _TrItemImageInfo(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem):
        super().__init__(tr_item.ts)
        self.tr_item = tr_item


def _create_detector_info(conf: RuleConf) -> DetectorInfo:
    schedule_info = ScheduleInfo(
        conf.schedule.min_time,
        conf.schedule.max_time,
        conf.schedule.days_to_bool)

    ticket_info = TicketInfo(conf.ticket.name)

    rule_info = RuleInfo(conf.name, schedule_info, ticket_info, {
        'speed-limit': conf.limits.speed_limit,
        'max-overspeed': conf.limits.over_speed,
        'types': list(conf.limits.vehicle_types)
    })

    return DetectorInfo(conf.detector_name, rule_info)


def _get_ra_speed_max_error_ua(lane_ra_speed: int) -> int:
    return 2 if lane_ra_speed < 100 else (lane_ra_speed * 0.02)


# noinspection PyUnusedLocal
def _get_ra_speed_max_error_uz(lane_ra_speed: int) -> int:
    return 0


# noinspection PyUnusedLocal
def _get_ra_speed_max_error_kz(lane_ra_speed: int) -> int:
    return 2


LaneToTrTrackItemList = List[Tuple[str, TrTrackItem]]


class Rule:
    def __init__(self,
                 deps: RuleDeps,
                 conf: RuleConf,
                 stats_name: str,
                 init_im_env: ImBuilderEnv,
                 fini_im_env: ImBuilderEnv,
                 best_im_env: ImBuilderEnv):

        self._deps = deps
        self._conf = conf
        self._info = _create_detector_info(conf)
        self._stats = RuleStats()

        self._stats_name = stats_name  # todo: remove

        self._schedule = Schedule(conf.schedule, deps.tz)

        self._init_im_env = init_im_env
        self._fini_im_env = fini_im_env
        self._best_im_env = best_im_env

        bylaws = CtrlUiconf.get_bylaws()
        if bylaws in ['ua', 'default']:
            self._get_max_error_fn = _get_ra_speed_max_error_ua
        elif bylaws == 'kz':
            self._get_max_error_fn = _get_ra_speed_max_error_kz
        elif bylaws == 'uz':
            self._get_max_error_fn = _get_ra_speed_max_error_uz
        else:
            eventlog.log_e(f'fail: unsupported (bylaws:{bylaws})')
            raise NotImplementedError()

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        json_handler(encode__rule_stats(self._stats_name, self._stats))

    def _test_suitable_class(self, plate_utf8: bytes, vehicle_class: str) -> bool:
        if vehicle_class not in self._conf.limits.vehicle_types:
            self._stats.class_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_utf8}, reason:class-unmatch, class:{vehicle_class})')
            return False

        return True

    def _test_suitable_schedule(self, plate_utf8: bytes, track: Trackset) -> bool:
        suitable = self._schedule.check_ts_usec(track.lr_track.min_ts) \
                   and self._schedule.check_ts_usec(track.lr_track.max_ts)

        if not suitable:
            self._stats.schedule_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_utf8}, reason:schedule-unmatch)')
            return False

        return True

    def _filter_tr_items_by_lane(self, tr_track: TrTrack, lanes_limit: Set[str]) -> List[TrTrackItem]:
        lane_items: List[TrTrackItem] = []
        for tr_item in tr_track.values():
            if tr_item.lr_item is None:
                continue

            if len(lanes_limit) == 0:
                lane_items.append(tr_item)
                continue

            lane_ref: Optional[str] = tr_item.lane_item.lane if tr_item.lane_item is not None else None

            if lane_ref in lanes_limit:
                lane_items.append(tr_item)

        return lane_items

    def check(self,
              trackset: Trackset,
              base_event: BaseEvent) -> Optional[Tuple[SpotSpeedWagaEvidence, TimeSpan]]:

        self._stats.total += 1
        plate_text_utf8: bytes = base_event.plate_utf8

        if not self._test_suitable_class(plate_text_utf8, base_event.best_wp.vehicle.mmr_class):
            return

        elif not self._test_suitable_schedule(plate_text_utf8, trackset):
            return

        spotspeeds = base_event.best_wp.vehicle.spotspeeds

        main_speed = spotspeeds.main.measured
        max_speed = self._conf.limits.speed_limit + self._conf.limits.over_speed
        max_error = self._get_max_error_fn(main_speed)

        if not spotspeeds.assert_overspeed(main_speed, max_speed + max_error):
            self._stats.raspeed_valid += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:speed-allowed, '
                           f'main-speed:{main_speed}, assert-speed:{max_speed + max_error})')
            return

        lane_items = self._filter_tr_items_by_lane(trackset.tr_track, self._conf.limits.lanes)
        if len(lane_items) < 3:
            self._stats.lane_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:lane-unmatch, '
                           f'source-items:{len(trackset.tr_track)}, target-items:{len(lane_items)})')
            return

        #
        # Create an evidence
        #

        self._stats.make_evidence_total += 1

        lr_imref = trackset.lr_track.imref
        image_infos: List[_TrItemImageInfo] = []
        for tr_item in lane_items:
            if not check_quad_dist_from_img_borders(tr_item.lr_item.plate_quad, lr_imref.width, lr_imref.height):
                image_infos.append(_TrItemImageInfo(tr_item))

        # TODO: how many images are required?
        if len(image_infos) < 2:
            self._stats.lane_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:lane-unmatch, '
                           f'source-items:{len(trackset.tr_track)}, target-items:{len(image_infos)})')
            return

        image_infos.sort(key=lambda ref: ref.ts)

        init_image_infos = image_infos[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)

        fini_image_infos = image_infos[:-6:-1]
        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)

        best_image_infos = image_infos  # TODO

        init_image_info = self._deps.vs.export_image_rgba(
            self._init_im_env.scene_rgba, init_image_infos, self._init_im_env.scene_dh)

        best_image_info = self._deps.vs.export_image_rgba(
            self._best_im_env.scene_rgba, best_image_infos, self._best_im_env.scene_dh)

        fini_image_info = self._deps.vs.export_image_rgba(
            self._fini_im_env.scene_rgba, fini_image_infos, self._fini_im_env.scene_dh)

        if None in [init_image_info, fini_image_info, best_image_info]:
            self._stats.evidence_failed_load_images += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:{"load-images"})')
            return

        return self._mk_evidence(
            base_event, trackset,
            spotspeeds.main.measured,
            spotspeeds.cam_2,
            self._conf.limits.speed_limit,
            init_image_info.tr_item,
            fini_image_info.tr_item,
            best_image_info.tr_item)

    def _mk_evidence(self,
                     base_event: BaseEvent,
                     trackset: Trackset,
                     main_speed_kmh: int,
                     control_speed_kmh: int,
                     speed_limit: int,
                     init_tr_item: TrTrackItem,
                     fini_tr_item: TrTrackItem,
                     best_tr_item: TrTrackItem) -> Optional[Tuple[SpotSpeedWagaEvidence, TimeSpan]]:

        # get best vehicle
        best_vehicle: Vehicle = base_event.best_wp.vehicle

        def mk_waypoint(services: RuleDeps,
                        im_env: ImBuilderEnv,
                        tr_item: TrTrackItem,
                        w_type: str = Waypoint.NAME_VIOLATION) -> ViolationWaypoint:

            waypoint_ts = im_env.scene_rgba.ts
            # waypoint_pos = Position(tr_item.ra_item.raw_f, tr_item.ra_item.raw_l)
            waypoint_pos = Position(14, 14)  # TODO
            wp_dir = trackset.tr_track.get_item_dir(tr_item.ts)

            wp_dir_alias = self._conf.footer_defvars.get(encode_dir_text(wp_dir))

            waypoint_origin = services.about.get_origin(waypoint_ts)

            if tr_item.lane_item is not None:
                waypoint_origin.location.set_lane(tr_item.lane_item.lane)

            footer_vars = ImFooterVars(
                services.tz, self._deps.about, waypoint_ts, defvars=self._conf.footer_defvars,
                location=waypoint_origin.location,
                geolocation=waypoint_origin.geolocation,
                plate_text=base_event.plate_ucode,
                spot_speed=main_speed_kmh,
                direction=wp_dir,
                spot_speed_cam_2=control_speed_kmh,
                spot_speed_ra_n=main_speed_kmh,
                speed_limit=speed_limit,
                spot_overspeed_wim=main_speed_kmh - speed_limit,
                spot_overspeed_main=main_speed_kmh - speed_limit,
                spot_speed_cam_n=best_vehicle.spotspeeds.cam_n,
                spot_speed_wim=best_vehicle.spotspeeds.wim,
                spot_speed_main=best_vehicle.spotspeeds.main,
                spot_speed_control=best_vehicle.spotspeeds.control)

            if w_type == Waypoint.NAME_VIOLATION:
                scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(waypoint_ts)
                im_blobs = im_env.builder.build(
                    im_env.scene_rgba, scene_dt_utc, waypoint_origin.geolocation,
                    footer_vars, trackset.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True)

                if len(im_blobs) != 3:
                    eventlog.log_w('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint_v = ViolationVehicle(
                    direction=wp_dir,
                    direction_alias=wp_dir_alias,
                    position=waypoint_pos,
                    plate_quads=[tr_item.lr_item.plate_quad])

                waypoint = ViolationWaypoint(
                    ViolationWaypoint.NAME_VIOLATION,
                    waypoint_ts, None, waypoint_origin.location, im_blobs,
                    waypoint_v)

            elif w_type == Waypoint.NAME_VIOLATION_BEST:
                scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(waypoint_ts)
                im_blobs = im_env.builder.build(
                    im_env.scene_rgba, scene_dt_utc, waypoint_origin.geolocation,
                    footer_vars, trackset.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True, thumb=True)

                if len(im_blobs) != 4:
                    eventlog.log_w('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint_v = ViolationVehicle(
                    direction=wp_dir,
                    direction_alias=wp_dir_alias,
                    position=waypoint_pos,
                    plate_quads=[tr_item.lr_item.plate_quad])

                waypoint = ViolationWaypoint(
                    ViolationWaypoint.NAME_VIOLATION_BEST,
                    waypoint_ts, None, waypoint_origin.location, im_blobs,
                    waypoint_v)

            else:
                eventlog.log_e('fail: unsupported')
                raise RuntimeError

            return waypoint

        try:
            init_wp = mk_waypoint(self._deps, self._init_im_env, init_tr_item)
            fini_wp = mk_waypoint(self._deps, self._fini_im_env, fini_tr_item)
            best_wp = mk_waypoint(self._deps, self._best_im_env, best_tr_item, Waypoint.NAME_VIOLATION_BEST)
        except RuntimeError:
            self._stats.evidence_failed_make_images += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                base_event.plate_utf8, 'make-images'))
            return None

        event_ts = self._fini_im_env.scene_rgba.ts
        direction = trackset.get_direction()

        evidence = SpotSpeedWagaEvidence(
            self._deps.about.get_origin(event_ts), event_ts, main_speed_kmh,
            control_speed_kmh, self._info, direction, self._conf.limits.speed_limit)

        evidence.add_waypoints([init_wp, best_wp, fini_wp])
        return evidence, TimeSpan(init_wp.main_ts, fini_wp.main_ts)
