from typing import Any, Callable, Dict, Final, Set

from edge.types.blob import BsChanRef, BlobProfileToBsChanMap
from edge.types.identity import Identity

from speedcontrol.trackset.trackset import Trackset
from speedcontrol.events.event import BaseEvent


class DetectorError(RuntimeError):
    pass


class IBaseDetector:

    def set_handler(self, event_handler: Callable[[Trackset, BaseEvent], None]) -> None:
        raise NotImplementedError

    def analyze(self, track: Trackset) -> None:
        raise NotImplementedError

    def destroy(self) -> None:
        pass

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        pass


class IDetector:
    TYPE_BASE = 'base'
    TYPE_BUSLANE = 'bus-lane'
    TYPE_BUSLANE_RULE = 'bus-lane~rule'
    TYPE_PARKING = 'parking'
    TYPE_PARKING_RULE = 'parking~rule'
    TYPE_REDLIGHT = 'red-light'
    TYPE_REDLIGHT_RULE = 'red-light~rule'
    TYPE_SECTIONSPEED = 'section-speed'
    TYPE_SECTIONSPEED_RULE = 'section-speed~rule'
    TYPE_SPOTSPEED = 'spot-speed'
    TYPE_SPOTSPEED_RULE = 'spot-speed~rule'
    TYPE_SPOTSPEED_WAGA = 'spot-speed-waga'
    TYPE_SPOTSPEED_WAGA_RULE = 'spot-speed-waga~rule'
    TYPE_TRANSIT = 'transit'
    TYPE_WANTED = 'wanted'
    TYPE_WANTED_RULE = 'wanted~rule'
    TYPE_WRONG_WAY = 'wrong-way'
    TYPE_WRONG_WAY_RULE = 'wrong-way~rule'
    TYPE_SEATBELT = 'seatbelt'
    TYPE_SEATBELT_RULE = 'seatbelt~rule'

    @property
    def identity(self) -> Identity:
        raise NotImplementedError()

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        raise NotImplementedError()

    def set_handler(self, event_handler: Callable[[Trackset, BaseEvent], None]) -> None:
        raise NotImplementedError()

    def analyze(self, trackset: Trackset, base_event: BaseEvent) -> None:
        raise NotImplementedError()

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        pass

    def destroy(self) -> None:
        pass


class DetectorUnitInfo:
    def __init__(self, identity: Identity, conf: Dict, footer_defvars: Dict[str, Any]):
        self.identity: Final[Identity] = identity
        self.conf: Final[Dict] = conf
        self.footer_defvars: Final[Dict[str, Any]] = footer_defvars


class DetectorGroups:
    def __init__(self):
        self.plate_absent: Set[IDetector] = set()
        self.plate_detected: Set[IDetector] = set()
        self.plate_lost: Set[IDetector] = set()
