from typing import Dict, List, Optional

from edge import eventlog
from edge.types.timespan import TimeSpan, decode__timespans
from edge.types.exceptions import JnodeError

from speedcontrol.services.types.leddar import LeddarServiceConf, ILeddarService


class _LeddarLog:
    def __init__(self, time: TimeSpan, active: List[TimeSpan]):
        self.active = active
        self.time = time

    def check(self, time_span: TimeSpan) -> List[TimeSpan]:
        spans = [x for x in self.active if x.check_overlap(time_span)]
        return spans


def decode__leddar_log(jnode: Dict, leddar_ref: str) -> _LeddarLog:
    """
    :raise JnodeError:
    """
    ts_jnode = jnode.get('ts')

    if type(ts_jnode) != list or len(ts_jnode) != 2:
        eventlog.log_e_param('#/ts')
        raise JnodeError

    time_interval = TimeSpan(*ts_jnode)

    leddar_list_jnode = jnode.get('leddar')
    if type(leddar_list_jnode) != list:
        eventlog.log_e_param('#/leddar')
        raise JnodeError

    for leddar_jnode in leddar_list_jnode:
        if leddar_jnode['ref'] != leddar_ref:
            continue

        active_jnode = leddar_jnode['active']
        active_intervals = decode__timespans(active_jnode)
        return _LeddarLog(time_interval, active_intervals)

    raise JnodeError


class _Service(ILeddarService):
    def __init__(self, conf: LeddarServiceConf):
        super().__init__()

        self._leddar_log: Optional[_LeddarLog] = None
        self._leddar_ref = conf.ref

    def ref(self) -> str:
        return self._leddar_ref

    def update(self, react_jnode: Dict) -> None:
        # noinspection PyBroadException
        try:
            self._leddar_log = decode__leddar_log(react_jnode, self._leddar_ref)
        except Exception:
            eventlog.log_x()
            return

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        if self._leddar_log is None:
            return []

        time_interval = TimeSpan(min_ts, max_ts)
        if not self._leddar_log.time.check_overlap(time_interval):
            # there is no stats for [cts ,lts]
            return []

        violation_list = self._leddar_log.check(time_interval)
        return violation_list


class _StubService(ILeddarService):
    def __init__(self):
        super().__init__()

    def update(self, react_jnode: Dict) -> None:
        pass

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        return []


def init_service(conf: Optional[LeddarServiceConf]) -> ILeddarService:
    return _StubService() if conf is None else _Service(conf)
