import ujson

from io import BytesIO
from typing import Dict, Final, Optional, Callable

from twisted.internet.defer import Deferred
from twisted.internet.protocol import Protocol, connectionDone
from twisted.python.failure import Failure
from twisted.web.client import Agent, Response, FileBodyProducer
from twisted.web.http_headers import Headers

from edge import eventlog
from edge.types.image import ImRgba
from edge.types.shape.rect import RectF, RectI

from speedcontrol.services.types.raas import RaasServiceConf, IRaasService
from speedcontrol.trackset.jnode.raas import RaasScene, decode__raas_scene
from speedcontrol.trackset.vrp import TrackVrp


class RaasError(RuntimeError):
    pass


class _ResponseBodyProto(Protocol):
    def __init__(self, handler: Deferred):
        """
        :param handler:
            Deferred - callbacks to str
        """
        self._handler = handler
        self._body_buffer = ''

    def dataReceived(self, body_data: str) -> None:
        self._body_buffer += body_data

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        self._handler.callback(self._body_buffer[: -1])  # strip '\0'


class _Service(IRaasService):
    def __init__(self, twisted_reactor, args: RaasServiceConf):
        super().__init__()

        self._agent: Final[Agent] = Agent(twisted_reactor)
        self._base_url: str = f'http://{args.raas_host}:{args.raas_port}/alpr/{args.raas_user}/'
        self._password: Final[bytes] = args.raas_pass.encode('utf-8')

    @staticmethod
    def _mk_signature(plate_image_bytes_io: BytesIO, secret: bytes) -> str:
        from hashlib import md5

        md5_hash = md5(plate_image_bytes_io.getvalue())
        md5_hash.update(BytesIO(secret).getvalue())
        signature = md5_hash.hexdigest()

        return signature

    def analyze(self, image_rgba: ImRgba, image_roi: RectI,
                handler: Callable[[TrackVrp | None], None]) -> None:

        mem_file = BytesIO()
        image_rgba.with_crop_image(image_roi, lambda crop_im: crop_im.save(mem_file, 'JPEG', quality=90))

        plate_image_signature: str = self._mk_signature(mem_file, self._password)

        mem_file.seek(0)
        qbody_producer = FileBodyProducer(mem_file)

        call_chain = self._agent.request(
            'POST', self._base_url + plate_image_signature,
            Headers({'User-Agent': ['EDGE'], 'Content-Type': ['image/jpeg']}),
            qbody_producer
        ).addCallbacks(
            callback=self._handle_reply_http_header_done,
            errback=self._handle_reply_http_header_fail
        ).addCallbacks(
            callback=self._handle_reply_http_body_done,
            errback=self._handle_reply_http_body_fail)

        # decode raas result

        def handle_decode_raas_scene_done(scene: RaasScene) -> None:
            best_vrp: Optional[TrackVrp] = scene.best_vrp_variant
            if best_vrp is None:
                eventlog.log_d('fail: decode raas scene (scene is empty)')
                handler(None)

            handler(best_vrp)

        def handle_decode_raas_scene_fail(failure: Failure) -> None:
            if not failure.check([RaasError]):
                eventlog.log_w(f'fail: decode raas scene ({failure})')

            handler(None)

        call_chain.addCallbacks(
            callback=handle_decode_raas_scene_done,
            errback=handle_decode_raas_scene_fail)

    @staticmethod
    def _handle_reply_http_header_done(response: Response) -> Deferred[str]:
        """
        :raise RaasError:
        """
        if response.code != 200:
            eventlog.log_w(f'fail: invalid-reply (http-code:{response.code})')
            raise RaasError()

        call_chain = Deferred()
        response.deliverBody(_ResponseBodyProto(call_chain))
        return call_chain

    @staticmethod
    def _handle_reply_http_header_fail(failure: Failure) -> None:
        """
        :raise RaasError:
        """
        eventlog.log_w(f'fail: invalid-query ({failure})')
        raise RaasError

    @staticmethod
    def _handle_reply_http_body_done(response_body: str) -> RaasScene:
        jnode: Dict = ujson.loads(response_body)
        return decode__raas_scene(jnode)

    @staticmethod
    def _handle_reply_http_body_fail(failure: Failure) -> None:
        """
        :raise RaasError:
        """
        if not failure.check([RaasError]):
            eventlog.log_w(f'fail: load response body ({failure})')

        raise RaasError


class _StubService(IRaasService):
    def __init__(self):
        super().__init__()
        self._scene: Final[RaasScene] = RaasScene(items=[])

    def analyze(self,
                image_rgba: ImRgba,
                image_roi: RectF,
                handler: Callable[[TrackVrp | None], None]) -> None:
        handler(None)


def init_service(twisted_reactor, conf: Optional[RaasServiceConf]) -> IRaasService:
    return _StubService() if conf is None else _Service(twisted_reactor, conf)
