from typing import Callable, Dict, List, Optional, Set, Tuple

from edge import eventlog
from edge.services.react import ReactSource
from edge.services.types.lane import ILaneService
from edge.types.shape.quad import QuadI

from speedcontrol.services.types.react_tr import ITrReactService, TrReactServiceConf
from speedcontrol.trackset.jnode.trackset import decode__trackset
from speedcontrol.trackset.trackset import Trackset


def _get_list_min_max_med(float_list: List[float]) -> Tuple[float, float, float]:
    min_value = 0.0
    max_value = 0.0
    med_value = 0.0

    float_list_len = len(float_list)
    if float_list_len > 0:
        float_list.sort()
        min_value = float_list[0]
        max_value = float_list[-1]
        med_value = float_list[float_list_len // 2]

    return min_value, max_value, med_value


class _ServiceStats:
    def __init__(self):
        self.lrra_track_count = int(0)
        self.lr_track_count = int(0)
        self.ra_track_count = int(0)
        self.track_count = int(0)

        self.near_plate_w_list: List[int] = []
        self.near_plate_h_list: List[int] = []
        self.far_plate_w_list: List[int] = []
        self.far_plate_h_list: List[int] = []

    def reset(self) -> None:
        self.near_plate_w_list[:] = []
        self.near_plate_h_list[:] = []
        self.far_plate_w_list[:] = []
        self.far_plate_h_list[:] = []

    def inc_plate_quad(self, catch_quad: QuadI, lost_quad: QuadI) -> None:
        catch_rect_w, catch_rect_h = catch_quad.wh
        lost_rect_w, lost_rect_h = lost_quad.wh

        if catch_rect_w * catch_rect_h >= lost_rect_w * lost_rect_h:
            self.near_plate_w_list.append(catch_rect_w)
            self.near_plate_h_list.append(catch_rect_h)

            self.far_plate_w_list.append(lost_rect_w)
            self.far_plate_h_list.append(lost_rect_h)

        else:
            self.near_plate_w_list.append(lost_rect_w)
            self.near_plate_h_list.append(lost_rect_h)

            self.far_plate_w_list.append(catch_rect_w)
            self.far_plate_h_list.append(catch_rect_h)

    def inc_track(self, trackset: Trackset) -> None:
        self.track_count += 1

        if bool(trackset.lr_track) and bool(trackset.ra_track):
            self.lrra_track_count += 1

        if bool(trackset.lr_track):
            self.lr_track_count += 1

        if bool(trackset.ra_track):
            self.ra_track_count += 1


class _Service(ITrReactService):
    def __init__(self, twisted_reactor, conf: TrReactServiceConf, lane_service: ILaneService):
        super().__init__()
        self._conf = conf
        self._stats = _ServiceStats()
        self._source = ReactSource(twisted_reactor)
        self.set_endpoints(conf.endpoints)
        self._source.set_handler(self._handle_tr_react)
        self._handler: Callable[[Trackset], None] | None = None
        self._lane_service = lane_service

    def destroy(self) -> None:
        self._source.destroy()

    def set_handler(self, handler: Callable[[Trackset], None]) -> None:
        assert self._handler is None
        self._handler = handler

    def set_endpoints(self, endpoints: Set[str]) -> None:
        self._source.set_endpoints(endpoints)

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        npw_min, npw_max, npw_med = _get_list_min_max_med(self._stats.near_plate_w_list)
        nph_min, nph_max, nph_med = _get_list_min_max_med(self._stats.near_plate_h_list)
        fpw_min, fpw_max, fpw_med = _get_list_min_max_med(self._stats.far_plate_w_list)
        fph_min, fph_max, fph_med = _get_list_min_max_med(self._stats.far_plate_h_list)

        json_handler({
            'type': 'tr-react',
            'name': 'tr-react.0',
            'stat': {
                'count': {
                    'lr': self._stats.lr_track_count,
                    'ra': self._stats.ra_track_count,
                    'lrra': self._stats.lrra_track_count,
                    'total': self._stats.track_count
                },
                'plate': {
                    'near-w': [npw_min, npw_med, npw_max],
                    'near-h': [nph_min, nph_med, nph_max],
                    'far-w': [fpw_min, fpw_med, fpw_max],
                    'far-h': [fph_min, fph_med, fph_max]
                }
            }
        })

        self._stats.reset()

    def _handle_tr_react(self, react_name: str, react_jnode: Dict) -> None:
        if self._handler is None:
            return

        try:
            trackset: Optional[Trackset] = decode__trackset(react_jnode)
        except (KeyError, ValueError) as x:
            eventlog.log_e(f'fail: unpack {react_name} <{x.message}>')
            return

        if trackset is not None:
            self._stats.inc_track(trackset)

            lr_track = trackset.lr_track
            if lr_track is not None:
                init_quad = lr_track[0].plate_quad
                fini_quad = lr_track[-1].plate_quad
                self._stats.inc_plate_quad(init_quad, fini_quad)

            self._handler(trackset)


def init_service(twisted_reactor, conf: TrReactServiceConf, lane_service: ILaneService) -> ITrReactService:
    return _Service(twisted_reactor, conf, lane_service)
