import time
import os

from collections import defaultdict
from typing import Dict, List

from edge.support import bs, fs
from edge import eventlog
from edge.types.stats import IStatsProvider

from speedcontrol.services.types.stats import StatsServiceConf, IEventCounters, IStatService


class BsCounters(bs.IBsCounters):
    def __init__(self):
        bs.IBsCounters.__init__(self)

        def _dict_factory():
            return {
                'blob-count': 0,
                'done-count': 0,
                'fail-count': 0
            }
        self._counters = defaultdict(_dict_factory)

    def inc_blob(self, chan: str) -> None:
        self._counters[chan]['blob-count'] += 1

    def inc_done(self, chan: str) -> None:
        self._counters[chan]['done-count'] += 1

    def inc_fail(self, chan: str) -> None:
        self._counters[chan]['fail-count'] += 1

    def serialize(self) -> Dict:
        return self._counters


class _EventCounters(IEventCounters):
    def __init__(self):
        self._fact_count = int(0)

    def inc_fact(self) -> None:
        self._fact_count += 1

    @property
    def fact_count(self) -> int:
        return self._fact_count


class _StatService(IStatService):
    def __init__(self, twisted_reactor, conf: StatsServiceConf):
        super().__init__()

        self._providers = list()

        self._event = _EventCounters()
        self._bs = BsCounters()

        self._stat_abspath = conf.stat_filepath
        self._stat_tmp_abspath = conf.stat_filepath + '.tmp'
        self._stat_tout_sec = conf.stat_tout_sec

        # reset on start
        twisted_reactor.callLater(0, self._report_counters)

        if conf.stat_tout_sec > 0:
            twisted_reactor.callLater(conf.stat_tout_sec, self._log_counters, twisted_reactor)

    @property
    def event(self) -> IEventCounters:
        return self._event

    @property
    def bs(self) -> bs.IBsCounters:
        return self._bs

    def add_providers(self, providers: List[IStatsProvider]) -> None:
        self._providers.extend(providers)

    def _log_counters(self, twisted_reactor) -> None:
        self._report_counters()

        twisted_reactor.callLater(
            self._stat_tout_sec, self._log_counters, twisted_reactor)

    def _report_counters(self) -> None:
        sys_utc_time = time.time()

        units = [
            {
                'type': 'stats',
                'name': 'stats',
                'stat': {
                    'mtime': [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                    'tout-sec': self._stat_tout_sec
                }
            }
        ]

        list(map(lambda provider: provider.collect_stats(units.append), self._providers))
        stat = {
            'units': units,
            'stat': {
                'mtime': [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                'tout-sec': self._stat_tout_sec
            },
            'event': {
                'count': self.event.fact_count
            },
            'bs': self.bs.serialize()
        }

        try:
            fs.write_json_file(stat, self._stat_tmp_abspath, True)
            os.rename(self._stat_tmp_abspath, self._stat_abspath)
        except IOError as x:
            eventlog.log_w(f'fail: write <{self._stat_abspath}> ({x})')


def init_service(twisted_reactor, args: StatsServiceConf) -> IStatService:
    return _StatService(twisted_reactor, args)
