from typing import Dict, List, Optional, Self, Tuple

from twisted.internet import inotify
from twisted.python.filepath import FilePath

from edge import eventlog
from edge.support import fs
from edge.types.timespan import TimeSpan, decode__timespans
from edge.types.exceptions import JnodeError

from speedcontrol.services.types.tl import TlServiceConf, ITlService


def custom_max(left: Optional[int | float], rigth: Optional[int | float]) -> Optional[int | float]:
    if None not in [left, rigth]:
        return max(left, rigth)
    elif left is not None:
        return left
    elif rigth is not None:
        return rigth


def custom_min(left: Optional[int | float], rigth: Optional[int | float]) -> Optional[int | float]:
    if None not in [left, rigth]:
        return min(left, rigth)
    elif left is not None:
        return left
    elif rigth is not None:
        return rigth


class _StubService(ITlService):
    def __init__(self):
        ITlService.__init__(self)

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        return []

    def get_times(self, ts: int) -> Tuple[Optional[int], Optional[int], Optional[int], Optional[int]]:
        return None, None, None, None


def _print_ti(string, ti_l: List[TimeSpan]) -> None:
    print(string)
    for ti in ti_l:
        print(ti.min_ts, '-', ti.max_ts)


class _TrafficLight:
    def __init__(self, ch_id: int,
                 red: List[TimeSpan],
                 amber: List[TimeSpan],
                 green: List[TimeSpan],
                 red_only: bool = False):

        self.ch_id = ch_id
        self.red = red
        self.amber = amber
        self.green = green

        self.prohibiting = red if red_only else self._get_prohibiting()

    def get_red_time(self, ts: int) -> Optional[int]:
        pr_list = [time_span for time_span in self.prohibiting if ts in time_span]

        if not len(pr_list):
            # current collor is green
            return None

        pr_item: TimeSpan = pr_list[0]

        # find red intervals in prohibited item
        red_spans = [span for span in self.red if span.check_overlap(pr_item)]

        # red started before ts
        red_spans = [span for span in red_spans if span.min_ts < ts]
        if not len(red_spans):
            return None

        # check "just red" instead of "any red"
        # just_red = any_red - any_amber
        any_red_time = ts - red_spans[0].min_ts

        post_amber_time = self.get_red_amber_time(ts) or 0
        return (any_red_time - post_amber_time) // 1000

    def get_green_amber_time(self, ts: int) -> Optional[int]:
        pr_list = [time_span for time_span in self.prohibiting if ts in time_span]

        if not len(pr_list):
            # current collor is green
            return None

        pr_item = pr_list[0]

        # find amber intervals in prohibited item
        amber_spans = [span for span in self.amber if span.check_overlap(pr_item)]

        # amber started before ts
        amber_spans = [span for span in amber_spans if span.min_ts < ts]

        if not len(amber_spans):
            return None

        min_red = self._search_min(self.red, min_ts=pr_item.min_ts, max_ts=pr_item.max_ts)
        min_amber = amber_spans[0].min_ts

        if min_red is None or ts < min_red:
            return (ts - min_amber) // 1000

        pre_amber = min_red - min_amber
        return pre_amber // 1000

    def get_red_amber_time(self, ts: int) -> Optional[int]:
        pr_list = [time_span for time_span in self.prohibiting if ts in time_span]

        if not len(pr_list):
            # current collor is green
            return None

        pr_item = pr_list[0]

        # find amber intervals in prohibited item
        amber_spans = [span for span in self.amber if span.check_overlap(pr_item)]

        # amber started before ts
        amber_spans = [span for span in amber_spans if span.min_ts < ts]
        if not len(amber_spans):
            return None

        min_red = self._search_min(self.red, min_ts=pr_item.min_ts, max_ts=pr_item.max_ts)
        if min_red and min_red < amber_spans[-1].min_ts:
            return (ts - amber_spans[-1].min_ts) // 1000

        else:
            return None  # ?

    @staticmethod
    def _search_max(items: List[TimeSpan],
                    min_ts: Optional[int] = None,
                    max_ts: Optional[int] = None
                    ) -> Optional[int]:

        if len(items) == 0:
            return None

        if max_ts is None:
            return items[-1].max_ts

        if min_ts is None:
            min_ts = 0

        for index in range(len(items) - 1, -1, -1):
            item = items[index]
            if min_ts < item.max_ts < max_ts:
                return item.max_ts

        return None

    @staticmethod
    def _search_min(items: List[TimeSpan],
                    min_ts: Optional[int] = None,
                    max_ts: Optional[int] = None
                    ) -> Optional[int]:

        if len(items) == 0:
            return None

        if min_ts is None:
            return items[0].min_ts

        if max_ts is None:
            # add 1 to include self last item in case of opened interval
            max_ts = items[-1].min_ts + 1

        for item in items:  # type: TimeSpan
            if min_ts < item.min_ts < max_ts:
                return item.min_ts

        return None

    def _get_prohibiting(self) -> List[TimeSpan]:
        prohibiting_items = []

        # if amber or red catch ts < green cts ts - add first item to prohibiting list
        min_time = None
        if len(self.amber) and len(self.red):
            min_time = min(self.amber[0].min_ts, self.red[0].min_ts)
        elif len(self.amber):
            min_time = self.amber[0].min_ts
        elif len(self.red):
            min_time = self.red[0].min_ts

        if min_time is not None and len(self.green) and min_time < self.green[0].min_ts:
            max_time_red = self._search_max(self.red, max_ts=self.green[0].min_ts)
            max_time_amber = self._search_max(self.amber, max_ts=self.green[0].min_ts)
            max_time = custom_max(max_time_amber, max_time_red)
            if None not in [max_time, min_time]:
                prohibiting_items.append(TimeSpan(min_time, max_time))

        # all spaces between green are prohibiting intervals
        for index in range(0, len(self.green) - 1, 1):
            max_time_red = self._search_max(self.red, max_ts=self.green[index + 1].min_ts)
            max_time_amber = self._search_max(self.amber, max_ts=self.green[index + 1].min_ts)
            max_time = custom_max(max_time_amber, max_time_red)

            min_time_red = self._search_min(self.red, min_ts=self.green[index].max_ts,
                                            max_ts=self.green[index + 1].min_ts)

            min_time_amber = self._search_min(self.amber, min_ts=self.green[index].max_ts,
                                              max_ts=self.green[index + 1].min_ts)

            min_time = custom_min(min_time_red, min_time_amber)
            if None not in [max_time, min_time]:
                prohibiting_items.append(TimeSpan(min_time, max_time))

        # if amber or red lost ts > green lost ts - add one more item to prohibiting list
        max_time = None
        if len(self.amber) and len(self.red):
            max_time = max(self.amber[-1].max_ts, self.red[-1].max_ts)
        elif len(self.amber):
            max_time = self.amber[-1].max_ts
        elif len(self.red):
            max_time = self.red[-1].max_ts

        # max_time = max(self.amber[-1].max_ts, self.red[-1].max_ts)
        if max_time is not None and len(self.green) and self.green[-1].max_ts < max_time:
            min_time_red = self._search_min(self.red, min_ts=self.green[-1].max_ts)
            min_time_amber = self._search_min(self.amber, min_ts=self.green[-1].max_ts)
            min_time = custom_min(min_time_red, min_time_amber)
            if None not in [max_time, min_time]:
                prohibiting_items.append(TimeSpan(min_time, max_time))

        return prohibiting_items

    def check(self, time_span: TimeSpan) -> List[TimeSpan]:
        spans = [x for x in self.prohibiting if x.check_overlap(time_span)]
        return spans


class TlLog:
    def __init__(self, time: TimeSpan, channels: List[_TrafficLight], red_only: bool = False):
        self.channels = channels
        self.time = time
        self.red_only = red_only

    @classmethod
    def deserialize(cls, tl_ident: int, jnode: Dict, red_only: bool = False) -> Self:
        """
        :raise JnodeError:
        """
        ts_jnode = jnode.get('ts')

        if type(ts_jnode) != list or len(ts_jnode) != 2:
            eventlog.log_e_param('#/ts')
            raise JnodeError

        time_interval = TimeSpan(*ts_jnode)

        traffic_lights_jnode = jnode.get('tl')
        if type(traffic_lights_jnode) != list:
            eventlog.log_e_param('#/tl')

        traffic_lights = []
        for tl_jnode in traffic_lights_jnode:
            if tl_ident != tl_jnode['id']:
                continue

            red = tl_jnode.get('red', [])

            amber = []
            green = []
            if not red_only:
                amber = tl_jnode.get('amber', [])
                green = tl_jnode.get('green', [])

            red_intervals = decode__timespans(red)
            amber_intervals = decode__timespans(amber)
            green_intervals = decode__timespans(green)

            traffic_light = _TrafficLight(tl_ident, red_intervals, amber_intervals, green_intervals, red_only)
            traffic_lights.append(traffic_light)

        if not traffic_lights:
            eventlog.log_e('fail: duplicate idents found')
            raise JnodeError

        return TlLog(time_interval, traffic_lights, red_only)


class _TlService(ITlService):
    def __init__(self, twisted_reactor, args: TlServiceConf):
        ITlService.__init__(self)

        self._twisted_reactor = twisted_reactor

        self._tl_log: Optional[TlLog] = None
        self._tl_ident = args.tl_ident
        self._red_only = args.red_only

        self._endpoint = args.endpoint
        self._inotify = inotify.INotify()
        self._inotify.startReading()
        self._reset_inotify_watch()

    def destroy(self) -> None:
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

    def _reset_inotify_watch(self) -> None:
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

        fs.silent_mkdir_r(self._endpoint)

        try:
            self._inotify.watch(
                FilePath(self._endpoint),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_react])
        except inotify.INotifyError as x:
            eventlog.log_d(f'fail: reset inotify watch ({x})')

    # noinspection PyUnusedLocal
    def _handle_react(self, ignored, filepath, mask) -> None:
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_w(f'endpoint was changed to {self._endpoint}')
            return

        # noinspection PyBroadException
        try:
            react_jnode = fs.load_json3(filepath.path, default=None)
            if react_jnode:
                self._tl_log = TlLog.deserialize(self._tl_ident, react_jnode, self._red_only)

        except JnodeError as x:
            return

        except Exception:
            eventlog.log_x()
            return

        finally:
            fs.silent_rm_r(filepath.path)

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        if self._tl_log is None:
            return []

        time_interval = TimeSpan(min_ts, max_ts)
        if not self._tl_log.time.check_overlap(time_interval):
            # there is no stats for [cts ,lts]
            return []

        ch_list = [ch for ch in self._tl_log.channels if ch.ch_id == self._tl_ident]

        if len(ch_list) != 1:
            eventlog.log_e('fail: find channel')
            return []

        ch_args: _TrafficLight = ch_list[0]

        violation_list = ch_args.check(time_interval)
        return violation_list

    def get_times(self, ts: int) -> Tuple[Optional[int], Optional[int], Optional[int], Optional[int]]:
        """
        :param ts:
            value in msec

        :returns:
            (green-amber-time, red-time, red-amber-time, halt-time)
            Values in msec
        """
        traffic_lights = [ch for ch in self._tl_log.channels if ch.ch_id == self._tl_ident]
        assert 1 == len(traffic_lights)

        tl: _TrafficLight = traffic_lights[0]

        green_amber_time = tl.get_green_amber_time(ts)
        red_time = tl.get_red_time(ts)
        red_amber_time = tl.get_red_amber_time(ts)

        if self._red_only:
            halt_time = red_time
        else:
            halt_time = 0
            if green_amber_time is not None:
                halt_time += green_amber_time

            if red_time is not None:
                halt_time += red_time

            if red_amber_time is not None:
                halt_time += red_amber_time

        return green_amber_time, red_time, red_amber_time, halt_time


def init_service(twisted_reactor, conf: Optional[TlServiceConf]) -> ITlService:
    if conf is None:
        rl = _StubService()
    else:
        rl = _TlService(twisted_reactor, conf)
    return rl
