import datetime
import struct
import time
import ujson

from typing import Final, Optional

from edge.image.helpers import ImHelpers
from edge.services.types.tz import ITzService
from edge.support.fs import silent_mkdir_r
from edge.support.queue import BlobFileTree
from edge.types.image import ImRgba

from speedcontrol.events.event import BaseEvent
from speedcontrol.events.jnode.event import encode__base_event
from speedcontrol.events.waypoint.measurement import BestWaypoint
from speedcontrol.services.types.trainer import TrainerServiceConf, ITrainerService
from speedcontrol.trackset.trackset import Trackset


class _StubTrainerService(ITrainerService):
    def __init__(self):
        super().__init__()

    def collect(self, base_event: BaseEvent, trackset: Trackset, scene_dh_data: ImRgba) -> None:
        pass


class _TrainerService(ITrainerService):
    def __init__(self, conf: TrainerServiceConf, tz_service: ITzService):
        super().__init__()

        silent_mkdir_r(conf.root_dir)

        self._conf: Final[TrainerServiceConf] = conf
        self._tz_service = tz_service
        self._imq = BlobFileTree(conf.root_dir, conf.min_df)

    def collect(self, base_event: BaseEvent, trackset: Trackset, scene_rgba: ImRgba) -> None:

        if trackset.lr_track and self._conf.lr in [TrainerServiceConf.LR_ABSENT]:
            return

        elif trackset.lr_track is None and self._conf.lr in [TrainerServiceConf.LR_PRESRNT]:
            return

        best_waypoint: BestWaypoint = base_event.best_wp
        if not self._conf.lanes or best_waypoint.location.lane in self._conf.lanes:
            image_struct_time: time.struct_time = time.gmtime(scene_rgba.ts // 1000000)
            image_time = datetime.time(image_struct_time.tm_hour, image_struct_time.tm_min)

            if self._conf.min_time < image_time < self._conf.max_time:
                image_jpeg: bytes = ImHelpers.mk_jpeg2(scene_rgba)
                json_bytes = bytes(ujson.dumps({
                    'event': encode__base_event(base_event, self._tz_service),
                    'trackset': trackset.jnode
                }), 'utf-8')

                json_size: bytes = struct.pack('I', len(json_bytes))

                dt_local: datetime = self._tz_service.map_ts_usec_to_datetime_local(scene_rgba.ts)
                blob_name = '.jpeg'
                if best_waypoint.vehicle.vrp_text_ucode:
                    blob_name = f'__{best_waypoint.vehicle.vrp_text_ucode}.jpeg'

                self._imq.add_blob(
                    dt_local,
                    blob_name=blob_name,
                    blob_bytes=image_jpeg + json_bytes + json_size)


def init_service(conf: Optional[TrainerServiceConf], tz_service: ITzService) -> ITrainerService:
    return _StubTrainerService() if conf is None else _TrainerService(conf, tz_service)
