from typing import Dict, Final, List, Optional, Tuple

from edge.types.dir import Dir
from edge.types.jnode.imageref import ImageRef
from edge.types.track import ITrack, LrTrackItem

from speedcontrol.trackset.jnode.vrp import TrackVrp


class LrTrack(ITrack, list):
    def __init__(self,
                 imref: ImageRef,
                 vrp: TrackVrp,
                 vrp_trace: List[LrTrackItem],
                 vrp_dir: Dir,
                 alarms_jnode: List[Dict] | None,
                 stat_detected: bool,
                 stat_lost: bool,
                 type_jnode: Optional[Dict],
                 seats_jnode: List[Dict] | None,
                 cam_speed_n: Optional[float]):
        """
        :param imref:
        :param vrp:
        :param vrp_trace:
        :param vrp_dir:
        :param alarms_jnode:
        :param stat_detected:
            Flag indicating that this LrTrack is a first snapshot of the track.
        :param stat_lost:
            Flag indicating that this LrTrack is a last snapshot of the track.
        :param type_jnode:
            TrTrack#/Lr/Vehicle/type
        :param seats_jnode:
            TrTrack#/Lr/Vehicle/seats
        :param cam_speed_n:
            The value provided by the camera using all track points
        """

        super().__init__(vrp_trace)

        self.imref: Final[ImageRef] = imref
        self.vrp: Final[TrackVrp] = vrp
        self.dir: Final[Dir] = vrp_dir
        self.alarms_jnode: Final[List[Dict] | None] = alarms_jnode
        self.stat_detected: Final[bool] = stat_detected
        self.stat_lost: Final[bool] = stat_lost
        self.seats_jnode: Final[List[Dict] | None] = seats_jnode
        self.cam_speed_n: Final[float | None] = cam_speed_n

        # {
        #     "info": [
        #         {
        #             "class": "CAR"
        #         }
        #     ],
        #     "weight": 36,
        #     "make": "HONDA",
        #     "weight-make": 75,
        #     "weight-class": 100,
        #     "model": "ACCORD",
        #     "class": "CAR"
        # }
        if type_jnode is None:
            self._vehile_type = None

        else:
            self._vehile_type = dict()
            if type_jnode['weight-class'] >= 0.0:
                self._vehile_type['info'] = [{
                    'class':  type_jnode['class']
                }]
                self._vehile_type['class'] = type_jnode['class']
                self._vehile_type['weight-class'] = type_jnode['weight-class']

            if type_jnode['weight-make'] >= 0.0:
                self._vehile_type['make'] = type_jnode['make']
                self._vehile_type['weight-make'] = type_jnode['weight-make']

                if type_jnode['weight'] >= 0.0:
                    self._vehile_type['model'] = type_jnode['model']
                    self._vehile_type['weight'] = type_jnode['weight']

    @property
    def vrp_center_trace(self) -> List[Tuple[int, int]]:
        return [item.plate_quad.center for item in self]

    @property
    def min_ts(self) -> int:
        return self[0].ts

    @property
    def max_ts(self) -> int:
        return self[-1].ts

    @property
    def type_jnode(self) -> Optional[Dict]:
        return self._vehile_type

    def get_item(self, ts) -> Optional[LrTrackItem]:
        for item in self:
            if item.ts == ts:
                return item
