from collections import defaultdict
from functools import cmp_to_key
from typing import Dict, Final, List, Tuple, Optional

from edge import eventlog
from edge.services import cam, lane
from edge.services.types.cam import ICamService
from edge.services.types.lane import ILaneService
from edge.support.service import ServicePack
from edge.types.exceptions import GenericError
from edge.types.imageref import ImageRef
from edge.types.shape.rect import RectI
from edge.types.trend import LrTrend, LrTrendItem

from tracker.types.kernel import KernelConf, ScaleConf
from tracker.services.track_matcher import LrRaMatchStore
from tracker.services.track_builder import RaTrackBuilder
from tracker.services.scene_emitter import SceneEmitter
from tracker.services.scene_provider import SceneProvider
from tracker.services.stats import StatService
from tracker.services.track_debouncer import TrackDebouncer
from tracker.services.track_emitter import init_service
from tracker.services.track_provider import TrackProvider
from tracker.services.types.scene_emitter import ISceneEmitter
from tracker.services.types.scene_provider import ISceneProvider
from tracker.services.types.track_builder import IRaTrackBuilder
from tracker.services.types.track_debouncer import ITrackDebouncer
from tracker.services.types.track_emitter import ITrackEmitter
from tracker.services.types.track_provider import ITrackProvider
from tracker.types.lr import LrScene, LrTrackSet, LrSceneItem, LrTrackStat
from tracker.types.lrra import LrRaSceneItem, LrRaScene, LrRaTrackset
from tracker.types.ra import RaSceneItem, RaScene, RaTrack


def _get_median_fl(ra_items: List[RaSceneItem]) -> Tuple[float, float]:
    ra_items.sort(key=lambda ra_item: ra_item.scene_f_cm)
    index: int = len(ra_items) // 2
    median_f = ra_items[index].scene_f_cm

    ra_items.sort(key=lambda ra_item: ra_item.scene_l_cm)
    index: int = len(ra_items) // 2
    median_l = ra_items[index].scene_l_cm

    return median_f, median_l


def _get_median_fl2(ra_items: List[RaSceneItem]) -> Tuple[float, float]:
    def _cmp(left: RaSceneItem, right: RaSceneItem) -> int:
        left_d = left.scene_f_cm * left.scene_f_cm + left.scene_l_cm * left.scene_l_cm
        right_d = right.scene_f_cm * right.scene_f_cm + right.scene_l_cm * right.scene_l_cm
        return right_d - left_d

    ra_items.sort(key=cmp_to_key(_cmp))
    index: int = len(ra_items) // 2
    median_f = ra_items[index].scene_f_cm
    median_l = ra_items[index].scene_l_cm

    return median_f, median_l


def _debouce_ra_items_pos(lr_trackset: LrTrackSet, ra_track: RaTrack) -> LrRaTrackset:
    lost_tout = int(10e6)  # 10 sec
    dead_tout = int(120e6)  # 2 min

    # Search for time intervals when VRP is in static position
    lr_trend = LrTrend()
    lr_trend.prolong(lr_trackset.lr_track.plate_trace, lost_tout, dead_tout)

    # Grouping of ra-items by time intervals, when VRP is in static position
    spanref_to_ra_items_map: Dict[int, List[RaSceneItem]] = defaultdict(list[RaSceneItem])
    ra_items: List[RaSceneItem] = []

    for ra_item in ra_track.trace:
        spanref: Optional[int] = None
        for trend_item in lr_trend:  # type: LrTrendItem
            if ra_item.frame_ts in trend_item.range:
                spanref = trend_item.range.min_ts
                spanref_to_ra_items_map[spanref].append(ra_item)
                break

        if spanref is None:
            ra_items.append(ra_item)

    # Debounce the position of the ra-items in each group.
    # In each group the ra-items must be static, because the VRP is static
    for span_items in spanref_to_ra_items_map.values():
        median_fl: Tuple[float, float] = _get_median_fl(span_items)
        for ra_item in span_items:  # type: RaSceneItem
            ra_item.scene_f_cm, ra_item.scene_l_cm = median_fl
            ra_items.append(ra_item)

    ra_items.sort(key=lambda x: x.frame_ts)
    ra_track = RaTrack.create(ra_items)
    return LrRaTrackset(lr_trackset, ra_track)


def _get_ra_track_items(ra_items: List[RaSceneItem],
                        ra_upscale: ScaleConf,
                        cam_service: ICamService) -> List[LrRaSceneItem]:

    tr_items: List[LrRaSceneItem] = []

    for ra_item in ra_items:
        x1, y1 = cam_service.mk_image_point(
            ra_item.scene_l_cm / 100.0 + ra_upscale.left,
            ra_upscale.top,
            ra_item.scene_f_cm / 100.0)

        x2, y2 = cam_service.mk_image_point(
            ra_item.scene_l_cm / 100.0 - ra_upscale.right,
            -ra_upscale.bottom,
            ra_item.scene_f_cm / 100.0)

        tr_items.append(LrRaSceneItem(None, ra_item, RectI(x1, y1, x2, y2)))

    return tr_items


def _match_ra_lr_items(lr_items: List[LrSceneItem], ra_items: List[LrRaSceneItem], systest: bool) -> LrRaScene:
    unused_lr_items: Dict[int, LrSceneItem] = dict([(x.oid, x) for x in lr_items])
    unused_ra_items: Dict[int, LrRaSceneItem] = dict([(x.ra_item.oid, x) for x in ra_items])

    lrra_items: List[LrRaSceneItem] = []
    for tr_ra_item in ra_items:
        ra_item = tr_ra_item.ra_item
        for lr_item in lr_items:
            if lr_item.dir == ra_item.speed_dir or systest:
                if lr_item.vrp.text_quad.center in tr_ra_item.ra_bbox:
                    unused_lr_items.pop(lr_item.oid, None)
                    unused_ra_items.pop(ra_item.oid, None)

                    tr_item = LrRaSceneItem(lr_item, ra_item, tr_ra_item.ra_bbox)
                    lrra_items.append(tr_item)

    return LrRaScene.create(
        lrra_items,
        list(unused_lr_items.values()),
        list(unused_ra_items.values()))


class BaseKernel:
    def __init__(
            self,
            services: ServicePack,
            scene_provider: ISceneProvider,
            track_provider: ITrackProvider,
            scene_emitters: List[ISceneEmitter],
            track_debouncer: ITrackDebouncer,
            lrra_match_store: LrRaMatchStore,
            ra_track_builder: IRaTrackBuilder,
            min_ra_track_len: int,
            ra_upscale: Optional[ScaleConf],
            systest: bool = False,
            plate_detected_mode: bool = False,
            cam_service: Optional[ICamService] = None):

        self._services: Final[ServicePack] = services

        self._min_ra_track_len = min_ra_track_len

        self._scene_react_producer = scene_provider
        self._scene_react_producer.set_handler(self._handle_scene_react)

        self._track_react_producer = track_provider
        self._track_react_producer.set_handler(self._handle_lr_track_react)

        self._lr_ra_match_store = lrra_match_store
        self._ra_track_builder: Final[IRaTrackBuilder] = ra_track_builder
        self._ra_track_builder.set_tout_handler(self._handle_ra_track_tout)

        self._scene_emitters: Final[List[ISceneEmitter]] = scene_emitters
        self._track_debouncer: ITrackDebouncer = track_debouncer

        self._ra_upscale = ra_upscale
        self._systest = systest
        self._plate_detected_mode = plate_detected_mode
        self._cam: Final[ICamService | None] = cam_service

    def destroy(self):
        self._services.destroy()

    def _handle_scene_react(self,
                            im_ref: Optional[ImageRef],
                            lr_scene: Optional[LrScene],
                            ra_scene: Optional[RaScene]):
        # Update _ra_track_store
        if ra_scene:
            self._ra_track_builder.add_scene(ra_scene.items)

        # Create the tr_scene from an avaliable lr_scene and ra_scene.
        if lr_scene and ra_scene:
            ra_items = _get_ra_track_items(ra_scene.items, self._ra_upscale, self._cam)
            tr_scene = _match_ra_lr_items(lr_scene, ra_items, self._systest)

            self._lr_ra_match_store.add(lr_scene.sender, tr_scene.tr_items)

        elif lr_scene:
            tr_scene = LrRaScene.create_lr(lr_scene)

        elif ra_scene:
            ra_items = _get_ra_track_items(ra_scene.items, self._ra_upscale, self._cam)
            tr_scene = LrRaScene.create_ra(ra_items)

        else:
            tr_scene = LrRaScene.create_empty()

        if im_ref:
            for consumer in self._scene_emitters:
                consumer.consume_scene(im_ref, tr_scene)

    def _handle_lr_track_react(self, lr_trackset: LrTrackSet) -> None:
        if not self._plate_detected_mode:
            if lr_trackset.lr_track.status != LrTrackStat.LOST:
                eventlog.log_d('skip not valid-lost track')
                return

        ra_track: Optional[RaTrack] = None
        ra_oids_desc: List[int] = self._lr_ra_match_store.get_ra_oids(
            lr_trackset.lr_track.source_id, lr_trackset.lr_track.oids)
        for ra_oid in ra_oids_desc:
            ra_track = self._ra_track_builder.get_track(ra_oid)
            if ra_track is not None:
                break

        if ra_track is None:
            tr_track = LrRaTrackset(lr_trackset, None)

        elif len(ra_track) < self._min_ra_track_len:
            tr_track = LrRaTrackset(lr_trackset, None)

        else:
            self._ra_track_builder.del_track(ra_track.oid)
            tr_track = _debouce_ra_items_pos(lr_trackset, ra_track)

        # noinspection PyBroadException
        try:
            self._track_debouncer.consume_track(tr_track)
        except GenericError:
            eventlog.log_w('fail')
        except Exception:
            eventlog.log_x()

    def _handle_ra_track_tout(self, ra_track: RaTrack) -> None:
        if len(ra_track) >= self._min_ra_track_len:
            tr_track = LrRaTrackset(None, ra_track)

            # noinspection PyBroadException
            try:
                self._track_debouncer.consume_track(tr_track)
            except GenericError:
                eventlog.log_w('fail')

            except Exception:
                eventlog.log_x()


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, conf: KernelConf):
        services = ServicePack()

        stats = StatService(
            twisted_reactor,
            scene_consumer_count=len(conf.scene_emitters),
            stat_filepath=conf.stats_service.stat_filepath,
            stat_tout_sec=conf.stats_service.stat_tout_sec)

        services.add(ILaneService, lane.init_service(conf.lane))

        services.add(ISceneProvider, SceneProvider(twisted_reactor, conf.scene_provider, stats.scene_provider))
        services.add(ITrackProvider, TrackProvider(twisted_reactor, conf.track_provider, stats.track_provider))

        next_scene_emitter_ref = 0
        scene_emitters: List[ISceneEmitter] = list()
        for scene_event_args in conf.scene_emitters:
            consumer = SceneEmitter(next_scene_emitter_ref, scene_event_args, stats.scene_event)
            scene_emitters.append(consumer)
            next_scene_emitter_ref += 1

        services.add(ITrackEmitter, init_service(conf.track_emitter, stats.track_emitter, services.get(ILaneService)))

        services.add(ITrackDebouncer, TrackDebouncer(
            twisted_reactor, services.get(ITrackEmitter), conf.track_emitter.debounce_tout, conf.image_store_args,
            stats.track_emitter))

        services.add(IRaTrackBuilder, RaTrackBuilder(twisted_reactor, conf.ra_track_store_tout, conf.systest))
        lr_ra_match_store = LrRaMatchStore(twisted_reactor, conf.lr_ra_oid_match_store_tout)

        if conf.cam is not None:
            services.add(ICamService, cam.init_service(conf.cam))

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', lambda: self.destroy())

        super().__init__(
            services,
            services.get(ISceneProvider),
            services.get(ITrackProvider),
            scene_emitters,
            services.get(ITrackDebouncer),
            lr_ra_match_store,
            services.get(IRaTrackBuilder),
            conf.min_ra_track_len,
            conf.ra_upscale_args,
            conf.systest,
            conf.plate_detected_mode,
            services.try_get(ICamService))

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stats

    def run(self):
        self._twisted_reactor.run()
