import os
from typing import Self

from twisted.internet import inotify
from twisted.python.filepath import FilePath

from edge import eventlog
from edge.support import fs
from edge.types.imageref import ImageRef
from edge.types.jnode.imageref import decode__image_ref_mc, decode__image_ref_mg

from tracker.services.stats import SceneReactCounters
from tracker.services.types.scene_provider import ISceneProvider, SceneProviderConf
from tracker.types.jnode.lr import decode__lr_scene
from tracker.types.jnode.ra import decode__ra_scene
from tracker.types.lr import LrSceneStat


class _ReactContext:
    def __init__(self, rm_delayed_call,
                 ignore_mc_scene_path: bool,
                 ignore_mg_scene_path: bool,
                 ignore_ra_scene_path: bool):
        """
        :param rm_delayed_call: twisted.internet.interfaces.IDelayedCall
        """

        self.rm_delayed_call = rm_delayed_call
        self._ignore_mc_scene_path = ignore_mc_scene_path
        self._ignore_mg_scene_path = ignore_mg_scene_path
        self._ignore_ra_scene_path = ignore_ra_scene_path

        self.mc_scene_path = None
        self.mg_scene_path = None
        self.lr_scene_path = None
        self.ra_scene_path = None

    def check_all(self) -> bool:
        if not self._ignore_mc_scene_path and not self.mc_scene_path:
            return False

        if not self._ignore_mg_scene_path and not self.mg_scene_path:
            return False

        if not self._ignore_ra_scene_path and not self.ra_scene_path:
            return False

        return self.lr_scene_path

    @classmethod
    def set_mc_scene_path_and_check_all(cls, mc_context: Self, mc_scene_abs_path: str) -> bool:
        mc_context.mc_scene_path = mc_scene_abs_path
        return mc_context.check_all()

    @classmethod
    def set_mg_scene_path_and_check_all(cls, mg_context: Self, mg_scene_abs_path: str) -> bool:
        mg_context.mg_scene_path = mg_scene_abs_path
        return mg_context.check_all()

    @classmethod
    def set_lr_scene_path_and_check_all(cls, lr_context: Self, lr_scene_abs_path: str) -> bool:
        lr_context.lr_scene_path = lr_scene_abs_path
        return lr_context.check_all()

    @classmethod
    def set_ra_scene_path_and_check_all(cls, ra_context: Self, ra_scene_abs_path: str) -> bool:
        ra_context.ra_scene_path = ra_scene_abs_path
        return ra_context.check_all()


class SceneProvider(ISceneProvider):
    # noinspection PyUnusedLocal
    def __init__(self, twisted_reactor, scene_react_args: SceneProviderConf, stat_counters: SceneReactCounters):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTime
        """

        ISceneProvider.__init__(self)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stat_counters

        self._mc_dir = scene_react_args.mc_dir
        self._mg_dir = scene_react_args.mg_dir
        self._lr_dir = scene_react_args.lr_dir
        self._ra_dir = scene_react_args.ra_dir
        self._wait_tout_sec = float(scene_react_args.tout_msec) / 1000.0

        self._react_ts_to_context_map = {}
        self._react_handler = None
        self._last_react_ts = 0
        self._inotify = inotify.INotify()
        self._inotify.startReading()

    def set_handler(self, handler) -> bool:
        # Allow to set the handler only once.
        if handler and not self._react_handler:

            if self._mc_dir:
                self._reset_inotify_watch(self._mc_dir, self._handle_mc_addscene_react)

            if self._mg_dir:
                self._reset_inotify_watch(self._mg_dir, self._handle_mg_addscene_react)

            if self._ra_dir:
                self._reset_inotify_watch(self._ra_dir, self._handle_ra_addscene_react)

            self._reset_inotify_watch(self._lr_dir, self._handle_lr_addscene_react)

            self._react_handler = handler
            return True

        return False

    # noinspection PyUnusedLocal
    def _handle_mc_addscene_react(self, ignored, filepath, mask) -> None:
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._mc_dir, self._handle_mc_addscene_react)
            eventlog.log_w(f'react dir was changed to {self._mc_dir}')
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_mc_scene_path_and_check_all)
        except Exception:
            eventlog.log_x()

    # noinspection PyUnusedLocal
    def _handle_mg_addscene_react(self, ignored, filepath, mask) -> None:
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._mg_dir, self._handle_mg_addscene_react)
            eventlog.log_w(f'react dir was changed to {self._mg_dir}')
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_mg_scene_path_and_check_all)
        except Exception:
            eventlog.log_x()

    # noinspection PyUnusedLocal
    def _handle_lr_addscene_react(self, ignored, filepath, mask) -> None:
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._lr_dir, self._handle_lr_addscene_react)
            eventlog.log_w(f'react dir was changed to {self._lr_dir}')
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_lr_scene_path_and_check_all)
        except Exception as x:
            eventlog.log_x()

    # noinspection PyUnusedLocal
    def _handle_ra_addscene_react(self, ignored, filepath, mask) -> None:
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._ra_dir, self._handle_ra_addscene_react)
            eventlog.log_w(f'react dir was changed to {self._ra_dir}')
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_ra_scene_path_and_check_all)
        except Exception as x:
            eventlog.log_x()

    def _handle_xx_addscene_react(self, xx_filepath, set_xx_scene_path_and_check_all_handler):
        if not (isinstance(xx_filepath, FilePath) and xx_filepath.isfile()):
            return

        react_ts = int(os.path.basename(xx_filepath.path))
        if react_ts < self._last_react_ts:
            fs.silent_rm_r(xx_filepath.path)
            return

        react_context = self._react_ts_to_context_map.get(react_ts, None)
        if react_context:
            all_available = set_xx_scene_path_and_check_all_handler(react_context, xx_filepath.path)
            if all_available:
                react_context.rm_delayed_call.cancel()
                self._handle_addscene_react(react_ts)
        else:
            rm_dalayed_call = self._twisted_reactor.callLater(
                self._wait_tout_sec, self._handle_addscene_react, react_ts)

            react_context = _ReactContext(rm_dalayed_call, self._mc_dir is None,
                                          self._mg_dir is None, self._ra_dir is None)
            set_xx_scene_path_and_check_all_handler(react_context, xx_filepath.path)
            self._react_ts_to_context_map[react_ts] = react_context

    def _handle_addscene_react(self, react_ts: int) -> None:
        react_context: _ReactContext = self._react_ts_to_context_map.pop(react_ts, None)
        # print(react_context)
        # print('mg_scene_path', react_context.mg_scene_path)
        # print('lr_scene_path', react_context.lr_scene_path)
        # print('ra_scene_path', react_context.ra_scene_path)
        # print('mc_scene_path', react_context.mc_scene_path)

        all_available = not react_context.rm_delayed_call.active()

        if react_ts > self._last_react_ts:
            # any of mg_scene, lr_scene and ra_scene are received during _tout
            self._last_react_ts = react_ts
        elif all_available:
            # all mg_scene, lr_scene and ra_scene are received during _tout, but a bit later then expected
            pass
        else:
            # it's too late to handle the add_scene react
            return

        #
        # deserialize scene image
        #

        im_ref = None
        # noinspection PyBroadException
        try:
            if react_context.mc_scene_path:
                mc_scene_jnode = fs.load_json3(react_context.mc_scene_path, default=None)
                if mc_scene_jnode:
                    im_ref = decode__image_ref_mc(mc_scene_jnode)
                    fs.silent_rm_r(react_context.mg_scene_path)
        except Exception:
            eventlog.log_x()

        finally:
            fs.silent_rm_r(react_context.mc_scene_path)

        if not im_ref:
            # noinspection PyBroadException
            try:
                if react_context.mg_scene_path:
                    mg_scene_jnode = fs.load_json3(react_context.mg_scene_path, default=None)
                    if mg_scene_jnode:
                        im_ref = decode__image_ref_mg(mg_scene_jnode)

            except Exception:
                eventlog.log_x()

            finally:
                fs.silent_rm_r(react_context.mg_scene_path)

        #
        # deserialize lr-scene
        #

        lr_scene = None
        # noinspection PyBroadException
        try:
            if react_context.lr_scene_path:
                lr_scene_jnode = fs.load_json3(react_context.lr_scene_path, default=None)
                if lr_scene_jnode:
                    lr_scene = decode__lr_scene(lr_scene_jnode)

                    if not im_ref:
                        im_ref = ImageRef(
                            lr_scene.imref.ts, lr_scene.imref.url, lr_scene.imref.codec,
                            lr_scene.imref.width, lr_scene.imref.height)

                    if lr_scene.stat == LrSceneStat.SUCCESS:
                        self._stat_counters.inc_lr_scene_item_count(len(lr_scene))
                    else:
                        lr_scene = None

        except Exception:
            eventlog.log_x()

        finally:
            fs.silent_rm_r(react_context.lr_scene_path)

        #
        # deserialize ra-scene
        #

        ra_scene = None
        try:
            if react_context.ra_scene_path:
                ra_scene_jnode = fs.load_json3(react_context.ra_scene_path, default=None)
                if ra_scene_jnode:
                    ra_scene = decode__ra_scene(ra_scene_jnode)
                    self._stat_counters.inc_ra_scene_item_count(len(ra_scene.items))

        except ValueError:
            eventlog.log_x()

        finally:
            fs.silent_rm_r(react_context.ra_scene_path)

        #
        # handle scene-react
        #

        if im_ref or lr_scene or ra_scene:
            # noinspection PyBroadException
            try:
                self._stat_counters.inc_scene_count()
                self._react_handler(im_ref, lr_scene, ra_scene)
            except Exception:
                eventlog.log_x()

    def _reset_inotify_watch(self, react_dir, handler) -> None:
        try:
            self._inotify.ignore(FilePath(react_dir))
        except KeyError:
            pass

        fs.silent_mkdir_r(react_dir)

        try:
            self._inotify.watch(
                FilePath(react_dir),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[handler])
        except inotify.INotifyError as x:
            eventlog.log_d(f'fail: reset inotify watch ({x})')
