import os

from time import time, gmtime, strftime
from typing import List

from edge import eventlog
from edge.services.types.service import IService
from edge.support import fs

from tracker.types.lrra import LrRaTrackset


class TrackReactCounters:
    def __init__(self):
        self.track_count = int(0)
        self.track_fail_count = int(0)

    def inc_track_count(self) -> None:
        self.track_count += 1

    def inc_track_fail_count(self) -> None:
        self.track_fail_count += 1


class SceneReactCounters:
    def __init__(self):
        self.scene_count = int(0)
        self.lr_scene_count = int(0)
        self.ra_scene_count = int(0)
        self.lr_scene_item_count = int(0)
        self.ra_scene_item_count = int(0)

    def inc_scene_count(self) -> None:
        self.scene_count += 1

    def inc_lr_scene_item_count(self, lr_scene_item_count) -> None:
        self.lr_scene_count += 1
        self.lr_scene_item_count += lr_scene_item_count

    def inc_ra_scene_item_count(self, ra_scene_item_count) -> None:
        self.ra_scene_count += 1
        self.ra_scene_item_count += ra_scene_item_count


class TrackEventCounters:
    def __init__(self):
        self.track_count = int(0)

        self.lrra_track_count = int(0)
        self.lr_track_count = int(0)
        self.ra_track_count = int(0)

        self.lr_track_dublicate_count = int(0)

        self.lrra_track_time = int(0)
        self.lr_track_time = int(0)
        self.ra_track_time = int(0)

    def inc_dublicate_count(self) -> None:
        self.lr_track_dublicate_count += 1

    def inc_track_count(self, tr_track: LrRaTrackset) -> None:
        self.track_count += 1

        if bool(tr_track.lr_trackset) and bool(tr_track.lr_trackset.lr_jnode) and bool(tr_track.ra_track):
            ctime = min(tr_track.lr_trackset.lr_jnode['PlateTrace'][0]['Ts'], tr_track.ra_track.min_ts)
            ltime = max(tr_track.lr_trackset.lr_jnode['PlateTrace'][-1]['Ts'], tr_track.ra_track.max_ts)

            self.lrra_track_time += ltime - ctime
            self.lrra_track_count += 1

        if bool(tr_track.lr_trackset) and bool(tr_track.lr_trackset.lr_jnode):
            self.lr_track_time += tr_track.lr_trackset.lr_jnode['PlateTrace'][-1]['Ts'] - \
                                  tr_track.lr_trackset.lr_jnode['PlateTrace'][0]['Ts']
            self.lr_track_count += 1

        if bool(tr_track.ra_track):
            self.ra_track_time += tr_track.ra_track.max_ts - tr_track.ra_track.min_ts
            self.ra_track_count += 1


class SceneEventCounters:
    def __init__(self, scene_consumer_count):
        self.scene_count_list = [int(0)] * scene_consumer_count

    def inc_scene_count(self, consumer_index: int) -> None:
        self.scene_count_list[consumer_index] += 1


class StatService(IService):
    def __init__(self, twisted_reactor, scene_consumer_count, stat_filepath, stat_tout_sec):
        self.track_provider = TrackReactCounters()
        self.scene_provider = SceneReactCounters()

        self.track_emitter = TrackEventCounters()
        self.scene_event = SceneEventCounters(scene_consumer_count)

        if stat_tout_sec:
            self._report_stat_counters(twisted_reactor, stat_filepath, stat_tout_sec)

    def _report_stat_counters(self, twisted_reactor, stat_filepath: str, stat_tout_sec: int) -> None:
        sys_utc_time = time()
        stat_file_lines: List[str] = [
            # header
            'stat.time-str: %s\n' % strftime('%d.%m.%Y %H:%M:%S', gmtime(sys_utc_time)),
            f'stat.time-sec: {int(round(sys_utc_time))}\n',
            f'stat.tout-sec: {stat_tout_sec}\n\n',

            # react-counters
            f'react.track-count: {self.track_provider.track_count}\n\n',
            f'react.track-fail-count: {self.track_provider.track_fail_count}\n\n',

            f'react.scene-count: {self.scene_provider.scene_count}\n',
            f'react.scene.lr-scene-count: {self.scene_provider.lr_scene_count}\n',
            f'react.scene.ra-scene-count: {self.scene_provider.ra_scene_count}\n',
            f'react.scene.lr-scene-item-count: {self.scene_provider.lr_scene_item_count}\n',
            f'react.scene.ra-scene-item-count: {self.scene_provider.ra_scene_item_count}\n\n',

            # event-counters
            f'event.track-count: {self.track_emitter.track_count}\n',
            f'event.track.lr-track-count: {self.track_emitter.lr_track_count}\n',
            f'event.track.lr-track-dup-count: {self.track_emitter.lr_track_dublicate_count}\n',
            f'event.track.ra-track-count: {self.track_emitter.ra_track_count}\n',
            f'event.track.lrra-track-count: {self.track_emitter.lrra_track_count}\n',

            # time-counters
            'event.track.lr-track-time: {:.3f}\n'.format(float(self.track_emitter.lr_track_time / 1000000.0)),
            'event.track.ra-track-time: {:.3f}\n'.format(float(self.track_emitter.ra_track_time / 1000000.0)),
            'event.track.lrra-track-time: {:.3f}\n'.format(float(self.track_emitter.lrra_track_time / 1000000.0))
        ]

        eventlog.log_d(f'tx-scene-count-list: {self.scene_event.scene_count_list}')

        try:
            tmp_filepath = stat_filepath + '.tmp'
            fs.write_text_file(stat_file_lines, tmp_filepath)
            os.rename(tmp_filepath, stat_filepath)
        except IOError as x:
            eventlog.log_w(f'fail: write <{stat_file_lines}> ({x})')

        twisted_reactor.callLater(
            stat_tout_sec, self._report_stat_counters, twisted_reactor, stat_filepath, stat_tout_sec)
