from typing import Callable, Dict, Final, List, Optional

from tracker.services.types.track_builder import IRaTrackBuilder
from tracker.types.ra import RaTrack, RaSceneItem


class _RaTrackContext:
    def __init__(self, ra_track: RaTrack, rm_delayed_call):
        self.ra_track: Final[RaTrack] = ra_track
        self.rm_delayed_call = rm_delayed_call
        self.rm_request = False


class RaTrackBuilder(IRaTrackBuilder):
    def __init__(self, twisted_reactor, tout_sec: int, systest: bool):
        self._twisted_reactor = twisted_reactor
        self._tout_sec = tout_sec
        self._systest: Final[bool] = systest

        self._oid_to_context_map: Dict[int, _RaTrackContext] = {}
        self._tout_handler: Callable[[RaTrack], None] | None = None

    def set_tout_handler(self, handler: Callable[[RaTrack], None]) -> None:
        self._tout_handler = handler

    def _handle_ra_track_tout(self, ra_oid) -> None:
        context: Optional[_RaTrackContext] = self._oid_to_context_map.get(ra_oid, None)
        if context and not context.rm_request:
            if self._tout_handler:
                self._tout_handler(context.ra_track)

        self._oid_to_context_map.pop(ra_oid, None)

    def add_scene(self, items: List[RaSceneItem]) -> None:
        for item in items:
            context: Optional[_RaTrackContext] = self._oid_to_context_map.get(item.oid, None)
            if context is None:
                # Create new ra_track.
                ra_track = RaTrack(item.oid, item)

                # Schedule to remove ra_track after _ra_track_tout_sec since the last update.
                rm_delayed_call = self._twisted_reactor.callLater(
                    self._tout_sec, self._handle_ra_track_tout, item.oid)

                self._oid_to_context_map[item.oid] = _RaTrackContext(ra_track, rm_delayed_call)

            else:
                # Update existing ra_track
                context.rm_delayed_call.reset(self._tout_sec)
                context.ra_track.append(item)

    def del_track(self, ra_oid: int) -> None:
        context: Optional[_RaTrackContext] = self._oid_to_context_map.get(ra_oid, None)
        if self._systest:
            context.rm_delayed_call.cancel()
            self._oid_to_context_map.pop(ra_oid, None)

        elif context:
            # ra_track will be removed automaticaly after _tout_sec timeout since the last update.
            # rm_request flag indicates that it's necessary to hide ra_track presence in the store for its clients.
            context.rm_request = True

    def get_track(self, ra_oid: int) -> Optional[RaTrack]:
        context: Optional[_RaTrackContext] = self._oid_to_context_map.get(ra_oid, None)
        if context and not context.rm_request:
            return context.ra_track
