from copy import deepcopy
from typing import Optional, Final, Dict

from edge import eventlog
from edge.support.queue import ImageFileQueue
from edge.types.exceptions import InternalError
from edge.types.imageref import ImageRef
from edge.types.jnode.imageref import decode__image_ref
from tracker.services.stats import TrackEventCounters
from tracker.services.types.track_debouncer import ITrackDebouncer, ImageQueueConf
from tracker.services.types.track_emitter import ITrackEmitter
from tracker.types.lr import LrTrackSet
from tracker.types.lrra import LrRaTrackset


class _ToutEntry:
    def __init__(self, track: LrRaTrackset, lr_imref: ImageRef, rm_delayed_call) -> None:
        self.tr_track = track
        self.lr_imref: ImageRef = lr_imref
        self.rm_delayed_call = rm_delayed_call


class TrackDebouncer(ITrackDebouncer):
    def __init__(self,
                 twisted_reactor,
                 next_consumer: ITrackEmitter,
                 debounce_tout: int,
                 imq_conf: Optional[ImageQueueConf],
                 stats: TrackEventCounters) -> None:

        self._twisted_reactor: Final = twisted_reactor
        self._next: Final[ITrackEmitter] = next_consumer
        self._debounce_tout: Final[int] = debounce_tout
        self._stats: Final[TrackEventCounters] = stats

        if imq_conf is None:
            self._im_store = None
        else:
            self._im_store = ImageFileQueue(
                twisted_reactor, None, imq_conf.endpoint,
                imq_conf.min_file_count,
                imq_conf.max_file_count,
                imq_conf.log_level)

        self._plate_text_to_tout_entry_map: Dict[str, _ToutEntry] = dict()

    def _handle_track_debounce_tout(self, plate_text: str) -> None:
        tout_entry: Optional[_ToutEntry] = self._plate_text_to_tout_entry_map.get(plate_text, None)
        if tout_entry is None:
            return

        try:
            self._next.consume_track(tout_entry.tr_track, tout_entry.lr_imref)
        finally:
            self._plate_text_to_tout_entry_map.pop(plate_text, None)

    def consume_track(self, next_tr_track: LrRaTrackset) -> None:
        """
        :raises RuntimeError
        """
        # tr-track doesn't contain the lr-track
        plate_text: Optional[str] = next_tr_track.lr_trackset and next_tr_track.lr_trackset.lr_track.plate_text
        if plate_text is None:
            self._next.consume_track(next_tr_track, None)
            return

        next_lr_trackset: LrTrackSet = next_tr_track.lr_trackset
        prev_entry: Optional[_ToutEntry] = self._plate_text_to_tout_entry_map.get(plate_text, None)

        if prev_entry is None:
            # add the new image to store
            im_jnode = next_lr_trackset.lr_jnode['Image']
            im_ref = decode__image_ref(im_jnode)
            if im_ref.url is not None and self._im_store:
                try:
                    store_img_ref = self._im_store.add_image_ref(im_ref)  # raises RuntimeError
                    im_ref.url = store_img_ref.url
                except RuntimeError:
                    eventlog.log_w(f'fail: ImageFileQueue::add_image_ref <{im_ref.url}>')
                    raise InternalError()

            assert plate_text is not None

            rm_delayed_call = self._twisted_reactor.callLater(
                self._debounce_tout, self._handle_track_debounce_tout, plate_text)

            self._plate_text_to_tout_entry_map[plate_text] = _ToutEntry(next_tr_track, im_ref, rm_delayed_call)

        else:
            self._stats.inc_dublicate_count()

            prev_lr_track: LrTrackSet = prev_entry.tr_track.lr_trackset

            # longer track is better
            if len(next_lr_trackset.lr_jnode['PlateTrace']) > len(prev_lr_track.lr_jnode['PlateTrace']):
                best_tr_track: LrRaTrackset = deepcopy(next_tr_track)
            else:
                best_tr_track: LrRaTrackset = deepcopy(prev_entry.tr_track)

            # add the new image to store
            im_jnode = best_tr_track.lr_trackset.lr_jnode['Image']
            im_ref = decode__image_ref(im_jnode)
            if im_ref.url is not None and self._im_store:
                try:
                    store_img_ref = self._im_store.add_image_ref(im_ref)  # raises RuntimeError
                    im_ref.url = store_img_ref.url

                    # remove the old one
                    self._im_store.del_image(prev_entry.lr_imref)

                except RuntimeError:
                    eventlog.log_w(f'fail: ImageFileQueue::add_image_ref <{im_ref.url}>')
                    raise InternalError()

                    # entry.tr_track always appears earlier than tr_track
            plate_trace = prev_lr_track.lr_jnode['PlateTrace'] + next_lr_trackset.lr_jnode['PlateTrace']
            best_tr_track.lr_trackset.lr_jnode['PlateTrace'] = plate_trace

            lr_track_refs = set(prev_lr_track.lr_jnode['Id'] + next_lr_trackset.lr_jnode['Id'])
            best_tr_track.lr_trackset.lr_jnode['Id'] = list(lr_track_refs)

            prev_entry.lr_imref = im_ref
            prev_entry.tr_track = best_tr_track
            # TODO : check ra_track's of entry.tr_track and tr_track (is it possible to have 2 r-track for 1 real car ?)
            prev_entry.rm_delayed_call.reset(secondsFromNow=self._debounce_tout)
