import os
from errno import ENOENT
from typing import Final, Optional

from edge import eventlog
from edge.services.types.lane import ILaneService
from edge.support import fs
from edge.support.dir import get_dir_track_lr
from edge.types.dir import Dir
from edge.types.imageref import ImageRef
from edge.types.shape.point import PointI, PointF
from edge.types.track import DirTrackItem, LaneTrack, DirTrack

from tracker.services.stats import TrackEventCounters
from tracker.services.types.track_emitter import ITrackEmitter, TrackEmitterConf
from tracker.types.jnode.lrra import encode__trackset
from tracker.types.lrra import LrRaTrackset
from tracker.types.ra import RaTrack
from tracker.types.lr import LrTrack


def create_lane_track_lrra(service: ILaneService, lrra_track: LrRaTrackset) -> Optional[LaneTrack]:

    if lrra_track.lr_trackset is not None:
        point_provider = ((item.ts, PointI(*item.plate_quad.center))
                          for item in lrra_track.lr_trackset.lr_track.plate_trace)
        return LaneTrack(*service.map_points_abs(point_provider))

    if lrra_track.ra_track is not None:
        point_provider = ((item.frame_ts, PointF(item.frame_l_rel, item.frame_t_rel))
                          for item in lrra_track.ra_track.trace)
        return LaneTrack(*service.map_points_rel(point_provider))


def create_dir_track_lrra(ra_track: Optional[RaTrack], lr_track: Optional[LrTrack]) -> Optional[DirTrack]:
    if ra_track is not None:
        dir_items = [
            DirTrackItem(item.frame_ts, item.speed_dir)
            for item in ra_track.trace if item.speed_dir in (Dir.CODE_A, Dir.CODE_R)
        ]
        if dir_items:
            return DirTrack(dir_items, ra_track.dir_2)

    if lr_track is not None:
        dir_items = get_dir_track_lr(lr_track.plate_trace, lr_track.dir)
        if dir_items:
            return DirTrack(dir_items, lr_track.dir)


class TrackEmitter(ITrackEmitter):
    def __init__(self, conf: TrackEmitterConf, stats: TrackEventCounters,
                 lane_service: ILaneService):
        super().__init__()

        self._stats = stats
        self._lane_service = lane_service

        self._event_dirs = conf.event_dirs
        for event_dirpath in self._event_dirs:
            fs.silent_mkdir_r(event_dirpath)

        self._event_file_prefix: Final[int] = os.getpid()
        self._event_file_seq: int = 0
        self._filter_conf = conf.ratrack_filter

    def consume_track(self, lrra_track: LrRaTrackset, lr_imref: Optional[ImageRef]) -> None:
        self._stats.inc_track_count(lrra_track)

        skip_ra = False
        if lrra_track.ra_track and lrra_track.lr_trackset:
            for filter_rule in self._filter_conf.rules:
                if lrra_track.lr_trackset.mmr_class in filter_rule.mmr_classes:
                    ra_speed: int = lrra_track.ra_track.speed_n_kmh
                    if ra_speed > filter_rule.max_speed:
                        skip_ra = True
                        break

        # noinspection PyBroadException
        try:
            lr_track: Optional[LrTrack] = lrra_track.lr_trackset and lrra_track.lr_trackset.lr_track
            dir_track: Optional[DirTrack] = create_dir_track_lrra(lrra_track.ra_track, lr_track)

            ra_track: Optional[RaTrack] = None
            if not skip_ra:
                ra_track = lrra_track.ra_track

            lane_track: Optional[LaneTrack] = create_lane_track_lrra(self._lane_service, lrra_track)
            trackset = encode__trackset(lr_imref, lrra_track.lr_trackset, ra_track, lane_track, dir_track)

            self._emit(trackset)

        except Exception:
            eventlog.log_x()

    def _emit(self, trackset: str) -> None:
        event_filename = f'{self._event_file_prefix}-{self._event_file_seq:d}'
        self._event_file_seq += 1

        for event_dir in self._event_dirs:
            try_count = 2
            track_dir_present = True
            event_abspath = os.path.join(event_dir, event_filename)
            while try_count > 0 and track_dir_present:
                try:
                    with open(event_abspath, 'w') as event_file:
                        event_file.write(trackset)

                    break

                except IOError as e:
                    if e.errno == ENOENT:
                        track_dir_present = False

                if not track_dir_present:
                    fs.silent_mkdir_r(event_dir)
                    track_dir_present = True
                try_count -= 1


def init_service(conf: TrackEmitterConf, stats: TrackEventCounters,
                 lane_service: ILaneService) -> ITrackEmitter:
    return TrackEmitter(conf, stats, lane_service)
