import collections
from typing import Dict, Optional, Tuple, List

from tracker.types.lrra import LrRaSceneItem

_RaOidToMatchCountMap = Dict[int, int]
_LrOidToRaOidToMatchCountMap = Dict[int, _RaOidToMatchCountMap]


class LrRaMatchStore:
    def __init__(self, twisted_reactor, match_tout_sec: int) -> None:
        self._twisted_reactor = twisted_reactor
        self._match_tout_sec = match_tout_sec

        self._lr_oid_to_delayed_call_map = dict()
        self._lr_oid_to_ra_oid_to_match_count_map: _LrOidToRaOidToMatchCountMap = collections.defaultdict(
            collections.Counter)

        self._lr_sender: Optional[Tuple[int, int]] = None

    def _handle_lr_oid_tout(self, lr_oid: int) -> None:
        self._lr_oid_to_ra_oid_to_match_count_map.pop(lr_oid, None)
        self._lr_oid_to_delayed_call_map.pop(lr_oid, None)

    def add(self, lr_sender: Tuple[int, int], lrra_items: List[LrRaSceneItem]) -> None:
        if self._lr_sender == lr_sender:
            for tr_item in lrra_items:
                # new lr_oid will be inserted.
                lr_oid = tr_item.lr_item.oid
                rm_delayed_call = self._lr_oid_to_delayed_call_map.get(lr_oid, None)
                if rm_delayed_call is None:
                    rm_delayed_call = self._twisted_reactor.callLater(
                        self._match_tout_sec, self._handle_lr_oid_tout, lr_oid)
                    self._lr_oid_to_delayed_call_map[lr_oid] = rm_delayed_call

                else:
                    rm_delayed_call.reset(self._match_tout_sec)

                self._lr_oid_to_ra_oid_to_match_count_map[lr_oid][tr_item.ra_item.oid] += 1

        else:
            self._lr_oid_to_ra_oid_to_match_count_map.clear()
            self._lr_oid_to_delayed_call_map.clear()
            self._lr_sender = lr_sender

    def get_ra_oids(self, lr_sender: Tuple[int, int], lr_oids: List[int]) -> List[int]:
        if self._lr_sender != lr_sender:
            return []

        final_ra_oid_to_match_count_map: _RaOidToMatchCountMap = collections.Counter()
        for lr_oid in lr_oids:
            ra_oid_to_match_count_map = self._lr_oid_to_ra_oid_to_match_count_map.get(lr_oid, None)
            if ra_oid_to_match_count_map:
                final_ra_oid_to_match_count_map += ra_oid_to_match_count_map

        # sort entries by count in desc order
        ra_oids_desc = sorted(final_ra_oid_to_match_count_map,
                              key=final_ra_oid_to_match_count_map.get,
                              reverse=True)

        return ra_oids_desc
