from twisted.internet import inotify
from twisted.python.filepath import FilePath

from edge import eventlog
from edge.support import fs

from tracker.services.stats import TrackReactCounters
from tracker.services.types.track_provider import TrackProviderConf, ITrackProvider
from tracker.types.jnode.lr import decode__lr_track


class TrackProvider(ITrackProvider):
    def __init__(self, twisted_reactor, track_react_args: TrackProviderConf, stat_counters: TrackReactCounters):
        ITrackProvider.__init__(self)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stat_counters

        self._react_dir = track_react_args.react_dir

        self._react_handler = None
        self._inotify = inotify.INotify()
        self._inotify.startReading()

    def set_handler(self, handler):
        self._react_handler = handler
        self._reset_inotify_watch()

    # noinspection PyUnusedLocal
    def _handle_track_react(self, ignored, filepath, mask) -> None:
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_w(f'react dir was changed to {self._react_dir}')
            return

        lr_track = None
        try:
            lr_track_jnode = fs.load_json3(filepath.path, default=None)
            if lr_track_jnode:
                lr_track = decode__lr_track(lr_track_jnode)

        except ValueError:
            eventlog.log_x()
            return

        finally:
            fs.silent_rm_r(filepath.path)

        if lr_track is None:
            self._stat_counters.inc_track_fail_count()
            return

        # noinspection PyBroadException
        try:
            self._stat_counters.inc_track_count()
            self._react_handler(lr_track)
        except Exception:
            eventlog.log_x()

    def _reset_inotify_watch(self) -> None:
        try:
            self._inotify.ignore(FilePath(self._react_dir))
        except KeyError:
            # it's ok, if filepath is not watched
            pass

        fs.silent_mkdir_r(self._react_dir)

        try:
            self._inotify.watch(
                FilePath(self._react_dir),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_track_react])
        except inotify.INotifyError:
            eventlog.log_x()
