from collections import defaultdict
from typing import Dict, Final, List, Optional, Self

import ujson

from edge import eventlog
from edge.services.types.lane import LaneServiceConf
from edge.types.exceptions import ParamsError
from edge.services.types.jnode.cam import decode__cam_service_conf
from edge.services.types.cam import CamServiceConf
from tracker.services.types.jnode.lane import decode__road_conf
from tracker.services.types.jnode.stats import StatsServiceConf
from tracker.services.types.scene_emitter import SceneEmitterConf
from tracker.services.types.scene_provider import SceneProviderConf
from tracker.services.types.track_emitter import TrackEmitterConf
from tracker.services.types.track_debouncer import ImageQueueConf
from tracker.services.types.track_provider import TrackProviderConf


class ScaleConf:
    def __init__(self, left: float, top: float, right: float, bottom: float):
        self.left: Final[float] = left
        self.top: Final[float] = top
        self.right: Final[float] = right
        self.bottom: Final[float] = bottom

    @classmethod
    def deserialize(cls, jnode: List[float]) -> Self:
        """
        :param jnode:
            main.conf#/ra-vehicle-ltrb
        """
        if len(jnode) != 4:
            eventlog.log_e_param('#/')
            raise ParamsError

        return cls(*jnode)

    def serialize(self) -> List[float]:
        return [self.left, self.top, self.right, self.bottom]


class KernelConf:
    DEF_LR_RA_OID_MATCH_STORE_TOUT_SEC = 5
    DEF_RA_TRACK_STORE_TOUT_SEC = 5
    DEF_MIN_RATRACK_LEN = 5

    def __init__(self, stats_service: StatsServiceConf,
                 scene_provider: SceneProviderConf,
                 track_provider: TrackProviderConf,
                 scene_emitters: List[SceneEmitterConf],
                 track_emitter: TrackEmitterConf,
                 lr_ra_oid_match_store_tout: int,
                 ra_track_store_tout: int,
                 min_ra_track_len: int,
                 image_store_args: Optional[ImageQueueConf],
                 lane_service: LaneServiceConf,
                 systest: bool = False,
                 plate_detected_mode: bool = False,
                 ra_upscale_args: Optional[ScaleConf] = None,
                 cam_service: Optional[CamServiceConf] = None):

        self.stats_service = stats_service

        self.scene_provider = scene_provider
        self.track_provider = track_provider

        self.scene_emitters = scene_emitters
        self.track_emitter = track_emitter

        self.lr_ra_oid_match_store_tout = lr_ra_oid_match_store_tout
        self.ra_track_store_tout = ra_track_store_tout
        self.min_ra_track_len = min_ra_track_len

        self.image_store_args = image_store_args
        self.systest = systest
        self.plate_detected_mode = plate_detected_mode
        self.ra_upscale_args = ra_upscale_args
        self.cam = cam_service
        self.lane = lane_service


def decode__kernel_conf(main_conf_jnode: Dict, misc_conf_jnode: Dict) -> KernelConf:
    # define stat-counters-args
    monitoring_jnode = main_conf_jnode.get('Monitoring')
    if not monitoring_jnode:
        raise ValueError('main.conf#/Monitoring')

    stat_counters_args = StatsServiceConf.deserialize(monitoring_jnode)

    # define others
    file_api_jnode = main_conf_jnode['FileApi']
    track_event_args = TrackEmitterConf.deserialize(file_api_jnode['TrackEvent'])
    scene_react_args = SceneProviderConf.deserialize(file_api_jnode['AddSceneReact'])
    track_react_args = TrackProviderConf.deserialize(file_api_jnode['AddTrackReact'])
    image_store_jnode = main_conf_jnode.get('store')
    image_store_args = ImageQueueConf.deserialize(image_store_jnode) if image_store_jnode else None

    road_conf_jnode = {}
    road_conf_filepath = main_conf_jnode.get('services', dict()).get('road.conf')
    if road_conf_filepath:
        with open(road_conf_filepath) as jnode:
            road_conf_jnode = ujson.load(jnode)

    # define others

    lane_service_args = decode__road_conf(road_conf_jnode)

    skip_rate_to_dir_set_map = defaultdict(set)
    for entry in file_api_jnode['SceneEvent']:
        skip_rate_to_dir_set_map[entry["SkipRate"]].add(entry["Dir"])

    scene_event_args_list = [SceneEmitterConf.deserialize(skip_rate_event_dirs[1], skip_rate_event_dirs[0])
                             for skip_rate_event_dirs in list(skip_rate_to_dir_set_map.items())]

    lr_ra_oid_match_store_tout = misc_conf_jnode.get(
        'lpr_track_caching_time', KernelConf.DEF_LR_RA_OID_MATCH_STORE_TOUT_SEC)

    ra_track_store_tout = misc_conf_jnode.get(
        'radar_track_caching_time', KernelConf.DEF_RA_TRACK_STORE_TOUT_SEC)

    min_ra_track_len = misc_conf_jnode.get(
        'min_valid_radar_track_length', KernelConf.DEF_MIN_RATRACK_LEN)

    systest = main_conf_jnode.get('systest', False)
    plate_detected_mode = main_conf_jnode.get('plate-detected-mode', False)

    ra_upscale_jnode = main_conf_jnode.get('ra-vehicle-ltrb')
    ra_upscale_args = None
    if ra_upscale_jnode:
        try:
            ra_upscale_args = ScaleConf.deserialize(ra_upscale_jnode)
        except ParamsError:
            eventlog.log_e_param('#/ra-vehicle-ltrb')
            raise ValueError()

    services_jnode = main_conf_jnode.get('services') or dict()
    cam_conf = None
    if ra_upscale_args is not None:
        cam_jnode = services_jnode.get('cam')
        if cam_jnode is None:
            eventlog.log_e_param('#/services/cam')
            raise ValueError()

        # radar terminal
        cam_conf = decode__cam_service_conf(cam_jnode)

    return KernelConf(
        stat_counters_args, scene_react_args, track_react_args, scene_event_args_list, track_event_args,
        lr_ra_oid_match_store_tout, ra_track_store_tout, min_ra_track_len, image_store_args,
        lane_service_args, systest, plate_detected_mode, ra_upscale_args, cam_conf)
