from enum import Enum
from typing import Dict, Final, List, Optional, Tuple

from edge.types.dir import Dir
from edge.types.imageref import ImageRef
from edge.types.shape.quad import QuadI
from edge.types.track import LrTrackItem


class LrVrp:
    """
    Vehicle registration plate
    """
    def __init__(self, plate_text: str, quad: QuadI) -> None:
        self.text_ucode: Final[str] = plate_text
        self.text_quad: Final[QuadI] = quad


class LrSceneItem:
    def __init__(self, item_ts: int, item_oid: int, item_vrp: LrVrp, item_dir: Dir, item_trace: List[int]) -> None:
        """
        :param item_ts:
            The tracking object timestamp.
        :param item_oid:
            The tracking object ID.
        :param item_vrp:
            The vehicle registration plate.
        :param item_dir:
            The direction of the vehicle's movement relative to the camera.
        :param item_trace:
        """
        self.ts: Final[int] = item_ts
        self.oid: Final[int] = item_oid
        self.vrp: Final[LrVrp] = item_vrp
        self.dir: Final[Dir] = item_dir
        self.trace: Final[List] = item_trace


class LrSceneStat(Enum):
    SUCCESS = 1
    TIMEOUT = 2
    FAILURE = 3


class LrScene(list[LrSceneItem]):
    def __init__(self, sender: Tuple[int, int], stat: LrSceneStat, imref: ImageRef, items: List[LrSceneItem]) -> None:
        super().__init__(items)

        self.sender: Final[Tuple[int, int]] = sender
        self.stat: Final[LrSceneStat] = stat

        self.imref: Final[ImageRef] = imref


class LrTrackStat(Enum):
    UNKNOWN = 0
    DETECTED = 1
    LOST = 2


class LrTrack:
    def __init__(self,
                 source_id: Tuple[int, int],
                 oids: List[int],
                 plate_text: str,
                 plate_trace: List[LrTrackItem],
                 status: LrTrackStat,
                 vrp_dir: Dir) -> None:

        self.status: Final[LrTrackStat] = status
        self.oids: Final[List[oids]] = oids
        self.dir: Final[Dir] = vrp_dir
        self.source_id: Final[Tuple[int, int]] = source_id
        self.plate_text: Final[str] = plate_text
        self.plate_trace: Final[List[LrTrackItem]] = plate_trace


class LrTrackSet:
    def __init__(self,
                 lr_track: LrTrack,
                 lr_jnode: Dict,
                 wa_jnode: Optional[Dict],
                 alarms_jnode: List[Dict] | None) -> None:

        self.lr_track: Final[LrTrack] = lr_track
        self.lr_jnode: Final[Dict] = lr_jnode
        self.wa_jnode: Final[Dict | None] = wa_jnode

        self.alarms_jnode: Final[List[Dict] | None] = alarms_jnode

    @property
    def mmr_class(self) -> str:
        type_jnode = self.lr_jnode.get('Vehicle', {}).get('type', {})
        return type_jnode.get('class', 'UNKNOWN') if type_jnode else 'UNKNOWN'
